// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.asmkey.gui;


import java.io.StringReader;
import java.util.Iterator;
import java.util.List;

import de.uka.ilkd.asmkey.proof.*;
import de.uka.ilkd.key.pp.AbbrevMap;
import de.uka.ilkd.key.gui.KeYMediator;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.parser.ParserException;
import de.uka.ilkd.key.parser.TermParserFactory;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.Node;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.util.Debug;


/** Class for loading and replaying proofs. Most of the code has been
 * copied from the package 'key.proof' and has been adapted for the
 * abstract proof representation (instead of the event driven proof
 * loading).
 *
 * @author Hubert Schmid
 */

final class ProofLoader {

    /** Return the SchemaVariable with the given name from the given
     * set. If no schema variable with this name exists, then null is
     * returned. */
    private static SchemaVariable lookupName(SetOfSchemaVariable vars, String name) {
        for (IteratorOfSchemaVariable i = vars.iterator(); i.hasNext();) {
            SchemaVariable sv = i.next();
            if (sv.name().toString().equals(name)) {
                return sv;
            }
        }
        return null;
    }

    /** The mediator used for loading proofs. */
    private final KeYMediator mediator;

    ProofLoader(KeYMediator mediator) {
        this.mediator = mediator;
    }

    /** Load and replay the given proof. */
    public void loadProof(ProofNode proof) {
        loadProofBranch(proof, mediator.getSelectedProof().root());
    }

    private void loadProofBranch(ProofNode proof, Node node) {
        // rules
        List rules = proof.getRules();
        for (int i = 0; i < rules.size(); ++i) {
            if (i > 0) {
                node = node.child(0);
            }
            ProofRule rule = (ProofRule) rules.get(i);
            loadProofRule(rule, node);
        }
        // child nodes
        List children = proof.getChildren();
        for (int i = 0; i < children.size(); ++i) {
            ProofNode child = (ProofNode) children.get(i);
            loadProofBranch(child, node.child(i));
        }
    }

    private void loadProofRule(ProofRule rule, Node node) {
        Goal goal = mediator.getSelectedProof().getGoal(node);

        PosInOccurrence pos =
            PosInOccurrence.findInSequent(goal.sequent(), rule.getFormula(), rule.getPosition());
        SetOfTacletApp applics =
            mediator.getTacletApplications(goal,
                                           rule.getName(),
                                           pos);
        if (applics.size() != 1) {
            throw new RuntimeException(rule.getName()
                                       + ": found "
                                       + applics.size()
                                       + " applications. Don't know what to do!");
        }
        TacletApp app = applics.iterator().next();

        SetOfSchemaVariable vars = app.uninstantiatedVars();

	if (rule instanceof ProofTaclet) {

	    // first pass: add variables
	    for (Iterator i = ((ProofTaclet)rule).getInstantiations().iterator(); i.hasNext();) {
		ProofInstantiation inst = (ProofInstantiation) i.next();
		SchemaVariable sv = lookupName(vars, inst.getVariable());
		if (sv.isVariableSV()) {
		    LogicVariable lv = new LogicVariable(new Name(inst.getVariable()), ((SortedSchemaVariable)sv).sort());
		    Term instance = TermFactory.DEFAULT.createVariableTerm(lv);
		    app = app.addCheckedInstantiation(sv, instance, true);
		}
	    }
	    
	    // second pass: add everything else
	    for (Iterator i = ((ProofTaclet)rule).getInstantiations().iterator(); i.hasNext();) {
		ProofInstantiation inst = (ProofInstantiation) i.next();
		SchemaVariable sv = lookupName(vars, inst.getVariable());
		if (sv.isVariableSV()) {
		    // ignore -- already done
		} else if (sv.isProgramSV()) {
		    // does not exists with ASMs
		    Debug.fail("not reached");
		} else if (sv.isDependingSV()) {
		    String varname = inst.getVariable();
		    String value = varname.substring(varname.indexOf('=') + 1);
		    Function c = new Function(new Name(value), ((SortedSchemaVariable)sv).sort(), new Sort[0]);
		    app = app.addInstantiation(sv, TermFactory.DEFAULT.createFunctionTerm(c), true);
		} else {
		    Namespace varNS = mediator.var_ns();
		    varNS = goal.getVariableNamespace(varNS);
		    varNS = app.extendVarNamespaceForSV(varNS, sv);
		    String text = inst.getTerm();
		    try {
			if(sv instanceof SortedSchemaVariable) {
			    Term term = TermParserFactory
				.createInstance().parse(new StringReader(text),
							((SortedSchemaVariable)sv).sort(),
							mediator.getServices(),
							varNS,
							mediator.func_ns(),
							mediator.sort_ns(),
							mediator.progVar_ns(),
							new AbbrevMap());
			    app = app.addCheckedInstantiation((SortedSchemaVariable)sv, term, true);
			}else{
			    Debug.fail(sv+" should be of type SortedSchemaVariable (or?)");
			}
		    } catch (ParserException pe) {
			throw new RuntimeException("While parsing " + text + ": " + pe);
		    }
		}
	    }
	    
	    ListOfIfFormulaInstantiation ifSeqFormulaList = SLListOfIfFormulaInstantiation.EMPTY_LIST;
	    for (Iterator i = ((ProofTaclet)rule).getIfSeqFormulas().iterator(); i.hasNext();) {
		int value = ((Integer) i.next()).intValue();
		ifSeqFormulaList = ifSeqFormulaList.append(new IfFormulaInstSeq(goal.sequent(), value));
	    }
	    app = app.setIfFormulaInstantiations(ifSeqFormulaList,
						 mediator.getServices(),
						 mediator.getUserConstraint().getConstraint());
	}

        app = app.createDependingFunctions(mediator.func_ns());

        goal.apply(app);
    }

}
