// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.asmkey.parser.ast;

import de.uka.ilkd.asmkey.util.ArrayUtil;
import de.uka.ilkd.key.parser.Location;

/** Implementation for abstract factory {@link AstFactory}. The
 * factory uses inner, anonymous classes to implement the various AST
 * nodes. They may only be accessed with the visitors.
 *
 * @author Hubert Schmid
 * @author Stanislas Nanchen 
 */

public final class AstFactoryImpl implements AstFactory {

    /* --- first order logic terms --- */

    public AstTerm createTerm(Location location, final AstOperator op, final AstTerm[] terms) {
        return new AstTerm(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitOperatorTerm(closure, op, (AstTerm[]) ArrayUtil.copy(terms));
                }

                /** for debug only */
                public String toString() {
                    return "[AstTerm:op=" + op + ",terms=" + ArrayUtil.toString(terms) + "]";
                }

            };
    }

    public AstTerm createTerm(Location location,
                              final AstQuantifier q,
                              final Identifier var,
                              final Identifier sort,
                              final AstTerm term) {
        return new AstTerm(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitQuantifierTerm(closure, q, var, sort, term);
                }

                /** for debug only */
                public String toString() {
                    return "[AstTerm:q=" + q + ",var=" + var + ",term=" + term + "]";
                }

            };
    }

    public AstTerm createTerm(Location location, final Identifier id, final AstTerm[] terms) {
        return new AstTerm(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitFunctionTerm(closure, id, (AstTerm[]) ArrayUtil.copy(terms));
                }

                /** for debug only */
                public String toString() {
                    return "[AstTerm:id=" + id + ",terms=" + ArrayUtil.toString(terms) + "]";
                }

            };
    }

    public AstTerm createSubstitutionTerm(Location location,
                                          final Identifier var,
                                          final Identifier sort,
                                          final AstTerm subst,
                                          final AstTerm term) {
        return new AstTerm(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitSubstitutionTerm(closure, var, sort, subst, term);
                }

                /** for debug only */
                public String toString() {
                    return "[AstTerm:var=" + var + ",sort=" + sort + ",subst=" + subst + ",term=" + term + "]";
                }

            };
    }

    public AstTerm createBigTerm(Location location,
				 final AstTerm term,
				 final boolean and_flag)
    {
	return new AstTerm(location) {
		
		public Object accept(AstTermVisitor visitor, Object closure) {
		    return visitor.visitBigTerm(closure, term, and_flag);
		}

		/** for debug only */
		public String toString() {
		    return "[AstTerm:and_flag=" + and_flag + ",term=" + term + "]";
		}
	    };
    }
    
    public AstTerm createBigTerm(Location location,
				 final Identifier id,
				 final AstTerm form,
				 final boolean and_flag,
				 final int type_flag,
				 final int prop_flag,
				 final Identifier[] rules)
    {
	return new AstTerm(location) {
		
		public Object accept(AstTermVisitor visitor, Object closure) {
		    return visitor.visitBigTerm(closure, id, form, and_flag, type_flag, prop_flag, rules);
		}

		/** for debug only */
		public String toString() {
		    return "[AstTerm:id=" + id + ",form=" + form + ",and_flag:" + 
			and_flag + ",type_flag=" + type_flag + ",prop_flag:" + prop_flag
			+ "rules=" + rules + "]";
		}
	    };
    }

    public AstTerm createFunctionTerm(Location location,
				      final Identifier id,
				      final AstTerm func,
				      final AstTerm form) {
	return new AstTerm(location) {
		
		public Object accept(AstTermVisitor visitor, Object closure) {
		    return visitor.visitFunctionTerm(closure, id, func, form);
		}

		/** for debug only */
		public String toString() {
		    return "[AstTerm:id=" + id + "func=" + func + ",form=" + form + "]";
		}
	    };
    }

    public AstTerm createAbbreviationTerm(Location location, final Identifier name) {
        return new AstTerm(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitAbbreviationTerm(closure, name);
                }

                /** for debug only */
                public String toString() {
                    return "[AstTerm:abbr=" + name + "]";
                }

            };
    }

    /* --- sequences and sets */

    public AstTerm createSetTerm(Location location, final AstTerm[] elems)
    {
	return new AstTerm(location) {
		
		public Object accept(AstTermVisitor visitor, Object closure) {
		    return visitor.visitSetTerm(closure, elems);
		}

		/** for debug only */
		public String toString() {
		    String elemsString = elems==null?"":elems.toString();
		    return "[AstTerm:set=:[" + elemsString + "]";
		}
	    };
    }

    public AstTerm createSequenceTerm(Location location, final AstTerm[] elems, final AstTerm tail)
    {
	return new AstTerm(location) {
		
		public Object accept(AstTermVisitor visitor, Object closure) {
		    return visitor.visitSequenceTerm(closure, elems, tail);
		}

		/** for debug only */
		public String toString() {
		    String elemsString = elems==null?"":elems.toString();
		    String tailString = tail==null?"":tail.toString();
		    return "[AstTerm:sequence=:<" + elemsString + ":" + tailString + ">";
		}
	    };
    }

    /* --- dynamic logic --- */

    public AstTerm createDlBox(Location location, final AstAsmRule rule, final AstTerm term, final AstTerm step) {
        return new AstTerm(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitBox(closure, rule, term, step);
                }

                /** for debug only */
                public String toString() {
                    return "[AstTerm(Box):rule=" + rule + ",term=" + term + "]";
                }

            };
    }

    public AstTerm createDlDiamond(Location location, final AstAsmRule rule, final AstTerm term, final AstTerm step) {
        return new AstTerm(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitDiamond(closure, rule, term, step);
                }

                /** for debug only */
                public String toString() {
                    return "[AstTerm(Diamond):rule=" + rule + ",term=" + term + "]";
                }

            };
    }

    public AstTerm createDlDef(Location location, final AstAsmRule rule, final AstTerm step) {
        return new AstTerm(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitDef(closure, rule, step);
                }

                /** for debug only */
                public String toString() {
                    return "[AstTerm(def):rule=" + rule + "]";
                }

            };
    }

    public AstTerm createDlUpd(Location location, final AstTerm lhs, final AstTerm rhs, final AstAsmRule rule, final AstTerm step) {
        return new AstTerm(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitUpd(closure, lhs, rhs, rule, step);
                }

                /** for debug only */
                public String toString() {
                    return "[AstTerm(upd):lhs=" + lhs + ",rhs=" + rhs + ",rule=" + rule + "]";
                }

            };
    }

    public AstTerm createDlCon(Location location, final AstAsmRule rule, final AstTerm step) {
        return new AstTerm(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitCon(closure, rule, step);
                }

                /** for debug only */
                public String toString() {
                    return "[AstTerm(Con):rule=" + rule + "]";
                }

            };
    }

    public AstTerm createDlInv(Location location, final AstTerm term, final AstAsmRule rule, final AstTerm step) {
        return new AstTerm(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitInv(closure, term, rule, step);
                }

                /** for debug only */
                public String toString() {
                    return "[AstTerm(inv):term=" + term + ",rule=" + rule + "]";
                }

            };
    }

    public AstTerm createDlJoinable(Location location, final AstAsmRule rule1, final AstAsmRule rule2, final AstTerm step) {
        return new AstTerm(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitJoinable(closure, rule1, rule2, step);
                }

                /** for debug only */
                public String toString() {
                    return "[AstTerm(joinable):rule1=" + rule1 + ",rule2=" + rule2 + "]";
                }

            };
    }

    public AstTerm createDlEqui(Location location, final AstAsmRule rule1, final AstAsmRule rule2, final AstTerm step) {
        return new AstTerm(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitEqui(closure, rule1, rule2, step);
                }

                /** for debug only */
                public String toString() {
                    return "[AstTerm(equi):rule1=" + rule1 + ",rule2=" + rule2 + "]";
                }

            };
    }

    public AstTerm createDlEquiS(Location location, final AstAsmRule rule1, final AstAsmRule rule2, final AstTerm step) {
        return new AstTerm(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitEquiS(closure, rule1, rule2, step);
                }

                /** for debug only */
                public String toString() {
                    return "[AstTerm(equiS):rule1=" + rule1 + ",rule2=" + rule2 + "]";
                }

            };
    }

    public AstTerm createDlEquiA(Location location, final AstAsmRule rule1, final AstAsmRule rule2, final AstTerm step) {
        return new AstTerm(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitEquiA(closure, rule1, rule2, step);
                }

                /** for debug only */
                public String toString() {
                    return "[AstTerm(equiA):rule1=" + rule1 + ",rule2=" + rule2 + "]";
                }

            };
    }

    public AstTerm createDlAccT(Location location, final AstTerm term1, final AstTerm term2, final AstTerm step) {
	return new AstTerm(location) {

		public Object accept(AstTermVisitor visitor, Object closure) {
		    return visitor.visitAccT(closure, term1, term2, step);
		}

		/** for debug only */
		public String toString() {
		    return "[AstTerm(accT):term1=" + term1 + ",term2=" + term2 + "]";
		}
		
	    };
    }

    public AstTerm createDlAccF(Location location, final AstTerm term, final AstTerm formula, final AstTerm step) {
	return new AstTerm(location) {

		public Object accept(AstTermVisitor visitor, Object closure) {
		    return visitor.visitAccF(closure, term, formula, step);
		}

		/** for debug only */
		public String toString() {
		    return "[AstTerm(accF):term=" + term + ",formula=" + formula + "]";
		}
		
	    };
    }

    public AstTerm createDlAcc(Location location, final AstTerm term, final AstAsmRule rule, final AstTerm step) {
	return new AstTerm(location) {
		
		public Object accept(AstTermVisitor visitor, Object closure) {
		    return visitor.visitAcc(closure, term, rule, step);
		}

		/** for debug only */
		public String toString() {
		    return "[AstTerm(acc):term=" + term + ",rule=" + rule + "]";
		}
	    };
    }
		
    /* --- taclets --- */

    public AstCondition createOperatorEqualsCondition(Location location,
                                                      final boolean equal,
                                                      final Identifier v,
                                                      final Identifier w) {
        return new AstCondition() {

                public void accept(AstConditionVisitor visitor) {
                    visitor.visitOperatorEquals(equal, v, w);
                }

                /** for debug only */
                public String toString() {
                    return "[Condition(op):equal=" + equal + ",v=" + v + ",w=" + w + "]";
                }

            };
    }

    public AstCondition createNotFreeCondition(Location location, final Identifier v, final Identifier w) {
        return new AstCondition() {

                public void accept(AstConditionVisitor visitor) {
                    visitor.visitNotFreeIn(v, w);
                }

                /** for debug only */
                public String toString() {
                    return "[Condition(not-free-in):v=" + v + ",w=" + w + "]";
                }

            };
    }

    public AstCondition createPureCondition(Location location, final Identifier v) {
        return new AstCondition() {

                public void accept(AstConditionVisitor visitor) {
                    visitor.visitPure(v);
                }

                /** for debug only */
                public String toString() {
                    return "[Condition(pure):v=" + v + "]";
                }

            };
    }

    public AstCondition createStaticCondition(Location location, final Identifier v) {
        return new AstCondition() {

                public void accept(AstConditionVisitor visitor) {
                    visitor.visitStatic(v);
                }

                /** for debug only */
                public String toString() {
                    return "[Condition(static):v=" + v + "]";
                }

            };
    }

    public AstCondition createStaticArgsCondition(Location location, final Identifier v) {
        return new AstCondition() {

                public void accept(AstConditionVisitor visitor) {
                    visitor.visitStaticArgs(v);
                }

                /** for debug only */
                public String toString() {
                    return "[Condition(staticargs):v=" + v + "]";
                }

            };
    }

    public AstCondition createDynamicCondition(Location location, final Identifier v) {
        return new AstCondition() {

                public void accept(AstConditionVisitor visitor) {
                    visitor.visitDynamic(v);
                }

                /** for debug only */
                public String toString() {
                    return "[Condition(dynamic):v=" + v + "]";
                }

            };
    }

    public AstCondition createAtomicCondition(Location location, final Identifier v) {
	return new AstCondition() {
		
		public void accept(AstConditionVisitor visitor) {
		    visitor.visitAtomic(v);
		}

		/** for debug only */
		public String toString() {
		    return "[Condition(atomic):v=" + v + "]";
		}
	    };
    }

    public AstCondition createCallCondition(Location location, final Identifier v) {
        return new AstCondition() {

                public void accept(AstConditionVisitor visitor) {
                    visitor.visitCall(v);
                }

                /** for debug only */
                public String toString() {
                    return "[Condition(call):v=" + v + "]";
                }

            };
    }

    public AstCondition createSkolemCondition(Location location, final Identifier v, final Identifier w) {
        return new AstCondition() {

                public void accept(AstConditionVisitor visitor) {
                    visitor.visitSkolem(v, w);
                }

                /** for debug only */
                public String toString() {
                    return "[Condition(skolem):v=" + v + ",w=" + w + "]";
                }

            };
    }

    public AstCondition createMayUpdateCondition(Location location, final Identifier r, final Identifier f) {
	return new AstCondition() {
		
		public void accept(AstConditionVisitor visitor) {
		    visitor.visitMayUpdate(true, r, f);
		}

		/** for debug only */
		public String toString() {
		    return "[Condition(mayUpdate):r=" + r + ",f=" + f + "]";
		}
	    };
    }

    public AstCondition createNotUpdateCondition(Location location, final Identifier r, final Identifier f) {
	return new AstCondition() {
		
		public void accept(AstConditionVisitor visitor) {
		    visitor.visitMayUpdate(false, r, f);
		}

		/** for debug only */
		public String toString() {
		    return "[Condition(notUpdate):r=" + r + ",f=" + f + "]";
		}
	    };
    }

    public AstCondition createMayAccessCondition(Location location, final Identifier r, final Identifier f) {
	return new AstCondition() {
		
		public void accept(AstConditionVisitor visitor) {
		    visitor.visitMayAccess(true, r, f);
		}

		/** for debug only */
		public String toString() {
		    return "[Condition(mayAccess):r=" + r + ",f=" + f + "]";
		}
	    };
    }

    public AstCondition createNotAccessCondition(Location location, final Identifier r, final Identifier f) {
	return new AstCondition() {
		
		public void accept(AstConditionVisitor visitor) {
		    visitor.visitMayAccess(false, r, f);
		}

		/** for debug only */
		public String toString() {
		    return "[Condition(notAccess):r=" + r + ",f=" + f + "]";
		}
	    };
    }

    public AstGoal createReplaceGoal(Location location,
                                     final String name,
                                     final AstTerm term,
                                     final AstSequent addSequent,
                                     final AstTacletDeclaration[] addRules) {
        return new AstGoal(location) {

                public void accept(AstGoalVisitor visitor) {
                    visitor.visitReplace(name, term, addSequent, addRules);
                }

                /** for debug only */
                public String toString() {
                    return "[AstGoal:term=" + term + "]";
                }

            };
    }

    public AstGoal createReplaceGoal(Location location,
                                     final String name,
                                     final AstSequent sequent,
                                     final AstSequent addSequent,
                                     final AstTacletDeclaration[] addRules) {
        return new AstGoal(location) {

                public void accept(AstGoalVisitor visitor) {
                    visitor.visitReplace(name, sequent, addSequent, addRules);
                }

                /** for debug only */
                public String toString() {
                    return "[AstGoal:sequent=" + sequent + "]";
                }

            };
    }

    public AstGoal createReplaceGoal(Location location,
				     final String name,
				     final AstSequent sequent) {
	return new AstGoal(location) {

		public void accept(AstGoalVisitor visitor) {
		    visitor.visitReplace(name, sequent);
		}
		
		/** for debug only */
		public String toString() {
		    return "[AstGoal:sequent=" + sequent + "]";
		}
	    };
    }

    public AstTacletModifier createTacletHeuristics(Location location, final Identifier[] ids) {
        return new AstTacletModifier(location) {

                public void accept(AstTacletModifierVisitor visitor) {
                    visitor.visitHeuristics(ids);
                }

                /** for debug only */
                public String toString() {
                    return "[AstTacletModifier:heuristics=" + ArrayUtil.toString(ids) + "]";
                }

            };
    }

    public AstTacletModifier createTacletNonInteractiveModifier(Location location) {
        return new AstTacletModifier(location) {

                public void accept(AstTacletModifierVisitor visitor) {
                    visitor.visitNonInteractive();
                }

                /** for debug only */
                public String toString() {
                    return "[AstTacletModifier:noninteractive]";
                }

            };
    }

    public AstTacletModifier createTacletRecursiveModifier(Location location) {
        return new AstTacletModifier(location) {

                public void accept(AstTacletModifierVisitor visitor) {
                    visitor.visitRecursive();
                }

                /** for debug only */
                public String toString() {
                    return "[AstTacletModifier:recursive]";
                }

            };
    }

    public AstTacletModifier createTacletDisplayNameModifier(Location location, final String displayName) {
        return new AstTacletModifier(location) {

                public void accept(AstTacletModifierVisitor visitor) {
                    visitor.visitDisplayName(displayName);
                }

                /** for debug only */
                public String toString() {
                    return "[AstTacletModifier:recursive]";
                }

            };
    }

    public AstTacletModifier createTacletHelpTextModifier(Location location, final String helpText) {
        return new AstTacletModifier(location) {

                public void accept(AstTacletModifierVisitor visitor) {
                    visitor.visitHelpText(helpText);
                }

                /** for debug only */
                public String toString() {
                    return "[AstTacletModifier:recursive]";
                }

            };
    }


    /* --- terms for asm rules --- */

    public AstAsmRule createAsmSeq(Location location, final AstAsmRule rule1, final AstAsmRule rule2) {
        return new AstAsmRule(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitSeq(closure, rule1, rule2);
                }

                /** for debug only */
                public String toString() {
                    return "[AstAsmRule(seq):rule1=" + rule1 + ",rule2=" + rule2 + "]";
                }

            };
    }

    public AstAsmRule createAsmPar(Location location, final AstAsmRule rule1, final AstAsmRule rule2) {
        return new AstAsmRule(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitPar(closure, rule1, rule2);
                }

                /** for debug only */
                public String toString() {
                    return "[AstAsmRule(par):rule1=" + rule1 + ",rule2=" + rule2 + "]";
                }

            };
    }

    public AstAsmRule createAsmSkip(Location location) {
        return new AstAsmRule(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitSkip(closure);
                }

                /** for debug only */
                public String toString() {
                    return "[AstAsmRule(skip)]";
                }

            };
    }

    public AstAsmRule createAsmAssign(Location location, final AstTerm lhs, final AstTerm rhs) {
        return new AstAsmRule(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitAssign(closure, lhs, rhs);
                }

                /** for debug only */
                public String toString() {
                    return "[AstAsmRule(assign):lhs=" + lhs + ",rhs=" + rhs + "]";
                }

            };
    }

    public AstAsmRule createAsmCall(Location location, final Identifier id, final AstTerm[] terms) {
        return new AstAsmRule(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitCall(closure, id, terms);
                }

                /** for debug only */
                public String toString() {
                    return "[AstAsmRule(call):id=" + id + ",terms=" + terms + "]";
                }

            };
    }

    public AstAsmRule createAsmIf(Location location, final AstTerm formula,
                                  final AstAsmRule thenRule, final AstAsmRule elseRule) {
        return new AstAsmRule(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitIf(closure, formula, thenRule, elseRule);
                }

                /** for debug only */
                public String toString() {
                    return "[AstAsmRule(if):formula=" + formula + ",thenRule=" + thenRule + ",elseRule" + elseRule + "]";
                }

            };
    }

    public AstAsmRule createAsmElseif(Location location, final AstTerm formula,
                                  final AstAsmRule thenRule, final AstAsmRule elseRule) {
        return new AstAsmRule(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitElseif(closure, formula, thenRule, elseRule);
                }

                /** for debug only */
                public String toString() {
                    return "[AstAsmRule(elseif):formula=" + formula + ",thenRule=" + thenRule + ",elseRule" + elseRule + "]";
                }

            };
    }
    
    public AstAsmRule createAsmLet(Location location, final Identifier var, final Identifier sort,
                                   final AstTerm term, final AstAsmRule rule) {
        return new AstAsmRule(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitLet(closure, var, sort, term, rule);
                }

                /** for debug only */
                public String toString() {
                    return "[AstAsmRule(let):var=" + var + ",sort=" + sort + ",term=" + term + ",rule=" + rule + "]";
                }

            };
    }

    public AstAsmRule createAsmForall(Location location,
                                      final Identifier var,
                                      final Identifier sort,
                                      final AstTerm formula,
                                      final AstAsmRule rule) {
        return new AstAsmRule(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitForall(closure, var, sort, formula, rule);
                }

                /** for debug only */
                public String toString() {
                    return "[AstAsmRule(forall):var="
                        + var
                        + ",sort="
                        + sort
                        + ",formula="
                        + formula
                        + ",rule="
                        + rule
                        + "]";
                }

            };
    }

    public AstAsmRule createAsmTry(Location location, final AstAsmRule tryRule, final AstAsmRule elseRule) {
        return new AstAsmRule(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitTry(closure, tryRule, elseRule);
                }

                /** for debug only */
                public String toString() {
                    return "[AstAsmRule(try):tryRule=" + tryRule + ",elseRule=" + elseRule + "]";
                }

            };
    }

    public AstAsmRule createAsmSchemaVariable(Location location, final Identifier id) {
        return new AstAsmRule(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitSchemaVariable(closure, id);
                }

                /** for debug only */
                public String toString() {
                    return "[AstAsmRule(variable):id=" + id + "]";
                }

            };
    }

    public AstAsmRule createAsmSubstitution(Location location,
                                            final Identifier var,
                                            final Identifier sort,
                                            final AstTerm subst,
                                            final AstAsmRule rule) {
        return new AstAsmRule(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitSubstitution(closure, var, sort, subst, rule);
                }

                /** for debug only */
                public String toString() {
                    return "[AstAsmRule(subst):var=" + var + ",sort=" + sort + ",subst=" + subst + ",rule=" + rule + "]";
                }

            };
    }

    public AstAsmRule createAsmRuleMeta(Location location, final Identifier id, final AstAsmRule rule) {
        return new AstAsmRule(location) {

                public Object accept(AstTermVisitor visitor, Object closure) {
                    return visitor.visitMeta(closure, id, rule);
                }

                /** for debug only */
                public String toString() {
                    return "[AstAsmRule(meta):id=" + id + ",rule=" + rule + "]";
                }

            };
    }

    /* --- declarations --- */

    public AstSchemaType createSchemaType(Location location, final PrimitiveSchemaType type) {
        return new AstSchemaType(location) {

                public void accept(AstSchemaTypeVisitor visitor) {
                    visitor.visitPrimitive(type);
                }

                public String toString() {
                    return "[SchemaType:type=" + type + "]";
                }

            };
    }

    public AstSchemaType createSchemaType(Location location,
                                          final Identifier sort, final boolean variable) {
        return new AstSchemaType(location) {

                public void accept(AstSchemaTypeVisitor visitor) {
                    visitor.visitComposed(sort, variable);
                }

                public String toString() {
                    return "[SchemaType:sort=" + sort + ",variable=" + variable + "]";
                }

            };
    }

    public AstSchemaType createDependingSchemaType(Location location, final Identifier sort) {
        return new AstSchemaType(location) {

                public void accept(AstSchemaTypeVisitor visitor) {
                    visitor.visitDepending(sort);
                }

                public String toString() {
                    return "[SchemaType:sort=" + sort + ",constant]";
                }

            };
    }

    public AstDeclaration createSortDeclaration(Location location, Identifier id, final boolean isGenericSort) {
        return new AstDeclaration(location, id) {

                public void accept(AstDeclarationVisitor visitor) {
                    visitor.visitSort(isGenericSort);
                }

                public String toString2() {
                    return ",sort";
                }

            };
    }

    public AstDeclaration createSchemaVariableDeclaration(Location location, Identifier id, final AstSchemaType type) {
        return new AstDeclaration(location, id) {

                public void accept(AstDeclarationVisitor visitor) {
                    visitor.visitSchemaVariable(type);
                }

                public String toString2() {
                    return ",type=" + type;
                }

            };
    }

    public AstDeclaration createPredicateDeclaration(Location location, Identifier id, final Identifier[] argTypes) {
        return new AstDeclaration(location, id) {

                public void accept(AstDeclarationVisitor visitor) {
                    visitor.visitPredicate((Identifier[]) ArrayUtil.copy(argTypes));
                }

                public String toString2() {
                    return ",args=" + ArrayUtil.toString(argTypes);
                }

            };
    }

    public AstDeclaration createDerivedPredicateDeclaration(Location location, Identifier id, 
							    final Identifier[] args,
							    AstTerm formula) {
	return null; /* new AstDeclaration(location, id) {
		
		public void accept(AstDeclarationVisitor visitor) {
		    visitor.visitDerivedPredicate((Identifier[]) ArrayUtil.copy(args), formula);
		}

		public String toString2() {
		    return ",args=" + ArrayUtil.toString(args) + ",formula=" + formula;
		}
		};*/
    }

    public AstDeclaration createFunctionDeclaration(Location location,
                                                    Identifier id,
                                                    final Identifier[] argTypes,
                                                    final Identifier returnType,
                                                    final boolean dynamic) {
        return new AstDeclaration(location, id) {

                public void accept(AstDeclarationVisitor visitor) {
                    visitor.visitFunction((Identifier[]) ArrayUtil.copy(argTypes), returnType, dynamic);
                }

                public String toString2() {
                    return ",args=" + ArrayUtil.toString(argTypes) + ",ret=" + returnType;
                }

            };
    }

    public AstDeclaration createDerivedFunctionDeclaration(Location location,
							   Identifier id,
							   final Identifier[] args,
							   final Identifier returnType,
							   final AstTerm term){
	return null; /*new AstDeclaration(location, id) {
		
		public void accept(AstDeclarationVisitor visitor) {
		    visitor.visitDerivedFunction((Identifier[]) ArrayUtil.copy(args), returnType, term);
		}

		public String toString2() {
		    return ",args=" + ArrayUtil.toString(args) + ",ret=" + returnType + ",term=" + term;
		}
		};*/
    }

    public AstDeclaration createHeuristicDeclaration(Location location, Identifier id) {
        return new AstDeclaration(location, id) {

                public void accept(AstDeclarationVisitor visitor) {
                    visitor.visitHeuristic();
                }

                public String toString2() {
                    return ",heuristic";
                }

            };
    }

    public AstDeclaration createProcedureDeclaration(Location location,
                                                     Identifier id,
                                                     final AstParameter[] params,
                                                     final AstAsmRule rule) {
        return new AstDeclaration(location, id) {

                public void accept(AstDeclarationVisitor visitor) {
                    visitor.visitProcedure((AstParameter[]) ArrayUtil.copy(params), rule);
                }

                public String toString2() {
                    return ",procedure";
                }

            };
    }

    public AstDeclaration createLemmaDeclaration(Location location,
						 Identifier id,
						 final AstTerm phi) {
	return new AstDeclaration(location, id) {

		public void accept(AstDeclarationVisitor visitor) {
		    visitor.visitLemma(phi);
		}

		public String toString2() {
		    return ",lemma";
		}

	    };
    }

}
