// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.asmkey.parser.tree;


import java.util.List;

import de.uka.ilkd.asmkey.logic.*;
import de.uka.ilkd.asmkey.parser.ast.*;
import de.uka.ilkd.asmkey.parser.ast.Operator;
import de.uka.ilkd.asmkey.parser.env.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.parser.ParserException;


/** Class to parse terms and asm rules. This class implements the
 * interfaces {@link de.uka.ilkd.asmkey.parser.ast.AstTermVisitor} and
 * {@link de.uka.ilkd.asmkey.parser.ast.AstAsmRuleVisitor} and
 * therefore contains a lot of functions. Only the functions {@link
 * #parse(AstTerm, List)}, {@link #parse(AstAsmRule, List)} and {@link
 * #parse(AstSequent)} should be called.
 *
 * Because terms have to be parsed recursively and bound logical
 * variables can differ in subterms, the visitor uses the closure
 * argument to pass a list of bound logical variables.
 *
 * SN: we use a AsmTermFactory to parse the special timed terms.
 *
 * @author Hubert Schmid
 * @author Stanislas Nanchen
 */

public final class TermParser implements AstTermVisitor {

    private static final Term[] EMPTY_TERMS = new Term[] {};

    /** The environment used for parsing terms. */
    private final TermEnvironment env;

    /** Shoft reference to the used term factory. */
    private final AsmTermFactory tf = AsmTermFactory.ASMDEFAULT;


    /** Create a term parser for the given environment. */
    public TermParser(TermEnvironment env) {
        this.env = env;
    }


    /* --- terms --- */

    /** Parse a term with the bound logic variables. The elements of
     * 'logicVariables' have to be of type {@link LogicVariable}. */
    public Term parse(AstTerm term, List logicVariables) throws ParserException {
        try {
            return parse2(term, new VariableStack(logicVariables));
        } catch (ParserExceptionWrapper ew) {
            throw ew.getParserException();
        }
    }

    /** Parses a term recursive. Hint: The method uses a different
     * name, because the signature 'Object' is to general and may be
     * called incidental, if {@link #parse(AstTerm, List)} is
     * wanted. */
    private Term parse2(AstTerm term, Object stack) {
        return (Term) term.accept(this, stack);
    }

    /** Parses and returns an array of terms. Calls {@link
     * #parse2(AstTerm, Object)} fo step, r each element. */
    private Term[] parse2(AstTerm[] terms, Object stack) {
        Term[] result = new Term[terms.length];
        for (int i = 0; i < terms.length; ++i) {
            result[i] = parse2(terms[i], stack);
        }
        return result;
    }

    /* --- Interface: AstTermVisitor --- */

    public Object visitOperatorTerm(Object closure, AstOperator op, AstTerm[] terms) {
        try {
            if (op.getOperator() == Operator.EQUALS) {
                //assert terms != null && terms.length == 2;
                return tf.createEqualityTerm(parse2(terms[0], closure),
                                             parse2(terms[1], closure));
            } else if (op.getOperator() == Operator.EQV) {
                //assert terms != null && terms.length == 2;
                return tf.createJunctorTerm(Op.EQV,
                                            parse2(terms, closure));
            } else {
                de.uka.ilkd.key.logic.op.Operator op2 = TermParserUtil.convertOp(env, op);
                if (terms == null) {
                    if (op2 instanceof Junctor) {
                        return tf.createJunctorTerm((Junctor) op2);
                    } else {
                        return tf.createFunctionTerm((Function) op2);
                    }
                } else {
                    if (op2 instanceof Junctor) {
                        return tf.createJunctorTerm((Junctor) op2,
                                                    parse2(terms, closure));
                    } else {
                        return tf.createFunctionTerm((Function) op2,
                                                     parse2(terms, closure));
                    }
                }
            }
        } catch (ParserException pe) {
            throw new ParserExceptionWrapper(pe);
        }
    }

    public Object visitQuantifierTerm(Object closure, AstQuantifier q,
                                      Identifier var, Identifier sort, AstTerm term) {
        try {
            if (sort == null) {
                // schema variable or naritylogicvariable.
		if (var.getText().charAt(0) == '%') {
		    LogicVariable v = new NArityLogicVariable(TermParserUtil.toName(var));
		    VariableStack stack = ((VariableStack) closure).push(v);
		    return createQuantifierTerm(q, v, parse2(term, stack));
		} else {
		    return createQuantifierTerm(q,
						(SortedSchemaVariable)EnvironmentUtil.getSchemaVariable(env, var),
						parse2(term, closure));
		}
            } else {
                LogicVariable v = new LogicVariable(TermParserUtil.toName(var),
						    EnvironmentUtil.getSort(env, sort));
		VariableStack stack = ((VariableStack) closure).push(v);
                return createQuantifierTerm(q, v, parse2(term, stack));
            }
        } catch (ParserException pe) {
            throw new ParserExceptionWrapper(pe);
        }
    }

    private Term createQuantifierTerm(AstQuantifier q, QuantifiableVariable v, Term term)
        throws ParserException
    {
        return tf.createQuantifierTerm(TermParserUtil.convertQuantifier(q), v, term);
    }

    public Object visitFunctionTerm(Object closure, Identifier id, AstTerm[] terms) {
        try {
            VariableStack stack = (VariableStack) closure;
            if (terms == null && NumberParser.isNumber(id.getText())) {
                /* operator name is a number. */
                try {
                    return NumberParser.parseNumber(env, id.getText());
                } catch (EnvironmentException ee) {
                    throw new ParserException(ee.getMessage(), id.getLocation());
                }
            } else if (env.containsMetaOp(id.getText())) {
                /* operator is a meta operator */
                MetaOperator op = EnvironmentUtil.getMetaOp(env, id);
                Term[] subs = terms == null
                    ? EMPTY_TERMS
                    : parse2(terms, closure);
                if (op instanceof Operator2) {
                    return ((Operator2) op).createTerm(null, subs);
                } else {
                    return tf.createMetaTerm(op, subs);
                }
            } else if (stack.contains(id.getText())) {
                /* operator is a logic (bound) variable */
                if (terms != null) {
                    throw new ParserException("Variable " + id + " can not have parameters.",
                                              id.getLocation());
                }
                return tf.createVariableTerm(stack.get(id.getText()));
            } else if (env.containsSchemaVariable(id.getText())) {
                /* operator is a schema variable */
                if (terms != null) {
                    throw new ParserException("Variable " + id + " can not have parameters.",
                                              id.getLocation());
                }
                return tf.createVariableTerm(EnvironmentUtil.getSchemaVariable(env, id));
            } else {
                de.uka.ilkd.key.logic.op.Operator op = EnvironmentUtil.getOperator(env, id);
                if (op instanceof FormalParameter) {
                    /* operator is a formal parameter */
                    if (terms != null) {
                        throw new ParserException(
                                                  "Formal parameter " + id + " can not have parameters.",
                                                  id.getLocation());
                    }
                    return ((FormalParameter) op).createTerm(null, EMPTY_TERMS);
                } else if (op instanceof Operator2) {
                    /* Operator2 should be called this way. */
                    return TermParserUtil.createTerm((Operator2) op);
                } else {
                    /* else call term factory */
		    if (terms == null) {
			if (op instanceof TimedDynamicFunction) {
			    throw new ParserException(
						      "The timed function" + id +
					      "needs at least one argument.",
						      id.getLocation());
			} else {
			    return tf.createFunctionTerm((Function) op);
			}
		    } else {
			return tf.createFunctionTerm((Function) op, parse2(terms, closure));
		    }
                }
            }
        } catch (ParserException pe) {
            throw new ParserExceptionWrapper(pe);
        }
    }

    public Object visitSubstitutionTerm(Object closure, Identifier var,
                                        Identifier sort, AstTerm subst, AstTerm term) {
        try {
            if (sort == null) {
                // schema variable
                return tf.createSubstitutionTerm(Op.SUBST,
                                                 (SortedSchemaVariable)EnvironmentUtil.getSchemaVariable(env, var),
                                                 parse2(subst, closure),
						 parse2(term, closure));
            } else {
                LogicVariable v = new LogicVariable(new Name(var.getText()),
                                                    EnvironmentUtil.getSort(env, sort));
                VariableStack stack = ((VariableStack) closure).push(v);
                return tf.createSubstitutionTerm(Op.SUBST, v,
                                                 parse2(subst, closure),
                                                 parse2(term, stack));
            }
        } catch (ParserException pe) {
            throw new ParserExceptionWrapper(pe);
        }
    }


    public Object visitBigTerm(Object closure, AstTerm term, boolean and_flag) {
	Operator2 big;
	
	if (and_flag) {
	    big = BigOperator.BIGAND;
	} else {
	    big = BigOperator.BIGOR;
	}
	return TermParserUtil.createTerm(big, parse2(term, closure));
    }

    public Object visitBigTerm(Object closure, Identifier id, AstTerm formula, boolean and_flag,
			       int type_flag, int prop_flag, Identifier[] rules)
    {
	Operator2 big;
	if (and_flag) {
	    big = BigOperator.bigAnd(TermParserUtil.toName(id),
				     type_flag,
				     prop_flag, 
				     TermParserUtil.toName(rules));
	} else {
	    big = BigOperator.bigOr(TermParserUtil.toName(id),
				    type_flag,
				    prop_flag, 
				    TermParserUtil.toName(rules));
	}
	return TermParserUtil.createTerm(big, parse2(formula, closure));
    }

    public Object visitFunctionTerm(Object closure, Identifier id, AstTerm func, AstTerm form) {
	Operator2 function = FunctionOperator.function(TermParserUtil.toName(id));
	return TermParserUtil.createTerm(function, parse2(func, closure), parse2(form, closure));
    }

    public Object visitAbbreviationTerm(Object closure, Identifier name) {
        try {
            return EnvironmentUtil.getAbbreviatedTerm(env, name);
        } catch (ParserException pe) {
            throw new ParserExceptionWrapper(pe);
        }
    }
    
    /* --- sequences and sets --- */

    public Object visitSetTerm (Object closure, AstTerm[] elems) {
	Term[] elems_t;
	
	if (elems == null) {
	    elems_t = null;
	} else {
	    elems_t = new Term[elems.length];
	    for (int i=0; i<elems.length; i++) {
		elems_t[i] = parse2(elems[i], closure);
	    }
	}
	return TermParserUtil.createSetTerm (elems_t);
    }

    public Object visitSequenceTerm (Object closure, AstTerm[] elems, AstTerm tail) {
	Term[] elems_t;
	Term tail_t;

	if (elems == null) {
	    elems_t=null;
	} else {
	    elems_t = new Term[elems.length];
	    for(int i=0; i<elems.length; i++) {
		elems_t[i] = parse2(elems[i], closure); 
	    }
	}

	if (tail == null) {
	    tail_t = null;
	} else {
	    tail_t = parse2(tail, closure);
	}

	return TermParserUtil.createSequenceTerm(elems_t, tail_t);
    }

    /* --- dynamic logic terms --- */

    public Object visitBox(Object closure, AstAsmRule rule, AstTerm term, AstTerm step) {
	return TermParserUtil.createTerm
	    ((step==null)?AsmOperator.BOX:AsmTimedOperator.TBOX(parse2(step, closure)),
	     parse2(rule, closure),
	     parse2(term, closure)
	     );
    }

    public Object visitDiamond(Object closure, AstAsmRule rule, AstTerm term, AstTerm step) {
	return TermParserUtil.createTerm
	    ((step==null)?AsmOperator.DIAMOND:AsmTimedOperator.TDIAMOND(parse2(step, closure)),
	     parse2(rule, closure),
	     parse2(term, closure)
	     );
    }

    public Object visitDef(Object closure, AstAsmRule rule, AstTerm step) {
	return TermParserUtil.createTerm
	    ((step==null)?AsmOperator.DEF:AsmTimedOperator.TDEF(parse2(step, closure)),
	     parse2(rule, closure)
	     );
    }

    public Object visitUpd(Object closure, AstTerm lhs, AstTerm rhs, AstAsmRule rule, AstTerm step) {
	return TermParserUtil.createTerm
	    ((step==null)?AsmOperator.UPD:AsmTimedOperator.TUPD(parse2(step, closure)),
	     parse2(lhs, closure),
	     parse2(rhs, closure),
	     parse2(rule, closure)
	     );
    }

    public Object visitCon(Object closure, AstAsmRule rule, AstTerm step) {
        return TermParserUtil.createTerm
	    ((step==null)?AsmOperator.CON:AsmTimedOperator.TCON(parse2(step,closure)),
	     parse2(rule, closure)
	     );
    }

    public Object visitInv(Object closure, AstTerm term, AstAsmRule rule, AstTerm step) {
        return TermParserUtil.createTerm
	    ((step==null)?AsmOperator.INV:AsmTimedOperator.TINV(parse2(step,closure)),
	     parse2(term, closure),
	     parse2(rule, closure)
	     );
    }

    public Object visitJoinable(Object closure, AstAsmRule rule1, AstAsmRule rule2, AstTerm step) {
        return TermParserUtil.createTerm
	    ((step==null)?AsmOperator.JOINABLE:AsmTimedOperator.TJOINABLE(parse2(step, closure)),
	     parse2(rule1, closure),
	     parse2(rule2, closure)
	     );
    }

    public Object visitEqui(Object closure, AstAsmRule rule1, AstAsmRule rule2, AstTerm step) {
        return TermParserUtil.createTerm
	    ((step==null)?AsmOperator.EQUI:AsmTimedOperator.TEQUI(parse2(step, closure)),
	     parse2(rule1, closure),
	     parse2(rule2, closure)
	     );
    }

    public Object visitEquiS(Object closure, AstAsmRule rule1, AstAsmRule rule2, AstTerm step) {
        return TermParserUtil.createTerm
	    ((step==null)?AsmOperator.EQUIS:AsmTimedOperator.TEQUIS(parse2(step, closure)),
	     parse2(rule1, closure),
	     parse2(rule2, closure)
	     );
    }

    public Object visitEquiA(Object closure, AstAsmRule rule1, AstAsmRule rule2, AstTerm step) {
        return TermParserUtil.createTerm
	    ((step==null)?AsmOperator.EQUIA:AsmTimedOperator.TEQUIA(parse2(step, closure)),
	     parse2(rule1, closure),
	     parse2(rule2, closure)
	     );
    }

    public Object visitAccT(Object closure, AstTerm term1, AstTerm term2, AstTerm step) {
	return TermParserUtil.createTerm
	    ((step==null)?AsmOperator.ACCT:AsmTimedOperator.TACCT(parse2(step, closure)),
	     parse2(term1, closure),
	     parse2(term2, closure)
	     );
    }

    public Object visitAccF(Object closure, AstTerm term, AstTerm formula, AstTerm step) {
	return TermParserUtil.createTerm
	    ((step==null)?AsmOperator.ACCF:AsmTimedOperator.TACCF(parse2(step, closure)),
	     parse2(term, closure),
	     parse2(formula, closure)
	     );
    }

    public Object visitAcc(Object closure, AstTerm term, AstAsmRule rule, AstTerm step) {
	return TermParserUtil.createTerm
	    ((step==null)?AsmOperator.ACC:AsmTimedOperator.TACC(parse2(step, closure)),
	     parse2(term, closure),
	     parse2(rule, closure)
	     );
    }

    /* --- ASM rules --- */

    /** Parse an ASM rule with the bound logic variables. The elements
     * of 'logicVariables' have to be of type {@link
     * LogicVariable}. */
    public Term parse(AstAsmRule rule, List logicVariables)
        throws ParserException
    {
        try {
            return parse2(rule, new VariableStack(logicVariables));
        } catch (ParserExceptionWrapper ew) {
            throw ew.getParserException();
        }
    }

    /** Parses an ASM rule recursive. Hint: The method uses a
     * different name, because the signature 'Object' is to general
     * and may be called incidental, if {@link #parse(AstAsmRule,
     * List)} is wanted. */
    private Term parse2(AstAsmRule rule, Object closure) {
        return (Term) rule.accept(this, closure);
    }

    /* --- Interface: AstAsmRuleVisitor --- */

    public Object visitSeq(Object closure, AstAsmRule rule1, AstAsmRule rule2) {
        return TermParserUtil.createTerm(AsmOperator.SEQ,
                                         parse2(rule1, closure),
                                         parse2(rule2, closure));
    }

    public Object visitPar(Object closure, AstAsmRule rule1, AstAsmRule rule2) {
        return TermParserUtil.createTerm(AsmOperator.PAR,
                                         parse2(rule1, closure),
                                         parse2(rule2, closure));
    }

    public Object visitSkip(Object closure) {
        return TermParserUtil.createTerm(AsmOperator.SKIP);
    }

    public Object visitAssign(Object closure, AstTerm lhs, AstTerm rhs) {
        return TermParserUtil.createTerm(AsmOperator.ASSIGN,
                                         parse2(lhs, closure),
                                         parse2(rhs, closure));
    }

    public Object visitCall(Object closure, Identifier id, AstTerm[] terms) {
        try {
            AsmProcedure p = EnvironmentUtil.getProcedure(env, id);
            if (terms == null) {
                return p.getOp().createTerm(null, EMPTY_TERMS);
            } else {
                return p.getOp().createTerm(null, parse2(terms, closure));
            }
        } catch (ParserException pe) {
            throw new ParserExceptionWrapper(pe);
        }
    }

    public Object visitIf(Object closure, AstTerm formula, AstAsmRule thenRule,
			      AstAsmRule elseRule) {
	return TermParserUtil.createTerm(AsmOperator.IF,
					   parse2(formula, closure),
					   parse2(thenRule, closure),
					   parse2(elseRule, closure));
    }

    public Object visitElseif(Object closure, AstTerm formula, AstAsmRule thenRule,
			      AstAsmRule elseRule) {
	return TermParserUtil.createTerm(AsmOperator.ELSEIF,
					   parse2(formula, closure),
					   parse2(thenRule, closure),
					   parse2(elseRule, closure));
    }

    public Object visitLet(Object closure, Identifier var, Identifier sort,
                           AstTerm term, AstAsmRule rule) {
        try {
            if (sort == null) {
                return TermParserUtil.createTerm(AsmOperator.LET,
                                                 (SortedSchemaVariable)EnvironmentUtil.getSchemaVariable(env, var),
                                                 parse2(term, closure),
                                                 parse2(rule, closure));
            } else {
                LogicVariable v = new LogicVariable(new Name(var.getText()), EnvironmentUtil.getSort(env, sort));
                VariableStack stack = ((VariableStack) closure).push(v);
                return TermParserUtil.createTerm(AsmOperator.LET, v, parse2(term, closure), parse2(rule, stack));
            }
        } catch (ParserException pe) {
            throw new ParserExceptionWrapper(pe);
        }
    }

    public Object visitForall(Object closure, Identifier var, Identifier sort,
                              AstTerm formula, AstAsmRule rule) {
        try {
            if (sort == null) {
                return TermParserUtil.createTerm(AsmOperator.FORALL,
                                                 (SortedSchemaVariable)EnvironmentUtil.getSchemaVariable(env, var),
                                                 parse2(formula, closure),
                                                 parse2(rule, closure));
            } else {
                LogicVariable v = new LogicVariable(new Name(var.getText()), EnvironmentUtil.getSort(env, sort));
                VariableStack stack = ((VariableStack) closure).push(v);
                return TermParserUtil.createTerm(AsmOperator.FORALL, v, parse2(formula, stack), parse2(rule, stack));
            }
        } catch (ParserException pe) {
            throw new ParserExceptionWrapper(pe);
        }
    }

    public Object visitTry(Object closure, AstAsmRule tryRule, AstAsmRule elseRule) {
        return TermParserUtil.createTerm(AsmOperator.TRY,
                                         parse2(tryRule, closure),
                                         parse2(elseRule, closure));
    }

    public Object visitSchemaVariable(Object closure, Identifier var) {
        try {
            if (env.containsSchemaVariable(var.getText())) {
                return tf.createVariableTerm(EnvironmentUtil.getSchemaVariable(env, var));
            } else {
                throw new ParserException("Variable " + var + " not declared.", var.getLocation());
            }
        } catch (ParserException pe) {
            throw new ParserExceptionWrapper(pe);
        }
    }

    public Object visitSubstitution(Object closure, Identifier var, Identifier sort,
                                    AstTerm subst, AstAsmRule rule) {
        try {
            if (sort == null) {
                // schema variable
                return tf.createSubstitutionTerm(Op.SUBST,
                                                 (SortedSchemaVariable)EnvironmentUtil.getSchemaVariable(env, var),
                                                 parse2(subst, closure),
                                                 parse2(rule, closure));
            } else {
                LogicVariable v = new LogicVariable(new Name(var.getText()),
                                                    EnvironmentUtil.getSort(env, sort));
                VariableStack stack = ((VariableStack) closure).push(v);
                return tf.createSubstitutionTerm(Op.SUBST, v,
                                                 parse2(subst, closure),
                                                 parse2(rule, stack));
            }
        } catch (ParserException pe) {
            throw new ParserExceptionWrapper(pe);
        }
    }

    public Object visitMeta(Object closure, Identifier id, AstAsmRule rule) {
        try {
            MetaOperator op = EnvironmentUtil.getMetaOp(env, id);
            return TermParserUtil.createTerm((Operator2) op,
                                             parse2(rule, closure));
        } catch (ParserException pe) {
            throw new ParserExceptionWrapper(pe);
        }
    }

    /* --- Sequent --- */

    /** Parse a sequent in the current environment. */
    public Sequent parse(AstSequent sequent) {
        VariableStack stack = new VariableStack();
        return Sequent.createSequent(semiSequent(parse2(toTerms(sequent.getAntList()), stack)),
                                     semiSequent(parse2(toTerms(sequent.getSucList()), stack)));
    }

    private AstTerm[] toTerms(List list) {
        return (AstTerm[]) list.toArray(new AstTerm[list.size()]);
    }

    private Semisequent semiSequent(Term[] terms) {
        Semisequent ss = Semisequent.EMPTY_SEMISEQUENT;
        for (int i = terms.length - 1; i >= 0; --i) {
            ss = ss.insertFirst(new ConstrainedFormula
		(terms[i], Constraint.BOTTOM)).semisequent();
        }
        return ss;
    }

}
