// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.asmkey.parser.tree;


import java.util.*;

import de.uka.ilkd.key.logic.op.LogicVariable;


/** This class represents a set of bound logical variables that is
 *  used during parsing terms.  The logical variables can be accessed
 *  by name.
 *
 * @author Hubert Schmid */

final class VariableStack {

    /** Stores the mappings between identifiers (strings) and {@link
     * LogicVariable}. */
    private final Map map;


    /** Create an empty set. */
    VariableStack() {
        this(new HashMap());
    }

    /** Create a set that contains all logical variables from the
     * given list. */
    VariableStack(List logicVariables) {
        this();
        if (logicVariables != null) {
            for (Iterator i = logicVariables.iterator(); i.hasNext();) {
                Object element = i.next();
                if (element instanceof LogicVariable) {
                    LogicVariable v = (LogicVariable) element;
                    map.put(v.name().toString(), v);
                } else {
                    throw new IllegalArgumentException("Only logic variables can be added to VariableStack: "
                                                       + element + "[" + element.getClass() + "]");
                }
            }
        }
    }

    private VariableStack(Map map) {
        this.map = map;
    }


    /** Create a new instance and add the given logic variable. */
    public VariableStack push(LogicVariable v) {
        Map copy = new HashMap(map);
        copy.put(v.name().toString(), v);
        return new VariableStack(copy);
    }

    /** Check if a logical variable with the given name exists. */
    public boolean contains(String id) {
        return map.containsKey(id);
    }

    /** Return a logical variable by name. The logical variable must
     * exist. */
    public LogicVariable get(String id) {
        return (LogicVariable) map.get(id);
    }

}
