// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.gui;

import java.util.Properties;
import java.util.LinkedList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.StringTokenizer;

/** This class encapsulates the information about the active
 * Heuristics and the maximum amount of heuristics steps before an
 * interactive step is required.
 */
public class GeneralSettings implements Settings {


    private static final String STUPID_MODE_KEY = "[General]StupidMode";
    private static final String SUGG_VARNAMES_KEY = "[General]SuggestiveVarNames";
    private static final String PROOF_ASSISTANT_MODE_KEY = "[General]ProofAssistant";

    /** minimize interaction is off by default */
    private boolean stupidMode = false;

    /** proof assistant is on by default */
    private boolean proofAssistantMode = true;
    
    private boolean suggestiveVarNames = false;


    private LinkedList listenerList = new LinkedList();


    // getter
    public boolean stupidMode() {
	return stupidMode;
    }

    public boolean proofAssistantMode() {
	return proofAssistantMode;
    }

    public boolean suggestiveVarNames() {
	return suggestiveVarNames;
    }
    

    // setter
    public void setStupidMode(boolean b) {
	stupidMode = b;
	fireSettingsChanged();
    }

    public void setProofAssistantMode(boolean b) {
	proofAssistantMode = b;
	fireSettingsChanged();
    }
    
    public void setSuggestiveVarNames(boolean b) {
	suggestiveVarNames = b;
	fireSettingsChanged();
    }


    /** gets a Properties object and has to perform the necessary
     * steps in order to change this object in a way that it
     * represents the stored settings
     */
    public void readSettings(Properties props) {
	String val = props.getProperty(STUPID_MODE_KEY);
	if (val != null) {
	    setStupidMode(Boolean.valueOf(val).booleanValue());
	} else {
	    setStupidMode(stupidMode);
	}

	val = props.getProperty(SUGG_VARNAMES_KEY);
	if (val != null) {
	    setSuggestiveVarNames(Boolean.valueOf(val).booleanValue());	
	} else {
	    setSuggestiveVarNames(suggestiveVarNames);	
	}

	val = props.getProperty(PROOF_ASSISTANT_MODE_KEY);
	if (val != null) {
	    setProofAssistantMode(Boolean.valueOf(val).booleanValue());	
	} else {
	    setProofAssistantMode(proofAssistantMode);	
	}
    }


    /** implements the method required by the Settings interface. The
     * settings are written to the given Properties object. Only entries of the form 
     * <key> = <value> (,<value>)* are allowed.
     * @param props the Properties object where to write the settings as (key, value) pair
     */
    public void writeSettings(Properties props) {
	props.setProperty(STUPID_MODE_KEY, "" + stupidMode);
	props.setProperty(SUGG_VARNAMES_KEY, "" + suggestiveVarNames);
	props.setProperty(PROOF_ASSISTANT_MODE_KEY, "" + proofAssistantMode);
    }

    /** sends the message that the state of this setting has been
     * changed to its registered listeners (not thread-safe)
     */
    protected void fireSettingsChanged() {
	Iterator it = listenerList.iterator();
	while (it.hasNext()) {
	    ((SettingsListener)it.next()).settingsChanged(new GUIEvent(this));
	}
    }

    /** 
     * adds a listener to the settings object 
     * @param l the listener
     */
    public void addSettingsListener(SettingsListener l) {
	listenerList.add(l);
    }

}
