// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.gui;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collections;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.pp.*;

public class InteractiveProver {

    /** the proof the interactive prover works on */
    private Proof proof;

    /** the user focused goal */
    private Goal focusedGoal;

    /** true iff in interactive mode */

    private boolean interactive = true;

    /** the central strategy processor including GUI signalling */
    private ApplyStrategy applyStrategy;

    /** list of proof listeners and interactive proof listeners */
    private List listenerList = Collections.synchronizedList(new ArrayList(10));


    /** listens to the current selected proof and node */
    private KeYSelectionListener selListener;

    /** the mediator */
    private KeYMediator mediator;
    
    private boolean resumeAutoMode = false;

 
    //private static Logger threadLogger = Logger.getLogger("key.threading");
    
    

    /** creates a new interactive prover object 
     */
    public InteractiveProver(KeYMediator mediator) {
	selListener = new InteractiveProverKeYSelectionListener();
	this.mediator = mediator;
	mediator.addKeYSelectionListener(selListener);
	applyStrategy = new ApplyStrategy(mediator);
    }

    /** returns the KeYMediator */
    KeYMediator mediator() {
	return mediator;
    }

    /** sets up a new proof 
     * @param p a Proof that contains the root of the proof tree
     */
    public void setProof(Proof p) {
	proof = p;
    }

    public void addAutoModeListener(AutoModeListener p) { 
	synchronized(listenerList) {
	    listenerList.add(p);
	}
    }

    public void removeAutoModeListener(AutoModeListener p) { 
	synchronized(listenerList) {	
	    listenerList.remove(p);
	}
    }

    /** fires the event that automatic execution has started */
    protected void fireAutoModeStarted(ProofEvent e) {
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((AutoModeListener)it.next()).
		    autoModeStarted(null);
	    }
	}
    }

    /** fires the event that automatic execution has stopped */
    private void fireAutoModeStopped(ProofEvent e) {
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((AutoModeListener)it.next()).
		    autoModeStopped(null);
	    }
	}
    }
    

    protected void fireAutoModeStopped() {
	fireAutoModeStopped(null);
    }


    public void setResumeAutoMode(boolean b) {
       resumeAutoMode = b;
    }
 

    public boolean resumeAutoMode() {
	return resumeAutoMode;
    }

    public boolean interactiveMode() {
	return interactive;
    }

    public void setInteractive(boolean interact) {
	interactive=interact;
    }


    /**
     * Apply a RuleApp and continue with update simplification or strategy 
     * application according to current settings.
     * @param app
     * @param goal
     */
    public void applyInteractive ( RuleApp app, Goal goal ) {
        ListOfGoal goalList = goal.apply(app);
        
        if (!getProof ().closed ()) {
            if ( resumeAutoMode () ) {
                startAutoMode ();
            } else {
                // only apply the update simplifier following an
                // interactive rule application
                Goal.applyUpdateSimplifier ( goalList );
                ReuseListener rl = mediator().getReuseListener();
                if (rl != null) {
                    rl.removeRPConsumedGoal(goal);
                    rl.addRPOldMarkersNewGoals(goalList);
                    if (rl.reusePossible()) {
                        mediator().indicateReuse(rl.getBestReusePoint());
                    } else {
                        mediator().indicateNoReuse();
                    }
                }
            }
        }
    }


    private int getMaxStepCount () {
        int rv = mediator ().getMaxAutomaticSteps();
        
        if ( Main.batchMode ) {
            //Allow much more steps in batchMode then in regular mode.
            rv *= 100;
        }
        
        return rv;
    }

    
    /** returns the proof the interactive prover is working on
     * @return the proof the interactive prover is working on
     */
    public Proof getProof() {
	return proof;
    }    

    /** starts the execution of rules with active strategy 
     * @param maxApplicationOfHeuristics maximal allowed automatic
     * applications of Taclets
     */
    public void startAutoMode () {
	startAutoMode( proof.openGoals () );
    }


    /** starts the execution of rules with active strategy. The
     * strategy will only be applied on the goals of the list that
     * is handed over and on the new goals an applied rule adds
     */
    public void startAutoMode ( ListOfGoal goals ) {
        if ( goals.isEmpty () ) {
            if ( Main.batchMode ) {
                // Everything is already proven.
                // Nothing to be saved. Exit successfully.
                System.exit ( 0 );
            } else {
                mediator ().popupWarning ( "No (more) goals", "Oops..." );
                return;
            }
        }
        
        if ( Main.batchMode ) {
            interactive = false;
        }
        
        applyStrategy.setProverTaskObserver(mediator().getProverTaskListener());
        
        applyStrategy.start ( goals, getMaxStepCount () );
    }
    
    /** stops the execution of rules */
    public void stopAutoMode () {
        applyStrategy.stop();
    }


    /**
     * collects all built-in rules that are applicable at the
     * given sequent position 'pos'.
     * @param pos the PosInSequent where to look for applicable rules 
     * @param userConstraint the user defined constraint
     */
    public ListOfBuiltInRule getBuiltInRule(PosInOccurrence pos, 
						 Constraint   userConstraint) {
	ListOfBuiltInRule rules = SLListOfBuiltInRule.EMPTY_LIST;
	IteratorOfRuleApp it = 
	    getInteractiveRuleAppIndex ().getBuiltInRule
	    (focusedGoal, pos, userConstraint).iterator();

	while (it.hasNext()) {
	    BuiltInRule r = (BuiltInRule) it.next().rule();
	    if (!rules.contains(r)) {
		rules = rules.prepend(r);
	    }
	}
	return rules;
    }

    /**
     * @return the <code>RuleAppIndex</code> of the goal currently focussed,
     *         after setting the index to unrestricted (non-automated mode)
     */
    private RuleAppIndex getInteractiveRuleAppIndex () {
        final RuleAppIndex index = focusedGoal.ruleAppIndex ();
        index.autoModeStopped ();
        return index;
    }

    /**
     * collects all built-in rule applications for the given rule that are
     * applicable at position 'pos' and the current user constraint
     * 
     * @param rule
     *            the BuiltInRule for which the applications are collected
     * @param pos
     *            the PosInSequent the position information
     * @param userConstraint
     *            the user defined constraint
     * @return a SetOfRuleApp with all possible rule applications
     */
    public SetOfRuleApp getBuiltInRuleApp(BuiltInRule rule, 
					  PosInOccurrence     pos,
					  Constraint       userConstraint) {
	
	SetOfRuleApp result = SetAsListOfRuleApp.EMPTY_SET;
	IteratorOfRuleApp it = getInteractiveRuleAppIndex ().
	    getBuiltInRule(focusedGoal, 
			   pos,
			   userConstraint).iterator();

	while (it.hasNext()) {
	    RuleApp app = it.next();
	    if (app.rule() == rule) {
		result = result.add(app);
	    }
	}

	return result;
    }
    
    /**
     * collects all applications of a rule given by its name at a give position in the sequent
     * @param name
     * 				the name of the BuiltInRule for which applications are collected.
     * @param pos
     * 				the position in the sequent where the BuiltInRule should be applied
     * @return
     * 				a SetOfRuleApp with all possible applications of the rule
     */
    protected SetOfRuleApp getBuiltInRuleAppsForName(String name, PosInOccurrence pos)
    {
    	SetOfRuleApp result = SetAsListOfRuleApp.EMPTY_SET;
		ListOfBuiltInRule match = SLListOfBuiltInRule.EMPTY_LIST;

		final Constraint userConstraint = mediator.getUserConstraint().getConstraint();

		//get all possible rules for current position in sequent
		ListOfBuiltInRule list = getBuiltInRule(pos, userConstraint);

		IteratorOfBuiltInRule iter = list.iterator();

		//find all rules that match given name
		while (iter.hasNext()) {
			BuiltInRule rule = iter.next();
			if (rule.name().toString().equals(name)) match = match.append(rule);
		}

		iter = match.iterator();

		//find all applications for matched rules
		while (iter.hasNext()) {
			result = result.union(getBuiltInRuleApp(iter.next(), pos, userConstraint));
		}

		return result;
    }
    
    /**
	 * collects all applicable NoFindTaclets of the current goal (called by the
	 * SequentViewer)
	 * 
	 * @return a list of Taclets with all applicable NoFindTaclets
	 */
    ListOfTacletApp getNoFindTaclet() {	
	return filterTaclet(getInteractiveRuleAppIndex ().
		       getNoFindTaclet(TacletFilter.TRUE,
				       mediator.getServices(),
				       mediator.getUserConstraint().getConstraint()));
    }    

    /** collects all applicable FindTaclets of the current goal
     * (called by the SequentViewer)
     * @return a list of Taclets with all applicable FindTaclets
     */

    ListOfTacletApp getFindTaclet(PosInSequent pos) {
	if (!pos.isSequent() && (focusedGoal!=null) && (pos!=null)) {
	    return filterTaclet(getInteractiveRuleAppIndex ().
			      getFindTaclet(TacletFilter.TRUE,
	 	                            pos.getPosInOccurrence(),
		                            mediator.getServices(),
					    mediator.getUserConstraint().getConstraint()));
	}
	return SLListOfTacletApp.EMPTY_LIST;
    }
    
    /** collects all applicable RewriteTaclets of the current goal
     * (called by the SequentViewer)
     * @return a list of Taclets with all applicable RewriteTaclets
     */
    ListOfTacletApp getRewriteTaclet(PosInSequent pos) {
	if (!pos.isSequent())  {
	    return filterTaclet(getInteractiveRuleAppIndex ().
		   getRewriteTaclet(TacletFilter.TRUE,
				    pos.getPosInOccurrence(),
				    mediator.getServices(),
				    mediator.getUserConstraint().getConstraint())); 
	}

	return SLListOfTacletApp.EMPTY_LIST;
    }



    /** collects all Taclet applications at the given position of the specified
     * taclet
     * @param pos the PosInOccurrence describing the position
     * @param taclet the Taclet
     * @return a list of all found rule applications of the given rule at
     * position pos  
     */
    protected SetOfTacletApp getAppsForName(Goal goal, String name, 
            PosInOccurrence pos) {
        return getAppsForName(goal, name, pos, TacletFilter.TRUE);
    }
    
    protected SetOfTacletApp getAppsForName(Goal goal, String name, 
                                            PosInOccurrence pos,
                                            TacletFilter filter) {


	SetOfTacletApp result = SetAsListOfTacletApp.EMPTY_SET;
        ListOfTacletApp fittingApps = SLListOfTacletApp.EMPTY_LIST;
        final RuleAppIndex index          = goal.ruleAppIndex();
	final Constraint   userConstraint =
            mediator.getUserConstraint().getConstraint();
	if ( pos == null ) {
            final IteratorOfNoPosTacletApp it =
                index.getNoFindTaclet ( filter,
                                        mediator.getServices(),
                                        userConstraint ).iterator ();
            while ( it.hasNext () )
                fittingApps = fittingApps.prepend ( it.next () );
        } else 
            fittingApps = index.getTacletAppAt ( filter,
                                                 pos, 
			                         mediator.getServices(),
					         userConstraint );

	IteratorOfTacletApp it = fittingApps.iterator();
	// filter fitting applications
	while (it.hasNext()) {
	    TacletApp app = it.next();
	    if (app.rule().name().toString().equals(name) ){
		result = result.add(app);
	    }
	}
//if (result.size()==0) System.err.println("Available was "+fittingApps);
	return result;
    }
    
    /** listener to KeYSelection Events in order to be informed if the
     * current proof changed
     */
    private class InteractiveProverKeYSelectionListener 
	implements KeYSelectionListener {

	/** focused node has changed */
	public void selectedNodeChanged(KeYSelectionEvent e) {
	    focusedGoal = e.getSource().getSelectedGoal();
	}

	/** the selected proof has changed (e.g. a new proof has been
	 * loaded) */ 
	public void selectedProofChanged(KeYSelectionEvent e) {
	    Debug.out("InteractiveProver: initialize with new proof");
            Proof newProof = e.getSource().getSelectedProof();
	    setProof(newProof); // this is null-safe
            if (newProof != null) {
	        focusedGoal = e.getSource().getSelectedGoal();
            } else {
                focusedGoal = null;
            }
	}

    }

    /**
     * takes NoPosTacletApps as arguments and returns a duplicate free list of
     * the contained TacletApps
     */
    private ListOfTacletApp filterTaclet(ListOfNoPosTacletApp tacletInstances) {
		java.util.HashSet applicableRules = new java.util.HashSet();
		ListOfTacletApp result = SLListOfTacletApp.EMPTY_LIST;
		IteratorOfNoPosTacletApp it = tacletInstances.iterator();		
		while (it.hasNext()) {
		    TacletApp app = it.next ();
	            
	            if (mediator().stupidMode()) {
	                ListOfTacletApp ifCandidates = app.findIfFormulaInstantiations(
	                   mediator().getSelectedGoal().sequent(),
	                   mediator().getServices(),
	                   mediator().getUserConstraint().getConstraint());
	                if (ifCandidates.size() == 0) continue; // skip this app
	            }
	
		    // for the moment, just remove taclets which are
		    // inconsistent with user constraint
		    if ( mediator ().getUserConstraint ().getConstraint ()
			 .join ( app.constraint () ).isSatisfiable () ) {
				Taclet taclet = app.taclet();
				if (!applicableRules.contains(taclet)) {
				    applicableRules.add(taclet);
				    result = result.prepend(app);
				}
		    }
		}
	
       	return result;
    }
    
}
