// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.gui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.AbstractListModel;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.border.TitledBorder;
import javax.swing.text.AttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;
import javax.swing.text.PlainDocument;

import de.uka.ilkd.key.logic.IteratorOfNamed;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Named;
import de.uka.ilkd.key.logic.Namespace;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.strategy.Strategy;
import de.uka.ilkd.key.strategy.StrategyFactory;

public class StrategySelector extends JDialog {

    private StrategyListModel strategyListModel;
    private List strategyList;
    private JTextField maxAutoStepsInputField;
    
    private JList strategies;
    
    private KeYMediator mediator;
    private Proof proof = null;
    private Namespace strategyNamespace;

    public StrategySelector ( JFrame parent, KeYMediator mediator ) {
        super ( parent, "Active strategy", true );
        
        this.mediator = mediator;
        proof = mediator.getProof();
        strategyNamespace = getAvailStrategies();
        
        createStrategyList();
        
        layoutStrategySelector();
        selectActiveStrategy();
        
        pack();
        setLocation(70,70);
        setVisible(true);
    }
    
    protected void createStrategyList() {
        IteratorOfNamed it = strategyNamespace.elements().iterator();
        strategyList = new LinkedList();
        
        while (it.hasNext()) {
            strategyList.add(it.next());
        }
    }
    
    private Namespace getAvailStrategies () {
        return StrategyFactory.getStrategyNamespace();
    }

    protected void selectActiveStrategy () {
        Name name = getStrategyOfProofName ();
        
        int index;
        for (index = 0; index < strategyList.size(); index++) {
            Named n = (Named) strategyList.get(index);
            if (name.equals(n.name())) {
                strategies.setSelectedIndex(index);
            }
        }
    }
    
    private Name getStrategyOfProofName () {
        return mediator.getStrategy();
    }

    /** lays out the selector */
    protected void layoutStrategySelector() {
        JPanel listPanel = createListPanel ();

        JPanel maxInput = createMaxInputPanel ();
        
        JPanel buttonPanel = createButtonPanel ();
        
        getContentPane().setLayout(new BoxLayout(getContentPane(), 
                BoxLayout.Y_AXIS));
        
        getContentPane().add(listPanel);
        getContentPane().add(maxInput);
        getContentPane().add(buttonPanel);
    }
    
    private JPanel createButtonPanel () {
        JButton okButton = new JButton("OK");
        okButton.setMnemonic(KeyEvent.VK_ENTER);
        okButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                // retrieve settings
                int maxSteps = Integer.parseInt(maxAutoStepsInputField.getText());
                Strategy activeStrategy = getActiveStrategy();
                
                applySettings ( maxSteps, activeStrategy );
                
                setVisible(false);
            }
            });
        
        JButton saveButton = new JButton("Save as Default");
        saveButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                // retrieve settings
                int maxSteps = Integer.parseInt(maxAutoStepsInputField.getText());
                Strategy activeStrategy = getActiveStrategy();
                
                applySettings ( maxSteps, activeStrategy );
                
                // save settings as default
                ProofSettings dflt=ProofSettings.THE_SETTINGS;          
                dflt.getStrategySettings().setMaxSteps(maxSteps);
                dflt.getStrategySettings().setStrategy(activeStrategy.name());
                //temporary solution, stores more than wanted %%%% 
                dflt.saveSettings();
                
                setVisible(false);
            }
            });
        
        JButton cancelButton = new JButton("Cancel");
        cancelButton.setMnemonic(KeyEvent.VK_ESCAPE);
        cancelButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                setVisible(false);
                dispose();
            }
            });
        
        JPanel buttonPanel = new JPanel();
        buttonPanel.add(okButton);
        buttonPanel.add(saveButton);
        buttonPanel.add(cancelButton);
        
        return buttonPanel;
    }

    protected void applySettings (int maxSteps, Strategy activeStrategy) {
        // apply settings
        mediator.setMaxAutomaticSteps(maxSteps);
        mediator.setStrategy(activeStrategy.name ());
    }

    private JPanel createMaxInputPanel () {
        JPanel maxInput = new JPanel();
        maxInput.setLayout(new BoxLayout(maxInput, BoxLayout.X_AXIS));
        maxAutoStepsInputField = new NumberInputField(mediator.getMaxAutomaticSteps(), 5);
        maxAutoStepsInputField.setMaximumSize(new Dimension(50,30));
        maxInput.add(new JLabel("Maximum automatic applications: "));
        maxInput.add(maxAutoStepsInputField);
        return maxInput;
    }

    private JPanel createListPanel () {
        strategyListModel = new StrategyListModel(strategyList);
        strategies = new JList(strategyListModel);
        strategies.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        strategies.setBackground(Color.white);
        
        JScrollPane listScrollPane = new
            JScrollPane(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, 
                JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        listScrollPane.getViewport().setView(strategies);
        listScrollPane.setBorder(new TitledBorder("Strategies"));
        
        listScrollPane.setPreferredSize(new Dimension(350, 150));
        listScrollPane.setMinimumSize(new Dimension(350, 150));
        
        JPanel listPanel = new JPanel();
        listPanel.add(listScrollPane);
        listPanel.setLayout(new BoxLayout(listPanel,BoxLayout.X_AXIS));
        
        return listPanel;
    }
    
    private Strategy getActiveStrategy () {
        int index = strategies.getSelectedIndex();
        Name name = (Name) strategies.getModel().getElementAt(index);
        StrategyFactory factory = (StrategyFactory) strategyNamespace.lookup(name);
        
        return factory.create(proof);
    }

    /** model of the heuristic lists used for JList*/
    static class StrategyListModel extends AbstractListModel {
    
        private List                    strategyList;

        private static final Comparator heuristicComparator = 
            new Comparator () {
                public int compare ( Object o1, Object o2 ) {
                    return ( ((Named)o1).name ().toString() )
                        .compareTo ( ((Named)o2).name ().toString() );
                }
            };

        StrategyListModel ( List strategyList ) {
            this.strategyList = strategyList;
            Collections.sort ( this.strategyList, heuristicComparator );
        }

        public int getSize () {
            return strategyList.size ();
        }

        public void remove ( Object obj ) {
            int index = strategyList.indexOf ( obj );
            strategyList.remove ( obj );

            fireIntervalRemoved ( this, index, index + 1 );
        }

        public void add ( Object obj ) {
            int s = strategyList.size ();
            strategyList.add ( obj );
            Collections.sort ( strategyList, heuristicComparator );
            fireIntervalAdded ( this, s, s + 1 );
            fireContentsChanged ( this, 0, s );
        }

        public void add ( Object[] obj ) {
            int s = strategyList.size ();
            for (int i = 0; i < obj.length; i++) {
                strategyList.add ( obj[i] );
            }
            Collections.sort ( strategyList, heuristicComparator );
            fireIntervalAdded ( this, s, s + obj.length );
            fireContentsChanged ( this, 0, s );
        }

        public void remove ( Object[] obj ) {
            for (int i = 0; i < obj.length; i++) {
                remove ( obj[i] );
            }
        }

        public Object getElementAt ( int i ) {
            return ((StrategyFactory)strategyList.get ( i )).name ();
        }

        public Object[] getAllElements () {
            return strategyList.toArray ();
        }
        

    }
    
    // INNER CLASS TO READ ONLY NUMBERS FOR MAX APPs
    static class NumberDocument extends PlainDocument {
    
        public void insertString(int offs, String str, AttributeSet a) 
            throws BadLocationException {
                if (str == null) {
                    return;
                }
                char[] upper = str.toCharArray();
                for (int i = 0; i < upper.length; i++) {
                if (upper[i] < '0' || upper[i] > '9') {
                    return;
                } 
            }
            super.insertString(offs, new String(upper), a);
        }
        
    }

    static class NumberInputField extends JTextField {
        public NumberInputField(int number, int cols) {
            super(cols);
            setText(""+number);
        }
    
        protected Document createDefaultModel() {
            return new NumberDocument();
        }
    }

}
