// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.java;

import de.uka.ilkd.key.java.recoderext.KeYCrossReferenceServiceConfiguration;
import de.uka.ilkd.key.java.recoderext.SchemaCrossReferenceServiceConfiguration;
import de.uka.ilkd.key.logic.VariableNamer;
import de.uka.ilkd.key.logic.InnerVariableNamer;
import de.uka.ilkd.key.logic.OuterVariableNamer;
import de.uka.ilkd.key.jml.Implementation2SpecMap;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.util.KeYExceptionHandler;
import de.uka.ilkd.key.util.KeYRecoderExcHandler;


/**
 * this is a collection of common services to the KeY prover. Services
 * include information on the underlying Java model and a converter to
 * transform Java program elements to logic (where possible) and back.
 */
public class Services{


    /** used to determine whether an expression is a compile-time 
     * constant and if so the type and result of the expression
     */
    private ConstantExpressionEvaluator cee;

    /** used to convert types, expressions and so on to logic elements
     * (in special into to terms or formulas)
     */
    private TypeConverter typeconverter;

    /**
     * the information object on the Java model
     */
    private JavaInfo javainfo;

    /**
     * mapping for jml specifications
     */
    private Implementation2SpecMap specMap;

    
    /**
     * used for variable renamings
     */
    private VariableNamer variablenamer;
    

    /**
     * the exception-handler
     */
    private KeYExceptionHandler exceptionHandler;

    /**
     * creates a new Services object with a new TypeConverter and a new
     * JavaInfo object with no information stored at none of these.
     */

    public Services(KeYExceptionHandler exceptionHandler){
	cee = new ConstantExpressionEvaluator(this);
        typeconverter = new TypeConverter(this);
	if(exceptionHandler == null){
	    this.exceptionHandler = new KeYRecoderExcHandler();
	}else{
	    this.exceptionHandler = exceptionHandler;
	}
        javainfo = new JavaInfo(new KeYProgModelInfo(typeconverter,exceptionHandler), this);
        variablenamer = new OuterVariableNamer();
	specMap = new Implementation2SpecMap(this);
    }

    public Services(){
	this((KeYExceptionHandler) null);
    }

    private Services(KeYCrossReferenceServiceConfiguration crsc, 
		     KeYRecoderMapping rec2key) {
	cee = new ConstantExpressionEvaluator(this);
	typeconverter = new TypeConverter(this);
	//	exceptionHandler = new KeYRecoderExcHandler();
	javainfo = new JavaInfo
	    (new KeYProgModelInfo(crsc, rec2key, typeconverter), this);
        variablenamer = new OuterVariableNamer();
    }


     /** THIS CONSTRUCTOR IS ONLY FOR TESTS!
      * creates a Services object that contains the given JavaInfo object, a
      * type converter is created
      * @param ji the JavaInfo to use
      */
     public Services(JavaInfo ji){
         javainfo = ji;
         typeconverter = new TypeConverter(this);
	 specMap = new Implementation2SpecMap(this);
	 exceptionHandler = new KeYRecoderExcHandler();
     }

    public KeYExceptionHandler getExceptionHandler(){
	return exceptionHandler;
    }

    public void setExceptionHandler(KeYExceptionHandler keh){
	exceptionHandler = keh;
    }

    /**
     * Returns the TypeConverter associated with this Services object.
     */
    public TypeConverter getTypeConverter(){
        return typeconverter;
    }

    private void setTypeConverter(TypeConverter tc) {
	typeconverter=tc;
    }


    /**
     * Returns the ConstantExpressionEvaluator associated with this Services object.
     */
    public ConstantExpressionEvaluator getConstantExpressionEvaluator(){
        return cee;
    }

    private void setConstantExpressionEvaluator(ConstantExpressionEvaluator ce) {
	cee = ce;
    }

    /**
     * Returns the JavaInfo associated with this Services object.
     */
    public JavaInfo getJavaInfo(){
        return javainfo;
    }

    private void setJavaInfo(JavaInfo ji) {
	javainfo=ji;
	if (ji.getServices()!=this) {
	    throw new IllegalStateException("Services are inconsistent.");
	}
    }


    /**
     * Returns the VariableNamer associated with this Services object.
     */
    public VariableNamer getVariableNamer(){
        return variablenamer;
    }
      
    /**
     * Returns the Implementation2SpecMap associated with this Services object.
     */
    public Implementation2SpecMap getImplementation2SpecMap(){
	return specMap;
    }

    private void setImplementation2SpecMap(Implementation2SpecMap map){
	specMap = map;
    }

    /**
     * creates a new services object containing a copy of the java info of
     * this object and a new TypeConverter (shallow copy)
     * @return the copy
     */
    public Services copy() {
	Debug.assertTrue
	    (!(getJavaInfo().getKeYProgModelInfo().getServConf() 
	       instanceof SchemaCrossReferenceServiceConfiguration),
	     "services: tried to copy schema cross reference service config.");
	Services s = new Services
	    (getJavaInfo().getKeYProgModelInfo().getServConf(),
	     getJavaInfo().getKeYProgModelInfo().rec2key().copy());
	s.setTypeConverter(getTypeConverter().copy(s));
	s.setImplementation2SpecMap(specMap.copy(s));
	s.setExceptionHandler(getExceptionHandler());
	return s;
    }

    /**
     * creates a new service object with the same ldt information 
     * as the actual one
     */
    public Services copyPreservesLDTInformation() {
	Debug.assertTrue
	    (!(javainfo.getKeYProgModelInfo().getServConf() 
	       instanceof SchemaCrossReferenceServiceConfiguration),
	     "services: tried to copy schema cross reference service config.");
	Services s = new Services(getExceptionHandler());
	s.setTypeConverter(getTypeConverter().copy(s));
	s.setImplementation2SpecMap(specMap.copy(s));
	return s;
    }
	
}
