// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.java;


import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.java.expression.*;
import de.uka.ilkd.key.java.expression.literal.*;
import de.uka.ilkd.key.java.expression.operator.*;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.logic.ldt.LDT;
import de.uka.ilkd.key.logic.sort.ArraySort;
import de.uka.ilkd.key.logic.sort.ArrayOfSort;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.sort.SortDefiningSymbols;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.ldt.*;
import recoder.service.ConstantEvaluator;
import java.util.HashMap;


public class TypeConverter {



    private Services services;
    private TermFactory termFab = TermFactory.DEFAULT;

    private LDT integerLDT;
    private LDT booleanLDT;
    private LDT[] models = new LDT[0];

    /** required for this handling */
    private final ThisReference thisReference = new ThisReference();
    

    TypeConverter(Services s){
        services = s;
    }

    public void init(LDT integerLDT, LDT booleanLDT) {
	this.integerLDT = integerLDT;
	this.booleanLDT = booleanLDT;
	this.models = new LDT[]{booleanLDT, integerLDT};
	Debug.out("Initialize LDTs: ", integerLDT, booleanLDT);
    }

    public LDT[] getModels() {
	return models;
    }

    public Services getServices() {
	return services;
    }

    public LDT getModelFor(Type t) {
	if (integerLDT == null || booleanLDT == null) {
	    return  null;
	}
	if (integerLDT.javaType().contains(t)) {
	    return integerLDT;
	} else if (booleanLDT.javaType().contains(t)) {
	    return booleanLDT;
	}
	Debug.out("typeconverter: No LDT found for ", t);
	return null;
    }

    public LDT getModelFor(Sort s) {
	if (integerLDT == null || booleanLDT == null) {
	    return  null;
	}
	if (integerLDT.targetSort() == s) {
	    return integerLDT;
	} else if (booleanLDT.targetSort() == s) {	    
	    return booleanLDT;
	}
	Debug.out("No LDT found for ", s);
	return null;
    }

    public NumberLDT getIntegerLDT() {
	return (NumberLDT)integerLDT;
    }

    public BooleanLDT getBooleanLDT() {
	return (BooleanLDT)booleanLDT;
    }
    

    HashMap mcrMap = new HashMap(10);
    
    private Term translateMetaClassReference(MetaClassReference mcr) {
//	    throw new IllegalArgumentException("Convert MCR to a constant");
//            return intLDT.translateLiteral( new CharLiteral('X'));
        String name = mcr.getTypeReference().getName().intern();
        
        if (mcrMap.containsKey(name)) return (Term) mcrMap.get(name);

        Sort dummySort = services.getJavaInfo().
            getObjectKeYJavaType().getSort();
        Term t = termFab.createFunctionTerm(new Function(new Name(name),
            dummySort, new ArrayOfSort()));
        mcrMap.put(name, t);
        return t;
    }

    private Term translateOperator
	(de.uka.ilkd.key.java.expression.Operator op,
	 LDT intLDT, LDT booleanLDT, ExecutionContext ec) {

	Term[] subs  = new Term[op.getArity()];
	if (op.getArity() >= 1) {
	    subs[0] = convertToLogicElement(op.getExpressionAt(0), ec);
	}
	if (op.getArity() == 2) {
	    subs[1] = convertToLogicElement(op.getExpressionAt(1), ec);	  
	}	
	
	LDT responsibleLDT = null;
	if (intLDT.isResponsible(op, subs)) {
	    responsibleLDT = intLDT;
	} else if (booleanLDT.isResponsible(op, subs)) {
	    responsibleLDT = booleanLDT;
	} else{
	    Debug.out("typeconverter: no data type model "+
		      "available to convert:", op, op.getClass());		
	    throw new IllegalArgumentException("TypeConverter could not handle"
					       +" this");
	}
	return termFab.createFunctionTerm
	    (responsibleLDT.getFunctionFor(op), subs);
    }
   

    private Term convertReferencePrefix(ReferencePrefix prefix, 
					ExecutionContext ec) {
	Debug.out("typeconverter: (prefix, class)", prefix, 
		  (prefix != null ? prefix.getClass() : null));		
	if (prefix instanceof FieldReference) {
	    return convertVariableReference((FieldReference) prefix, ec);
	} else if (prefix instanceof MetaClassReference) {
	    Debug.out("typeconverter: "+
		      "WARNING: metaclass-references not supported yet");
	    throw new IllegalArgumentException("TypeConverter could not handle"
					       +" this");
	} else 	if (prefix instanceof ProgramVariable) {
	    // the base case: the leftmost item is a local variable
	    return termFab.createVariableTerm
		((ProgramVariable)prefix);
	} else 	if (prefix instanceof VariableReference) {
	    Debug.out("typeconverter: "+
		      "variablereference:", (((VariableReference)prefix).getProgramVariable()));
	    return termFab.createVariableTerm
		(((VariableReference)prefix).getProgramVariable());
	}  else if (prefix instanceof ArrayReference) {	   
	    return convertArrayReference((ArrayReference)prefix, ec);
	} else if (prefix instanceof ThisReference) {	 
	    return convertToLogicElement(ec.getRuntimeInstance());
	} else {
	    Debug.out("typeconverter: WARNING: unknown reference prefix:", 
		      prefix, prefix.getClass());
	    throw new IllegalArgumentException("TypeConverter failed to convert "
					       + prefix);
	}	
    }

    public Term convertVariableReference(VariableReference fr,
					 ExecutionContext ec) {	       
	Debug.out("TypeConverter: FieldReference: ",fr);
	ReferencePrefix prefix = fr.getReferencePrefix();
	ProgramVariable var = fr.getProgramVariable();
	Term sub = null;
	if (var.isStatic()) {
	    return termFab.createVariableTerm(var);
	} else if (prefix == null) {
	    if (var.isMember()) {
		return termFab.createAttributeTerm
		    (var, convertReferencePrefix(thisReference, ec));
	    }
	    return termFab.createVariableTerm(var); 
	} else if (!(prefix instanceof PackageReference) ) {
	    return termFab.createAttributeTerm(var, 
					       convertReferencePrefix(prefix, ec));
	} 
	Debug.out("typeconverter: Not supported reference type (fr, class):",
		  fr, fr.getClass());
	throw new IllegalArgumentException
	    ("TypeConverter could not handle this");	
    }
    

    public Term convertArrayReference(ArrayReference ar, 
				      ExecutionContext ec){
        Term[] index = new Term[ar.getDimensionExpressions().size()];
        Term t = convertToLogicElement(ar.getReferencePrefix(), ec);
         for(int i=0;i<index.length;i++) index[i]=
             convertToLogicElement(
             ar.getDimensionExpressions().getExpression(i), ec);
        return termFab.createArrayTerm(ArrayOp.getArrayOp
                   ((ArraySort)t.sort()), t, index);
    }

    private Term convertToInstanceofTerm(Instanceof io, 
					 ExecutionContext ec) {
	Term obj=convertToLogicElement(io.getChildAt(0), ec);
	KeYJavaType type=((TypeReference)io.getChildAt(1)).getKeYJavaType();
	SortDefiningSymbols s=(SortDefiningSymbols)(type.getSort());
	InstanceofSymbol func
	    = (InstanceofSymbol)s.lookupSymbol(InstanceofSymbol.NAME);
	return termFab.createFunctionTerm(func, obj);
    }

    public Term convertToLogicElement(ProgramElement pe, 
				      ExecutionContext ec) {
	return convertToLogicElement(pe, 
				     integerLDT, 
				     booleanLDT, 
				     ec);
    }

    public Term convertToLogicElement(ProgramElement pe) {
	return convertToLogicElement(pe, 
				     integerLDT, 
				     booleanLDT, 
				     null);	
    }

    public Term convertToLogicElement(ProgramElement pe, 
				      LDT intLDT, 
				      LDT booleanLDT,
				      ExecutionContext ec) {
	Debug.out("typeconverter: called for:", pe, pe.getClass());
	if (pe instanceof ProgramVariable) {	    
	    return termFab.createFunctionTerm((ProgramVariable)pe);
	} else if (pe instanceof FieldReference) {
	    return convertVariableReference((FieldReference)pe, ec);
	} else if (pe instanceof VariableReference) {
	    return convertVariableReference((VariableReference)pe, ec);
	} else if (pe instanceof ArrayReference){
	    return convertArrayReference((ArrayReference)pe, ec);
	} else if (pe instanceof IntLiteral) {
	    return intLDT.translateLiteral((IntLiteral)pe);
	} else if (pe instanceof ParenthesizedExpression) {
	    return convertToLogicElement
		(((ParenthesizedExpression)pe).getChildAt(0), 
		 intLDT, booleanLDT, ec);
	} else if (pe instanceof CharLiteral) {
	    return intLDT.translateLiteral((CharLiteral)pe);
	} else if (pe instanceof Negative 
		   && ((Negative)pe).getChildAt(0) instanceof IntLiteral) {
	    String val = ((IntLiteral)((Negative)pe).getChildAt(0)).getValue();
	    if (val.charAt(0)=='-') {
		return intLDT.translateLiteral
		    (new IntLiteral(val.substring(1)));
	    } else {
		return intLDT.translateLiteral
		    (new IntLiteral("-"+val));
	    }
	} else if (pe instanceof Negative 
		   && ((Negative)pe).getChildAt(0) instanceof LongLiteral ) {
	    String val = ((LongLiteral)
			  ((Negative)pe).getChildAt(0)).getValue();
	    if (val.charAt(0)=='-') {
		return intLDT.translateLiteral
		    (new LongLiteral(val.substring(1)));
	    } else {
		return intLDT.translateLiteral
		    (new LongLiteral("-"+val));
	    }
	} else if (pe instanceof BooleanLiteral) {
	    return booleanLDT.translateLiteral((BooleanLiteral) pe);
	} else if (pe instanceof NullLiteral) {
	    return services.getJavaInfo().getNullConst();
	} else if (pe instanceof LongLiteral) {
	    return intLDT.translateLiteral((LongLiteral)pe);
	} else if (pe instanceof ThisReference) {
	    return convertReferencePrefix((ThisReference)pe, ec);
	} else if (pe instanceof PassiveExpression) {
	    return convertToLogicElement(((PassiveExpression)pe).getChildAt(0), 
					 intLDT, booleanLDT, ec);
	} else if (pe instanceof Instanceof) {
	    return convertToInstanceofTerm((Instanceof)pe, ec);
	} else if (pe instanceof de.uka.ilkd.key.java.expression.Operator) {
	    return translateOperator
		((de.uka.ilkd.key.java.expression.Operator)pe,
		 intLDT, booleanLDT, ec);
	} else if (pe instanceof PrimitiveType) {
	    throw new IllegalArgumentException("TypeConverter could not handle"
					       +" this primitive type");
	} else if (pe instanceof MetaClassReference) {
            return translateMetaClassReference((MetaClassReference)pe);
        }
	throw new IllegalArgumentException
	    ("TypeConverter: Unknown or not convertable ProgramElement " + pe+
             " of type "+pe.getClass());
    }

    private boolean isArithmeticOperator
	(de.uka.ilkd.key.java.expression.Operator op) {
	if (op instanceof Divide || op instanceof Times || 
	    op instanceof Plus || op instanceof Minus || 
	    op instanceof Modulo || op instanceof ShiftLeft ||
	    op instanceof ShiftRight || op instanceof BinaryAnd || 
	    op instanceof BinaryNot || op instanceof BinaryOr || 
	    op instanceof BinaryXOr || op instanceof Negative || 
	    op instanceof PreIncrement || op instanceof PostIncrement ||
	    op instanceof PreDecrement || op instanceof PostDecrement) {
	    return true;
	} 
	return false;
    }


    private boolean isCompareOperator
	(de.uka.ilkd.key.java.expression.Operator op) {
	if (op instanceof GreaterThan || op instanceof GreaterOrEquals ||
	    op instanceof LessThan || op instanceof LessOrEquals || 
	    op instanceof Equals || op instanceof LogicalNot
	    || op instanceof LogicalAnd || op instanceof LogicalOr) {
	    return true;
	}
	return false;
    }


    private boolean isBooleanOperator
	(de.uka.ilkd.key.java.expression.Operator op) {
	if (op instanceof LogicalAnd || op instanceof LogicalOr ||
	    op instanceof LogicalNot) {
	    return true;
	}
	return false;
    }


    /**
     * performs binary numeric promotion on the argument types
     */
    public KeYJavaType getPromotedType(KeYJavaType type1, 
				       KeYJavaType type2) {
	if ((type1.getName().equals("boolean") &&
	     type2.getName().equals("boolean")))
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_BOOLEAN);
	if ((type1.getName().equals("byte") ||
	     type1.getName().equals("short")||
	     type1.getName().equals("char") ||
	     type1.getName().equals("int")) &&
	    (type2.getName().equals("byte") ||
	     type2.getName().equals("short")||
	     type2.getName().equals("char") ||
	     type2.getName().equals("int")))
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_INT);
	if ((type2.getName().equals("long")) &&
	    (type1.getName().equals("byte") ||
	     type1.getName().equals("short")||
	     type1.getName().equals("int")  ||
	     type1.getName().equals("char") ||
	     type1.getName().equals("long"))) 
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_LONG);
	if ((type1.getName().equals("long")) &&
	    (type2.getName().equals("byte") ||
	     type2.getName().equals("short")||
	     type2.getName().equals("char") ||
	     type2.getName().equals("int")  ||
	     type2.getName().equals("long")))
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_LONG);	    
	throw new RuntimeException("Could not determine promoted type "+
				   "of "+type1+" and "+type2);
    }


    // this method performs unary numeric promotion on the arguments
    public KeYJavaType getPromotedType(KeYJavaType type1) {
	if (type1.getName().equals("boolean"))
	    // not really numeric ...
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_BOOLEAN);
	if (type1.getName().equals("byte") ||
	    type1.getName().equals("short")||
	    type1.getName().equals("char") ||
	    type1.getName().equals("int"))
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_INT);
	if (type1.getName().equals("long"))
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_LONG);
	throw new RuntimeException("Could not determine promoted type "+
				   "of "+type1);
    }


    public KeYJavaType getBooleanType() {
	return booleanLDT.getKeYJavaType(PrimitiveType.JAVA_BOOLEAN);
    }

    public Sort getPrimitiveSort(Type t) {
        LDT result = getModelFor(t);
	Debug.out("LDT found", t, result);
        return (result == null ? null : result.targetSort());
    }

    /** convenience method. Does the same as 
     * <code>getKeYJavaType(e, null)</code> and should only be used if there is
     * no ExecutionContext available.
     */
    public KeYJavaType getKeYJavaType(Expression e){
	return getKeYJavaType(e, null);
    }

    /** returns the KeYJavaType of <code>e</code>
     * @param ec the ExecutionContext of <code>e</code>. If the Type of 
     * <code>e</code> doesn't depend on the context (which is usually the case)
     * <code>ec</code> may be null. 
     */
    public KeYJavaType getKeYJavaType(Expression e, ExecutionContext ec) {
	if(e instanceof ThisReference){
	    return ec.getTypeReference().getKeYJavaType();
	}	
	return e.getKeYJavaType(services, ec);
    }

    public Expression convertToProgramElement(Term term) {
	if (term.op()==Op.NULL) {
	    return NullLiteral.NULL;
	} else if (term.op() instanceof Function) {
	    for (int i=0; i<models.length; i++) {
		if (models[i].hasLiteralFunction((Function)term.op())) {
		    return models[i].translateTerm(term, null);	       
		}
	    }
	}
	ExtList children=new ExtList();
	for (int i=0; i<term.arity(); i++) {
	    children.add(convertToProgramElement(term.sub(i)));
	}
	if (term.op() instanceof ProgramInLogic) {
	    return ((ProgramInLogic)term.op()).convertToProgram(term, children);
	} else if (term.op() instanceof Function) {
	    final int arity = term.arity();
	    for (int i=0; i<models.length; i++) {		
		if (models[i].containsFunction((Function)term.op())) {
		    return models[i].translateTerm(term, children);
		}  
	    }
	} 
	throw new RuntimeException("Cannot convert term to program: "+term
				   +" "+term.op().getClass());
    }

    public KeYJavaType getKeYJavaType(Term t) {
// 	if (t.op() instanceof ProgramInLogic) {
// 	    return ((ProgramInLogic)t.op()).getKeYJavaType(t);
// 	}
	if (t.sort() instanceof de.uka.ilkd.key.logic.sort.ObjectSort) {
	    return services.getJavaInfo().getKeYJavaType(t.sort());
	}
	KeYJavaType boolKJT = services.getJavaInfo().getKeYJavaType
	    (PrimitiveType.JAVA_BOOLEAN);
	if (t.sort()==boolKJT.getSort()) {
	    return boolKJT;
	}
	KeYJavaType intKJT = services.getJavaInfo().getKeYJavaType
	    (PrimitiveType.JAVA_INT);
	if (t.sort()==intKJT.getSort()) {
	    return intKJT;            //Attention: this is not always
				      //true since the mapping Sort ->
				      //KJT is no function! See bug
				      //#240
	}
	return getKeYJavaType(convertToProgramElement(t));
    }

    public KeYJavaType getKeYJavaType(Type t) {
	if ( t instanceof KeYJavaType )
	    return (KeYJavaType)t;
	LDT model = getModelFor(t);
	if (model != null) {
	    return model.getKeYJavaType(t);
	} else {
	    Debug.out("javainfo: no predefined model for type ",t);
	    return null;
	}	
    }


    /** These methods are taken from recoder (and modified) */

    public boolean isWidening(PrimitiveType from, PrimitiveType to) {
	// we do not handle null's
	if (from == null || to == null) return false;
	// equal types can be coerced
	if (from == to ) return true;
	// boolean types cannot be coerced into something else
	if (from == PrimitiveType.JAVA_BOOLEAN ||
	    to == PrimitiveType.JAVA_BOOLEAN)	    
	    return false;
	// everything else can be coerced to a double
	if (to == PrimitiveType.JAVA_DOUBLE) return true;
	// but a double cannot be coerced to anything else
	if (from == PrimitiveType.JAVA_DOUBLE) return false;
	// everything except doubles can be coerced to a float
	if (to == PrimitiveType.JAVA_FLOAT) return true;
	// but a float cannot be coerced to anything but float or double
	if (from == PrimitiveType.JAVA_FLOAT) return false;	
	// everything except float or double can be coerced to a long
	if (to == PrimitiveType.JAVA_LONG) return true;
	// but a long cannot be coerced to anything but float, double or long
	if (from == PrimitiveType.JAVA_LONG) return false;
	// everything except long, float or double can be coerced to an int
	if (to == PrimitiveType.JAVA_INT) return true;
	// but an int cannot be coerced to the remaining byte, char, short
	if (from == PrimitiveType.JAVA_INT) return false;
	// between byte, char, short, only one conversion is admissible
	return (from == PrimitiveType.JAVA_BYTE &&
		to == PrimitiveType.JAVA_SHORT);
    }

    public boolean isWidening(ClassType from, ClassType to) {
	return isWidening ( getKeYJavaType ( from ),
			    getKeYJavaType ( to   ) );
    }

    public boolean isWidening(ArrayType from, ArrayType to) {
	KeYJavaType toBase   =  to.getBaseType().getKeYJavaType();
	if (toBase == 
	    services.getJavaInfo().getObjectKeYJavaType() ) { // seems incorrect
	    return true;
	}
	KeYJavaType fromBase = from.getBaseType().getKeYJavaType();
	if (toBase.getJavaType () instanceof PrimitiveType) {
	    return toBase == fromBase;
	}
	return isWidening(fromBase, toBase);
    }
    
    public boolean isWidening(Type from, Type to) {
	if ( from instanceof KeYJavaType )
	    return isWidening ( (KeYJavaType)from,
				getKeYJavaType ( to ) );
	if ( to   instanceof KeYJavaType )
	    return isWidening ( getKeYJavaType ( from ),
				(KeYJavaType)to );

	if (from instanceof ClassType) {
	    return isWidening ( getKeYJavaType ( from ),
				getKeYJavaType ( to   ) );
	} else if (from instanceof PrimitiveType) {
	    if (to instanceof PrimitiveType) {
		return isWidening((PrimitiveType)from, (PrimitiveType)to);
	    }
	} else if (from instanceof ArrayType) {
	    if (to instanceof ClassType) {
		if (getKeYJavaType ( to ) == 
		    services.getJavaInfo().getObjectKeYJavaType() ) {
		    return true;
		}
		if (to.getFullName().equals("java.lang.Cloneable") || 
		    to.getFullName().equals("Cloneable")) {
		    return true;
		}
		if (to.getFullName().equals("java.io.Serializable") || 
		    to.getFullName().equals("Serializable")) {
		    return true;
		}
		return false;
	    } else if (to instanceof ArrayType) {
		return isWidening((ArrayType)from, (ArrayType)to);
	    }
	}
	return false;
    }    

    // this also handles class types
    public boolean isWidening(KeYJavaType from, KeYJavaType to) {
	Type a = from.getJavaType ();
	Type b = to  .getJavaType ();

	if ( a instanceof ClassType || a == null ) {
	    return
		from.getSort ().extendsTrans ( to.getSort () ) ||
		( a == NullType.JAVA_NULL &&
		  b instanceof ArrayType );
	} else {
	    if ( b == null )
		return
		    to == services.getJavaInfo().
		    getObjectKeYJavaType () &&
		    a instanceof ArrayType;
	    else
		return isWidening ( a, b );
	}
    }


    public boolean isImplicitNarrowing
	(Expression expr, PrimitiveType to) {
	
	int minValue, maxValue;
	if (to == PrimitiveType.JAVA_BYTE) {
	    minValue = Byte.MIN_VALUE;
	    maxValue = Byte.MAX_VALUE;
	} else if (to == PrimitiveType.JAVA_CHAR) {
	    minValue = Character.MIN_VALUE;
	    maxValue = Character.MAX_VALUE;
	} else if (to == PrimitiveType.JAVA_SHORT) {
	    minValue = Short.MIN_VALUE;
	    maxValue = Short.MAX_VALUE;
	} else {
	    return false;
	}

	ConstantExpressionEvaluator cee = 
	    services.getConstantExpressionEvaluator();

	ConstantEvaluator.EvaluationResult res = 
	    new ConstantEvaluator.EvaluationResult();

	if (!cee.isCompileTimeConstant(expr, res) || 
	    res.getTypeCode() != ConstantEvaluator.INT_TYPE) {
	    return false;
	}
	int value = res.getInt();
	return (minValue <= value) && (value <= maxValue);	
    }

    public boolean isIdentical ( Type from, Type to ) {
	from = getKeYJavaType ( from );
	to   = getKeYJavaType ( to   );
	return from == to;
    }

    public boolean isAssignableTo ( Type from, Type to ) {
	return isIdentical ( from, to ) || isWidening ( from, to );
    }

    public boolean isAssignableTo ( Expression expr, Type to, ExecutionContext ec ) {
	return
	    ( to instanceof PrimitiveType &&
	      isImplicitNarrowing ( expr, (PrimitiveType)to ) ) ||
	    isIdentical ( expr.getKeYJavaType ( getServices (), ec ), to ) ||
	    isWidening  ( expr.getKeYJavaType ( getServices (), ec ), to );
    }

    // this also handles class types
    public boolean isNarrowing(KeYJavaType from, KeYJavaType to) {
	Type a = from.getJavaType ();
	Type b = to  .getJavaType ();

	if ( a instanceof ClassType || a == null ) {
	    return
		to.getSort ().extendsTrans ( from.getSort () ) ||
		( from == services.
		  getJavaInfo().getObjectKeYJavaType () &&
		  a instanceof ArrayType );
	} else {
	    if ( b == null )
		return false;
	    else
		return isNarrowing ( a, b );
	}
    }

    public boolean isNarrowing(PrimitiveType from, PrimitiveType to) {
	// we do not handle null's
	if (from == null || to == null) return false;

	if (from == PrimitiveType.JAVA_BYTE) {
	    return (to == PrimitiveType.JAVA_CHAR );
	}
	if (from == PrimitiveType.JAVA_SHORT) {
	    return (to == PrimitiveType.JAVA_BYTE ||
		    to == PrimitiveType.JAVA_CHAR);
	}
	if (from == PrimitiveType.JAVA_CHAR) {
	    return (to == PrimitiveType.JAVA_BYTE ||
		    to == PrimitiveType.JAVA_SHORT);
	}
	if (from == PrimitiveType.JAVA_INT) {
	    return (to == PrimitiveType.JAVA_BYTE ||
		    to == PrimitiveType.JAVA_SHORT ||
		    to == PrimitiveType.JAVA_CHAR);
	}
	if (from == PrimitiveType.JAVA_LONG) {
	    return (to == PrimitiveType.JAVA_BYTE ||
		    to == PrimitiveType.JAVA_SHORT ||
		    to == PrimitiveType.JAVA_CHAR ||
		    to == PrimitiveType.JAVA_INT);
	}
	if (from == PrimitiveType.JAVA_FLOAT) {
	    return (to == PrimitiveType.JAVA_BYTE ||
		    to == PrimitiveType.JAVA_SHORT ||
		    to == PrimitiveType.JAVA_CHAR ||
		    to == PrimitiveType.JAVA_INT ||
		    to == PrimitiveType.JAVA_LONG);
	}
	if (from == PrimitiveType.JAVA_DOUBLE) {
	    return (to == PrimitiveType.JAVA_BYTE ||
		    to == PrimitiveType.JAVA_SHORT ||
		    to == PrimitiveType.JAVA_CHAR ||
		    to == PrimitiveType.JAVA_INT ||
		    to == PrimitiveType.JAVA_LONG ||
		    to == PrimitiveType.JAVA_FLOAT);
	}
	return false;
    }

    public boolean isNarrowing(ClassType from, ClassType to) {
	return isNarrowing ( getKeYJavaType ( from ),
			     getKeYJavaType ( to   ) );
    }

    public boolean isNarrowing(ArrayType from, ArrayType to) {
	KeYJavaType toBase   = to.getBaseType().getKeYJavaType();
	KeYJavaType fromBase = from.getBaseType().getKeYJavaType();
	return
	    isReferenceType ( toBase   ) &&
	    isReferenceType ( fromBase ) &&
	    isNarrowing(fromBase, toBase);
    }

    public boolean isNarrowing(Type from, Type to) {
	if ( from instanceof KeYJavaType )
	    return isNarrowing ( (KeYJavaType)from,
				getKeYJavaType ( to ) );
	if ( to   instanceof KeYJavaType )
	    return isNarrowing ( getKeYJavaType ( from ),
				(KeYJavaType)to );

	if (from instanceof ClassType) {
	    return isNarrowing ( getKeYJavaType ( from ),
				 getKeYJavaType ( to   ) );
	} else if (from instanceof PrimitiveType) {
	    if (to instanceof PrimitiveType) {
		return isNarrowing((PrimitiveType)from, (PrimitiveType)to);
	    }
	} else if (from instanceof ArrayType) {
	    if (to instanceof ArrayType) {
		return isNarrowing((ArrayType)from, (ArrayType)to);
	    }
	}
	return false;
    }

    public boolean isCastingTo ( Type from, Type to ) {
	// there is currently no interface handling

	// identity conversion
	if ( isIdentical ( from, to ) )
	    return true;

	// conversions between numeric types are always possible
	if ( isArithmeticType ( from ) &&
	     isArithmeticType ( to   ) )
	    return true;

	// all widening conversions
	if ( isWidening ( from, to ) )
	    return true;

	// narrowing
	return isNarrowing ( from, to );
    }

    public boolean isArithmeticType ( Type t ) {
	if ( t instanceof KeYJavaType )
	    t = ((KeYJavaType)t).getJavaType ();
	return
	    t == PrimitiveType.JAVA_BYTE   ||
	    t == PrimitiveType.JAVA_SHORT  ||
	    t == PrimitiveType.JAVA_INT    ||
	    t == PrimitiveType.JAVA_CHAR   ||
	    t == PrimitiveType.JAVA_LONG   ||
	    t == PrimitiveType.JAVA_FLOAT  ||
	    t == PrimitiveType.JAVA_DOUBLE;
    }

    public boolean isIntegralType ( Type t ) {
	if ( t instanceof KeYJavaType )
	    t = ((KeYJavaType)t).getJavaType ();
	return
	    t == PrimitiveType.JAVA_BYTE   ||
	    t == PrimitiveType.JAVA_SHORT  ||
	    t == PrimitiveType.JAVA_INT    ||
	    t == PrimitiveType.JAVA_CHAR   ||
	    t == PrimitiveType.JAVA_LONG;
    }

    public boolean isReferenceType ( Type t ) {
	if ( t instanceof KeYJavaType )
	    t = ((KeYJavaType)t).getJavaType ();
	return
	    // there is currently no interface handling
	    t == null ||
	    ( t instanceof ClassType && !( t instanceof NullType ) ) ||
	    t instanceof ArrayType;
    }

    public boolean isNullType ( Type t ) {
	if ( t instanceof KeYJavaType )
	    t = ((KeYJavaType)t).getJavaType ();
	return
	    t == NullType.JAVA_NULL;
    }

    public boolean isBooleanType ( Type t ) {
	if ( t instanceof KeYJavaType )
	    t = ((KeYJavaType)t).getJavaType ();
	return
	    t == PrimitiveType.JAVA_BOOLEAN;
    }

    public TypeConverter copy(Services services) {
	TypeConverter tc = new TypeConverter(services);
	tc.init(integerLDT, booleanLDT);
	return tc;
    }

}
