// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//


package de.uka.ilkd.key.java.reference;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.expression.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.logic.ProgramElementName;
import de.uka.ilkd.key.logic.op.ProgramMethod;
import de.uka.ilkd.key.logic.op.SortedSchemaVariable;
import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.java.visitor.Visitor;
import de.uka.ilkd.key.util.Debug;

/**
 *  Method reference.
 *  @author <TT>AutoDoc</TT>
 */
public class MethodReference extends JavaNonTerminalProgramElement
                             implements MemberReference, ReferencePrefix, 
                             ReferenceSuffix, ExpressionStatement, 
                             TypeReferenceContainer, NameReference {
 
    /**
       Access path.
    */
    protected final ReferencePrefix prefix;

    /**
     *      Name.
     */
    protected final MethodName name;

    /**
     *      Arguments.
     */
    protected final ArrayOfExpression arguments;
    
    public MethodReference(ArrayOfExpression args, MethodName n, 
			   ReferencePrefix p) {
	this.prefix=p;
	name = n;
	Debug.assertTrue(name != null, "Tried to reference unnamed method.");
	this.arguments=args;
    }

    public MethodReference(ExtList children, MethodName n, ReferencePrefix p) {
	this(new ArrayOfExpression((Expression[]) 
				   children.collect(Expression.class)), 
	     n, p);
    }


    public SourceElement getFirstElement() {
        return (prefix == null) 
	    ? getChildAt(0).getFirstElement() : prefix.getFirstElement();
    }


    /**
     *      Get reference prefix.
     *      @return the reference prefix.
     */
    public ReferencePrefix getReferencePrefix() {
        return prefix;
    }

    public MethodReference setReferencePrefix(ReferencePrefix p) {
        return new MethodReference(arguments, name, p);
    }

    /**
     *      Returns the number of children of this node.
     *      @return an int giving the number of children of this node
     */

    public int getChildCount() {
        int result = 0;
        if (prefix     != null) result++;
        if (name       != null) result++;
        if (arguments  != null) result += arguments.size();
        return result;
    }

    /**
     *      Returns the child at the specified index in this node's "virtual"
     *      child array
     *      @param index an index into this node's "virtual" child array
     *      @return the program element at the given position
     *      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     *                 of bounds
     */
    public ProgramElement getChildAt(int index) {
        if (prefix != null) {
            if (index == 0) return prefix;
            index--;
        }
        if (name != null) {
            if (index == 0) return name;
            index--;
        }
        if (arguments != null) {
	    return arguments.getExpression(index);
        }
        throw new ArrayIndexOutOfBoundsException();
    }


    public int getChildPositionCode(ProgramElement child) {
        // role 0: prefix
        // role 1: name
        // role 3 (IDX): parameters
        if (prefix == child) {
            return 0;
        }
        if (name == child) {
            return 1;
        }
        if (arguments != null) {
            int index = getArrayPos(arguments,child);
            if (index >= 0) {
                return ((index << 4) | 2);
            }
        }
        return -1;
    }


    /**
     *      Get the number of type references in this container.
     *      @return the number of type references.
     */
    public int getTypeReferenceCount() {
        return (prefix instanceof TypeReference) ? 1 : 0;
    }

    /*
      Return the type reference at the specified index in this node's
      "virtual" type reference array.
      @param index an index for a type reference.
      @return the type reference with the given index.
      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
      of bounds.
    */
    public TypeReference getTypeReferenceAt(int index) {
        if (prefix instanceof TypeReference && index == 0) {
            return (TypeReference)prefix;
        }
        throw new ArrayIndexOutOfBoundsException();
    }


    /**
     *      Get the number of expressions in this container.
     *      @return the number of expressions.
     */
    public int getExpressionCount() {
        int result = 0;
        if (prefix instanceof Expression) result += 1;
        if (arguments != null) {
            result += arguments.size();
        }
        return result;
    }

    /*
      Return the expression at the specified index in this node's
      "virtual" expression array.
      @param index an index for an expression.
      @return the expression with the given index.
      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
      of bounds.
    */
    public Expression getExpressionAt(int index) {
        if (prefix instanceof Expression) {
            if (index == 0) {
                return (Expression)prefix;
            }
            index -= 1;
        }
        if (arguments != null) {
            return arguments.getExpression(index);
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    /**
     *      Get name.
     *      @return the string.
     */
    public final String getName() {
        return (name == null) ? null : name.toString();
    }

    /**
     *      Get identifier.
     *      @return the identifier.
     */
    public ProgramElementName getProgramElementName() {
	if (name instanceof ProgramElementName) {
	    return (ProgramElementName) name; 	
	} else if (name instanceof SortedSchemaVariable) {
	    return (((SortedSchemaVariable)name).getProgramElementName());
	} else return null;
    }

    /**
     *      Get arguments.
     *      @return the expression array wrapper.
     */
    public ArrayOfExpression getArguments() {
        return arguments;
    }

    /**
     *      Gets index-th argument    
     *      @return the expression 
     */
    public Expression getArgumentAt(int index) {
        if (arguments != null) {
            return arguments.getExpression(index);
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    /**
     * returns the KeY representation of the method %% todo: check
     * this method (e.g. concerning statically bound private methods)
     */
    public ProgramMethod method(Services services, 
				ExecutionContext ec) {
	KeYJavaType referencePrefixType = null;
	KeYJavaType contextType = null;
	ReferencePrefix methodPrefix = prefix;
	if (methodPrefix == null) {
	    methodPrefix = 
		(ec.getRuntimeInstance() == null ? 
		 ec.getTypeReference() : ec.getRuntimeInstance());  
	} 
	if (methodPrefix instanceof Expression) {
	    referencePrefixType = services.getTypeConverter().
		getKeYJavaType((Expression) methodPrefix, ec);
	} else if (methodPrefix instanceof TypeReference) {
	    referencePrefixType 
		= ((TypeReference)methodPrefix).getKeYJavaType();
	} else {
	    Debug.fail("Unknown prefix of method reference");
	}
	if (ec != null){
	    if (ec.getRuntimeInstance() == null) {
		contextType = ec.getTypeReference().getKeYJavaType();	
	    } else {
		contextType = services.getTypeConverter().getKeYJavaType
		    ((Expression)ec.getRuntimeInstance());
	    }
	} else {
	    contextType = referencePrefixType; //Can this case occur?
	}
	return method(services, referencePrefixType, contextType, ec);
    }

    /**
     * determines the arguments types and constructs a signature of the current
     * method
     */
    public ListOfKeYJavaType getMethodSignature(Services services,
						ExecutionContext ec) {
	ListOfKeYJavaType signature = SLListOfKeYJavaType.EMPTY_LIST;
	if (arguments != null) {
	    for (int i = arguments.size()-1; i>=0; i--) {
		signature = signature.prepend(services.getTypeConverter().
					      getKeYJavaType(getArgumentAt(i), 
							     ec));
	    }
	}
	return signature;
    }

    /**
     * returns the KeY representation of the method.
     * Returns null if there is no implementation
     */
    public ProgramMethod method(Services services, 
				KeYJavaType refPrefixType,
				KeYJavaType context) {
	return services.getJavaInfo().getProgramMethod
	    (refPrefixType, name.toString(), 
	     getMethodSignature(services, null), context);
    }

    /**
     * returns the KeY representation of the method.
     * Returns null if there is no implementation
     */
    public ProgramMethod method(Services services, 
                KeYJavaType refPrefixType,
                KeYJavaType context,
                ExecutionContext ec) {
    return services.getJavaInfo().getProgramMethod
        (refPrefixType, name.toString(), 
         getMethodSignature(services, ec), context);
    }
    
    public ProgramMethod method(Services services, 
				KeYJavaType refPrefixType,
				ExecutionContext ec) {
	return services.getJavaInfo().getProgramMethod
	    (refPrefixType, name.toString(), 
	     getMethodSignature(services, ec), 
	     ec.getTypeReference().getKeYJavaType());
    }

    private boolean implicit() {
	return getProgramElementName().toString().charAt(0)=='<';
    }


    public MethodName getMethodName() {
	return name;
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnMethodReference(this);
    }

    public void prettyPrint(PrettyPrinter p) throws java.io.IOException {
        p.printMethodReference(this);
    }

    public KeYJavaType getKeYJavaType(Services javaServ, 
				      ExecutionContext ec) {	
	return method(javaServ, ec).getKeYJavaType();
		      
    }

    public KeYJavaType getKeYJavaType(Services javaServ) {	
	return getKeYJavaType();
    }

    public KeYJavaType getKeYJavaType() {
	Debug.fail("");
	return null;
    }

}
