// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//


package de.uka.ilkd.key.java.statement;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.logic.ProgramElementName;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.java.visitor.Visitor;
import de.uka.ilkd.key.logic.ProgramPrefix;

/**
 *  Labeled statement.
 */

public class LabeledStatement extends JavaStatement 
    implements StatementContainer, 
	       NamedProgramElement, 
	       ProgramPrefix {

    /**
     *      Name.
     */

    protected final Label name;

    /**
     *      Body.
     */

    protected final Statement body;


    /**
     * Constructor for the transformation of COMPOST ASTs to KeY.
     * @param children the children of this AST element as KeY classes.
     * May contain: 	a Label (as name of the label)
     * 	        	a Statement (as body of the labeled statement)
     * 		        Comments	
     */
    public LabeledStatement(ExtList children, Label label, PositionInfo pos) {
	super(children);
	name=label;
	PositionInfo posInfo = pos;

	body=(Statement)children.get(Statement.class);
    }


    /**
     *      Labeled statement.
     */
    public LabeledStatement() {
	super();
	name=null;
	body=null;
    }

    /**
     *      Labeled statement.
     *      @param id an identifier.
     */

    public LabeledStatement(Label name) {
	this.name=name;
	body=new EmptyStatement();
    }

    /**
     *      Labeled statement.
     *      @param id a Label.
     *      @param statement a statement.
     */

    public LabeledStatement(Label id, Statement statement) {
        this.name=id;
        body=statement;
    }


    public SourceElement getFirstElement() {
        return getChildAt(0).getFirstElement();
    }

    public SourceElement getLastElement() {
        return body.getLastElement();
    }


    /**
     *      Get name.
     *      @return the string.
     */

    public final String getName() {
        return (name == null) ? null : name.toString();
    }

    /**
     *      Get identifier.
     *      @return the identifier.
     */

    public Label getLabel() {
        return name;
    }


    /**
     *      Get identifier.
     *      @return the identifier.
     */
    public ProgramElementName getProgramElementName() {
	if ((name instanceof ProgramElementName) || (name==null)) {
	    return (ProgramElementName) name;
	} 
	Debug.out("labeledstatement: SCHEMAVARIABLE IN LABELEDSTATEMENT");
	return null;
    }

    /**
     *      Get body.
     *      @return the statement.
     */
    public Statement getBody() {
        return body;
    }

    /**
     *      Returns the number of children of this node.
     *      @return an int giving the number of children of this node
     */

    public int getChildCount() {
        int result = 0;
        if (name != null) result++;
        if (body != null) result++;
        return result;
    }

    /**
     *      Returns the child at the specified index in this node's "virtual"
     *      child array
     *      @param index an index into this node's "virtual" child array
     *      @return the program element at the given position
     *      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     *                 of bounds
     */

    public ProgramElement getChildAt(int index) {
        int len;
        if (name != null) {
            if (index == 0) return name;
            index--;
        }
        if (body != null) {
            if (index == 0) return body;
            index--;
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    public int getChildPositionCode(ProgramElement child) {
        // role 0: name
        // role 1: body
        if (name == child) {
            return 0;
        }
        if (body == child) {
            return 1;
        }
        return -1;
    }

    /**
     *      Get the number of statements in this container.
     *      @return the number of statements.
     */

    public int getStatementCount() {
        return (body != null) ? 1 : 0;
    }

    /**
     *       Return the statement at the specified index in this node's
     *       "virtual" statement array.
     *       @param index an index for a statement.
     *       @return the statement with the given index.
     *       @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     *       of bounds.
     */

    public Statement getStatementAt(int index) {
        if (body != null && index == 0) {
            return body;
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnLabeledStatement(this);
    }

    public void prettyPrint(PrettyPrinter p) throws java.io.IOException {
        p.printLabeledStatement(this);
    }

    /** testing if programelements are equal modulo renaming abstract
     * from names. Therefore declaration of label names have to be
     * mapped to the same abstract name. This is done here.
     */
    public boolean equalsModRenaming(SourceElement se, 
				     NameAbstractionTable nat) {
	if (!(se instanceof LabeledStatement)) {
	    return false;
	}
	nat.add(getLabel(), ((LabeledStatement)se).getLabel());
	super.equalsModRenaming(se, nat);
	if (((LabeledStatement)se).getChildCount()!=getChildCount()) {
	    return false;
	}	
	for (int i = 0; i<getChildCount(); i++) {
	    if (!getChildAt(i).equalsModRenaming
		(((NonTerminalProgramElement)se).getChildAt(i), nat)) {
		return false;
	    }
	}
	return true;
    }
    
    public int hashCode(){
    	int result = 17;
    	result = 37 * result + getChildCount();
    	for (int i = 0; i<getChildCount(); i++) {
    		result = 37 * result + getChildAt(i).hashCode();
    	}
    	return result;
    }
    
    public boolean equals(Object o){
    	return super.equals(o);
    }

}
