// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.java.visitor;

import de.uka.ilkd.key.java.SourceElement;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.Expression;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.rule.TacletForTests;
import java.util.Iterator;
import java.util.HashSet;
import java.util.Set;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.java.TypeConverter;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.reference.ExecutionContext;
import de.uka.ilkd.key.java.reference.TypeReference;
import de.uka.ilkd.key.java.reference.MethodReference;
import de.uka.ilkd.key.java.reference.TypeRef;
import de.uka.ilkd.key.java.abstraction.IteratorOfKeYJavaType;
import de.uka.ilkd.key.java.abstraction.ListOfKeYJavaType;
import de.uka.ilkd.key.logic.op.ProgramMethod;
import de.uka.ilkd.key.java.reference.ReferencePrefix;
import de.uka.ilkd.key.logic.ProgramElementName;
import de.uka.ilkd.key.util.Debug;


/** 
 * Walks through a java AST in depth-left-fist-order. 
 * This walker is used to collect all Types in a program.
 */
public class ProgramTypeCollector extends JavaASTVisitor {

    private HashSet result = new HashSet();
    private HashSet programMethods = new HashSet();
    private KeYJavaType type;
    private Services services;
    private ProgramVariable self;
    private Set alreadyVisitedProgramMethods;
    
    public ProgramTypeCollector(ProgramElement root, ProgramVariable self, 
				KeYJavaType type, Services services,
				Set alreadyVisitedProgramMethods) {
	super(root);
	this.type = type;
	this.self=self;
        this.services = services;
	this.alreadyVisitedProgramMethods =
	    alreadyVisitedProgramMethods;
    }

    /** the action that is performed just before leaving the node the
     * last time 
     */
    protected void doAction(ProgramElement node) {
	node.visit(this);
    }

    /** starts the walker*/
    public void start() {	
	walk(root());	
    }

    public HashSet getResult() { 
	return result;
    }    

    public HashSet getProgramMethods() {
	return programMethods;
    }
    
    public String toString() {
	return result.toString();
    }


    public void performActionOnTypeReference(TypeReference x) {
	result.add(((TypeReference) x).getKeYJavaType());
    }


    public void performActionOnMethodReference(MethodReference x) {
	MethodReference currentMR = (MethodReference) x;
	ProgramVariable currentSelf = self;
	KeYJavaType currentType = type;
	ReferencePrefix referencePre = currentMR.getReferencePrefix();

	if (referencePre != null) {
	    if (referencePre instanceof ProgramVariable) {
		currentSelf = (ProgramVariable) referencePre;
		currentType = currentSelf.getKeYJavaType();
	    }
	    if (referencePre instanceof TypeReference) {
		currentType = ((TypeReference) referencePre).
		    getKeYJavaType();
		currentSelf = new ProgramVariable
		    (new ProgramElementName("x_" + 
					    referencePre.toString()),
		     currentType);
	    }
	    if (referencePre instanceof Expression) {
		currentType = services.getTypeConverter().getKeYJavaType
		    ((Expression) referencePre,
		     new ExecutionContext(new TypeRef(type), self));
		currentSelf = new ProgramVariable
		    (new ProgramElementName("x_" + 
					    referencePre.toString()),
		     currentType);		   
	    }
	
	    ListOfKeYJavaType imps = 
		services.getJavaInfo().getKeYProgModelInfo()
		.findImplementations
		(currentType,
		 currentMR.getMethodName().toString(),
		 currentMR.getMethodSignature
		 (services,
		  new ExecutionContext(new TypeRef(currentType), 
				       currentSelf)));

	    IteratorOfKeYJavaType impsIt = imps.iterator();
	    while (impsIt.hasNext()) {
		currentType = impsIt.next();
		    
		ProgramMethod currentPM =
		    services.getJavaInfo().getProgramMethod
		    (currentType,
		     currentMR.getMethodName().toString(),
		     currentMR.getMethodSignature(services,
						  new ExecutionContext
						  (new TypeRef(currentType),
						   currentSelf)),
		     currentSelf.getKeYJavaType());
		//System.out.println("pm: " + currentPM);
		    
		if (!alreadyVisitedProgramMethods.contains(currentPM)) {
		    alreadyVisitedProgramMethods.add(currentPM);
		    //System.out.println("pm building new ptc: " + currentPM);
		    if (services.getJavaInfo().getKeYJavaType
			(currentPM.getContainedInKeYJavaType()) != null) {
			result.add
			    (services.getJavaInfo().getKeYJavaType
			     (currentPM.getContainedInKeYJavaType()));
		    }
			
		    ProgramTypeCollector mCollector = 
			new ProgramTypeCollector
			(currentPM, currentSelf, 
			 currentType, services,
			 alreadyVisitedProgramMethods);
			
		    mCollector.start();
		    Iterator it = mCollector.getResult().iterator();
		    while (it.hasNext()) {
			result.add(it.next());
		    }
		}
		    
		// ACHTUNG: getProgramMethods ist unntig in der jetzigen Implementierung
		// Allerdings ist keine berprfung ob ggf. eine programmethod mehrfach wiederholt wird

		/*
		  it = mCollector.getProgramMethods().iterator();
		      
		  Set addToMethodRefs = mCollector.getMethodRefs();
		  Iterator addToMethodRefsIt = addToMethodRefs.iterator();
		  while (addToMethodRefsIt.hasNext()) {
		  MethodReference mR = (MethodReference) addToMethodRefsIt.next();
		  // only add those methodReferences which have not been visited
		  // to the set of methodReferences which need to be visited
		  if (!visitedMethodRefs.contains(mR)) {
		  methodRefs.add(mR);
		  }
		  }
		*/
	    }
	}
    }

    protected void doDefaultAction(SourceElement x) {
        if (x instanceof Expression) {
	    final KeYJavaType expressionType = 
		services.getTypeConverter().getKeYJavaType
		( (Expression)x, 
		  new ExecutionContext(new TypeRef(type), self));
	    Debug.assertTrue(expressionType != null, 
			     "Could not determine type of " + x);
	    result.add(expressionType);
	}
    }
}

