// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.expression.operator.PostIncrement;
import de.uka.ilkd.key.java.Statement;
import de.uka.ilkd.key.java.StatementBlock;
import de.uka.ilkd.key.rule.UpdateSimplifier;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.proof.*;
import junit.framework.*;


public class TestVariableNamer extends TestCase {

    private Proof proof;
    private ProgramVariable x1, x2, y;
    private ConstrainedFormula formula;
    private PosInOccurrence posOfFind;


    public TestVariableNamer(String name) {
	super(name);
    }


    private ConstrainedFormula constructFormula(ProgramVariable containedVar) {
    	Statement statement = new PostIncrement(containedVar);
    	StatementBlock statementBlock = new StatementBlock(statement);
    	JavaBlock javaBlock = JavaBlock.createJavaBlock(statementBlock);

	TermFactory termFactory = TermFactory.DEFAULT;
	Term subterm = termFactory.createJunctorTerm(Op.TRUE);
	Term term = termFactory.createDiamondTerm(javaBlock, subterm);

	return new ConstrainedFormula(term);
    }


    private Goal constructGoal(ConstrainedFormula formula, Proof proof) {
    	Semisequent empty = Semisequent.EMPTY_SEMISEQUENT;
    	Semisequent ante = empty.insert(0, formula).semisequent();
	Semisequent succ = empty;

    	Sequent seq = Sequent.createSequent(ante, succ);
	Node node = new Node(proof, seq);

	TacletIndex tacletIndex = new TacletIndex();
	BuiltInRuleAppIndex builtInRuleAppIndex = new BuiltInRuleAppIndex(null);
	RuleAppIndex ruleAppIndex = new RuleAppIndex(tacletIndex,
						     builtInRuleAppIndex);

	return new Goal(node, ruleAppIndex);
    }


    private void addGlobal(Goal goal, ProgramVariable globalVar) {
 	SetOfProgramVariable globals = goal.getGlobalProgVars().add(globalVar);
	goal.setGlobalProgVars(globals);
    }


    private boolean inGlobals(Goal goal, ProgramVariable globalVar) {
    	IteratorOfProgramVariable it = goal.getGlobalProgVars().iterator();
	while(it.hasNext()) {
	    if(it.next() == globalVar) {
	    	return true;
	    }
	}
	return false;
    }


    public void setUp() {
    	proof = new Proof();
	UpdateSimplifier sus = new UpdateSimplifier();
	proof.setSimplifier(sus);

	KeYJavaType keyJavaType = new KeYJavaType(Sort.FORMULA);

    	ProgramElementName name = new ProgramElementName("x");
    	x1 = new ProgramVariable(name, keyJavaType);
	x2 = new ProgramVariable(name, keyJavaType);

    	ProgramElementName name2 = new ProgramElementName("y");
    	y = new ProgramVariable(name2, keyJavaType);

	formula = constructFormula(x2);

	posOfFind = new PosInOccurrence(formula, PosInTerm.TOP_LEVEL, proof);
    }
    
    
    private void testTemporaryRename(VariableNamer vn) {
    	ProgramVariable v = vn.rename(y);
	assertFalse(v.getProgramElementName().getProgramName().equals("y"));
	
	ConstrainedFormula cf = constructFormula(v);
	Goal g = constructGoal(cf, proof);
	PosInOccurrence pio = new PosInOccurrence(cf, 
						  PosInTerm.TOP_LEVEL, 
						  proof);
	
	v = vn.rename(v, g, pio);
	assertTrue(v.getProgramElementName().getProgramName().equals("y"));
    }


    public void testInnerRename() {
 	Goal goal = constructGoal(formula, proof);
	Sequent originalSequent = goal.sequent();
	VariableNamer vn = new InnerVariableNamer();
	ProgramVariable v, v2;

	v = vn.rename(y, goal, posOfFind);
	assertTrue(v.getProgramElementName().getProgramName().equals("y"));
	assertTrue(goal.sequent().equals(originalSequent));

   	v = vn.rename(x1, goal, posOfFind);
	assertTrue(v.getProgramElementName().getProgramName().equals("x"));
	assertTrue(goal.sequent().equals(originalSequent));

	addGlobal(goal, v);
	v2 = vn.rename(x2, goal, posOfFind);
	assertFalse(v2.getProgramElementName().getProgramName().equals("x"));
	assertFalse(goal.sequent().equals(originalSequent));
	assertTrue(inGlobals(goal, v));
	
	testTemporaryRename(vn);
    }


    public void testOuterRename() {
 	Goal goal = constructGoal(formula, proof);
	Sequent originalSequent = goal.sequent();
	VariableNamer vn = new OuterVariableNamer();
	ProgramVariable v, v2;

	v = vn.rename(y, goal, posOfFind);
	assertTrue(v.getProgramElementName().getProgramName().equals("y"));
	assertTrue(goal.sequent().equals(originalSequent));

   	v = vn.rename(x1, goal, posOfFind);
	assertTrue(v.getProgramElementName().getProgramName().equals("x"));
	assertTrue(goal.sequent().equals(originalSequent));

	addGlobal(goal, v);
	v2 = vn.rename(x2,goal, posOfFind);
	assertTrue(v2.getProgramElementName().getProgramName().equals("x"));
	assertFalse(goal.sequent().equals(originalSequent));
	assertFalse(inGlobals(goal, v));
	
	testTemporaryRename(vn);
    }
}
