// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

/**
 * Creates names to be used for newly introduced variables, and takes care
 * of renaming other variables in this situation if necessary (i.e. in case 
 * of naming conflicts)
 * Depending on the implementing class, this could mean changing either the
 * name of the new variable ("inner renaming") or of the other variable
 * ("outer renaming").
 */

package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.java.ArrayOfProgramElement;
import de.uka.ilkd.key.java.SourceElement;
import de.uka.ilkd.key.java.Statement;
import de.uka.ilkd.key.java.statement.MethodFrame;
import de.uka.ilkd.key.java.annotation.Annotation;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.RuleAppIndex;
import de.uka.ilkd.key.proof.TacletIndex;
import de.uka.ilkd.key.proof.ProgVarReplacer;
import de.uka.ilkd.key.rule.inst.*;
import de.uka.ilkd.key.rule.IteratorOfNoPosTacletApp;
import de.uka.ilkd.key.rule.ListOfNoPosTacletApp;
import de.uka.ilkd.key.rule.NoPosTacletApp;
import de.uka.ilkd.key.rule.SetAsListOfNoPosTacletApp;
import de.uka.ilkd.key.rule.SetOfNoPosTacletApp;
import de.uka.ilkd.key.util.Debug;
import java.util.*;


public abstract class VariableNamer {

    /**
     * shared counter used for creating unique names when no context 
     * information is available
     */
     private int tempCounter = 0;


    /**
     * returns the next counter value to be used for a variable with the passed
     * basename
     * that is, the maximum counter already existing for the passed 
     * basename in the passed goal, plus 1; or 0 if no variable with this 
     * basename yet exists
     */
    protected int getCounter(String basename, Goal goal) {
        int result = -1;

        IteratorOfProgramVariable it = goal.getGlobalProgVars().iterator();
        while(it.hasNext()) {
            ProgramElementName name = it.next().getProgramElementName();
	    int c;

            if(name instanceof PermIndProgramElementName &&
	       ((IndProgramElementName)name).getBaseName().equals(basename)) {
	    	c = ((IndProgramElementName)name).getIndex();
            } else if(name.getProgramName().equals(basename)) { //necessary?
	    	c = 0;
	    } else {
	    	c = -1;
	    }
	    
	    if(c > result) {
            	result = c;
            }
        }

	result++;
        return result;
    }


    /**
     * returns a NameCreationInfo representing the method stack of a
     * PosInOccurrence
     */
    protected NameCreationInfo getMethodStack(PosInOccurrence posOfFind) {
        ListOfProgramMethod list = SLListOfProgramMethod.EMPTY_LIST;

        Term term = posOfFind.subTerm();
        JavaBlock jb = term.javaBlock();

        SourceElement element = jb.program();
        while(element != element.getFirstElement()) {
            element = element.getFirstElement();

            if(element instanceof MethodFrame) {
                MethodFrame frame = (MethodFrame)element;
                ProgramMethod method = frame.getProgramMethod();
                if(method != null) {
                    list = list.append(method);
                }
            }
        }

        return new MethodStackInfo(list);
    }


    /**
     * determines the basename of a given ProgramElementName (i.e. the name
     * minus any counters affixed by the renaming mechanism)
     */
    protected String getBaseName(ProgramElementName varname) {
	String result;

	if(varname instanceof IndProgramElementName) {
	    result = ((IndProgramElementName)varname).getBaseName();
	} else {
	    result = varname.getProgramName();
	}

	return result;
    }


    /**
     * creates a ProgramElementName object to be used for permanent renamings
     */
    protected ProgramElementName createName(String name,
    					    int index,
					    NameCreationInfo creationInfo) {
	return new PermIndProgramElementName(name, index, creationInfo);
    }
    
    
    /**
     * replaces variables in a goal's sequent
     */
    protected void replaceInSequent(Map map, Goal goal) {
	ProgVarReplacer pvr = new ProgVarReplacer(map);
	SequentChangeInfo sci = pvr.replace(goal.sequent());
	goal.setSequent(sci);
    }
    
    
    /**
     * replaces variables in a goal's global variables
     */
    protected void replaceInGlobals(Map map, Goal goal) {
    	SetOfProgramVariable s = goal.getGlobalProgVars();

    	Iterator it = map.entrySet().iterator();
	while(it.hasNext()) {
	    Map.Entry mapping = (Map.Entry)it.next();
	    ProgramVariable var1 = (ProgramVariable)(mapping.getKey());
	    ProgramVariable var2 = (ProgramVariable)(mapping.getValue());
	    if(s.contains(var1)) {
	    	s = s.remove(var1);
	    	s = s.add(var2);
	    }
	}
	
	goal.setGlobalProgVars(s);
    }
    
    
    
    /**
     * helper for replaceInTacletApps()
     * replaces variables in an SVInstantiations
     */
    private SVInstantiations replaceInsts(Map map, SVInstantiations insts) {
    	SVInstantiations result = insts;
	ProgVarReplacer pvr = new ProgVarReplacer(map);
		
    	IteratorOfEntryOfSchemaVariableAndInstantiationEntry it;
	it = insts.pairIterator();
	while(it.hasNext()) {
	    EntryOfSchemaVariableAndInstantiationEntry e = it.next();
	    SchemaVariable sv     = e.key();
	    InstantiationEntry ie = e.value();
	    Object inst = ie.getInstantiation();

	    if(ie instanceof ContextInstantiationEntry) {
		Statement s = (Statement) inst;
		Statement newS = pvr.replace(s);
		ContextInstantiationEntry cie = (ContextInstantiationEntry) ie;
		result = result.replace(cie.prefix(), 
					cie.suffix(),
					cie.activeStatementContext(),
					newS);
	    } else if(ie instanceof OperatorInstantiation) {
	    	/*nothing to be done (currently)*/
	    } else if(ie instanceof ProgramInstantiation) {
		Statement s = (Statement) inst;
		Statement newS = pvr.replace(s);
		result = result.replace(sv, newS);
	    } else if(ie instanceof ProgramListInstantiation) {
		ArrayOfProgramElement a = (ArrayOfProgramElement) inst;
		int size = a.size();
		Statement[] array = new Statement[size];
		
		for(int i = 0; i < size; i++) {
		    Statement s = (Statement) a.getProgramElement(i);
		    array[i] = pvr.replace(s);
		}
		
		ArrayOfProgramElement newA = new ArrayOfProgramElement(array);
		result = result.replace(sv, newA);	
	    } else if(ie instanceof TermInstantiation) {
		Term t = (Term) inst;
		Term newT = pvr.replace(t);
		result = result.replace(sv, newT);
	    } else {
		Debug.fail("unexpected subtype of InstantiationEntry");
	    }
	}
	
	return result;
    }
    
    
    
    /**
     * replaces variables in a goal's taclet apps
     */
    protected void replaceInTacletApps(Map map, Goal goal) {
	RuleAppIndex ruleAppIndex = goal.ruleAppIndex();
	TacletIndex tacletIndex = ruleAppIndex.tacletIndex();
	final ListOfNoPosTacletApp noPosTacletApps = tacletIndex.
	getPartialInstantiatedApps();
	SetOfNoPosTacletApp appsToBeRemoved, appsToBeAdded;
	appsToBeRemoved = SetAsListOfNoPosTacletApp.EMPTY_SET;
	appsToBeAdded   = SetAsListOfNoPosTacletApp.EMPTY_SET;	
	IteratorOfNoPosTacletApp it = noPosTacletApps.iterator();
	while(it.hasNext()) {
	    NoPosTacletApp noPosTacletApp = it.next();
	    SVInstantiations insts = noPosTacletApp.instantiations();
	    
	    SVInstantiations newInsts = replaceInsts(map, insts);
	    
	    if(!newInsts.equals(insts)) {
	    	NoPosTacletApp newNoPosTacletApp
		 = NoPosTacletApp.createNoPosTacletApp(noPosTacletApp.taclet(),
		    				       newInsts,
						       noPosTacletApp.constraint(),
						       noPosTacletApp.newMetavariables(),
						       noPosTacletApp.ifFormulaInstantiations());
		appsToBeRemoved = appsToBeRemoved.add(noPosTacletApp);
		appsToBeAdded   = appsToBeAdded.add(newNoPosTacletApp);
	    }
	}
	
	tacletIndex.removeTaclets(appsToBeRemoved);
	tacletIndex.addTaclets(appsToBeAdded);
    }

    /**
     * renames variable var by giving it a unique, temporary name
     */
    public ProgramVariable rename(ProgramVariable var) {
    	String basename = var.getProgramElementName().getProgramName();
	ProgramElementName newname
		= new TempIndProgramElementName(basename, tempCounter++, null);
	return new ProgramVariable(newname, var.getKeYJavaType());
    }


    /**
     * renames variables after introduction of new variable var,
     * according to the renaming strategy
     * returns the (possibly) renamed version of var
     */
    public abstract ProgramVariable rename(ProgramVariable var,
                                           Goal goal,
                                           PosInOccurrence posOfFind);



    /**
     * ProgramElementName carrying an additional index
     */
    private abstract static class IndProgramElementName
    			    extends ProgramElementName {
        private final int index;

        IndProgramElementName(String name,
	                      int index,
			      NameCreationInfo creationInfo) {
            super(name, creationInfo);
            this.index = index;
        }

        public int getIndex() {
            return index;
        }

	public String getBaseName() {
	    return super.toString();
	}

	public String toString() {
	    return getProgramName();
	}

        public boolean equals(Object o) {
	    if(!super.equals(o) || o.getClass() != getClass()) {
                return false;
            }
            IndProgramElementName name = (IndProgramElementName)o;
            return index == name.getIndex();
        }

        public int compareTo(Object o) {
            return getProgramName().compareTo(((Name)o).toString());
        }

	public int hashCode() {
	    return getProgramName().hashCode();
	}

	/**
	 *@return the annotations.
	 */
	public Annotation[] getAnnotations(){
	    return new Annotation[0];
	}

	public int getAnnotationCount(){
	    return 0;
	}
    }


    /**
     * indexed ProgramElementName used for temporary name proposals
     */
    private static class TempIndProgramElementName
                   extends IndProgramElementName {
    	TempIndProgramElementName(String name,
	                          int index,
				  NameCreationInfo creationInfo) {
            super(name, index, creationInfo);
        }

        public String getProgramName() {
            return super.getProgramName() + "#" + getIndex();
        }
    }


    /**
     * indexed ProgramElementName used for regular name proposals
     */
    private static class PermIndProgramElementName
                   extends IndProgramElementName {
        PermIndProgramElementName(String name,
	                          int index,
				  NameCreationInfo creationInfo) {
            super(name, index, creationInfo);
        }

	public String getProgramName() {
            return super.getProgramName() +
	    	   (getIndex() == 0 ? "" : "_" + getIndex());
        }
    }
}
