// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.logic.op.ArrayOfQuantifiableVariable;
import de.uka.ilkd.key.logic.op.LogicVariable;
import de.uka.ilkd.key.logic.op.Modality;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.op.QuantifiableVariable;
import de.uka.ilkd.key.logic.op.SetOfQuantifiableVariable;
import de.uka.ilkd.key.logic.op.UpdateOperator;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.rule.soundness.SVSkolemFunction;
import de.uka.ilkd.key.util.Debug;

public class WaryClashFreeSubst extends ClashFreeSubst {

    /** depth of recursion of the <code>apply</code> method */
    private int                       depth            = 0;
    /** the formula should be prepended with a quantifier */
    private boolean                   createQuantifier = false;

    /** variable with which the original variable should be
     * substituted below modalities */
    private QuantifiableVariable      newVar           = null;
    private Term                      newVarTerm       = null;

    /** variables occurring within the original term and within the
     * term to be substituted */
    private SetOfQuantifiableVariable warysvars            = null;
   
    /** for <code>createQuantifier==true</code>, should an existential
     * or an universal quantifier be created */
    private boolean                   allQuantifier;

    public WaryClashFreeSubst ( QuantifiableVariable v,
				Term                 s,
				boolean              p_allQuantifier ) {
	super ( v, s );
	allQuantifier = p_allQuantifier;
	warysvars = null;
    }

    /** substitute <code>s</code> for <code>v</code> in <code>t</code>,
     * avoiding collisions by replacing bound variables in
     * <code>t</code> if necessary.
     */
    public Term apply(Term t) {
	Term res;

	if ( depth == 0 ) {
	    Debug.assertTrue ( t.sort () == Sort.FORMULA,
			       "WaryClashFreeSubst can only be applied" +
			       " to formulas" );
	    
	    if ( !getSubstitutedTerm ().isRigid () )
		findUsedVariables ( t );
	}

	++depth;
	try {
	    res = super.apply ( t );
	} finally { --depth; }

	if ( depth == 0 && createQuantifier )
	    res = createQuantifierTerm ( res );

	return res;
    }

    /**
     * Determine a set of variables that do already appear within
     * <code>t</code> or the substituted term, and whose names should
     * not be used for free variables
     */
    private void findUsedVariables ( Term t ) {
	VariableCollectVisitor vcv;

	vcv   = new VariableCollectVisitor ();
	getSubstitutedTerm ().execPostOrder ( vcv );
	warysvars = vcv.vars ();

	vcv   = new VariableCollectVisitor ();
	t                    .execPostOrder ( vcv );
	warysvars = warysvars.union ( vcv.vars () );	
    }

    /**
     * Create a new logic variable to be used for substitutions below
     * modalities
     */
    private void createVariable () {
	if ( !createQuantifier ) {
	    createQuantifier = true;
	    
	    newVar     = newVarFor ( getVariable (), warysvars );
	    newVarTerm = tf.createVariableTerm ( (LogicVariable)newVar );
	}	
    }

    /**
     * Prepend the term <code>t</code> with the equation
     * <code>newVar=substitutedTerm</code> and quantify
     * <code>newVar</code> existentially
     */
    private Term createQuantifierTerm ( Term t ) {
	Term eq = tf.createEqualityTerm ( newVarTerm,
					  getSubstitutedTerm () );
	Term res;
	if ( allQuantifier ) {
	    Term co = tf.createJunctorTerm    ( Op.IMP, eq, t );
	    res     = tf.createQuantifierTerm ( Op.ALL, newVar, co );
	} else {
	    Term co = tf.createJunctorTerm    ( Op.AND, eq, t );
	    res     = tf.createQuantifierTerm ( Op.EX, newVar, co );
	}

	return res;
    }

    /** substitute <code>s</code> for <code>v</code> in
     * <code>t</code>, avoiding collisions by replacing bound
     * variables in <code>t</code> if necessary.  It is
     * assumed, that <code>t</code> contains a free occurrence of
     * <code>v</code>. */
    protected Term apply1(Term t) {
	// don't move to a different modality level
	if ( !getSubstitutedTerm ().isRigid () ) {
	    if ( t.op () instanceof Modality )
		return applyOnModality ( t );		
	    if ( t.op () instanceof UpdateOperator )
		return applyOnUpdate   ( t );
	    if ( t.op () instanceof SVSkolemFunction )
		return applyToSVSkolem ( t );
	}
	return super.apply1 ( t );
    }

    /**
     * Apply the substitution (that replaces a variable with a
     * non-rigid term) on t, which has a modality as top-level
     * operator. This is done by creating a (top-level) existential
     * quantifier. This method is only called from <code>apply1</code>
     * for substitutions with non-rigid terms
     *
     * PRECONDITION: <code>warysvars != null</code>
     */
    private Term applyOnModality ( Term t ) {
	return applyBelowModality ( t );
    }

    /**
     * Apply the substitution (that replaces a variable with a
     * non-rigid term) on t, which has an update operator as top-level
     * operator. This is done by creating a (top-level) existential
     * quantifier. This method is only called from <code>apply1</code>
     * for substitutions with non-rigid terms
     *
     * PRECONDITION: <code>warysvars != null</code>
     */
    private Term applyOnUpdate ( Term t ) {
	int arity = t.arity();

	// only the last child is below the update
	if ( !t.sub ( arity - 1 ).freeVars ().contains ( getVariable () ) )
	    return super.apply1 ( t );

	createVariable ();

	Term[] newSubterms = new Term[arity];
	ArrayOfQuantifiableVariable[] newBoundVars =
	    new ArrayOfQuantifiableVariable[arity];

	int i;
	for ( i = 0; i < arity - 1; i++ ) {
	    // the update operator doesn't bind any variables
	    newSubterms [i] = apply ( t.sub ( i ) );
	    newBoundVars[i] = t.varsBoundHere ( i ); // == {}
	}

	newSubterms [i] = applyBelowModality ( t.sub ( i ) );
	newBoundVars[i] = t.varsBoundHere ( i ); // == {}

	return tf.createTerm(t.op(),
			     newSubterms,
			     newBoundVars,
			     t.javaBlock(),
			     null);
    }

    /**
     * Apply the substitution to a term/formula below a modality or
     * update
     */
    private Term applyBelowModality ( Term t ) {
	createVariable ();
	
	ClashFreeSubst cfs = new ClashFreeSubst ( getVariable (),
						  newVarTerm );
	return cfs.apply ( t );
    }

    /**
     * Apply the substitution to a non-rigid skolem function
     */
    private Term applyToSVSkolem ( Term t ) {
    	return applyBelowModality(t);
    }

}
