// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.logic.ldt;

import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.java.expression.*;
import de.uka.ilkd.key.java.abstraction.Type;
import de.uka.ilkd.key.java.abstraction.ListOfType;
import de.uka.ilkd.key.java.expression.operator.*;
import de.uka.ilkd.key.java.expression.literal.*;
import de.uka.ilkd.key.java.Expression;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.op.Function;


/** This class inherits from LDT and implements all method that are
 * necessary to handle integers, shorts and bytes.
 * It loads the needed rule file and offers method to convert java
 * number types to their logic counterpart
 */
public abstract class NumberLDT extends LDT {

    /** the function symbols for the integer operations. 
     * Used as shortcuts. 
     */

    protected Function greaterThan;    
    protected Function greaterOrEquals;
    protected Function lessThan;
    protected Function lessOrEquals;
    protected Function plus;
    protected Function minus;
    protected Function times;
    protected Function divide;
    protected Function modulo;
    protected Function shiftLeft;
    protected Function shiftRight;
    protected Function binaryAnd;
    protected Function binaryNot;
    protected Function binaryOr;
    protected Function binaryXOr;
    protected Function negative;
    protected Function unaryMinusJint;
    protected Function unaryMinusJlong;
    protected Function addJint;
    protected Function addJlong;
    protected Function subJint;
    protected Function subJlong;
    protected Function mulJint;
    protected Function mulJlong;
    protected Function modJint;
    protected Function modJlong;
    protected Function divJint;
    protected Function divJlong;
    protected Function shiftrightJint;
    protected Function shiftrightJlong;
    protected Function shiftleftJint;
    protected Function shiftleftJlong;
    protected Function unsignedshiftrightJint;
    protected Function unsignedshiftrightJlong;
    protected Function orJint;
    protected Function orJlong;
    protected Function andJint;
    protected Function andJlong;
    protected Function xorJint;
    protected Function xorJlong;
    protected Function negJint;
    protected Function negJlong;
    protected Function numbers;
    protected Function negativeNumber;

    /** the  unction symbols used to represent literals */
    protected Function sharp;
    protected Function charID;
    protected Function numberSymbol[] = new Function[10];


    public NumberLDT(Name name) {
	super(name);
    }

    public Function numberSymbol() {
	return numbers;
    }

    public void init(Namespace func_ns, Sort targetSort, ListOfType type) {
	super.init(func_ns, targetSort, type);
	// the numeric literals
	sharp = addFunction((Function) func_ns.lookup(new Name("#")));
	// character literals
	charID = addFunction((Function) func_ns.lookup(new Name("C")));      
	for (int i = 0; i < 10; i++) {
	    numberSymbol[i] = addFunction((Function)func_ns.lookup(new Name(""+i)));	    
	}
	numbers          = addFunction((Function)func_ns.lookup(new Name("Z")));	

	// predicates
	greaterThan     = addFunction((Function)func_ns.lookup(new Name("gt")));
	greaterOrEquals = addFunction((Function)func_ns.lookup(new Name("geq")));
	lessThan        = addFunction((Function)func_ns.lookup(new Name("lt")));
	lessOrEquals    = addFunction((Function)func_ns.lookup(new Name("leq")));
    }

    /** returns the function symbol for the given operation 
     * null if no function is found for the given operator
     * @return  the function symbol for the given operation 
    */
    public Function getFunctionFor
	(de.uka.ilkd.key.java.expression.Operator op) {
	if (op instanceof GreaterThan) {
	    return greaterThan;
	} else if (op instanceof GreaterOrEquals) {
	    return greaterOrEquals;
	} else if (op instanceof LessThan) {
	    return lessThan;
	} else if (op instanceof LessOrEquals) {
	    return lessOrEquals;
	} else if (op instanceof Divide) {
	    return divide;
	} else if (op instanceof Times) {
	    return times;
	} else if (op instanceof Plus) {
	    return plus;
	} else if (op instanceof Minus) {
	    return minus;
	} else if (op instanceof Modulo) {
	    return modulo;
	} else if (op instanceof ShiftLeft) {
	    return shiftLeft;
	} else if (op instanceof ShiftRight) {
	    return shiftRight;
	} else if (op instanceof BinaryAnd) {
	    return binaryAnd;
	} else if (op instanceof BinaryNot) {
	    return binaryNot;
	} else if (op instanceof BinaryOr) {
	    return binaryOr;
	} else if (op instanceof BinaryXOr) {
	    return binaryXOr;
	} else if (op instanceof Negative) {
	    return negative;
	}
	return null;
    }

    /** translates a given integer literal to its logic counterpart 
     * @param lit the Literal to be translated (has to be an
     * IntLiteral of an LongLiteral
     * @result the Term that represent the given integer in its logic
     * form
     */ 
    public Term translateLiteral(Literal lit) {
	int length=0;
	boolean minusFlag = false;
	Debug.assertTrue(lit instanceof IntLiteral || 
			 lit instanceof LongLiteral ||
			 lit instanceof CharLiteral,
			 "Literal '"+lit+"' is not an integer literal.");

	char[] int_ch=null;
	Term result = TermFactory.DEFAULT.createFunctionTerm(sharp);
	Function identifier=numbers;
	if (lit instanceof CharLiteral) {
	    lit = new IntLiteral(""+ (int)(((CharLiteral)lit)
					   .getCharValue()) ) ;
	    identifier = charID;
	}

	String literalString = ""; 
       	if (lit instanceof IntLiteral) {
	    literalString = ((IntLiteral)lit).getValue();
	} else {
	    Debug.assertTrue(lit instanceof LongLiteral);
	    literalString = ((LongLiteral)lit).getValue();
	}

	if (literalString.charAt(0) == '-') {
	    minusFlag = true;	    
	    literalString = 
		literalString.substring(1);
	}
       	if (lit instanceof IntLiteral) {
	    if (literalString.startsWith("0x")) {
                try {
                    int i = Integer.parseInt
			(literalString.substring(2),16);
                    int_ch=(""+i).toCharArray();
                }catch(NumberFormatException nfe) {
                    Debug.fail("Not a hexadecimal constant!");
		}
            } else{
	        int_ch = literalString.toCharArray();
            }
	    length=int_ch.length; 
	} else if (lit instanceof LongLiteral) {
	    if (literalString.startsWith("0x")) {
		try {
		    // long literals have an 'L' as last sign; we have
		    // to skip it 
                    final long l = Long.parseLong
			(literalString.substring(2, 
						 literalString.length() - 1), 
			 16);
                    int_ch=(""+l).toCharArray();
                } catch(NumberFormatException nfe) {
                    Debug.fail("Not a hexadecimal constant!");
		}
		length = int_ch.length; 
	    } else {
		// long literals have an 'L' as last sign; skip it
		int_ch = literalString.toCharArray();
		length = int_ch.length - 1; 
	    }
	}
	for (int i = 0; i < length; i++) {
	    result = TermFactory.DEFAULT.createFunctionTerm
		(numberSymbol[int_ch[i]-'0'], result);
	}
	if (minusFlag) {
	    result = TermFactory.DEFAULT.createFunctionTerm
		(negativeNumber, result);
	}

 	result = TermFactory.DEFAULT.createFunctionTerm
 	    (identifier, result);

	Debug.out("integerldt: result of translating literal (lit, result):", 
		  lit, result);

	return result;

    }

    public Function getNumberTerminator(){
        return sharp;
    }

    public Function getNumberSymbol(){
        return numbers;
    }

    public Function getCharSymbol(){
        return charID;
    }

    private boolean isNumberLiteral(Function f) {
	char c=f.name().toString().charAt(0);
	return (c-'0'>=0) && (c-'0'<=9);
    }

    public boolean hasLiteralFunction(Function f) {
	return containsFunction(f) && (f.arity()==0 || isNumberLiteral(f));
    }

    public Expression translateTerm(Term t, ExtList children) {
	if (!containsFunction((Function) t.op())) return null;
	Function f = (Function)t.op();	
	if (isNumberLiteral(f) || f == numbers || f ==charID) {	    
	    StringBuffer sb = new StringBuffer("");
	    Term it = t;
	    if (f == charID || f == numbers) {
		it = it.sub(0); 
		f = (Function)it.op();	    
	    }
	    while (isNumberLiteral(f)) {
		sb.insert(0, f.name().toString().charAt(0));
		it=it.sub(0);
		f = (Function)it.op();	    
	    }
	    // numbers must end with a sharp
	    if (f == sharp) {
		return new IntLiteral(sb.toString());
	    }
	}
	throw new RuntimeException("NumberLDT: Cannot convert term to program: "
				   +t);
    }

} 
