// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
//

package de.uka.ilkd.key.logic.op;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.abstraction.PrimitiveType;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.ldt.LDT;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.sort.PrimitiveSort;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.rule.MatchConditions;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import de.uka.ilkd.key.logic.Namespace;

/** this class implements the interface for
 * MetaOperators. MetaOperators are used to do complex term
 * transformation when applying a taclet. Often these transformation
 * caanot be described with the taclet scheme (or trying to do so would
 * result in a huge number of rules)
 */
public abstract class AbstractMetaOperator extends Op implements MetaOperator {

    public static final Sort METASORT = new PrimitiveSort(new Name("Meta"));
    private int arity;

    private TermFactory termFab = TermFactory.DEFAULT; 

    public AbstractMetaOperator(Name name, int arity) {
	super(name);
	this.arity = arity;
    }

    /**
     * checks whether the top level structure of the given @link Term
     * is syntactically valid, given the assumption that the top level
     * operator of the term is the same as this Operator. The
     * assumption that the top level operator and the term are equal
     * is NOT checked.  
     * @return true iff the top level structure of
     * the @link Term is valid.
     */
    public boolean validTopLevel(Term term) {
	// a meta operator accepts almost everything
	return term.op() instanceof AbstractMetaOperator;
    }


    /**
     * determines the sort of the {@link Term} if it would be created using this
     * Operator as top level operator and the given terms as sub terms. The
     * assumption that the constructed term would be allowed is not checked.
     * @param term an array of Term containing the subterms of a (potential)
     * term with this operator as top level operator
     * @return sort of the term with this operator as top level operator of the
     * given substerms
     */
    public Sort sort(Term[] term) {
	return METASORT;
    }

    /** @return arity of the Operator as int */
    public int arity() {
	return arity;
    }

    /** @return String representing a logical integer literal 
     *  in decimal representation
     */
    public static String convertToDecimalString(Term term, Services services) {
      	String result = "";
	boolean neg = false;
	Operator top = term.op();
	LDT intModel = services.getTypeConverter().
	    getModelFor(PrimitiveType.JAVA_INT);
	Namespace intFunctions = intModel.functions();
	Operator numbers = (Operator)intFunctions.lookup(new Name("Z"));
	Operator base = (Operator)intFunctions.lookup(new Name("#"));
	Operator minus =(Operator) intFunctions.lookup(new Name("neg"));
	
	// check whether term is really a "literal"
	boolean number=false;
	if (!top.name().equals(numbers.name())){
	    Debug.out("abstractmetaoperator: Cannot convert to number:", term);
	    throw (new NumberFormatException());
	}
	term = term.sub(0);
	top = term.op();

	while (top.name().equals(minus.name())){
	    neg=!neg;
	    term = term.sub(0);
	    top = term.op();
	}

	while (! top.name().equals(base.name())){
	    result = top.name()+result;
	    term = term.sub(0);
	    top = term.op();
	}
	
	if (neg)
	    return "-"+result;
	else
	    return result;
    }
    
    

    /** (non-Javadoc)
     * by default meta operators do not match anything 
     * @see de.uka.ilkd.key.logic.op.Operator#match(Operator, de.uka.ilkd.key.rule.MatchConditions, de.uka.ilkd.key.java.Services)
     */
    public MatchConditions match(Operator op, MatchConditions mc,
            Services services) {
        return null;
    }
    /** calculates the resulting term. */
    public abstract Term calculate(Term term, SVInstantiations svInst, Services services);

}
