// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.logic.op;

import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.sort.ProgramSVSort;


/**
 * This is an if-else operator (T0 ?= T1) T2 /-> T3 .
 * It replaces the former CondTermOp, the notation of the conditional term
 * operator <code> (ref1?=ref2).attr /-> t </code> can be used further but this
 * form is just an abbreviation for <code> (ref1?=ref2) ref1.attr /-> t </code>
 */
public class IfElse extends Op {

    
    /** 
     * creates the default if-else operator
     */
    IfElse() {
	super(new Name("if-else"));
    }

    /** 
     * creates an if-else operator of the given name 
     */
    IfElse(Name name) {
	super(name);
    }

    /**
     * @return true iff the arity equals 4 and if the following sort conditions
     * hold:
     *   1. in the sort hierarchy the sorts of subterm 0 and 1 have to occur in the
     *      same line
     *   2. the sorts of the terms in the 'then' and 'else' part of the term
     *      have to be the same one (may be weakened later)
     */
    public boolean validTopLevel(Term term) {

	Sort s0 = term.sub(0).sort();
	Sort s1 = term.sub(1).sort();
	Sort s2 = term.sub(2).sort();
	Sort s3 = term.sub(3).sort();

	if (term.arity() != arity() || 
	    !(s0.extendsTrans(s1) || s1.extendsTrans(s0) ||
	      s0 instanceof ProgramSVSort || s1 instanceof ProgramSVSort) ||   
	    !(s2.extendsTrans(s3) || s3.extendsTrans(s2) || 
	      s2 instanceof ProgramSVSort || s3 instanceof ProgramSVSort)) {
	    return false;
	}
	return true;
    }

    /**
     * determines the sort of the {@link Term} if it would be created using this
     * Operator as top level operator and the given terms as sub terms. The
     * assumption that the constructed term would be allowed is not checked.
     * @param term an array of Term containing the subterms of a (potential)
     * term with this operator as top level operator
     * @return sort of the term with this operator as top level operator of the
     * given substerms
     * @throws IllegalArgumentException if the given 'then' and 'else' part of the term
     * have different sorts
     */
    public Sort sort(Term[] term) {
	Sort s2 = term[2].sort();
	Sort s3 = term[3].sort();
	if (s2.extendsTrans(s3) || (s2 instanceof ProgramSVSort)) {
	    return s3;
	} if (s3.extendsTrans(s2) || (s3 instanceof ProgramSVSort)) {
	    return s2;
	} else 
	    throw new IllegalArgumentException
		("Cannot determine sort of invalid term (sort of the " +
		 "'then' and 'else' part are not compatible\n " +
		 "(" + term[2].sort() + " and " +
		 term[3].sort() + ")");
    }

    /** @return arity of the if-else operator as int */
    public int arity() {
	return 4;
    }
}
