// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.logic.op;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.logic.sort.Sort;
import java.io.IOException;
import de.uka.ilkd.key.logic.JavaBlock;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.ProgramElementName;
import de.uka.ilkd.key.rule.MatchConditions;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.java.visitor.Visitor;

public class ModalOperatorSV implements SchemaVariable {
    /** creates a new SchemaVariable. That is used as placeholder for
     * modal operators.
     * @param name the Name of the SchemaVariable
     * @param arity the arity of the modal operators matched by this SV
     * @param modalities modal operators matched by this SV
     */    
    private int arity;
    private Modality[] modalities;
    private Name name;
    
    ModalOperatorSV(Name name, int arity, Modality[] modalities) {
	this.arity = arity;
	this.modalities = modalities;
	this.name = name;
    }
    
    // Operator.java methods

    public Name name(){
	return name;
    }

    /**
     * checks whether the top level structure of the given @link Term
     * is syntactically valid, given the assumption that the top level
     * operator of the term is the same as this Operator. The
     * assumption that the top level operator and the term are equal
     * is NOT checked.  
     * @return true iff the top level structure of
     * the {@link Term} is valid.
     */
    public boolean validTopLevel(Term term){
	if (term.arity() != this.arity) return false;
	boolean result = true;
	for(int i=0;i<term.arity(); i++)
	    result = result && term.sub(i).sort().equals(Sort.FORMULA);
	return result;
    }

    /**
     * determines the sort of the {@link Term} if it would be created using this
     * Operator as top level operator and the given terms as sub terms. The
     * assumption that the constructed term would be allowed is not checked.
     * @param term an array of Term containing the subterms of a (potential)
     * term with this operator as top level operator
     * @return sort of the term with this operator as top level operator of the
     * given substerms
     */
    public Sort sort(Term[] term){
	return Sort.FORMULA;
    }


   /** @return arity of the Operator as int */
    public int arity(){
	return this.arity;
    }

    /**
     * @return true if the value of "term" having this operator as
     * top-level operator and may not be
     * changed by modalities
     */
    public boolean isRigid (Term term) {
	return (term.javaBlock () == JavaBlock.EMPTY_JAVABLOCK);
    }

    
    
    /**
     * (non-Javadoc)
     * @see de.uka.ilkd.key.logic.op.Operator#match(Operator, de.uka.ilkd.key.rule.MatchConditions, de.uka.ilkd.key.java.Services)
     */
    public MatchConditions match(Operator op, MatchConditions mc,
            Services services) {        
        if (!(op instanceof Modality)) {
            Debug.out("FAILED. ModalitySV matches only modalities (template, orig)",
                    this, op);
            return null;
        }  
        final Modality modalOp = (Modality)op;
        boolean matchedModal = false;
        for(int i=0; i<modalities.length; i++) {
            if (modalities[i] == modalOp) {
                // add a match condition schemamodal -> o
                Modality o=(Modality) 
                   mc.getInstantiations().getInstantiation(this);
                if (o == null) {
                    return mc.
                      setInstantiations(mc.getInstantiations().add(this, modalOp));
                } else {
                    if(o != modalOp) {
                        Debug.out("FAILED. Already instantiated with a different modality.");
                        return null;
                    }
                }                           
            }
        }
        Debug.out("FAILED. template is schema modal,"
                +" term is a modality, but not a matching one");
        return null; // template is schema modal, term is a modality, but not a matching one     
    }
    // SchemaVariable methods

    public boolean isModalOperatorSV() {
	return true;
    }
    
    public Modality[] modalities() {
	return modalities;
    }
 
    public boolean isRigid () {
	return false;
    }
    
    /** toString */
    public String toString() {
	return name() + " (modal operator)";
    }

    public boolean equalsModRenaming(SourceElement se, 
				     NameAbstractionTable nat) {
	return se == this;
    }
    
    
    public Class matchType() {
	return Modality.class;
    }

    /** returns true iff this SchemaVariable is used to match
     * bound (quantifiable) variables 
     * @return true iff this SchemaVariable is used to match
     * bound (quantifiable) variables 
     */

    public boolean isVariableSV() {
	return false;
    }

    /** returns true iff this SchemaVariable is used to match
     * a term but not a formula
     * @return true iff this SchemaVariable is used to match
     * a term but not a formula
     */
    public  boolean isTermSV() {
	return false;
    }

    /** returns true iff this SchemaVariable is used to match
     * a formula 
     * @return true iff this SchemaVariable is used to match
     * a formula
     */
    public boolean isFormulaSV() {
	return false;
    }

    /** returns true iff this SchemaVariable is used to match
     * a part of a program 
     * @return true iff this SchemaVariable is used to match
     * a part of a program
     */
    public boolean isProgramSV() {
	return false;
    }

    /**
     * returns true if the schemavariable can match a list of syntactical
     * elements 
     * @return true if the schemavariable can match a list of syntactical
     * elements 
     */
    public boolean isListSV() {
	return false;
    }

    /** returns true iff this SchemaVariable is used to match (create)
     * a skolem term
     * @return true iff this SchemaVariable is used to match (create)
     * a skolem term
     */
    public boolean isDependingSV() {
	return false;
    }

    // Stuff needed for ProgramSV 
    /** @returns comments if the schemavariable stands for programm
     * construct and has comments attached to it (not supported yet)
     */
    public Comment[] getComments() {
	return new Comment[0];
    }

    public KeYJavaType getKeYJavaType(){
	return null;
    }

    public SourceElement getFirstElement() {
	return null;
    }

    public SourceElement getLastElement() {
	return null;
    }

    public Position getStartPosition() {
	return Position.UNDEFINED;
    }

    public Position getEndPosition() {
	return Position.UNDEFINED;
    }

    public Position getRelativePosition(){
	return Position.UNDEFINED;
    }

    public ReferencePrefix getReferencePrefix(){
	return null;
    }

    public ReferencePrefix setReferencePrefix(ReferencePrefix r) {
	return null;
    }


    public int getDimensions() {
	return 0;
    }


    public int getTypeReferenceCount() {
	return 0;
    }

    public TypeReference getTypeReferenceAt(int index) {
	return null;
    }

    public PackageReference getPackageReference() {
	return null;
    }

    public int getExpressionCount() {
	return 0;
    }

    public Expression getExpressionAt(int index) {
	return null;
    }

    public int getChildCount() {
	return 0;
    }

    public ProgramElement getChildAt(int index) {
	return null;
    }
    
    public int getIndexOfChild(ProgramElement child)  {
	return 0;
    }
    
    public int getChildPositionCode(ProgramElement child) {
	return 0;
    }

    public int getIndexOfChild(int positionCode) {
	return 0;
    }

    public int getRoleOfChild(int positionCode) {
	return 0;
    }

    public int getStatementCount() {
	return 0;
    }

    public int size() {
	return 0;
    }

    public ArrayOfExpression getUpdates() {
	return null;
    }

    public ArrayOfLoopInitializer getInits() {
	return null;
    }

    public Statement getStatementAt(int i) {
	return null;
    }

    public ProgramElementName getProgramElementName() {
	return new ProgramElementName(toString());
    }

    public String getName() {
	return name().toString();
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnSchemaVariable(this);
    }

    /** this pretty printer method is for the program pretty printer
     * and needs not to be overwritten by ProgramSV but at the moment
     * it is not 
     */
    public void prettyPrint(PrettyPrinter w) throws IOException {
	w.printSchemaVariable(this);
    }

    public KeYJavaType getKeYJavaType(Services javaServ) {
	return null;
    }

}
