// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
/**
 * Created on 18.01.2005
 */
package de.uka.ilkd.key.logic.op;

import java.util.HashMap;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.sort.ArrayOfSort;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.rule.MatchConditions;
import de.uka.ilkd.key.util.Debug;

/**
 * Non Rigid Fucntion with explicit dependencies. This means a list of 
 * location of whose and only whose value the interpretation of the 
 * function symbol depends.
 * (May be generalized to more than locations in future)
 */
public class NRFunctionWithExplicitDependencies extends NonRigidFunction {

    // HACK. Need more general framework for creating such functions see also
    // ArrayOp, AttributeOp
    private static HashMap pool = new HashMap();
    
    /**
     * retrieves the non-rigid function with the given name and dependencies
     * or returns null if no such function symbol exists
     * @param name the Name of the function symbol to look for
     * @param dependencies the ArrayOfLocation with the dependencies
     * @return the non-rigid function symbol 
     */
    public static NRFunctionWithExplicitDependencies
	getSymbol(Name name, ArrayOfLocation dependencies) {
        HashMap mapDep2Op = (HashMap)pool.get(name);
        NRFunctionWithExplicitDependencies op = 
            (NRFunctionWithExplicitDependencies) mapDep2Op.get(name);
        return op;
    }
    
    /**
     * retrieves the non-rigid function with the given name and dependencies
     * and creates one if not available
     * @param name the Name of the function symbol to look for
     * @param sort the Sort of the function symbol
     * @param argSorts the array of Sorts for the arguments
     * @param dependencies the array of Locations describing the dependencies
     * @return the non-rigid function symbol 
     */
    public static NRFunctionWithExplicitDependencies
    	getSymbol(Name name, Sort sort, Sort[] argSorts, Location[] dependencies) {        
        return getSymbol(name, sort,
                new ArrayOfSort(argSorts), new ArrayOfLocation(dependencies));
    }
        
    /**
     * retrieves the non-rigid function with the given name and dependencies
     * and creates one if not available
     * @param name the Name of the function symbol to look for
     * @param sort the Sort of the function symbol
     * @param argSorts the ArrayOfSort for the arguments
     * @param dependencies the ArrayOfLocation with the dependencies
     * @return the non-rigid function symbol 
     */
    public static NRFunctionWithExplicitDependencies
    	getSymbol(Name name, Sort sort, 
    	        ArrayOfSort argSorts, ArrayOfLocation dependencies) {

        HashMap mapDep2Op = (HashMap)pool.get(name);
        if (mapDep2Op == null) {
            mapDep2Op = new HashMap();
            pool.put(name, mapDep2Op);
        }
        NRFunctionWithExplicitDependencies op = 
            (NRFunctionWithExplicitDependencies) mapDep2Op.get(dependencies);
        
        if (op == null) {
            op = new NRFunctionWithExplicitDependencies
            	(name, sort, argSorts, dependencies);
            mapDep2Op.put(name, op);
        }
                
        return op;
    }    
    
    /**
     * the meaning of this function symbol depends on the values of the
     * locations contained in this array
     */
    private final ArrayOfLocation dependencies;
    
    /**    
     * creates a non rigid function with given signaturen and dependencies
     * @param name the Name of the non-rigid function symbol
     * @param sort the Sort of the symbol
     * @param argSorts the array of Sort defining the argument sorts
     * @param dependencies the array of Locations whose values influence 
     * the meaning of this symbol
     */ 
    private NRFunctionWithExplicitDependencies
    	(Name name, Sort sort, Sort[] argSorts, Location[] dependencies) {
        this(name, sort, new ArrayOfSort(argSorts), 
                new ArrayOfLocation(dependencies));       
    }

    /**
     * creates a non rigid function with given signaturen and dependencies
     * @param name the Name of the non-rigid function symbol
     * @param sort the Sort of the symbol
     * @param argSorts the ArrayOfSort defining the argument sorts
     * @param dependencies the ArrayOfLocation whose values influence 
     * the meaning of this symbol
     */
    private NRFunctionWithExplicitDependencies(Name name, Sort sort,
            ArrayOfSort argSorts, ArrayOfLocation dependencies) {
        super(name, sort, argSorts);
        this.dependencies = dependencies;
    }

    /**
     * two non-rigid function symbols can be matched if their list of
     * dependencies match 
     */
    public MatchConditions match(Operator op, MatchConditions mc, Services services) {
        MatchConditions result = mc;
        if (this == op) return result;
        if (op.getClass() != getClass()) {
            Debug.out("FAILED matching. Incomaptible operators " +
            "(template, operator)", this, op);
            return null;
        }      
        final NRFunctionWithExplicitDependencies nrFunc = 
            (NRFunctionWithExplicitDependencies)op;
        if (dependencies.size() == nrFunc.dependencies.size()) {
            for (int i = 0, locs = dependencies.size(); i<locs; i++) {
                result = dependencies.getLocation(i).
                match(nrFunc.dependencies.getLocation(i), result, services);
                if (result == null) { // fail fast
                    Debug.out
                    ("FAILED. NRFuncWithExplicitDependences mismatch " +
                            "(template, operator)", this, nrFunc);
                    return null;
                }
            }
            return result;
        }
        Debug.out("FAILED matching. NRWithExplicitDependencies match " +
                "failed because of incompatible locations (template, operator)", 
                this, op);
        return null;
    }
    
    /**
     * returns the list of locations this function depends on    
     * @return the array of locations this function depends on
     */
    public ArrayOfLocation dependencies() {
        return dependencies;
    }
    
    /**
     * toString
     */
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append(name().toString());
        sb.append("[");
        for (int i = 0; i<dependencies.size(); i++) {
            sb.append(dependencies.getLocation(i));
            if (i<dependencies.size()-1) {
                sb.append(",");
            }
        }
        sb.append("]");
        sb.append("(");
        for (int i = 0; i<argSort().size(); i++) {
            sb.append(argSort().getSort(i));
            if (i<argSort().size()-1) {
                sb.append(",");
            }
        }
        sb.append("):");
        sb.append(sort());        
        return sb.toString();
    }
}
