// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.logic.op;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.rule.MatchConditions;
import de.uka.ilkd.key.rule.metaconstruct.arith.*;
import de.uka.ilkd.key.rule.metaconstruct.*;
import de.uka.ilkd.key.util.Debug;

import java.io.ObjectStreamException;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Term;


/** 
 * This class contains logical connectives, most of them derive
 * from this class. If you want to create a term with a connective
 * that is stored in this class you have to take this one, because
 * these operators are handled as singleton so that e.g. 
 * these operators are compared using equality on references not on
 * names. 
 */
public abstract class Op implements Operator {
    
   
    // OPERATORS
    /** the ususal 'negation' operator '-' */
    public static final Junctor NOT = new Junctor(new Name("not"),1);	
    /** 
     * the ususal 'and' operator '/\' (be A, B formulae then 'A /\ B'
     * is true if and only if A is true and B is true 
     */
    public static final Junctor AND = new Junctor(new Name("and"),2);
    /** 
     * the ususal 'or' operator '\/' (be A, B formulae then 'A \/ B'
     * is true if and only if A is true or B is true 
     */
    public static final Junctor OR = new Junctor(new Name("or"),2);
    /**
     * the ususal 'implication' operator '->' (be A, B formulae then
     * 'A -> B' is true if and only if A is false or B is true 
     */
    public static final Junctor IMP = new Junctor(new Name("imp"),2);
    /** 
     * the ususal 'equivalence' operator '<->' (be A, B formulae then       
     * 'A <->  B' is true if and only if A and B have the same truth
     * value 
     */ 
    public static final Equality EQV = new Equality(new Name("equiv"),
						    Sort.FORMULA);
    /** the ususal 'equality' operator '=' */
    public static final Equality EQUALS = new Equality(new Name("equals"));
    /** the ususal 'forall' operator 'all' (be A a formula then       
     * 'all x.A' is true if and only if for all elements d of the
     * universe A{x<-d} (x substitued with d) is true */     
    public static final Quantifier ALL = new Quantifier(new Name("all"));
    /** the ususal 'exists' operator 'ex' (be A a formula then       
     * 'ex x.A' is true if and only if there is at least one elements
     * d of the universe such that A{x<-d} (x substitued with d) is true */     
    public static final Quantifier EX = new Quantifier(new Name("exist"));
    /** the diamond operator of dynamic logic. A formula
     * <alpha;>Phi can be read as after processing the program alpha
     * there exists a state such that Phi holds. */    
    public static final Modality DIA = new Modality(new Name("diamond"));
    /** the box operator of dynamic logic. A formula
     * [alpha;]Phi can be read as 'In all states reachable
     * processing the program alpha the formula Phi holds'.*/
    public static final Modality BOX = new Modality(new Name("box"));
    /** the throughout operator of dynamic logic. A formula
     * [[alpha;]]Phi can be read as 'In all intermediate states during
     * processing the program alpha the formula Phi holds'.*/
    public static final Modality TOUT = new Modality(new Name("throughout"));
    /**
     * Diamond operator used for a transaction that is going to commit 
     */
    public static final Modality DIATRC = new Modality(new Name("diamond_trc"));
    /**
     * Box operator used for a transaction that is going to commit 
     */
    public static final Modality BOXTRC = new Modality(new Name("box_trc"));
    /**
     * Throughout operator used for a transaction that is going to commit 
     */
    public static final Modality TOUTTRC = new Modality(new Name("throughout_trc"));
    /**
     * Diamond operator used for a transaction that is going to abort
     */
    public static final Modality DIATRA = new Modality(new Name("diamond_tra"));
    /**
     * Box operator used for a transaction that is going to abort
     */
    public static final Modality BOXTRA = new Modality(new Name("box_tra"));
    /**
     * Throughout operator used for a transaction that is going to abort
     */
    public static final Modality TOUTTRA = new Modality(new Name("throughout_tra"));
    /** the substitution operator {var<-term}. {x<-d}A(x) means
     * replace all free occurrences of variable x in A with d. */	
    public static final SubstOp SUBST = new SubstOp(new Name("subst"));
    /** the wary substitution operator {var<-term}'. {x<-d}'A(x) means
     * replace all free occurrences of variable x in A with d, however
     * without replacing x with a non-rigid A below modalities */
    /** create an existential quantifier if needed */
    public static final SubstOp WARY_SUBST_EX =
	new WarySubstOp(new Name("warySubstEx"), false);
    /** create an universal quantifier if needed */
    public static final SubstOp WARY_SUBST_ALL =
	new WarySubstOp(new Name("warySubstAll"), true);
    /** the true constant */
    public static final Junctor TRUE = new Junctor(new Name("true"),0);
    /** the false constant */
    public static final Junctor FALSE = new Junctor(new Name("false"),0);
    /** the null pointer */
    public static final Function NULL = new Function(new Name("null"),
						     Sort.NULL, 
						     new Sort[0]);
    public static final ListOp LIST = new ListOp(new Name("TermList"));

    /** represents the everything-keyword used in modifier sets.*/
    public static final Function EVERYTHING = 
    new Function(new Name("everything"), Sort.TERMLIST, new Sort[0]);

    /** represents the nothing-keyword used in modifier sets.*/
    public static final Function NOTHING = 
    new Function(new Name("nothing"), Sort.TERMLIST, new Sort[0]);

    /** the 'if-else' operator */
    public static final IfElse IF_ELSE = new IfElse();

    /** control operator required for specification computation */
    public static final Junctor COMPUTE_SPEC_OP = new ComputeSpecOp();

    public static final MetaLength META_LENGTH = new MetaLength();

    public static final MetaNext META_NEXT = new MetaNext();

    public static final MetaCreated META_CREATED = new MetaCreated();

    public static final MetaTransient META_TRANSIENT = new MetaTransient();

    /** the shadow operator for transactions */

    public static final MetaShadow META_SHADOW = new MetaShadow();

    /** used to add integers */
    public static final MetaAdd META_ADD = new MetaAdd();
    public static final MetaSub META_SUB = new MetaSub();
    public static final MetaMul META_MUL = new MetaMul();
    public static final MetaDiv META_DIV = new MetaDiv();
    public static final MetaJDiv META_JDIV = new MetaJDiv();
    public static final MetaMod META_MOD = new MetaMod();
    public static final MetaJMod META_JMOD = new MetaJMod();
    public static final MetaLess META_LESS = new MetaLess();
    public static final MetaGreater META_GREATER = new MetaGreater();
    public static final MetaLeq META_LEQ = new MetaLeq();
    public static final MetaGeq META_GEQ = new MetaGeq();
    public static final MetaEqual META_EQ = new MetaEqual();
    public static final MetaJavaIntUnaryMinus META_INT_UNARY_MINUS = new
	MetaJavaIntUnaryMinus();
    public static final MetaJavaLongUnaryMinus META_LONG_UNARY_MINUS = new
	MetaJavaLongUnaryMinus();
    public static final MetaJavaIntAdd META_INT_ADD = new
	MetaJavaIntAdd();
    public static final MetaJavaIntSub META_INT_SUB = new
	MetaJavaIntSub();
    public static final MetaJavaIntMul META_INT_MUL = new
	MetaJavaIntMul();
    public static final MetaJavaIntDiv META_INT_DIV = new
	MetaJavaIntDiv();
    public static final MetaJavaIntMod META_INT_MOD = new
	MetaJavaIntMod();
    public static final MetaJavaIntAnd META_INT_AND = new
	MetaJavaIntAnd();
    public static final MetaJavaIntOr META_INT_OR = new
	MetaJavaIntOr();
    public static final MetaJavaIntXor META_INT_XOR = new
	MetaJavaIntXor();
    public static final MetaJavaIntComplement META_INT_COMPLEMENT = new
	MetaJavaIntComplement();
    public static final MetaJavaIntShiftRight META_INT_SHIFTRIGHT =
	new MetaJavaIntShiftRight();
    public static final MetaJavaIntShiftLeft META_INT_SHIFTLEFT = new
	MetaJavaIntShiftLeft();
    public static final MetaJavaIntUnsignedShiftRight
	META_INT_UNSIGNEDSHIFTRIGHT = new
	MetaJavaIntUnsignedShiftRight();

    public static final MetaJavaLongAdd META_LONG_ADD = new
	MetaJavaLongAdd();
    public static final MetaJavaLongSub META_LONG_SUB = new
	MetaJavaLongSub();
    public static final MetaJavaLongMul META_LONG_MUL = new
	MetaJavaLongMul();
    public static final MetaJavaLongDiv META_LONG_DIV = new
	MetaJavaLongDiv();
    public static final MetaJavaLongMod META_LONG_MOD = new
	MetaJavaLongMod();
    public static final MetaJavaLongAnd META_LONG_AND = new
	MetaJavaLongAnd();
    public static final MetaJavaLongOr META_LONG_OR = new
	MetaJavaLongOr();
    public static final MetaJavaLongXor META_LONG_XOR = new
	MetaJavaLongXor();
    public static final MetaJavaLongComplement META_LONG_COMPLEMENT = new
	MetaJavaLongComplement();
    public static final MetaJavaLongShiftRight META_LONG_SHIFTRIGHT =
	new MetaJavaLongShiftRight();
    public static final MetaJavaLongShiftLeft META_LONG_SHIFTLEFT =
	new MetaJavaLongShiftLeft();
    public static final MetaJavaLongUnsignedShiftRight
	META_LONG_UNSIGNEDSHIFTRIGHT = new
	MetaJavaLongUnsignedShiftRight();
    public static final MetaInLong META_IN_LONG = new
	MetaInLong();
    public static final MetaInInt META_IN_INT = new
	MetaInInt();
    public static final MetaInChar META_IN_CHAR = new
	MetaInChar();
    public static final MetaInShort META_IN_SHORT = new
	MetaInShort();
    public static final MetaInByte META_IN_BYTE = new
	MetaInByte();
    public static final MetaModuloByte META_MODULO_BYTE = new
	MetaModuloByte();
    public static final MetaModuloShort META_MODULO_SHORT = 
	new MetaModuloShort();
    public static final MetaModuloInteger META_MODULO_INTEGER = 
	new MetaModuloInteger();
    public static final MetaModuloLong META_MODULO_LONG = new MetaModuloLong();

    public static final StaticAnalysis STATIC_ANALYSIS = new StaticAnalysis();

    public static final WhileInvRule WHILE_INV_RULE = new WhileInvRule();

    public static final IntroduceNewAnonymousUpdateOp INTRODUCE_NEW_ANON_UPDATE = 
        new IntroduceNewAnonymousUpdateOp();

    public static final ResolveQuery RESOLVE_QUERY = new ResolveQuery();

    public static final ConstantValue CONSTANT_VALUE = new ConstantValue();

    /** metaconstructs for OCL simplification */
    public static final MetaAllSubtypes METAALLSUBTYPES = new 
	MetaAllSubtypes();

    protected final Name name;
   	
    protected Op(Name name) {
	this.name=name;
    }
    
    protected boolean equalsForResolve(Operator op) {
	if (op.name().equals(this.name())) {
	    if (op.getClass() == this.getClass()) {
		if (op instanceof Junctor) {
		    if (((Junctor)op).arity() == ((Junctor)this).arity()) {
			return true;
		    } 
		} else {
		    return true;
		}
	    } 
	}
	return false;
    }

    
    protected Object readResolve() 
	throws ObjectStreamException {	

	Op[] op = {Op.NOT, Op.AND, Op.OR, Op.IMP,
		   Op.ALL, Op.EX,
		   Op.DIA, Op.BOX, Op.TOUT,
		   Op.DIATRC, Op.BOXTRC, Op.TOUTTRC,
		   Op.DIATRA, Op.BOXTRA, Op.TOUTTRA,
                   Op.SUBST, Op.TRUE,
		   Op.FALSE};
	for (int i=0;i<op.length;i++) {
	    if (equalsForResolve(op[i])) {
		return op[i];
	    }
	}
	return this;
    }
    
    /**
     * Returns a modality corresponding to a string
     * @param str name of the modality to return
     */
    public static Modality getModality(String str) {
        if("diamond".equals(str)) return Op.DIA;
        if("box".equals(str)) return Op.BOX;
        if("throughout".equals(str)) return Op.TOUT;
        if("diamond_trc".equals(str)) return Op.DIATRC;
        if("box_trc".equals(str)) return Op.BOXTRC;
        if("throughout_trc".equals(str)) return Op.TOUTTRC;
        if("diamond_tra".equals(str)) return Op.DIATRA;
        if("box_tra".equals(str)) return Op.BOXTRA;
        if("throughout_tra".equals(str)) return Op.TOUTTRA;
        return null;
    }

    public Name name() {
	return name;
    }
	
    public String toString() {
	return name().toString();
    }

    /**
     * @return true if the value of "term" having this operator as
     * top-level operator and may not be changed by modalities
     */
    public boolean isRigid (Term term) {
	return term.hasRigidSubterms ();
    }
        
    /** 
     * implements the default operator matching rule which means 
     * that the compared object have to be equal otherwise
     * matching fails
     */
    public MatchConditions match(Operator op, MatchConditions mc,
            Services services) {
        if (op == this) {
            return mc;
        }
        Debug.out("FAILED. Operators are different(template, candidate)", this, op);
        return null;
    }
}
