// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//


package de.uka.ilkd.key.logic.op;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.sort.GenericSort;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.sort.ArrayOfSort;
import de.uka.ilkd.key.logic.sort.SortDefiningSymbols;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.rule.MatchConditions;
import de.uka.ilkd.key.rule.inst.GenericSortCondition;
import de.uka.ilkd.key.rule.inst.SortException;
import de.uka.ilkd.key.util.Debug;

public class SortDependingFunction extends Function
    implements SortDependingSymbol {
    
    private final Name kind;
    private final Sort sortDependingOn;

    /** creates a Function 
     * @param name String with name of the function
     * @param sort the Sort of the function (result type)
     * @param kind name of the kind this object belongs to
     * @param sortDependingOn sort this object is depending on, this
     * is the difference between various objects of the same kind
     * @return argsorts array of type Sort with the allowed sorts of
     * the function arguments */   
    public SortDependingFunction ( Name   name,
				   Sort   sort,
				   Sort[] argSorts,
				   Name   kind,
				   Sort   sortDependingOn ) {
	super ( name, sort, argSorts );
	this.kind            = kind;
	this.sortDependingOn = sortDependingOn;
    }

    public SortDependingFunction ( Name        name,
				   Sort        sort,
				   ArrayOfSort argSorts,
				   Name        kind,
				   Sort        sortDependingOn ) {
	super ( name, sort, argSorts );
	this.kind            = kind;
	this.sortDependingOn = sortDependingOn;
    }

    /**
     * @return the sort this object has been instantiated for
     */
    public Sort    getSortDependingOn () {
	return sortDependingOn;
    }

    /**
     * Compares "this" and "p"
     * @param p object to compare
     * @return true iff this and p are instances of the same kind of
     * symbol, but possibly instantiated for different sorts
     */
    public boolean isSimilar          ( SortDependingSymbol p ) {
	return getKind ().equals ( p.getKind () );
    }

    /**
     * Assign a name to this term symbol, independant of concrete
     * instantiations for different sorts
     * @return the kind of term symbol this object is an instantiation
     * for
     */
    public Name    getKind            () {
	return kind;
    }

    /**
     * Get the instance of this term symbol defined by the given sort
     * "p_sort"
     * @return a term symbol similar to this one, or null if this
     * symbol is not defined by "p_sort"
     *
     * POSTCONDITION: result==null || (this.isSimilar(result) &&
     * result.getSortDependingOn()==p_sort)
     */
    public SortDependingSymbol getInstanceFor ( SortDefiningSymbols p_sort ) {
	return p_sort.lookupSymbol ( getKind () );
    }

    
    
    /* (non-Javadoc)
     * @see de.uka.ilkd.key.logic.op.Operator#match(de.uka.ilkd.key.logic.op.Operator, de.uka.ilkd.key.rule.MatchConditions, de.uka.ilkd.key.java.Services)
     */
    public MatchConditions match(Operator op, MatchConditions mc,
            Services services) {      
        if (this.getClass() != op.getClass()) {
            Debug.out("FAILED. Given operator cannot be matched by a sort" +
            		"depending function (template, orig)", this, op);
            return null;
        } else if (getSortDependingOn() instanceof GenericSort) {         
            MatchConditions matchCond = mc;
            final GenericSortCondition c = 
                GenericSortCondition.createCondition( this, (SortDependingFunction)op);            
            if ( c == null ) {
                Debug.out("FAILED. Generic sort condition");
                return null; //FAILED;
            } else {
                try {
                    matchCond = matchCond
                    .setInstantiations ( matchCond.getInstantiations ().
                            add ( c ) );
                } catch ( SortException e ) {
                    Debug.out("FAILED. Sort mismatch.");
                    return null; //FAILED;
                }
            }        
            return matchCond;
        } else {
            return super.match(op, mc, services);
        }
    }
}
