// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.ocl.ocldltranslator.transform;

import java.util.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.*;

import de.uka.ilkd.key.ocl.ocldltranslator.*;
import de.uka.ilkd.key.ocl.ocldltranslator.heuristics.*;
import de.uka.ilkd.key.ocl.ocldltranslator.control.*;
import de.uka.ilkd.key.ocl.ocldltranslator.normalize.*;

/** 
 * A class collecting all constant symbols for the allInstances feature.
 *
 * It is used to get the axiomatic description for ALL such function symbols
 * since the translation omits such axiomatizations for the @pre-forms.
 *
 * A flag controls whether the '@pre' form or the non-'@pre'-form will
 * be collected.
 */ 
 
public class AllInstancesCollector extends Visitor implements TranslationConstants {
    
   
    private java.util.Set allInstancesSymbols;
    private boolean atPreForm;
    private ProverFacade facade;

    /**
     * Sets up an AllInstances-Collector that used the given facade object in order to
     * created non-@pre-function symbols if necessary.
     */

    public AllInstancesCollector(ProverFacade facade){
	this.facade = facade; 
    }

    /**
     * Starts collecting all the symbols used to represent the OCLs 'allInstances' feature that
     * occur in the given formula.
     * For the @pre-form the function symbol in non-'@pre'-form will be added !
     * @param formula - the formula that will be searched.
     * @param atPreForm - if true (false) only the (non-)-'@pre'-forms will be collected.
     * @return a set containing the specified function symbols.
     */

    public Set collectAllInstancesSymbols(Term formula, boolean atPreForm){
    
       	allInstancesSymbols = new HashSet(); 
	this.atPreForm = atPreForm;

	// Analyze the formula ...
	formula.execPostOrder(this);
	
	return allInstancesSymbols;
       
    }


    // Implementation of Visitor ...
    

    public void visit(Term visited) {
	
	 
	boolean changed = false;
	Function nonAtPreFunctionSymbol=null;
	
	if (visited.op() instanceof Function){
	    Function currentFunction = (Function) visited.op();
	    String funcNameString = currentFunction.name().toString();

	    int index = funcNameString.indexOf("::");
	    if (index != -1){
		funcNameString = funcNameString.substring(index + 2);
	    }
	   
	    if (funcNameString.startsWith(ALL_INSTANCES)){
		// We found an occurence of an allInstances-symbol ...
		
		// System.err.println("Found allInstances symbol : " + currentFunction);
		
		 if(atPreForm) {
		    // We are looking for @pre-occurrences only ...
		    
		    if (funcNameString.endsWith("@pre")){
			// we have an @pre-occurence, therefore we have to get the corresponding non-@pre function.
			// We use the facade object in order to garantuee that if such a symbol has been generated before,
			// we get exactly this already existing object ...

			 

			String nonAtPreFuncName = funcNameString.substring(0, funcNameString.length() - 4);
			nonAtPreFunctionSymbol = facade.getFunctionForFeature(nonAtPreFuncName,
									      new Sort[]{},
									      currentFunction.sort());
			// Add this symbol to the result ...
			allInstancesSymbols.add(nonAtPreFunctionSymbol);

			// System.err.println("Added non atPreForm symbol : " + nonAtPreFunctionSymbol);
		    }
		    
		} else {
		    // We are looking for non-@pre-occurrences only ...
		    
		    if (!funcNameString.endsWith("@pre")){
			// we have a non-@pre-occurence, therefore we add this non-@pre function to the result.
			
			allInstancesSymbols.add(currentFunction);
			
			// System.err.println("Added non same symbol : " + currentFunction);
			
		    }
		}
	 
	    }

	}

    }
}
