// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.ocl.ocldltranslator.transform;

import java.util.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.op.*;

import de.uka.ilkd.key.ocl.ocldltranslator.*;
import de.uka.ilkd.key.ocl.ocldltranslator.heuristics.*;
import de.uka.ilkd.key.ocl.ocldltranslator.control.*;
import de.uka.ilkd.key.ocl.ocldltranslator.normalize.*;

/** 
 * Only for use in the generation of the translation of an OCL constaint !!!
 *
 * Abstract base class for classes handling occurences of @pre-tagged
 * function symbols in the DL formula generated as the translation
 * of a pre/post-constraint.
 */ 
 
public class SimpleAtPreTransformer extends Visitor implements KeyAtPreTransformer {
    
    private Term visited;
    private Term computedResult;
    private java.util.Set alreadyHandledSymbols;
    private Vector additionalPreTerms;
    protected java.util.Map funcNameSymbolMap;
    protected java.util.Set nonrigidFunctionSymbols;
    protected java.util.Map atPreTaggedAttributeFunctions;
    protected java.util.Map atPreTaggedMethodFunctions;
    protected KeyOclConstraintTranslator cTranslator;

    /**
     * the stack contains the subterms that will be added in the next step of
     * execPostOrder in Term in order to build the new term. A boolean value
     * between or under the subterms on the stack indicate that a term using
     * these subterms should build a new term instead of using the old one,
     * because one of its subterms has been built, too.
     */
    private Stack subStack; //of Term (and Boolean)
    private TermFactory tf=TermFactory.DEFAULT;
    private Boolean newMarker=new Boolean(true);
    


    private Term[] neededSubs(int n) {
	boolean newTerm=false;
	Term[] result=new Term[n];
	for (int i=n-1; i>=0; i--) {
	    Object top=subStack.pop();
	    if (top==newMarker){
		newTerm=true; 
		top=subStack.pop();
	    }
	    result[i]=(Term) top;
	}
	if (newTerm && (subStack.empty() || subStack.peek()!=newMarker) ) {
	    subStack.push(newMarker);
	}
	return result;
    }




    // Implementation of AtPreTransformer ...

     
    /**
     * Sets up the transformer with the necessary informations.
     * This method has to be called before 'transformAtPre' is called ! 
     * @param nonrigidFunctionSymbols - the nonrigid function symbols in the given term.
     * @param prepostfSymbolMap - maps all names of function symbols onto the corresponding function symbol
     * @param atPreTaggedAttributeFunctions - maps function symbols for @pre-tagged attributes onto the 
     *                                        corresponding (non-tagged) attribute operator
     * @param constraitTranslator - the constraint translator which was used to construct the term to be transformed.
     */

    public void initTransformer(Set nonrigidFunctionSymbols, 
				Map prepostfSymbolMap, 
				Map atPreTaggedAttributeFunctions,
				Map atPreTaggedMethodFunctions,
				KeyOclConstraintTranslator constraitTranslator){
	this.funcNameSymbolMap = prepostfSymbolMap;
	this.nonrigidFunctionSymbols = nonrigidFunctionSymbols;
	this.atPreTaggedAttributeFunctions = atPreTaggedAttributeFunctions;
	this.atPreTaggedMethodFunctions = atPreTaggedMethodFunctions;
	this.cTranslator = constraitTranslator;

    }


    /**
     * Starts the transformation of @pre-tagged function symbols in the given term
     * @param prePostTermWithDiamond - the term to be transformed.
     * @param nonrigidFunctionSymbols - the nonrigid function symbols in the given term.
     * @param prepostfSymbolMap - maps all names of function symbols onto the corresponding function symbol
     * @param constraitTranslator - the constraint translator which was used to construct the term to be transformed.
     */

    public void transformAtPre(TranslationResult translationPre, 
			       TranslationResult translationPost){
    
	KeyTranslationResult preTranslation = (KeyTranslationResult) translationPre;
	KeyTranslationResult postTranslation = (KeyTranslationResult) translationPost;

	computedResult=null;
	subStack = new Stack(); // of Term
	
	
	additionalPreTerms = new Vector(); 

	alreadyHandledSymbols = new HashSet(); 

	// Now get the prePart and the postPart of the Term ...

	// Determine the preTerm and the postTerm ...
	Term preTerm = preTranslation.getTerm();
	Term postTerm = postTranslation.getTerm();
	Term postTermWithAxioms = cTranslator.convertTranslationResultToFormula(postTranslation, Op.IMP);

	// Analyze the postTerm (with axioms!) and compute which additional terms 
	// have to be added to the preTerm ...
	postTermWithAxioms.execPostOrder(this);
	
	// Now add the computed terms conjunctively to the preTerm ...
	// But be careful: Add the new formulas conjunctively to the translation term itself
	// but NOT on the conversion of the preTranslation to a formula (IMPLICATION !)...
	Term newPreTerm = preTerm;
	Term currentTerm;
	for (Enumeration e = additionalPreTerms.elements() ; e.hasMoreElements();){
	    currentTerm = (Term) e.nextElement();
	    newPreTerm = tf.createJunctorTermAndSimplify(Op.AND,newPreTerm, currentTerm);
	}
	
	// Change the preTranslation accordingly ...
	preTranslation.setTerm(newPreTerm);

       
    }



    // Implementation of Visitor ...
    

    public void visit(Term visited) {
	
	 // System.err.println("Currently VISITING : " + visited);


	boolean changed = false;
	Function newFunctionSymbol=null;
	

	if (visited.op() instanceof Function){
	    Function currentFunction = (Function) visited.op();
	    
	    Function newFunction;
	    String funcNameString = currentFunction.name().toString();
	  
	    if(!alreadyHandledSymbols.contains(currentFunction) &&  atPreTaggedAttributeFunctions.containsKey(currentFunction)){
		// We have an @pre-tagged funtion symbol for some attribute or method
		// of some model-type.

		// Therefore we have to add the following equation in the prePart ...
		// - For some attribute a in class C:
		//                  all' o:C (o.a = a@pre(o))
	

		
		ProgramVariable attributeOp 
		    = (ProgramVariable) atPreTaggedAttributeFunctions.get
		           (currentFunction);
		     
		Term funcTerm;
		Term attrTerm;
		
		if (attributeOp.isStatic()) {		
		    funcTerm = tf.createFunctionTerm(currentFunction);
		    attrTerm = tf.createVariableTerm(attributeOp);		    
		} else {
		    Sort appliedToSort = currentFunction.argSort(0);
		    LogicVariable objVar = new LogicVariable(new Name("o"),appliedToSort);
		    Term objTerm = tf.createVariableTerm(objVar);
		    funcTerm = tf.createFunctionTerm(currentFunction, objTerm);
		    attrTerm = tf.createAttributeTerm(attributeOp, objTerm);		    
		}


		Term eqTerm = tf.createEqualityTerm(funcTerm, attrTerm);

	
		// Finally build all-closure (important: by using constraintTranslator, because
		// of special quantors !!!)
		    
		Term allTerm = cTranslator.createAllClosureOfTerm(eqTerm);
		
		// System.err.println("Adding new term to preTerm : " + allTerm);
		
		// add this new term ...
		additionalPreTerms.add(allTerm);
		
		// Store current symbol in order to omit multiple handling of the same thing ...
		alreadyHandledSymbols.add(currentFunction);
		    
		
		
	    } else if (!alreadyHandledSymbols.contains(currentFunction) 
		       &&  atPreTaggedMethodFunctions.containsKey(currentFunction)) {
		Term[] argterms=new Term[currentFunction.arity()];
		Namespace varNs=new Namespace();
		for (int i=0; i<currentFunction.arity(); i++) {
		    String n=currentFunction.argSort(i).name().toString().substring(0,3).toLowerCase();
		    int j=0;
		    Name name;
		    do {
			name=new Name(n+(j++));
		    }
		    while (varNs.lookup(name)!=null);
		    LogicVariable lv=new LogicVariable(name, currentFunction.argSort(i));
		    argterms[i]=tf.createVariableTerm(lv);
		}
		ProgramMethod meth = (ProgramMethod) atPreTaggedMethodFunctions.get(currentFunction);
		Term funcTerm;
		Term methTerm;
		Term eqTerm;		
		funcTerm = tf.createFunctionTerm(currentFunction, argterms);
		methTerm = tf.createFunctionTerm(meth, argterms);
		eqTerm = tf.createEqualityTerm(funcTerm, methTerm);

	
		// Finally build all-closure (important: by using constraintTranslator, because
		// of special quantors !!!)
		    
		Term allTerm = cTranslator.createAllClosureOfTerm(eqTerm);
		
		// System.err.println("Adding new term to preTerm : " + allTerm);
		
		// add this new term ...
		additionalPreTerms.add(allTerm);
		
		// Store current symbol in order to omit multiple handling of the same thing ...
		alreadyHandledSymbols.add(currentFunction);
		
	    } else {
		// We have some other function symbol ...

		if(!alreadyHandledSymbols.contains(currentFunction) && funcNameString.endsWith("@pre")){
		    // we have an @pre-function, therefore we have to create a new defining term in the preTerm ...
		    
		    // Check wheter we already have a corresponding function symbol in the
		    // map from the preconditionFormulas ...
		    String searchFuncName = funcNameString.substring(0, funcNameString.length() - 4);
		    
		    // System.err.println("Connecting " + funcNameString + "; with : " + searchFuncName);
		    
		    
		    if (funcNameSymbolMap.containsKey(searchFuncName)){
			newFunction = (Function) funcNameSymbolMap.get(searchFuncName);
			// System.err.println("FOUND SYMBOL WITH SAME NAME => REUSING SYMBOL :" + newFunction);
		    } else {
			// the corresponding function has not been used in
			// the formula therefore we have to create a new such symbol
			// and store it in the map ...
			Name newFuncName;
			Name currentFuncName = currentFunction.name();
			newFuncName = new Name(searchFuncName);
			
			newFunction = new Function(newFuncName,
						   currentFunction.sort(),
						   currentFunction.argSort());
			
			funcNameSymbolMap.put(searchFuncName,newFunction);
			
			// System.err.println("NOT FOUND ! CREATING NEW SYMBOL :" + newFunction);
			
		    }
		    
		    newFunctionSymbol = newFunction;
		    
		    // Now we have to create a formula connecting the @pre symbol and the new symbol ...
		    // all x_1,..,x_n: f@pre(x_11,...x_n) = f(x_11,...x_n)
		    
		    ArrayOfSort argSorts = currentFunction.argSort();
		    NewVariableDeliverer varDeliverer = new NewVariableDeliverer();
		    Term[] qVarTerms = new Term[argSorts.size()];
		    LogicVariable currentVar;
		    for (int i=0;i<argSorts.size(); i++){
			currentVar = varDeliverer.getNewVariableForSort(argSorts.getSort(i));
			qVarTerms[i] = tf.createVariableTerm(currentVar);
		    }
		    
		    // Now build the two function terms ...
		    
		    Term funcTerm1 = tf.createFunctionTerm(currentFunction, qVarTerms);
		    Term funcTerm2 = tf.createFunctionTerm(newFunctionSymbol, qVarTerms);
		    
		    Term eqTerm = tf.createEqualityTerm(funcTerm1,funcTerm2);
		    
		    // Finally build all-closure (important: by using constraintTranslator, because
		    // of special quantors !!!)
		    
		    Term newTerm = cTranslator.createAllClosureOfTerm(eqTerm);
		    
		    // System.err.println("Adding new term to preTerm : " + newTerm);
		    
		    // add this new term ...
		    additionalPreTerms.add(newTerm);
		    
		    // Store current symbol in order to omit multiple handling of the same thing ...
		    alreadyHandledSymbols.add(currentFunction);
		    
		    
		}
	    }
	
	    
	}

	if (changed && (subStack.empty() || subStack.peek()!=newMarker)) {
	    subStack.push(newMarker);
	}
	
	
	ArrayOfQuantifiableVariable boundVars 
	    = ((visited.varsBoundHere(0).size() == 0) ?
	       visited.varsBoundHere(1) :
	       visited.varsBoundHere(0)); 
	
	
	Term[] neededsubs = neededSubs(visited.arity());
	if (changed || (!subStack.empty() && 
			subStack.peek() == newMarker)) {

	    //System.err.println("Create new term with : " );
	    //System.err.println("func symbol ==  : " + newFunctionSymbol );
	    //System.err.println("boundVars ==  : " + boundVars);

	    Operator newOp;
	    if (changed) {
		// Take the newly generated function symbol ...
		newOp = newFunctionSymbol;
	    } else {
		// Take the existing one ...
		newOp = visited.op();
	    }

	    subStack.push(tf.createTerm(newOp, 
					neededsubs,
					boundVars,
					visited.javaBlock(),
					visited.ruleVar()));
	} else {
	    subStack.push(visited);
	}		
	
    }

    
}
