// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.parser.jml;

import java.io.StringReader;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import de.uka.ilkd.key.java.Comment;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.java.annotation.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.java.visitor.JavaASTCollector;
import de.uka.ilkd.key.java.statement.LoopStatement;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.jml.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.op.ProgramMethod;
import de.uka.ilkd.key.proof.init.ProofInputException;

/** 
 * provides methods for parsing the specifications of a single class/interface.
 */
public class JMLSpecBuilder{
    
    private TypeDeclaration cd; 
    private Services services; 
    private NamespaceSet nss; 
    private TermFactory tf;
    private boolean parsedMethods = false; 
    private boolean parsedClass = false;
    private HashMap method2specs = new HashMap();
    private HashMap method2term = null;
    private JMLClassSpec classSpec = null;
    private ArithOpProvider aOP = null;
    //contains jml-comments in cd
    private Set jmlComments = null;
    //contains jml-comments without model method declarations.
    private Set jmlCommentsWithoutMM = null;
    //contains jml-comments without model method  and model field declarations.
    private Set jmlCommentsWithoutMFMM = null;

    public JMLSpecBuilder(TypeDeclaration cd, Services services, 
			  NamespaceSet nss, TermFactory tf){
	this.cd = cd;
	this.services = services;
	this.nss = nss;
	this.tf = tf;
	aOP = new DefaultArithOpProvider(nss.functions());
	Comment[] comments = cd.getComments();
	if(comments != null){
	    jmlComments = new HashSet();
	    for(int i=0; i<comments.length; i++){
		if(comments[i].containsJMLSpec()){
		    jmlComments.add(comments[i].getJMLSpec());
		}
	    }
	}
    }

    // obsolete: only used if one wants to build the proofobl. for a single
    // method via buildProofObligation(pm).
    public JMLSpecBuilder(Services services, NamespaceSet nss, TermFactory tf){
	this.services = services;
	this.nss = nss;
	this.tf = tf;
	parsedClass = parsedMethods = true;
	aOP = new DefaultArithOpProvider(nss.functions());
    }

    /**
     * returns a term containing the proofobligations for the methodspecs
     * of <code>pm</code> and the invariant and constraint proofobligation
     * for the containing class. 
     */
    public Term buildProofObligation(ProgramMethod pm){
	Implementation2SpecMap ism = services.getImplementation2SpecMap();
	SetOfJMLMethodSpec specs = ism.getSpecsForMethod(
	    pm.getMethodDeclaration());
	classSpec = ism.getSpecForClass(pm.getContainedInKeYJavaType());
	return makeConjunction(specs, pm.getMethodDeclaration());
    }

    /** 
     * parses jml model fields declared in <code>cd</code>. 
     * Can't be called before parseModelMethodDecls() has been called
     */
    public void parseModelFieldDecls() throws ProofInputException {
	if(jmlCommentsWithoutMFMM == null && jmlCommentsWithoutMM != null){
	    jmlCommentsWithoutMFMM = new HashSet();
	    Iterator it = jmlCommentsWithoutMM.iterator();
	    while(it.hasNext()){
		String s = (String) it.next();
		parseFieldDecl(s, jmlCommentsWithoutMFMM, classSpec);
	    }
	}else{
	    throw new ProofInputException(jmlCommentsWithoutMFMM == null?
					  "Tried to parse model field "+
					  "declarations before parsing "+
					  "model method declarations":
					  "already parsed model field "+
					  "declarations for class/interface "+
					  cd);
	}
    } 

    /** 
     * parses jml model methods declared in <code>cd</code>. 
     */
    public void parseModelMethodDecls() throws ProofInputException {
	if(jmlCommentsWithoutMM == null){
	    classSpec = new JMLClassSpec(services, cd, tf, nss);
	    jmlCommentsWithoutMM = new HashSet();
	    Iterator it = jmlComments.iterator();
	    while(it.hasNext()){
		String s = (String) it.next();
		parseMethodDecl(s, jmlCommentsWithoutMM, classSpec);
	    }
	}else{
	    throw new ProofInputException("already parsed model method "+
					  "declarations for class/interface "+
					  cd);
	}
    } 

    /**
     * parses jml specification like invariants, constraints or represents
     * in <code>cd</code>. parseModelMethodDecls() and parseModelFieldDecls()
     * have to be called first.
     */
    public void parseJMLClassSpec() throws ProofInputException {
	if(!parsedClass && jmlCommentsWithoutMFMM != null){
	    Iterator it = jmlCommentsWithoutMFMM.iterator();
	    while(it.hasNext()){
		String s = (String) it.next();
		StringReader stream = new StringReader(s);
		KeYJMLParser parser = 
		    new KeYJMLParser(new KeYJMLLexer(stream),
				     "no file, parsing comment from "+
				     cd.getName(),
				     services, nss,
				     TermFactory.DEFAULT, cd, 
				     null, classSpec, aOP);
		try{
		    parser.parseClassSpec();
		}catch(NotSupportedExpressionException e){
		    services.getExceptionHandler().reportException(e);
		} catch (antlr.ANTLRException e) {
		    throw new ProofInputException(e);
		}
	    }
	    parsedClass = true;
	}else{
	    if(jmlCommentsWithoutMFMM == null){
		throw new ProofInputException("tried to parse class spec "+
					      "without parsing model "+
					      "declarations for class "+
					      cd+" first");
	    }
	}
    }

    public NamespaceSet namespaces(){
	return nss;
    }

    private void parseMethodDecl (String s, Set otherSpecs, JMLClassSpec cs) 
	throws ProofInputException{
	String[] rest = 
	    s.split("(p\\w+ +)(static +)?(pure +)?(helper +)?(model +)"+
		    "(p\\w+ +)?(non_null +)?\\n?(static +)?(pure +)?(helper +)?"+
		    "\\n?\\w+\\s+\\w+\\s*\\n?@?\\s*\\("+
		    "\\n?(\\s*\\w+ \\w+,?\\s*\\n?@?\\s*)*\\)(throws +\\w+ +)?;"
		    , 2);
	if(rest.length == 2){
	    String methDecl = s.substring(rest[0].length(), 
					  s.length()-rest[1].length());
	    StringReader stream = new StringReader(methDecl);
	    String nonMethSpec = 
		rest[0].split("public +[a-z]+_behavior" , 2)[0];
	    String methSpec = 
		"/*@ "+rest[0].substring(nonMethSpec.length())+" @*/";
	    otherSpecs.add(nonMethSpec);
	    KeYJMLParser parser = 
		new KeYJMLParser(new KeYJMLLexer(stream),
				 "parsing comment from "+
				 cd.getName(),
				 services, nss,
				 TermFactory.DEFAULT, cd, 
				 null, cs, aOP);
	    try{
		parser.parseMethodDecl(methSpec);
	    }catch(NotSupportedExpressionException e){
		services.getExceptionHandler().reportException(e);
	    } catch (antlr.ANTLRException e) {
		throw new ProofInputException(e);
	    }
	    parseMethodDecl(rest[1], otherSpecs, cs);
	}else{
	    otherSpecs.add(s);
	}
    }

    private void parseFieldDecl (String s, Set otherSpecs, JMLClassSpec cs) 
	throws ProofInputException{
	
	int mIndex = s.indexOf("model");
	if(mIndex == -1) mIndex = s.indexOf("ghost");
	if(mIndex != -1){
	    int start = s.substring(0, mIndex+1).lastIndexOf("\n");
	    int end = s.indexOf(";", mIndex);
	    otherSpecs.add(s.substring(0, start+1));
	    StringReader stream = 
		new StringReader(s.substring(start+1, end+1));
	    KeYJMLParser parser = 
		new KeYJMLParser(new KeYJMLLexer(stream),
				 "parsing comment from "+
				 cd.getName(),
				 services, nss,
				 TermFactory.DEFAULT, cd, 
				 null, cs, aOP);
	    try{
		parser.parseFieldDecl();
	    }catch(NotSupportedExpressionException e){
		services.getExceptionHandler().reportException(e);
	    } catch (antlr.ANTLRException e) {
		throw new ProofInputException(e);
	    }	  
	    parseMethodDecl(s.substring(end+1), otherSpecs, cs);
	}else{
	    otherSpecs.add(s);
	}
    }

    /**
     * parses the specifications attached as comments to <code>m</code>.
     */
    protected SetOfJMLMethodSpec parseJMLMethodSpec(Method m) 
	throws ProofInputException {
	SetOfJMLMethodSpec result = SetAsListOfJMLMethodSpec.EMPTY_SET;
	if(!(m instanceof MethodDeclaration)){
	    return result;
	}
	MethodDeclaration md = (MethodDeclaration) m;
	Comment[] comments = md.getComments();
	if(comments != null){
	    for(int i=0; i<comments.length; i++){
		if(comments[i].containsJMLSpec() && 
		   !UsefulTools.isClassSpec(comments[i])){
		    int errorCount = 
			services.getExceptionHandler().getExceptions().size();
		    String s = comments[i].getJMLSpec();
		    StringReader stream = new StringReader(s);
		    KeYJMLParser parser = 
			new KeYJMLParser(new KeYJMLLexer(stream),
					 "no file, parsing comment from "+
					 md.getName()+" in class "+
					 cd.getName(),
					 services, nss,
					 TermFactory.DEFAULT, cd, 
					 md, classSpec, aOP);
		    try{
			parser.methodspecification();
		    }catch(NotSupportedExpressionException e){
			services.getExceptionHandler().reportException(e);
		    } catch (antlr.ANTLRException e) {
			throw new ProofInputException(e);
		    }
		    if(services.getExceptionHandler().getExceptions().size() >
		       errorCount){
			//if the specification for md contains unsupported
			// expressions we discard it.
			services.getImplementation2SpecMap().
			    clearSpecsForMethod(md);
		    }
		    result = result.union(parser.getSpecs());
		}
	    }
	}
	parseLoopInvariants(md);
	return result;
    }

    private void parseLoopInvariants(MethodDeclaration md)
	throws ProofInputException {
	JavaASTCollector coll = 
	    new JavaASTCollector(md, de.uka.ilkd.key.java.statement.
				 LoopStatement.class);
	coll.start();
	ListOfProgramElement l = coll.getNodes();
	IteratorOfProgramElement it = l.iterator();
	LoopStatement loop;
	while(it.hasNext()){
	    loop = (LoopStatement) it.next();
	    Comment[] comments = loop.getComments();
	    if(comments != null){
		for(int i=0; i<comments.length; i++){
		    if(comments[i].containsJMLSpec()){
			int errorCount = services.getExceptionHandler().
			    getExceptions().size();
			String s = comments[i].getJMLSpec();
			StringReader stream = new StringReader(s);
			KeYJMLParser parser = 
			    new KeYJMLParser(new KeYJMLLexer(stream),
					     "no file, parsing comment from loop"+
					     "in method "+md.getName()+" in class "+
					     cd.getName(),
					     services, nss,
					     TermFactory.DEFAULT, cd, 
					     md, classSpec, aOP);
			try{
			    parser.parseLoopInvariant(loop);
			}catch(NotSupportedExpressionException e){
			    services.getExceptionHandler().reportException(e);
			} catch (antlr.ANTLRException e) {
			    throw new ProofInputException(e);
			}
			if(services.getExceptionHandler().getExceptions()
			   .size() > errorCount){
			    //if the specification for loop contains 
			    // unsupported expressions we discard it.
			    services.getImplementation2SpecMap().
				clearSpecsForLoop(loop);
			}
		    }
		}
	    }
	}
    }

    /**
     * returns a mapping from methods to terms that are conjunctions
     * of the proofobligations generated from the methodspecs of these methods
     * obsolete (but for testing reasons still there).
     */
    public HashMap buildProofObligations() throws ProofInputException{
	if(!parsedMethods || !parsedClass){
	    parseSpecs();
	}
	if(method2term != null){
	    return method2term;
	}else{
	    method2term = new HashMap();
	    Iterator it = method2specs.keySet().iterator();
	    while(it.hasNext()){
		Object m = it.next();
		Term t = makeConjunction((SetOfJMLMethodSpec) 
					 method2specs.get(m), 
					 (MethodDeclaration) m);
		if(t!=null){
		    method2term.put(m, t);
		}
	    }
	}
	return method2term;
    }

    /**
     * returns a conjunction of the proofobligations for <code>md</code>
     * generated from <code>specs</code>.
     * obsolete (but for testing reasons still there).
     */
    private Term makeConjunction(SetOfJMLMethodSpec specs, 
				 MethodDeclaration md){
	Term result = null;
	Term a, inv, invPre;
	JMLMethodSpec spec;
	if(specs != SetAsListOfJMLMethodSpec.EMPTY_SET){
	    IteratorOfJMLMethodSpec it = specs.iterator();
	    spec = (JMLMethodSpec) it.next();
	    result = spec.createDLFormula(false, false);
	    inv = classSpec.getPreservesInvariantBehavior(md, false);
	    invPre = spec.getCompletePre(false, false);
	    while(it.hasNext()){
		spec = (JMLMethodSpec) it.next();
		a = spec.createDLFormula(false, false);
		invPre = tf.createJunctorTerm(
		    Op.OR,new Term[]{spec.getCompletePre(false, false), 
				     invPre});
		if(a != null){
		    if(result.op() != Op.TRUE){
			result = tf.createJunctorTerm(
			    Op.AND,new Term[]{result,a});
		    }else{
			result = a;
		    }
		}
	    }
	    if(inv != null){
		inv = tf.createJunctorTerm(Op.IMP,new Term[]{invPre, inv});
	    }
	    if(result != null && result.op() != Op.TRUE){
		if(inv != null){
		    result = 
			tf.createJunctorTerm(Op.AND,new Term[]{result, inv});
		}
	    }else{
		result = inv;
	    }
	// the method has no specification, but maybe it has to preserve
	// invariants, etc.
	// implicit methods need a special treatment.
	}else if(!md.getName().startsWith("<") && (
	       !md.isStatic() &&
	       (classSpec.getInstanceInvariants()!=null ||
		classSpec.getInstanceConstraints()!=null)
	       || classSpec.getStaticConstraints()!=null)
		 || classSpec.getStaticInvariants()!=null
		 && (!md.getName().startsWith("<") || 
		     md.getName().equals("<clinit>"))){
	    result = classSpec.getPreservesInvariantBehavior(md, false);
	}
	return result;
    }

    /**
     * returns a map from methods to the specs which have been parsed by this
     * JMLSpecBuilder.
     */
    public HashMap getMethod2Specs(){
	try{
	    parseSpecs();
	}catch(ProofInputException e){
	    e.printStackTrace();
	}
	return method2specs;
    }

    /**
     * returns a map with the model methods and their specs
     * parsed by this JMLSpecBuilder.
     */
    public HashMap getModelMethod2Specs(){
	try{
	    parseSpecs();
	}catch(ProofInputException e){
	    e.printStackTrace();
	}
	return classSpec.buildModelMethod2Specs();
    }

    /** 
     * Parses the specs of the methods declared in <code>cd</code>
     */
    public void parseJMLMethodSpecs() throws ProofInputException {
	if(parsedMethods) return;
	KeYJavaType kjt = services.getJavaInfo().getKeYJavaType(cd);
	ListOfMethod ms = 
	    services.getJavaInfo().getKeYProgModelInfo().getMethods(kjt); 
//	ListOfMethod ms = tds.getTypeDeclaration(j).
//      getAllMethods(initConfig.getServices());
	IteratorOfMethod it = ms.iterator();
	while(it.hasNext()){
	    MethodDeclaration m = (MethodDeclaration)it.next();
	    if(m != null){
		method2specs.put(m, parseJMLMethodSpec(m));
	    }
	}
	ms = services.getJavaInfo().getKeYProgModelInfo().getConstructors(kjt);
	it = ms.iterator();
	while(it.hasNext()){
	    MethodDeclaration m = (MethodDeclaration)it.next();
	    if(m != null){
		method2specs.put(m, parseJMLMethodSpec(m));
	    }
	}
	IteratorOfNamed mmit = 
	    classSpec.getModelMethods().allElements().iterator();
	while(mmit.hasNext()){
	    MethodDeclaration m = ((ProgramMethod) mmit.next()).
		getMethodDeclaration();
	    parseJMLMethodSpec(m);
	}
	parsedMethods = true;
    }

    private void parseSpecs() throws ProofInputException{
	parseJMLClassSpec();
	parseJMLMethodSpecs();
    }
}	

