// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

/* -*-antlr-*- */
header {

package de.uka.ilkd.key.parser;

import java.io.*;
import java.util.List;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

import antlr.*;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.sort.oclsort.OclSort;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.metaconstruct.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.java.JavaReader;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.Recoder2KeY;
import de.uka.ilkd.key.java.SchemaRecoder2KeY;
import de.uka.ilkd.key.java.SchemaJavaReader;
import de.uka.ilkd.key.java.visitor.DeclarationProgramVariableCollector;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.proof.mgt.*;
import de.uka.ilkd.key.proof.init.Includes;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.java.visitor.*;
import de.uka.ilkd.key.pp.AbbrevMap;
import de.uka.ilkd.key.java.recoderext.KeYCrossReferenceServiceConfiguration;
import de.uka.ilkd.key.util.*;
import de.uka.ilkd.key.logic.sort.ProgramSVSort;

}

/** 
 * A parser for first-order terms and formulae.
 */  

class ProblemParser extends TacletParser;
options {
    importVocab=KeYLexer;
    k = 1;
    defaultErrorHandler = true;
}


{
    
    
    static final Sort[] AN_ARRAY_OF_SORTS = new Sort[0];
    static final Term[] AN_ARRAY_OF_TERMS = new Term[0];
    protected AbbrevMap scm = new AbbrevMap();

    // these variables are set if a file is read in step by
    // step. This used when reading in LDTs because of cyclic
    // dependencies.
    protected boolean skip_functions;
    protected boolean skip_predicates;
    protected boolean skip_sorts;
    protected boolean skip_schemavariables;
    protected boolean skip_rulesets;
    protected boolean skip_taclets;
    protected boolean onlyWith=false;
    protected boolean parse_includes = false;
    protected Includes includes = new Includes();
    protected UseInsteadMap useInstead=new UseInsteadMap(); 

    // if this is used then we can capture parts of the input for later use
    protected DeclPicker capturer = null;

    protected KeYExceptionHandler keh = null;

    protected int lineOffset=0;
    protected int colOffset=0;

    protected ProgramMethod pm = null;

    protected SetOfTaclet taclets = SetAsListOfTaclet.EMPTY_SET; 
    protected HashMap taclet2Builder;   
    protected Choice defaultChoice;
    protected SetOfChoice activatedChoices = SetAsListOfChoice.EMPTY_SET;
    protected HashSet usedChoiceCategories = new HashSet();
    protected HashMap category2Default = new HashMap();

    private static ListOfRulevariable NO_RULE_VARIABLES =
        SLListOfRulevariable.EMPTY_LIST;
    private Term a=null;

    protected TermFactory tf = TermFactory.DEFAULT;
    boolean schemaMode; 

    public boolean errors = false;
    
    private ParserConfig schemaConfig;
    private ParserConfig normalConfig;
    
    // the current active config
    private ParserConfig parserConfig;

    private ContractSet contracts = new ContractSet();

    public ProblemParser(TokenStream lexer, 
                         String filename,
                         ParserConfig schemaConfig,
                         ParserConfig normalConfig,
                         HashMap taclet2Builder,
                         SetOfTaclet taclets) { 
        this(lexer);
        setFilename(filename);
        if (lexer instanceof DeclPicker) {
            this.capturer = (DeclPicker) lexer;
        }
        if (normalConfig!=null) defaultChoice = normalConfig.getDefaultChoice(); 
        this.schemaConfig = schemaConfig;
        this.normalConfig = normalConfig;       
        switchToNormalMode();
        this.taclet2Builder=taclet2Builder;
        this.taclets = taclets;
        if(normalConfig != null){
            this.keh = normalConfig.services().getExceptionHandler();
        }else{
            this.keh = new KeYRecoderExcHandler();
        }
    }

    protected void setSkips(boolean svars, boolean funcs, 
                boolean preds, boolean sorts,
                boolean rsets, boolean tacs) {

       skip_schemavariables = svars;
       skip_functions       = funcs;
       skip_predicates      = preds;
       skip_sorts           = sorts;
       skip_rulesets        = rsets;
       skip_taclets         = tacs;
    }

    public void reportError(RecognitionException ex){
        keh.reportException(ex);
    }

    protected void skipFuncs() {
        skip_functions = true;
    }
    
    protected KeYJavaType getTypeByClassName(String s){
        KeYJavaType kjt = null;
        try {
            kjt = getJavaInfo().getTypeByClassName(s);
        } catch(RuntimeException e){
            return null;
        } 
        return kjt;
    }

    protected boolean isPackage(String name){
        try {
            return getJavaInfo().isPackage(name);
        } catch(RuntimeException e){        
            // may be thrown in cases of invalid java identifiers
            return false;
        } 
    }

    protected void skipPreds() {
        skip_predicates = true;
    }

    protected void skipTaclets() {
        skip_taclets = true;
    }

    protected void skipVars() {
        skip_schemavariables = true;
    }

    protected void skipSorts() {
        skip_sorts = true;
    }

    protected void skipRuleSets() {
        skip_rulesets = true;
    }
    
    /** returns the java information object */
    public JavaInfo getJavaInfo() {
        return parserConfig.javaInfo();
    }


    /** parses a problem but without reading the declarations of
     * sorts, functions and predicates. These have to be given
     * explicitly.
     * @param functions() the Namespace with the function and predicate
     * symbols to be used
     * @param sorts() the Namespace with the sorts to be used
     * @param ruleSets() the Namespace with the rule sets to be used
     * the rule sets of the current problem file will be added 
     */ 
    public Term parseTacletsAndProblem() 
    throws antlr.RecognitionException, antlr.TokenStreamException{
        // restore known state
        setSkips(false, false, false, false, false, false);
        // skip decls
        skipSorts(); 
        skipFuncs();
        skipPreds();    
        
        return problem();
    }

    /**
     * returns the ProgramMethod parsed in the jml_specifications section.
     */
    public ProgramMethod getProgramMethod(){
        return pm;
    }

    /** parses a problem but without reading the declarations of
     * sorts, functions and predicates. These have to be given
     * explicitly.
     * @param functions() the Namespace with the function and predicate
     * symbols to be used
     * @param sorts() the Namespace with the sorts to be used
     * @param heuristics() the Namespace with the heuristics to be used
     * the heuristics of the current problem file will be added 
     */ 
    public Term parseProblem() 
        throws antlr.RecognitionException, antlr.TokenStreamException
        
        {
            // restore known state
            setSkips(false, false, false, false, false, false);
            // skip decls
            skipSorts(); 
            skipFuncs();
            skipPreds();    
            skipTaclets();
            return problem();
        }

    public void parseIncludes()
        throws antlr.RecognitionException, antlr.TokenStreamException
        {
            parse_includes=true;
            problem();
        }

    public void parseWith()
        throws antlr.RecognitionException, antlr.TokenStreamException
        {
            onlyWith=true;
            problem();
        }

    public void addSort(Sort s) {
        sorts().add(s);
    }

    public void addSorts(Namespace ns) {
        IteratorOfNamed it=ns.elements().iterator();
        while (it.hasNext()) {
            addSort((Sort)it.next());
        }
    }

    public void addFunction(Function f) {
        functions().add(f);
    }


    public void addFunctions(Namespace ns) {
        IteratorOfNamed it=ns.elements().iterator();
        while (it.hasNext()) {
            addFunction((Function)it.next());
        }
    }

    public Services getServices() {
        return parserConfig.services();
    }

    public NamespaceSet namespaces() {
        return parserConfig.namespaces();
    }

    public Namespace sorts() {
        return namespaces().sorts();
    }

    public Namespace variables() {
        return namespaces().variables();
    }

    public Namespace functions() {
        return namespaces().functions();
    }

    public Namespace ruleSets() {
        return namespaces().ruleSets();
    }

    public Namespace programVariables() {
        return namespaces().programVariables();
    }

    public Namespace choices(){
        return namespaces().choices();
    }

    public Named lookup(Name n) {
        final Namespace[] lookups = {
            normalConfig.namespaces().functions(), 
            schemaConfig.namespaces().functions(), 
            normalConfig.namespaces().variables(), 
            schemaConfig.namespaces().variables(), 
            schemaConfig.namespaces().programVariables()
        };

        for (int i = 0; i<lookups.length; i++) {
            if (lookups[i].lookup(n) != null) {
                return lookups[i].lookup(n);
            }
        }
        return null;    
    }

    public SetOfTaclet getTaclets(){
        return taclets;
    }

    public ContractSet getContracts(){
        return contracts;
    }

    public HashMap getCategory2Default(){
        return category2Default;
    }

    public void setOnlyWith(boolean onlyWith){
        this.onlyWith = onlyWith;
        parse_includes = false;
    }

    public SetOfChoice getActivatedChoices(){
        return activatedChoices;
    }


    private void switchToNormalMode() {
        schemaMode   = false;
        parserConfig = normalConfig;
    }

    private boolean inSchemaMode() {
        return schemaMode;
    }

    private void switchToSchemaMode() {
        schemaMode   = true;
        parserConfig = schemaConfig;    
    }

    protected int getLine() {
        int line = -1;
        try {
            line = LT(0).getLine()+lineOffset;
        } 
        catch (TokenStreamException e) {
            System.err.println("No further token in stream");
        }
        return line;
    }   

    protected int getColumn() {
        int     col = -1;
        try {
            col = LT(0).getColumn()+colOffset;
        } 
        catch (TokenStreamException e) {
            System.err.println("No further token in stream");
        }
        return col;
    }       

    protected TermSymbol progVar(Token id,Sort s) throws NotDeclException{
        TermSymbol v = (TermSymbol)
            programVariables().lookup(new Name(id.getText()));

        if (v == null) {
            throw new NotDeclException("program variable", id, getFilename());
        }
        return v;
    }

    protected static HashSet progVars(JavaBlock jb) {
        if (jb == JavaBlock.EMPTY_JAVABLOCK) {
            return new HashSet();
        }   
        DeclarationProgramVariableCollector pvc
            = new DeclarationProgramVariableCollector(jb.program());
        pvc.start();
        return pvc.result();
    }


    protected void bindProgVars(HashSet progVars) {
        namespaces().setProgramVariables(new Namespace(programVariables()));
        Iterator it=progVars.iterator();
        while (it.hasNext()) {
            programVariables().add((IProgramVariable)it.next());
        }
    }

    protected void unbindProgVars() {
        namespaces().setProgramVariables(programVariables().parent());
    }

    public Choice getDefaultChoice(){
        return defaultChoice;
    }

    public Includes getIncludes(){
        return includes;
    }

    protected void addInclude(String filename, boolean relativePath, boolean ldt){
        InputStream input=null;
        RuleSource source=null;
        if(relativePath){
            int end = getFilename().lastIndexOf(File.separator);
            int start = 0;
            filename = filename.replace('/', File.separatorChar);
            filename = filename.replace('\\', File.separatorChar);
            if(getFilename().startsWith("file:")){
                start = 5;
            }
            File path=new File(getFilename().substring(start,end+1)+filename);
            try{ 
                source = RuleSource.initRuleFile(path.toURL()); 
            }catch(java.net.MalformedURLException e){
                System.err.println("Exception due to malformed URL of file "+
                                   filename+"\n " +e);
            }
        }else{
            source = RuleSource.initRuleFile(filename+".key"); 
        }
        if(ldt){
            includes.putLDT(filename, source);
        }else{
            includes.put(filename, source);
        }
    }
    
    protected LogicVariable bindVar(Token id,Sort s) {
        LogicVariable v=new LogicVariable(new Name(id.getText()), s);
        namespaces().setVariables(variables().extended(v));
        return v;
    }
    
    protected void unbindVars() {
        namespaces().setVariables(variables().parent());
    }

    public Term createAttributeTerm(Term prefix, 
                                    TermSymbol attribute,
                                    int shadowPrimes) {
        Term result = prefix;
        if (!inSchemaMode()) {
            if (((ProgramVariable)attribute).isStatic()){
                result = tf.createVariableTerm((ProgramVariable)attribute);
            } else {
                if (shadowPrimes > 0) {
                    Term shadownum = TermParser.toZNotation("" + shadowPrimes, functions(), tf);
                    result = tf.createShadowAttributeTerm((ProgramVariable)attribute, 
                                                          result, shadownum);
                } else {
                    result = tf.createAttributeTerm((ProgramVariable)attribute, result);
                }
            }
        } else {
            if(shadowPrimes>0){
                Term shadownum = TermParser.toZNotation("" + shadowPrimes, functions(), tf);
                result = tf.createShadowAttributeTerm((SchemaVariable)attribute, result, shadownum);         
            } else
                result = tf.createAttributeTerm((SchemaVariable)attribute, result);         
        }

        return result;
    }

    public TermSymbol getAttribute(Sort prefixSort, 
                                   String attributeName)
        throws antlr.SemanticException  {
        final JavaInfo javaInfo = getJavaInfo();
        TermSymbol v = null;
        if (!inSchemaMode()) {
            if (attributeName.indexOf(':') != -1) {
                v = javaInfo.getAttribute
		    (attributeName, javaInfo.getKeYJavaType
		     (attributeName.substring(0, attributeName.indexOf(':'))));
            } else {
                v = javaInfo.getAttribute(attributeName,(ObjectSort)prefixSort);
            }
        } else {
            v = (SortedSchemaVariable)variables().lookup(new Name(attributeName));
        }
        if ( v == null && !"length".equals(attributeName) ) {
            throw new NotDeclException
                ("Attribute ", attributeName, getFilename(), 
                 getLine(), getColumn());

        }

        return v;
    }

    public de.uka.ilkd.key.logic.op.Location[] extractLocations(List locNames)
        throws antlr.SemanticException {
        de.uka.ilkd.key.logic.op.Location[] vars = 
            new de.uka.ilkd.key.logic.op.Location[locNames.size()];
        for (int i = 0; i<vars.length; i++) {
            final String locName = (String)locNames.get(i);
	    vars[i] = (SortedSchemaVariable)
		schemaConfig.namespaces().variables().
		    lookup(new Name(locName));
	    if (vars[i] == null) {
		Object o = programVariables().lookup(new Name(locName));
		if (o != null) {
		    vars[i] = (de.uka.ilkd.key.logic.op.Location) o;
		} else {
		    vars[i] = (de.uka.ilkd.key.logic.op.Location) 
			getAttribute(null, locName);
		}
	    }	    
	}
        return vars;
    }

    protected Term termForParsedVariable(ParsableVariable v) 
    throws antlr.SemanticException 
    {
        if ( v instanceof LogicVariable ) {
        	return tf.createVariableTerm((LogicVariable)v);
        } else if ( inSchemaMode() && v instanceof SchemaVariable ) {
        	return tf.createVariableTerm((SchemaVariable)v);
        } else if ( v instanceof ProgramVariable ) {
        	return tf.createVariableTerm((ProgramVariable)v);
        }
	String errorMessage = "";
        if ( inSchemaMode() ) {
       	  errorMessage += v +" is not a program, logic or schema variable";
        } else {
          errorMessage += v +" is not a logic or program variable";
        }

	throw new KeYSemanticException
		(errorMessage, getFilename(), getLine(), getColumn());
    }



    private TacletBuilder createTacletBuilderFor
        (Object find, boolean sameUpdatePrefix) 
        throws InvalidFindException {
        if ( sameUpdatePrefix && !( find instanceof Term ) )
            throw new InvalidFindException
                ("\"sameUpdateLevel\" may only be used for rewrite taclets:"+find,
                 getFilename(), getLine(), getColumn());
        if ( find == null ) {
            return new NoFindTacletBuilder();
        } else if ( find instanceof Term ) {
            return new RewriteTacletBuilder().setFind((Term)find)
                .setSameUpdatePrefix(sameUpdatePrefix);
        } else if ( find instanceof Sequent ) {
            Sequent findSeq = (Sequent) find;
            if ( findSeq == Sequent.EMPTY_SEQUENT ) {
                return new NoFindTacletBuilder();
            } else if (findSeq.antecedent().size() == 1
                       && findSeq.succedent().size() == 0 ) {
                return new AntecTacletBuilder().
                    setFind(findSeq.antecedent().get(0).formula());
            } else if (   findSeq.antecedent().size() == 0
                          && findSeq.succedent().size() == 1 ) {
                return new SuccTacletBuilder().
                    setFind(findSeq.succedent().get(0).formula());
            } else {
                throw new InvalidFindException
                    ("Unknown find-sequent (perhaps null?):"+findSeq,
                     getFilename(), getLine(), getColumn());    
            }
        } else {
            throw new InvalidFindException
                    ("Unknown find class type: " + find.getClass().getName(),
                     getFilename(), getLine(), getColumn());
        }
    }       



    private void addGoalTemplate(TacletBuilder b,
                                 String id,
                                 Object rwObj,
                                 Sequent addSeq,
                                 ListOfTaclet addRList,
                                 SetOfSchemaVariable pvs,
                                 SetOfChoice soc) 
        throws SemanticException
        {
            TacletGoalTemplate gt = null;
            if ( rwObj == null ) {
                // there is no replacewith, so we take
                gt = new TacletGoalTemplate(addSeq,
                                            addRList,
                                            NO_RULE_VARIABLES,
                                            pvs);
            } else {
                if ( b instanceof NoFindTacletBuilder ) {
                    // there is a replacewith without a find.
                    throw
                        new UnfittingReplacewithException
                        ("Replacewith without find", getFilename(),
                         getLine(), getColumn());
                } else if ( b instanceof SuccTacletBuilder
                            || b instanceof AntecTacletBuilder ) {
                    if ( rwObj instanceof Sequent ) {
                        gt = new AntecSuccTacletGoalTemplate(addSeq,
                                                             addRList,
                                                             NO_RULE_VARIABLES,
                                                             (Sequent)rwObj, pvs); 
                    } else {
                        throw new UnfittingReplacewithException
                            ("Replacewith in a Antec-or SuccTaclet has "+
                             "to contain a sequent (not a term)", 
                             getFilename(), getLine(), getColumn());
                    }
                } else if ( b instanceof RewriteTacletBuilder ) {
                    if ( rwObj instanceof Term ) {
                        gt = new RewriteTacletGoalTemplate(addSeq,
                                                           addRList,
                                                           NO_RULE_VARIABLES,
                                                           (Term)rwObj, pvs);  
                    } else {
                        throw new UnfittingReplacewithException
                            ("Replacewith in a RewriteTaclet has "+
                             "to contain a term (not a sequent)", 
                             getFilename(), getLine(), getColumn());
                    }
                }
            }
            gt.setName(id);
            b.addTacletGoalTemplate(gt);
            if(soc != null) b.addGoal2ChoicesMapping(gt,soc);
        }

    protected JavaBlock getJavaBlock(Token t) throws antlr.SemanticException {
        String s=t.getText();
        de.uka.ilkd.key.java.JavaReader jr=null;
        try {
            if (inSchemaMode()) {
                jr = new SchemaRecoder2KeY(parserConfig.services(), 
                    parserConfig.namespaces());
                ((SchemaJavaReader)jr).setSVNamespace(variables());
                return jr.readBlockWithEmptyContext(s);
            } else {    
                jr = new Recoder2KeY(parserConfig.services(), 
                    parserConfig.namespaces());
                return jr.readBlockWithProgramVariables(programVariables(), s);    
            }
        } catch (de.uka.ilkd.key.java.PosConvertException e) {
            lineOffset = e.getLine()-1;
            colOffset  = e.getColumn();
            throw new JavaParserException
		(e.getMessage(), t, getFilename(), lineOffset, colOffset);                        
        } catch (de.uka.ilkd.key.java.ConvertException e) {
            if (s.length()>10) {
                s=s.substring(0,10)+"...";
            }
            if (e.parseException()!=null
            &&  e.parseException().currentToken != null
            &&  e.parseException().currentToken.next != null) {       
                lineOffset=e.parseException().currentToken.next.beginLine;
                colOffset=e.parseException().currentToken.next.beginColumn;
                e.parseException().currentToken.next.beginLine=getLine()-1;
                e.parseException().currentToken.next.beginColumn=getColumn();
                throw new JavaParserException(e.parseException().getMessage(), t, 
                    getFilename(), -1, -1);  // row/columns already in text
            } 
            if (e.proofJavaException()!=null
            &&  e.proofJavaException().currentToken != null
            &&  e.proofJavaException().currentToken.next != null) {       
                lineOffset=e.proofJavaException().currentToken.next.beginLine-1;
                colOffset=e.proofJavaException().currentToken.next.beginColumn;
                e.proofJavaException().currentToken.next.beginLine=getLine();
                e.proofJavaException().currentToken.next.beginColumn=getColumn();
                throw new JavaParserException(e.proofJavaException().getMessage(), 
                    t, getFilename(), lineOffset, colOffset); 
            }   
            throw new JavaParserException(e.getMessage(), t, getFilename());
        }
    }


    
    public void schema_var_decl(String name,Sort s,
                                boolean makeVariableSV,
                                boolean makeDependingSV,
                                boolean rigid,
                                boolean listModifier)
        {
            if (!skip_schemavariables) {
                SchemaVariable v;
                if ( s == Sort.FORMULA ) {
                    v = SchemaVariableFactory.createFormulaSV
                        (new Name(name), listModifier, rigid);
                } else if ( s instanceof ProgramSVSort ) {
                    v = SchemaVariableFactory.createProgramSV
                        (new ProgramElementName(name),(ProgramSVSort) s, listModifier);
                } else {
                    if ( makeVariableSV ) {
                        v=SchemaVariableFactory.createVariableSV
                            (new Name(name), s, listModifier);
                    } else if ( makeDependingSV ) {
                        v = SchemaVariableFactory.createDependingSV
                            (new Name(name), s, listModifier);
                    } else { v = SchemaVariableFactory.createTermSV
                            (new Name(name), s, listModifier, rigid);
                    }
                }
                if (inSchemaMode()) {
                    variables().add(v);
                }
            }

        }
    


    public void testLiteral(String l1, String l2)
        throws KeYSemanticException
        {
            if (!l1.equals(l2)){
                throw new KeYSemanticException
                    ("Expecting '"+l1+"', found '"+l2+"'.",getFilename(),getLine(),getColumn());
            };
        }


    protected Operator lookupVarfuncId(String varfunc_name, 
                    PairOfTermArrayAndBoundVarsArray args) throws NotDeclException {
        Term[] argTerms = null;
        if (args != null) {
            argTerms = args.getTerms();
        }
        return lookupVarfuncId(varfunc_name, argTerms);
    }

    /** looks up a function, (program) variable or static query of the 
     * given name varfunc_id and the argument terms args in the namespaces 
     * and java info. 
     * @param args is null iff no argument list is given, for instance `f', 
     * and is an array of size zero, if an empty argument list was given,
     * for instance `f()'.
     */
    protected Operator lookupVarfuncId(String varfunc_name, Term[] args) 
        throws NotDeclException{
        // case 1: variable
        Operator v = (TermSymbol) variables().lookup(new Name(varfunc_name));
        if (v != null && args==null) {
            return v;
        }
        // case 2: function
        v = (TermSymbol) functions().lookup(new Name(varfunc_name));
        if (v != null) { // we allow both args==null (e.g. `c')
                         // and args.length=0 (e.g. 'c())' here 
            return v;
        }
        // case 3: static query
        if (varfunc_name.lastIndexOf("::")>0 && args!=null) {
            String classname = varfunc_name.substring
                (0, varfunc_name.lastIndexOf("::"));
            varfunc_name = varfunc_name.substring
                (varfunc_name.lastIndexOf("::")+2);
            v = getServices().getJavaInfo().getProgramMethodTerm
                (null, varfunc_name, args, classname).op();
            if (v != null) {
                return v;
            }
        }
        // case 4: program variable
        v = (ProgramVariable) programVariables().lookup
            (new ProgramElementName(varfunc_name));
        if (v != null && args==null) {
            return v;
        }
        String[] argTypeStrings = null;
        if (args!=null) {
            argTypeStrings = new String[args.length];
            for (int i=0; i<args.length; i++) {
                argTypeStrings[i] = args[i].sort().toString();
            }
        }
        if (args==null) {
            throw new NotDeclException
                ("(program) variable or constant", varfunc_name,
                 getFilename(), getLine(), getColumn());
        } else {
            throw new NotDeclException
                ("function or static query", varfunc_name,
                 getFilename(), getLine(), getColumn());
        }
    }
    
}

top : {} a=problem;

problem returns [ Term a = null ]
{
    Taclet s = null;
    SetOfChoice choices=SetAsListOfChoice.EMPTY_SET;
    Choice c = null;
    String ctext = null;
    String st = null;
    String[] varDecls = null;
    Namespace funcNSForActivatedChoices = new Namespace();
    Term fma = null;
    Term modifTerm = null;
    SetOfTerm termSet = null;
    String contractName = null;
    String pref = null;
}
    :
	{ if (capturer != null) capturer.mark(); }
        (pref = preferences)
        { if ((pref!=null) && (capturer != null)) capturer.mark(); }
        
        ( st = java )
        
        decls
        { 
            if(parse_includes || onlyWith) return null;
            switchToNormalMode(); 
            variables().add(parserConfig.namespaces().programVariables()); 
            IteratorOfChoice it = activatedChoices.iterator(); 
            while(it.hasNext()){
                 funcNSForActivatedChoices=(funcNSForActivatedChoices.
                                            extended(it.next().funcNS().allElements()));
             } 
             funcNSForActivatedChoices=funcNSForActivatedChoices.
                 extended(defaultChoice.
                          funcNS().allElements()); 
        }
        ( "contracts"
            LBRACE {
                switchToNormalMode();
                 IteratorOfChoice it = choices.iterator();
                 Namespace funcNSForRules = funcNSForActivatedChoices;
                 while(it.hasNext()){
                     c=it.next();
                     funcNSForRules = 
                         funcNSForRules.extended(c.funcNS().allElements());
                 }
                 namespaces().setFunctions(funcNSForRules); 
            }
            (   
                fma = formula ("modifies" {termSet = SetAsListOfTerm.EMPTY_SET;}
                    (modifTerm=accessterm 
                        {termSet=termSet.add(modifTerm);}
                    (COMMA modifTerm=accessterm 
                        {termSet=termSet.add(modifTerm);})*)? )?
                {contractName = null;}
                (DISPLAYNAME name:STRING_LITERAL {contractName = name.getText();})?
                { //Syntax check fma
                    contracts.add(new DLHoareTripleContract(fma, termSet,
                        ruleSets(), contractName, getServices()));
                } SEMI )*
            RBRACE 
        )*
        (  "rules" 
            (LPAREN choices=option[choices]
                (COMMA choices=option[choices])* RPAREN )? LBRACE
            { 
                switchToSchemaMode(); 
                 IteratorOfChoice it = choices.iterator();
                 Namespace funcNSForRules = funcNSForActivatedChoices;
                 while(it.hasNext()){
                     c=it.next();
                     funcNSForRules = 
                         funcNSForRules.extended(c.funcNS().allElements());
                 }
                 namespaces().setFunctions(funcNSForRules); 
            }
            ( 
                s = taclet[choices] SEMI
                {
                    try {
                        if (!skip_taclets) {
                            taclets = taclets.addUnique(s);
                        }
                    } catch(de.uka.ilkd.key.collection.NotUniqueException e) {
                        throw new KeYSemanticException
				("Cannot add taclet \"" + s.name() + 
				 "\" to rule base as a with the same "+
				 "name already exists.", 
				  getFilename(), getLine(), getColumn());
                    }
                }
            )*
            RBRACE {choices=SetAsListOfChoice.EMPTY_SET;}
        ) *
        (PROBLEM LBRACE 
            {switchToNormalMode(); if (capturer != null) capturer.capture();}
//            (   
                a = formula 
//            |
//                jml_spec
//            )
            RBRACE)?
   ;

/*jml_spec
{
    ListOfType sig = SLListOfType.EMPTY_LIST; 
    String type = "";
    KeYJavaType classType = null;
}:
      JML_SPECIFICATIONS 
        LBRACE 
        id:IDENT {type = id.getText();}
        (DOT id2:IDENT {type += "."+id2.getText();})+
        {
            classType = getJavaInfo().getTypeByName(type.substring(0,
                    type.lastIndexOf(".")));
        }
        LPAREN 
        (id3:IDENT 
            {
                sig = sig.append(getJavaInfo().getTypeByName(id3.getText()));
            }
        )? 
        (COMMA id4:IDENT 
            {
                sig = sig.append(getJavaInfo().getTypeByName(id4.getText()));
            }
        )*  
        RPAREN 
        {
            pm = getJavaInfo().getProgramMethod(classType, id2.getText(), sig,
                                                classType);
        } 
        RBRACE
    ;
*/

java returns [String s = null]
{ StringBuffer b=new StringBuffer(); } :
// the following contortion is done to remove the "java" keyword from the lexer
        ((javaId:IDENT{testLiteral("java",javaId.getText());} LBRACE (l:STRING_LITERAL {b.append(l.getText());} 
                | SLASH {b.append("/");})+
                {s=b.toString(); } RBRACE))?
    ;

preferences returns [String s = null]:
	( KEYSETTINGS LBRACE
		(str:STRING_LITERAL {s = str.getText();})?
		RBRACE )?
	;

proof [ProblemLoader prl] :
        ( PROOF LBRACE
            ( pseudosexpr[prl] )+ 
            RBRACE
        )?
    ;

pseudosexpr [ProblemLoader prl] { char eid='0'; } :
        LPAREN (eid=expreid
            (str:STRING_LITERAL { prl.beginExpr(eid,str.getText()); } )? 
            ( pseudosexpr[prl] )* ) ?
        { prl.endExpr(eid); }
        RPAREN   
    ;

expreid returns [ char eid = '0' ]:
   FORMULA { eid='f'; }
   |
   TERM    { eid='t'; }
   |
   RULE    { eid= 'r'; }
   |
   BUILTIN { eid='n'; }
   |
   BRANCH  { eid='b'; }
   |
   INST    { eid='i'; }
   |
   IFSEQFORMULA { eid='q'; }
   |
   INTERACTIVE  { eid='a'; }
   |
   HEUR     { eid='h'; }
   |
   IDENT
   |
   KEYLOG { eid='l'; }
   |
   KEYLOGUSER { eid='u'; }
   |
   KEYLOGVERSION {eid='v'; }
   |
   KEYLOGSETTINGS {eid='s'; }
;
