// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

/* -*-antlr-*- */
header {

package de.uka.ilkd.key.parser;

import java.io.*;
import antlr.*;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.*; 
import de.uka.ilkd.key.logic.sort.oclsort.OclSort;   
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.metaconstruct.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.proof.RuleSource;
import de.uka.ilkd.key.proof.init.Includes;

import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.SchemaJavaReader;
import de.uka.ilkd.key.java.SchemaRecoder2KeY;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.abstraction.ListOfKeYJavaType;
import de.uka.ilkd.key.java.abstraction.SLListOfKeYJavaType;
import de.uka.ilkd.key.java.visitor.DeclarationProgramVariableCollector;
import de.uka.ilkd.key.util.*;
import java.util.List;
import java.util.LinkedList;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import de.uka.ilkd.key.pp.AbbrevMap;
}

/**
 * A parser for Taclets, based on the term/formula grammar terms.g
 */  

class TacletParser extends TermParser;
options {
    importVocab=KeYLexer;
    k = 1;
    defaultErrorHandler=true;
}

{
            
    static final Sort[] AN_ARRAY_OF_SORTS = new Sort[0];
    static final Term[] AN_ARRAY_OF_TERMS = new Term[0];

    protected HashMap category2Default = new HashMap();
    protected HashMap taclet2Builder;
    protected boolean parse_includes = false;
    protected Includes includes = new Includes();
    protected boolean onlyWith=false;
    protected SetOfChoice activatedChoices = SetAsListOfChoice.EMPTY_SET;
    protected HashSet usedChoiceCategories = new HashSet();

    private NamespaceSet nss;
    private AbbrevMap scm = new AbbrevMap();

    protected KeYExceptionHandler keh = null;

    // these variables are set if a file is read in step by
    // step. This used when reading in LDTs because of cyclic
    // dependencies.
    private boolean skip_functions;
    private boolean skip_predicates;
    private boolean skip_sorts;
    private boolean skip_schemavariables;
    private boolean skip_rulesets;
    private Choice defaultChoice;

    private TermFactory tf;
    private SchemaJavaReader javaReader;
    private Services services;

    private boolean errors = false;

    public TacletParser(TokenStream lexer,
                        String filename,
                        TermFactory tf,
                        SchemaJavaReader jr,
                        Services services,  
                        NamespaceSet nss,
                        Choice defaultChoice,
                        HashMap taclet2Builder) {
        this(lexer,2);
        setFilename(filename);
        this.tf = tf;
        this.javaReader = jr;
        this.services = services;
        this.nss  = nss;    
        switchToSchemaMode();
        this.defaultChoice = defaultChoice;
        this.taclet2Builder = taclet2Builder;
        this.keh = services.getExceptionHandler();
    }

    public TacletParser(TokenStream lexer,
                        String filename,
                        TermFactory tf,
                        Services services,
                        NamespaceSet nss,
                        Choice defaultChoice) {
        this(lexer, filename, tf,
             new SchemaRecoder2KeY(services, nss), services, nss, defaultChoice, new HashMap());
    }

    public TacletParser(TokenStream lexer,
                        String filename,
                        TermFactory tf,
                        Services services,
                        NamespaceSet nss) {
        this(lexer, filename, tf,
             new SchemaRecoder2KeY(services, nss), services, nss, 
             new Choice(new Name("Default")), new HashMap());
    }

    public void reportError(RecognitionException ex){
        keh.reportException(ex);
    }

    public Namespace sorts() {
        return nss.sorts();
    }

    public Namespace choices() {
        return nss.choices();
    }

    public Namespace functions() {
        return nss.functions();
    }

    public Namespace ruleSets() {
        return nss.ruleSets();
    }

    public Namespace variables() {
        return nss.variables();
    }

    private Namespace programVariables() {
        return nss.programVariables();
    }

    public NamespaceSet namespaces() {
        return nss;
    }

    public Named lookup(Name n) {
	final Namespace[] lookups = {
		functions(), variables(), 
		programVariables()
	};
	for (int i = 0; i<lookups.length; i++) {
	    if (lookups[i].lookup(n) != null) {
		return lookups[i].lookup(n);
            }
	}
	return null;	
    }

    protected boolean isPackage(String name){
        try {
            return getJavaInfo().isPackage(name);
        } catch(RuntimeException e){	    
	    // may be thrown in case of invalid an java identifier
            return false;
        } 
    }

    public Term createAttributeTerm(Term prefix, 
				    TermSymbol attribute,
				    int shadowPrimes) {
	Term result = prefix;
	if (!inSchemaMode()) {
	    if (((ProgramVariable)attribute).isStatic()){
		result = tf.createVariableTerm((ProgramVariable)attribute);
	    } else {
		if (shadowPrimes > 0) {
		    Term shadownum = TermParser.toZNotation("" + shadowPrimes, functions(), tf);
		    result = tf.createShadowAttributeTerm((ProgramVariable)attribute, 
                                                           	   result, shadownum);
		} else {
		    result = tf.createAttributeTerm((ProgramVariable)attribute, result);
		}
	    }
	} else {
	    if(shadowPrimes>0){
		Term shadownum = TermParser.toZNotation("" + shadowPrimes, functions(), tf);
                result = tf.createShadowAttributeTerm((SchemaVariable)attribute, result, shadownum);         
	    } else
                result = tf.createAttributeTerm((SchemaVariable)attribute, result);         
	}

	return result;
    }

    public TermSymbol getAttribute(Sort prefixSort, 
				   String attributeName) 
	throws antlr.SemanticException 	{

	    final JavaInfo javaInfo = getJavaInfo();
	    TermSymbol v = null;
            if (!inSchemaMode()) {
		if (attributeName.indexOf(':') != -1) {
		   v = javaInfo.getAttribute(attributeName, 
			javaInfo.getKeYJavaType(attributeName.substring(0, 
						attributeName.indexOf(':'))));
	        } else {
		    v = javaInfo.getAttribute(attributeName, (ObjectSort) prefixSort);
                }
            } else {
                v = (SortedSchemaVariable)variables().lookup(new Name(attributeName));
            }
            if ( v == null && !"length".equals(attributeName) ) {
                throw new NotDeclException
                    ("Attribute ", attributeName, getFilename(), 
			getLine(), getColumn());
	}

	return v;
    }

    public de.uka.ilkd.key.logic.op.Location[] extractLocations(List locNames)
        throws antlr.SemanticException {
	de.uka.ilkd.key.logic.op.Location[] vars = 
	    new de.uka.ilkd.key.logic.op.Location[locNames.size()];
	for (int i = 0; i<vars.length; i++) {
	    final String locName = (String)locNames.get(i);
	    Object o = programVariables().lookup(new Name(locName));
	    if (o != null) {
		vars[i] = (de.uka.ilkd.key.logic.op.Location) o;
	    } else {
		vars[i] = (de.uka.ilkd.key.logic.op.Location) getAttribute(null, locName);
	    }
	}
	return vars;
    }


    // taclets are always parsed in in schema mode
    public void switchToSchemaMode() {
    }

    private boolean inSchemaMode() {
        return true;
    }

    public void switchToNormalMode() {
    }

    private TacletBuilder createTacletBuilderFor
        (Object find, boolean sameUpdatePrefix) 
        throws InvalidFindException {
        if ( sameUpdatePrefix && !( find instanceof Term ) )
            throw new InvalidFindException
                ("\"sameUpdateLevel\" may only be used for rewrite taclets:" + find,
                 getFilename(), getLine(), getColumn());
        if ( find == null ) {
            return new NoFindTacletBuilder();
        } else if ( find instanceof Term ) {
            return new RewriteTacletBuilder().setFind((Term)find)
                .setSameUpdatePrefix(sameUpdatePrefix);
        } else if ( find instanceof Sequent ) {
            Sequent findSeq = (Sequent) find;
            if ( findSeq == Sequent.EMPTY_SEQUENT ) {
                return new NoFindTacletBuilder();
            } else if (   findSeq.antecedent().size() == 1
                          && findSeq.succedent().size() == 0 ) {
                Term findFma = findSeq.antecedent().get(0).formula();
                return new AntecTacletBuilder().setFind(findFma);
            } else if (   findSeq.antecedent().size() == 0
                          && findSeq.succedent().size() == 1 ) {
                Term findFma = findSeq.succedent().get(0).formula();
                return new SuccTacletBuilder().setFind(findFma);
            } else {
                throw new InvalidFindException
                    ("Unknown find-sequent (perhaps null?):"+findSeq,
                     getFilename(), getLine(), getColumn());
            }
        } else {
            throw new InvalidFindException
                    ("Unknown find class type: " + find.getClass().getName(),
                     getFilename(), getLine(), getColumn());
        }
    }       

    protected static HashSet progVars(JavaBlock jb) {
        DeclarationProgramVariableCollector pvc
            = new DeclarationProgramVariableCollector(jb.program());
        pvc.start();
        return pvc.result();
    }


    protected void bindProgVars(HashSet progVars) {
        namespaces().setProgramVariables(new Namespace(programVariables()));
        Iterator it=progVars.iterator();
        while (it.hasNext()) {
            programVariables().add((IProgramVariable)it.next());
        }
    }

    protected void unbindProgVars() {
        namespaces().setProgramVariables(programVariables().parent());
    }

    protected KeYJavaType getTypeByClassName(String s){
        KeYJavaType kjt = null;
        try {
            kjt = getJavaInfo().getTypeByClassName(s);
        } catch(RuntimeException e){
            return null;
        } 
        return kjt;
    }


    protected void addInclude(String filename, boolean relativePath, boolean ldt){
        InputStream input=null;
        RuleSource source=null;
        if(relativePath){
            int end = getFilename().lastIndexOf(File.separator);
            int start = 0;
            filename = filename.replace('/', File.separatorChar);
            filename = filename.replace('\\', File.separatorChar);
            if(getFilename().startsWith("file:")){
                start = 5;
            }
            File path=new File(getFilename().substring(start,end+1)+filename);
            try{ 
                source = RuleSource.initRuleFile(path.toURL()); 
            }catch(java.net.MalformedURLException e){
                System.err.println("Exception due to malformed URL of file "+
                                   filename+"\n " +e);
            }
        }else{
            source = RuleSource.initRuleFile(filename+".key"); 
        }
        if(ldt){
            includes.putLDT(filename, source);
        }else{
            includes.put(filename, source);
        }
    }

    
    private static ListOfRulevariable NO_RULE_VARIABLES =
        SLListOfRulevariable.EMPTY_LIST;

    private void addGoalTemplate(TacletBuilder b,
                                 String id,
                                 Object rwObj,
                                 Sequent addSeq,
                                 ListOfTaclet addRList,
                                 SetOfSchemaVariable pvs,
                                 SetOfChoice soc) 
        throws SemanticException
        {
            TacletGoalTemplate gt = null;
            if ( rwObj == null ) {
                // there is no replacewith, so we take
                gt = new TacletGoalTemplate(addSeq,
                                            addRList,
                                            NO_RULE_VARIABLES,
                                            pvs);
            } else {
                if ( b instanceof NoFindTacletBuilder ) {
                    // there is a replacewith without a find.
                    throw 
                        new UnfittingReplacewithException
                        ("Replacewith without find", getFilename(),
                         getLine(), getColumn());
                } else if ( b instanceof SuccTacletBuilder
                            || b instanceof AntecTacletBuilder ) {
                    if ( rwObj instanceof Sequent ) {
                        gt = new AntecSuccTacletGoalTemplate(addSeq,
                                                             addRList,
                                                             NO_RULE_VARIABLES,
                                                             (Sequent)rwObj,
                                                             pvs);  
                    } else {
                        throw new UnfittingReplacewithException
                            ("Replacewith in a Antec-or SuccTaclet has "+
                             "to contain a sequent (not a term)", 
                             getFilename(), getLine(), getColumn());
                    }
                } else if ( b instanceof RewriteTacletBuilder ) {
                    if ( rwObj instanceof Term ) {
                        gt = new RewriteTacletGoalTemplate(addSeq,
                                                           addRList,
                                                           NO_RULE_VARIABLES,
                                                           (Term)rwObj,
                                                           pvs);  
                    } else {
                        throw new UnfittingReplacewithException
                            ("Replacewith in a RewriteTaclet has "+
                             "to contain a term (not a sequent)", 
                             getFilename(), getLine(), getColumn());
                    }
                }
            }
            gt.setName(id); 
            b.addTacletGoalTemplate(gt);
            if(soc != null) b.addGoal2ChoicesMapping(gt,soc);
        }
    


    protected int getLine() {
        int line = -1;
        try {
            line = LT(0).getLine();
        } 
        catch (TokenStreamException e) {
            System.err.println("No further token in stream");
        }
        return line;
    }   

    protected int getColumn() {
        int col = -1;
        try {
            col = LT(0).getColumn();
        } 
        catch (TokenStreamException e) {
            System.err.println("No further token in stream");
        }
        return col;
    }   

    protected Term termForParsedVariable(ParsableVariable v) 
        throws antlr.SemanticException 
        {
            if ( v instanceof LogicVariable ) {
                return tf.createVariableTerm((LogicVariable)v);
            } else if ( v instanceof SchemaVariable ) {
                return tf.createVariableTerm((SchemaVariable)v);
            } else if ( v instanceof ProgramVariable ) {
                return tf.createVariableTerm((ProgramVariable)v);
                //      } else if ( v instanceof ParsableVariable ) {
                //          return tf.createVariableTerm((ParsableVariable)v);

            } else {
                throw new KeYSemanticException
                    (""+ v +" is not a logic or schema or program variable.", 
			getFilename(), getLine(), getColumn());
            }
        }

    protected TermSymbol progVar(Token id,Sort s) throws NotDeclException{
        TermSymbol v = (TermSymbol)
            programVariables().lookup(new Name(id.getText()));

        if (v == null) {
            throw new NotDeclException("program variable", id, getFilename());
        }
        return v;
    }

    protected LogicVariable bindVar(Token id,Sort s) {
        LogicVariable v=new LogicVariable(new Name(id.getText()), s);
        namespaces().setVariables(variables().extended(v));
        return v;
    }
    
    protected void unbindVars() {
        namespaces().setVariables(variables().parent());
    }

    protected JavaBlock getJavaBlock(Token t) throws antlr.SemanticException {
        String s=t.getText();
        if (!inSchemaMode()) {
            try {
                return javaReader.readBlockWithProgramVariables(programVariables(), s);
            } catch (Exception e) {
                if (s.length()>10) {
                    s=s.substring(0,10)+"...";
                }
                throw new JavaParserException("Reading Java block " + s + 
					      " failed. Reason: " + e.getMessage()); 
            } 

        }
        javaReader.setSVNamespace(variables());
        return javaReader.readBlockWithEmptyContext(s);
    }

    /** returns the java information object */
    public JavaInfo getJavaInfo() {
        return getServices().getJavaInfo();
    }

    public Services getServices() {
        return services;
    }




       
    public void schema_var_decl(String name,Sort s,
                                    boolean makeVariableSV,
                                    boolean makeDependingSV,
                                    boolean rigid,
                                    boolean listModifier)
	{
              if (!skip_schemavariables) {
                  SchemaVariable v;
                  if ( s == Sort.FORMULA ) {
                      v = SchemaVariableFactory.createFormulaSV
                          (new Name(name), listModifier, rigid);
                      } else if ( s instanceof ProgramSVSort ) {
                          v = SchemaVariableFactory.createProgramSV
                         (new ProgramElementName(name),(ProgramSVSort) s, listModifier);
                      } else {
                        if ( makeVariableSV ) {
                             v=SchemaVariableFactory.createVariableSV
                             (new Name(name), s, listModifier);
                        } else if ( makeDependingSV ) {
                               v = SchemaVariableFactory.createDependingSV
                             (new Name(name), s, listModifier);
                        } else { v = SchemaVariableFactory.createTermSV
                              (new Name(name), s, listModifier, rigid);
                        }
                     }
                     if (inSchemaMode()) {
                         variables().add(v);
                     }
               }

         }
     


    public void testLiteral(String l1, String l2)
    throws KeYSemanticException
    {
     if (!l1.equals(l2)){
        throw new KeYSemanticException
        ("Expecting '"+l1+"', found '"+l2+"'.",getFilename(), getLine(), getColumn());
	};
    }

    protected Operator lookupVarfuncId(String varfunc_name, 
                    PairOfTermArrayAndBoundVarsArray args) throws NotDeclException {
        Term[] argTerms = null;
        if (args != null) {
            argTerms = args.getTerms();
        }
        return lookupVarfuncId(varfunc_name, argTerms);
    }

    /** looks up a function, (program) variable or static query of the 
     * given name varfunc_id and the argument terms args in the namespaces 
     * and java info. 
     * @param args is null iff no argument list is given, for instance `f', 
     * and is an array of size zero, if an empty argument list was given,
     * for instance `f()'.
     */
    protected Operator lookupVarfuncId(String varfunc_name, Term[] args) 
                                                    throws NotDeclException{
        // case 1: variable
        Operator v = (TermSymbol) variables().lookup(new Name(varfunc_name));
        if (v != null && args==null) {
            return v;
        }
        // case 2: function
        v = (TermSymbol) functions().lookup(new Name(varfunc_name));
        if (v != null) { // we allow both args==null (e.g. `c')
                         // and args.length=0 (e.g. 'c())' here 
            return v;
        }
        // case 3: static query
        if (varfunc_name.lastIndexOf("::")>0 && args!=null) {
            String classname = varfunc_name.substring
            (0, varfunc_name.lastIndexOf("::"));
            varfunc_name = varfunc_name.substring
            (varfunc_name.lastIndexOf("::")+2);
            v = getServices().getJavaInfo().getProgramMethodTerm
                (null, varfunc_name, args, classname).op();
            if (v != null) {
                return v;
            }
        }
        // case 4: program variable
        v = (ProgramVariable) programVariables().lookup
            (new ProgramElementName(varfunc_name));
        if (v != null && args==null) {
            return v;
        }
        if (args==null) {
            throw new NotDeclException
            ("(program) variable or constant", varfunc_name,
                getFilename(), getLine(), getColumn());
        } else {
            throw new NotDeclException
            ("function or static query", varfunc_name,
                getFilename(), getLine(), getColumn());
        }
    }


}

top {Taclet r;} : r=taclet[SetAsListOfChoice.EMPTY_SET];

taclet[SetOfChoice choices] returns [Taclet r] 
{ 
    Sequent ifSeq = Sequent.EMPTY_SEQUENT;
    Object  find = null;
    r = null;
    TacletBuilder b = null;
    boolean sameUpdatePrefix = false;
}
    : 
        name:IDENT 
        (LPAREN choices=option[choices](COMMA choices=option[choices])* RPAREN)? 
        LBRACE
        ( "if" LPAREN ifSeq=seq RPAREN ) ?
        ( "find" LPAREN find = termorseq RPAREN
            ( "sameUpdateLevel" { sameUpdatePrefix = true; } ) ? ) ?
        { 
            b = createTacletBuilderFor(find, sameUpdatePrefix);
            b.setName(new Name(name.getText()));
            b.setIfSequent(ifSeq);
        }
        ( "varcond" LPAREN varexplist[b] RPAREN ) ?
        goalspecs[b]
        modifiers[b]
        RBRACE
        { 
            b.setChoices(choices);
            r = b.getTaclet(); 
            taclet2Builder.put(r,b);
        }
    ;

modifiers[TacletBuilder b] :
        ( rulesets[b] 
        | "noninteractive" { 
                //      b.setNoninteractive(true);  
                // "noninteractive" (as it is now) is confusing
                // dropped this completely until a better solution (->Uwe)
            }
        | "recursive"      { b.setRecursive(true); }
        | DISPLAYNAME dname:STRING_LITERAL 
            {b.setDisplayName(dname.getText());}
        | HELPTEXT htext:STRING_LITERAL
            {b.setHelpText(htext.getText());}
        ) *
    ;

seq returns [Sequent s] {Semisequent ant,suc; s = null; } : 
        ant=semisequent SEQARROW suc=semisequent 
        { s = Sequent.createSequent(ant, suc); }
    ;

termorseq returns [Object o]
{
    Term head = null;
    Sequent s = null;
    Semisequent ss = null;
    o = null; 
}
    :
        head=term ( COMMA s=seq | SEQARROW ss=semisequent ) ?
        {
            if ( s == null ) {
                if ( ss == null ) {
                    // Just a term
                    o = head;
                } else {
                    // A sequent with only head in the antecedent.
                    Semisequent ant = Semisequent.EMPTY_SEMISEQUENT;
                    ant = ant.insertFirst(
                                          new ConstrainedFormula(head, Constraint.BOTTOM)).semisequent();
                    o = Sequent.createSequent(ant,ss);
                }
            } else {
                // A sequent.  Prepend head to the antecedent.
                Semisequent newAnt = s.antecedent();
                newAnt = newAnt .insertFirst(
                                             new ConstrainedFormula(head, Constraint.BOTTOM)).semisequent();
                o = Sequent.createSequent(newAnt,s.succedent());
            }
        }
    |
        SEQARROW ss=semisequent
        {
            o = Sequent.createSequent(Semisequent.EMPTY_SEMISEQUENT,ss);
        }
    ;

semisequent returns [Semisequent ss]
{ 
    Term head = null;
    ss = Semisequent.EMPTY_SEMISEQUENT; 
}
    :
        /* empty */ | 
        head=term ( COMMA ss=semisequent ) ? 
        { ss = ss.insertFirst(new ConstrainedFormula(head, Constraint.BOTTOM)).semisequent(); }
    ;

varexplist[TacletBuilder b] : varexp[b] ( COMMA varexp[b] ) * ;

varexp[TacletBuilder b]
{
    ParsableVariable x;
    ParsableVariable y; 
    ParsableVariable z;
    Sort s = null;
    boolean negation = false;
}
    :

        (
            "typeof" LPAREN y=varId RPAREN  
		(x=varId "new" 
            		{ b.addVarsNew((SchemaVariable) x, (SchemaVariable) y, false); }
		| "not" (
			"same" "typeof" LPAREN x=varId RPAREN
            		{ b.addVariableCondition
			   (new TypeComparisionCondition((SchemaVariable) y, (SchemaVariable) x, 
				TypeComparisionCondition.NOT_SAME)); }
			|
			"compatible" "typeof" LPAREN x=varId RPAREN 
            		{ b.addVariableCondition
			   (new TypeComparisionCondition((SchemaVariable) y, (SchemaVariable) x, 
				TypeComparisionCondition.NOT_COMPATIBLE)); }
			)
		)
        | "elemtypeof" LPAREN y=varId  RPAREN  x=varId "new"
            { b.addVarsNew((SchemaVariable) x, (SchemaVariable) y, true); }
        | (sortId_check[true] varId "new") => s=sortId_check[true]  x=varId "new"
            {b.addVarsNew((SchemaVariable) x, s); }

                
	| (IDENT varId varId varId) => 
	   (mId:IDENT { testLiteral("methodreference",mId.getText());} 
	    x=varId y=varId z=varId ("not" { negation = true; } )? sId1:IDENT {
                    testLiteral("static",sId1.getText());
		    b.addVariableCondition(new StaticMethodCondition
                                           (negation, (SchemaVariable)x, (SchemaVariable)y,
                                            (SchemaVariable)z));
                    negation = false; 
                })
	| x = varId
            ( "not" (
                    "free" "in" y=varId 
                    { b.addVarsNotFreeIn((SchemaVariable)x,(SchemaVariable)y); }
                | ( {"static".equals(LT(1).getText())}?
                    sId2:IDENT
                    { testLiteral("static",sId2.getText());
		      b.addVariableCondition(new StaticReferenceCondition
                                             ((SchemaVariable) x, true));}		
		    | 
                    sId4:IDENT y=varId 
                    { testLiteral("sameLiteral",sId4.getText());
		      b.addVariableCondition(new TestLiteral
                                             ((SchemaVariable) x, (SchemaVariable) y));
		    }
                ) 
	    )
            | sId3:IDENT
                { testLiteral("static",sId3.getText());
		   b.addVariableCondition
                  (new StaticReferenceCondition((SchemaVariable) x,
                                                false));}
            | "new" "depending" "on" y=varId
                { b.addVarsNewDependingOn((SchemaVariable)x,(SchemaVariable)y); }
            | "smaller" "than" y=varId
                { b.addVariableCondition(
                        new OrderingCondition ((SchemaVariable)x,(SchemaVariable)y)); }
            | "isquery" 
                { b.addVariableCondition(new TestQuery((SchemaVariable)x));}
            | "hasSort" s=any_sortId_check[true]
               { if ( !( s instanceof GenericSort ) )
                   throw new GenericSortException ( "sort",
                                                    "Generic sort expected", s,
                                                    getFilename (), getLine (), getColumn () );
                 if ( !( x instanceof SortedSchemaVariable
                         && ( (SortedSchemaVariable)x ).sort () ==
                                ProgramSVSort.EXPRESSION ) )
                   throw new KeYSemanticException("Expected schema variable of kind EXPRESSION, " +
			"but is " + x , getFilename(), getLine(), getColumn());
                 b.addVariableCondition(new JavaTypeToSortCondition ((SchemaVariable)x, (GenericSort)s)); }

            
	    )        
        )
    ;

goalspecs[TacletBuilder b] :
        "close" "goal"
    | goalspecwithoption[b] ( SEMI goalspecwithoption[b] )* ;

goalspecwithoption[TacletBuilder b]
{
    SetOfChoice soc = SetAsListOfChoice.EMPTY_SET;
} :
        ((LPAREN soc=option[soc](COMMA soc=option[soc])* RPAREN 
                LBRACE
                goalspec[b,soc] 
                RBRACE)
        |  
            goalspec[b,null] 
        )
    ;

option[SetOfChoice soc] returns [SetOfChoice result=null]
{
    Choice c = null;
} :
        cat:IDENT COLON choice:IDENT
        {
            c = (Choice) choices().lookup(new Name(cat.getText()+":"+choice.getText()));
            if(c==null){
                throw new NotDeclException
			("Option", choice, getFilename());
            }else{
                result = soc.add(c);
            }
        }
    ;

goalspec[TacletBuilder b, SetOfChoice soc] 
{
    Object rwObj = null;
    Sequent addSeq = Sequent.EMPTY_SEQUENT;
    ListOfTaclet addRList = SLListOfTaclet.EMPTY_LIST;
    SetOfSchemaVariable addpv = SetAsListOfSchemaVariable.EMPTY_SET;
    String name = null;
}
    :
        (id:STRING_LITERAL COLON {name = id.getText();})?
        (   ( rwObj = replacewith
                (addSeq=add)? 
                (addRList=addrules)? 
                (addpv=addprogvar)?
            )
        | ( addSeq=add (addRList=addrules)? )
        | ( addRList=addrules )
        )
        {
            addGoalTemplate(b,name,rwObj,addSeq,addRList,addpv,soc);
        }
        
    ;

replacewith returns [Object o] { o = null; } :
        "replacewith" LPAREN o=termorseq RPAREN;

add returns [Sequent s] { s = null;} :
        "add" LPAREN s=seq RPAREN;

addrules returns [ListOfTaclet lor] { lor = null; } :
        "addrules" LPAREN lor=tacletlist RPAREN;

addprogvar returns [SetOfSchemaVariable pvs] {pvs = null; } :
        "addprogvars" LPAREN pvs=pvset RPAREN;

tacletlist returns [ListOfTaclet lor]
{ 
    Taclet head = null;
    lor = SLListOfTaclet.EMPTY_LIST; 
}
    :
        head=taclet[SetAsListOfChoice.EMPTY_SET]   
        ( /*empty*/ | COMMA lor=tacletlist ) { lor = lor.prepend(head); }
    ;

pvset returns [SetOfSchemaVariable pvs] 
{
    ParsableVariable pv = null;
    pvs = SetAsListOfSchemaVariable.EMPTY_SET;
}
    :
        pv=varId
        ( /*empty*/ | COMMA pvs=pvset ) { pvs = pvs.add
                                          ((SchemaVariable)pv); };



rulesets[TacletBuilder b] :
        "heuristics" LPAREN ruleset[b] ( COMMA ruleset[b] ) * RPAREN ;

ruleset[TacletBuilder b] :
        id:IDENT
        {   
            RuleSet h = (RuleSet) ruleSets().lookup(new Name(id.getText()));
            if (h == null) {
                throw new NotDeclException("ruleset", id, getFilename());
            }
            b.addRuleSet(h);
        }
    ;


// This is overridden from terms.g to allow SchemaVariables in quantifiers.


bindingterm returns [Term a = null]
{}
    :
        a = sv_or_local_bindingterm
    ;

sv_or_local_bindingterm returns [Term a = null]
{
    Operator op = null;
    QuantifiableVariable v = null;
    TermSymbol ts = null;
    Term a1;
    Sort s = null;
    boolean bound = false;
}
    :
        (( 
                "all" { op = Op.ALL; }
            |   "ex"  { op = Op.EX;  }
            )
            v0:IDENT
            (   
                COLON s=sortId 
                {
                    v = bindVar(v0,s);
                    bound = true;
                }
            |
                /*empty*/
                {
                    ts = (TermSymbol) variables().lookup(new Name(v0.getText()));   
                    if (   ts == null 
                           || ts instanceof LogicVariable ) {
                        throw new KeYSemanticException("Quantified variables need a sort.", 
				getFilename(), getLine(), getColumn());
                    }
                    if ( ! (ts instanceof SchemaVariable && 
                            ((SchemaVariable)ts).isVariableSV())) {
                        throw new KeYSemanticException
				(ts+" is not allowed in a quantifier.", 
				 getFilename(), getLine(), getColumn());
                    }
                    v = (SortedSchemaVariable) ts;
                    bound = false;
                }
            )
            DOT a1 = term60 
            { 
                a = tf.createQuantifierTerm((Quantifier)op, v, a1);
                if (bound) {unbindVars();}
            })
    |
        a = update_or_substitution
  ; exception
        catch [TermCreationException ex] {
              keh.reportException
		(new KeYSemanticException
			(ex.getMessage(), getFilename(), getLine(), getColumn()));
        }



//This is overridden from terms.g to allow SchemaVariables.
//"TermWithBoundVars" is used here instead of "Term" to allow for
//syntax needed for OCL simplification.
argument returns [TermWithBoundVars p = null]
{
    List listVars = new LinkedList();
    SortedSchemaVariable var = null;
    ArrayOfQuantifiableVariable vars = null;
    SortedSchemaVariable[] arrayVars = null;
    Term a1;

    LogicVariable v = null;
    LogicVariable[] logArrayVars = null;
    Sort s = null;
}
	:
        (
            BACKSLASH v0:IDENT 
            (
                //Takes care of parsing the "problem" part, i.e. interprets
                //variables as logical variables.
                COLON s=sortId
                {
                    v = bindVar(v0, s);
                    listVars.add(v);
                }
                (
                    BACKSLASH v1:IDENT COLON s=sortId
                    {
                        v = bindVar(v1, s);
                        listVars.add(v);
                    }
                )* SLASH a1=term60
                {
                    logArrayVars = (LogicVariable[])listVars.toArray(new LogicVariable[listVars.size()]);
                    vars = new ArrayOfQuantifiableVariable(logArrayVars);
                    p = new TermWithBoundVars(vars, a1);
                    unbindVars();
                }
            |
                //Takes care of parsing the "rules" part, i.e. interprets variables
                //as schema variables.
                {
                    var = (SortedSchemaVariable)variables().lookup(new Name(v0.getText())); 
                    listVars.add(var);
                }
                (
                    BACKSLASH v2:IDENT
                    {
                        var = (SortedSchemaVariable)variables().lookup(new Name(v2.getText())); 
                        listVars.add(var);
                    }
                )* SLASH a1=term60
                {
                    arrayVars = (SortedSchemaVariable[])listVars.toArray(new SortedSchemaVariable[listVars.size()]);
                    vars = new ArrayOfQuantifiableVariable(arrayVars);
                    p = new TermWithBoundVars(vars, a1);
                }
            )
	    | a1=term60
            {
                p = new TermWithBoundVars(vars, a1);
            }
        )
	;


metaId returns [Operator v = null] 
    :
      META_SHADOW 
        { v = Op.META_SHADOW ; }
    | CONSTANT_VALUE
        { v = Op.CONSTANT_VALUE; }
    | META_INT_UNARY_MINUS 
        { v = Op.META_INT_UNARY_MINUS ; }
    | META_LONG_UNARY_MINUS 
        { v = Op.META_LONG_UNARY_MINUS ; }
    | META_ADD 
        { v = Op.META_ADD; }
    | META_SUB
        { v = Op.META_SUB; }
    | META_MUL
        { v = Op.META_MUL; }
    | META_DIV 
        { v = Op.META_DIV; }
    | META_JDIV 
        { v = Op.META_JDIV; }
    | META_MOD
        { v = Op.META_MOD; }
    | META_JMOD
        { v = Op.META_JMOD; }
    | META_LESS
        { v = Op.META_LESS; }
    | META_GREATER
        { v = Op.META_GREATER; }
    | META_LEQ
        { v = Op.META_LEQ; }
    | META_GEQ
        { v = Op.META_GEQ; }
    | META_EQ
        { v = Op.META_EQ; }
    | META_INT_ADD 
        { v = Op.META_INT_ADD; }
    | META_INT_SUB 
        { v = Op.META_INT_SUB; }
    | META_INT_MUL 
        { v = Op.META_INT_MUL; }
    | META_INT_DIV 
        { v = Op.META_INT_DIV; }
    | META_INT_MOD 
        { v = Op.META_INT_MOD; }
    | META_INT_AND 
        { v = Op.META_INT_AND; }
    | META_INT_OR
        { v = Op.META_INT_OR; }
    | META_INT_XOR
        { v = Op.META_INT_XOR; }
    | META_INT_COMPLEMENT
        { v = Op.META_INT_COMPLEMENT; }
    | META_INT_SHIFTRIGHT
        { v = Op.META_INT_SHIFTRIGHT; }
    | META_INT_SHIFTLEFT
        { v = Op.META_INT_SHIFTLEFT; }
    | META_INT_UNSIGNEDSHIFTRIGHT
        { v = Op.META_INT_UNSIGNEDSHIFTRIGHT; }
    | META_LONG_ADD 
        { v = Op.META_LONG_ADD; }
    | META_LONG_SUB 
        { v = Op.META_LONG_SUB; }
    | META_LONG_MUL 
        { v = Op.META_LONG_MUL; }
    | META_LONG_DIV 
        { v = Op.META_LONG_DIV; }
    | META_LONG_MOD 
        { v = Op.META_LONG_MOD; }
    | META_LONG_AND 
        { v = Op.META_LONG_AND; }
    | META_LONG_OR
        { v = Op.META_LONG_OR; }
    | META_LONG_XOR
        { v = Op.META_LONG_XOR; }
    | META_LONG_COMPLEMENT
        { v = Op.META_LONG_COMPLEMENT; }
    | META_LONG_SHIFTRIGHT
        { v = Op.META_LONG_SHIFTRIGHT; }
    | META_LONG_SHIFTLEFT
        { v = Op.META_LONG_SHIFTLEFT; }
    | META_LONG_UNSIGNEDSHIFTRIGHT
        { v = Op.META_LONG_UNSIGNEDSHIFTRIGHT; }
    | META_IN_LONG 
        { v = Op.META_IN_LONG; }
    | META_IN_INT
        { v = Op.META_IN_INT; }
    | META_IN_CHAR
        { v = Op.META_IN_CHAR; }
    | META_IN_SHORT
        { v = Op.META_IN_SHORT; }
    | META_IN_BYTE 
        { v = Op.META_IN_BYTE; }
    | META_MODULO_BYTE
        { v = Op.META_MODULO_BYTE; }
    | META_MODULO_SHORT
        { v = Op.META_MODULO_SHORT; }
    | META_MODULO_INTEGER
        { v = Op.META_MODULO_INTEGER; }
    | META_MODULO_LONG 
        { v = Op.META_MODULO_LONG; }    
    | STATIC_ANALYSIS
        { v = Op.STATIC_ANALYSIS; }
    | WHILE_INV_RULE
        { v = Op.WHILE_INV_RULE; }
    | RESOLVE_QUERY
        { v = Op.RESOLVE_QUERY; }
    | META_LENGTH
        { v = Op.META_LENGTH; }
    | META_NEXT
        { v = Op.META_NEXT; }
    | META_CREATED
        { v = Op.META_CREATED; }
    | META_TRANSIENT
        { v = Op.META_TRANSIENT; }
    | INTRODUCE_NEW_ANON_UPDATE {
	{ v= Op.INTRODUCE_NEW_ANON_UPDATE; }
    }

    //For OCL simplification
    | META_ALLSUBTYPES
        {v = Op.METAALLSUBTYPES;}
    ;

metaTerm returns [Term result = null]
{
    LinkedList al = new LinkedList();
    Operator vf = null;
    Term t = null;
} 
    :
        (vf = metaId
            LPAREN 
            t = term
            {
                al.add(t);
            }
            ( COMMA 
                t = term
                {
                    al.add(t);
                }   
            )* RPAREN                      
            {   
                result = tf.createMetaTerm((MetaOperator)vf,
                                           (Term[])al.toArray(AN_ARRAY_OF_TERMS));
            }         
        ) 
 ; exception
     catch [TermCreationException ex] {
         keh.reportException
  	    (new KeYSemanticException
			(ex.getMessage(), getFilename(), getLine(), getColumn()));
        }


// This is overridden from terms.g to allow SchemaVariables in attributes.


update_or_substitution returns [Term result = null] 
{} :
        
        (
             LBRACE IDENT 
             (
                 COLON sortId 
             |
                 logicTermReEntry
             )
         ) => result = substitutionterm 

    | result = updateterm 
 ; 


substitutionterm returns [Term result = null] 
{ Term a1, a2 = null;
Sort s = null;
TermSymbol ts;
QuantifiableVariable v = null;
boolean bound = false;
boolean waryAll = false;
boolean waryEx = false;
} :

        LBRACE v1:IDENT
        (COLON s=sortId {
                bound = true;
            } | 
            { functions().lookup(new Name(v1.getText())) == null }?             
            {
                ts = (TermSymbol) variables().lookup(new Name(v1.getText()));
                if (   ts == null
                       || ts instanceof LogicVariable ) {
                    throw new KeYSemanticException("Substituted variables need a sort.", 
				getFilename(), getLine(), getColumn());
                }
                if ( ! (ts instanceof SchemaVariable &&
                        ((SchemaVariable)ts).isVariableSV()) ) {
                    throw new KeYSemanticException
			(ts+" is not allowed as bound variable in a substitution term.",
			getFilename(), getLine(), getColumn());
                }
                v = (SortedSchemaVariable) ts;
                bound = false;
            })  
	(  
        a1=logicTermReEntry
        ( "waryAll" { waryAll = true; } |
            "waryEx"  { waryEx  = true; } )?
        RBRACE
        { 
            if (bound) {v = bindVar(v1,s);} 
        } 
        a2=term60
        { 
            if ( waryAll ) {
                result = tf.createSubstitutionTerm ( Op.WARY_SUBST_ALL,
                                                     v, a1, a2 );
            } else if ( waryEx ) {
                result = tf.createSubstitutionTerm ( Op.WARY_SUBST_EX,
                                                     v, a1, a2 );
            } else {
                result = tf.createSubstitutionTerm ( Op.SUBST,
                                                     v, a1, a2 );
            }
            if (bound) {unbindVars();}
        }

    )
   ; exception
        catch [TermCreationException ex] {
              keh.reportException
		(new KeYSemanticException
			(ex.getMessage(), getFilename(), getLine(), getColumn()));
        }



shortCond[Term cond1, Term cond2] returns [Term result = null]
{
  Term c_then = null;
  ParsableVariable v = null;
}:        
        DOT v = varId CONDUP c_then = logicTermReEntry {
            if (v instanceof SortedSchemaVariable) {
                result = tf.createIfElseTerm((SortedSchemaVariable)v, cond1, 
                                        cond2, c_then);
            } else {
                result = tf.createIfElseTerm((ProgramVariable)v, cond1, 
                                             cond2, c_then);
            }
        }
 ; exception
        catch [TermCreationException ex] {
              keh.reportException
		(new KeYSemanticException
			(ex.getMessage(), getFilename(), getLine(), getColumn()));
        }



// entry point overwritten in taclets.g
specialTerm returns [Term result = null] 
{
    Operator vf = null;
}:
    result = metaTerm
    |
        vf = expr_op 
            {   
                result = tf.createFunctionTerm((Function)vf, AN_ARRAY_OF_TERMS);
            }                 
   ; exception
        catch [TermCreationException ex] {
              keh.reportException
		(new KeYSemanticException
			(ex.getMessage(), getFilename(), getLine(), getColumn()));
        }



