// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//
//

/**
 * Replaces program variables in a sequent.
 */

package de.uka.ilkd.key.proof;

import de.uka.ilkd.key.java.Statement;
import de.uka.ilkd.key.java.StatementBlock;
import de.uka.ilkd.key.java.visitor.ProgVarReplaceVisitor;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.util.Debug;
import java.util.ArrayList;
import java.util.Map;

public class ProgVarReplacer {

    Map map;
    TermFactory fact;


    /**
     * creates a ProgVarReplacer that replaces program variables as specified 
     * by the map parameter
     */
    public ProgVarReplacer(Map m) {
        fact = TermFactory.DEFAULT;
        map = m;
    }
    
    
    /**
     * merges "next" into "base"
     * assumption: 
     *   "next" is the result of replacing in "base" the formula at position 
     *   "idx" by calling Semisequent.replace()
     *   (this implies that "next" contains exactly one removed and one added
     *   formula)
     */
    private void mergeSemiCIs(SemisequentChangeInfo base, 
                              SemisequentChangeInfo next, 
                              int idx) {
        Debug.assertTrue(next.modifiedFormulas().isEmpty());
        
        IteratorOfConstrainedFormula remIt = next.removedFormulas().iterator();
        Debug.assertTrue(remIt.hasNext());
        ConstrainedFormula remCf = remIt.next();
        Debug.assertFalse(remIt.hasNext());
        base.removedFormula(idx, remCf);     
        
        IteratorOfConstrainedFormula addIt = next.addedFormulas().iterator();
        Debug.assertTrue(addIt.hasNext());
        ConstrainedFormula addCf = addIt.next();
        Debug.assertFalse(addIt.hasNext());
        base.addedFormula(idx, addCf);
       
        base.setFormulaList(next.getFormulaList());
        base.setSemisequent(next.semisequent());
    }


    /**
     * carries out the replacements on a sequent
     */
    public SequentChangeInfo replace(Sequent s) {
        SemisequentChangeInfo anteCI = replace(s.antecedent());
        SemisequentChangeInfo succCI = replace(s.succedent());
        
        Semisequent newAntecedent = anteCI.semisequent();
        Semisequent newSuccedent  = succCI.semisequent();

        Sequent newSequent = Sequent.createSequent(newAntecedent, 
                                                   newSuccedent);
                                               
        SequentChangeInfo result = SequentChangeInfo.createSequentChangeInfo
                                              (anteCI, succCI, newSequent, s);
        return result;
    }        


    /**
     * carries out the replacements on a semisequent
     */
    public SemisequentChangeInfo replace(Semisequent s) {
    	SemisequentChangeInfo result = new SemisequentChangeInfo();
        result.setFormulaList(s.toList());
        result.setSemisequent(s);

        for(int i = 0; i < s.size(); i++) {
            ConstrainedFormula oldcf = s.get(i);
            ConstrainedFormula newcf = replace(oldcf);
            
            if(!oldcf.equals(newcf)) {
                SemisequentChangeInfo semiCI 
                                      = result.semisequent().replace(i, newcf);
                mergeSemiCIs(result, semiCI, i);
            }
        }

        return result;
    }


    /**
     * carries out the replacements on a constrained formula
     */
    public ConstrainedFormula replace(ConstrainedFormula cf) {
        ConstrainedFormula result;
        result = new ConstrainedFormula(replace(cf.formula()),
                                        cf.constraint());
        return result;
    }
   
    /**
     * replaces the locations inside an update     
     */
    private UpdateOperator replace(UpdateOperator updateOp) {       
        UpdateOperator result = updateOp; 
        final Location[] locs = new Location[updateOp.locationCount()];
        boolean toReplace = false;
        for(int i = 0, size = updateOp.locationCount(); i<size; i++) {
            Location loc = updateOp.location(i);
            if (loc instanceof ProgramVariable &&
                    !((ProgramVariable)loc).isMember()) {
                ProgramVariable newVar = (ProgramVariable)map.get(loc);
                if(newVar != null) {
                    locs[i] = newVar;
                    toReplace = true;
                } else {
                    locs[i] = loc;
                }
            } else {
                locs[i] = loc;
            }
        }
        if (toReplace) {
            result = UpdateOperator.createUpdateOp(locs);
        }   
        return result;
    }
    
    /**
     * carries out the replacements on a term
     */
    public Term replace(Term t) {
        Term result = t;
        
        if(t.op() instanceof ProgramVariable        
           && !((ProgramVariable)t.op()).isMember()) {
            ProgramVariable newVar = (ProgramVariable)map.get(t.op());
            if(newVar != null) {
                result = fact.createVariableTerm(newVar);
            }
        }  else {        
            Term subTerms[] = new Term[t.arity()];        
            ArrayList vars = new ArrayList();
            
            for(int i = 0; i < t.arity(); i++) {
                subTerms[i] = replace(t.sub(i));
                
                ArrayOfQuantifiableVariable vbh = t.varsBoundHere(i);
                for(int j = 0; j < vbh.size(); j++) {
                    vars.add(vbh.getQuantifiableVariable(j));
                }
            }

            QuantifiableVariable v1[] = new QuantifiableVariable[0];
            v1 = (QuantifiableVariable[])(vars.toArray(v1));

            JavaBlock jb = t.javaBlock();
            if (!jb.isEmpty()) {
                jb = JavaBlock.createJavaBlock
                    ((StatementBlock)replace((Statement)jb.program()));
            }
            
            final Operator op = t.op() instanceof UpdateOperator ?
                    replace((UpdateOperator)t.op()) : t.op(); 
            final Term newTerm = fact.createTerm(op, subTerms, v1, jb, null);
            if(!newTerm.equals(t)) {
                result = newTerm;
            }
        }
        
        return result;
    }
    
    
    /**
     * carries out the replacements on a statement
     */
    public Statement replace(Statement s) {
        ProgVarReplaceVisitor pvrv = new ProgVarReplaceVisitor(s, map, false);
	pvrv.start();
	return (Statement) pvrv.result();
    }
}
