// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.proof;

import java.util.*;
import de.uka.ilkd.key.gui.LDTSettings;
import de.uka.ilkd.key.pp.AbbrevMap;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.strategy.Strategy;
import de.uka.ilkd.key.strategy.StrategyFactory;
import de.uka.ilkd.key.proof.mgt.ProofEnvironment;
import de.uka.ilkd.key.proof.mgt.BasicTask;
import de.uka.ilkd.key.proof.mgt.ProofCorrectnessMgt;
import de.uka.ilkd.key.proof.mgt.DefaultProofCorrectnessMgt;
import de.uka.ilkd.key.rule.UpdateSimplifier;

/**
 * A proof is represented as a tree whose nodes are created by
 * applying rules on the current (open) goals and dividing them up
 * into several new subgoals. To distinguish between open goals (the
 * ones we are working on) and closed goals or inner nodes we restrict
 * the use of the word goal to open goals which are a subset of the
 * proof tree's leaves.  This proof class represents a proof and
 * encapsulates its tree structure.  The {@link Goal} class represents
 * a goal with all needed extra information, and methods to apply
 * rules.  Furthermore it offers services that deliver the open goals,
 * namespaces and several other informations about the current state
 * of the proof.
 */
public class Proof implements Named {

    /** name of the proof */
    private Name name;

    /** the root of the proof */
    private Node root;

    /** 
     * list with prooftree listeners of this proof 
     * attention: firing events makes use of array list's random access
     * nature
     */
    private List listenerList = new ArrayList(10);
    
    /** list with the open goals of the proof */ 
    private ListOfGoal openGoals = SLListOfGoal.EMPTY_LIST;

    /** during closure this can be set by "subTreeCompletelyClosed" */
    private Node closedSubtree = null;
    
    /** declarations &c, read from a problem file or otherwise */
    private String problemHeader = "";

    /** the update simplifier (may be moved to nodes)*/
    private UpdateSimplifier upd_simplifier;

    /** name maps of variables, functions, sorts, heuristics and choices */
    private NamespaceSet namemap;

    /** the choices activated in this proof */
    private SetOfChoice activatedChoices;

    /** the LDT-settings of this Proof*/
    private LDTSettings ldts;

    /** the java information object: JavaInfo+TypeConverter */
    private Services services;

    /** maps the Abbreviations valid for this proof to their corresponding terms.*/
    private AbbrevMap abbreviations = new AbbrevMap();

    /** User constraint */
    private ConstraintTableModel userConstraint = new ConstraintTableModel ();    

    /** Deliverer for new metavariables */
    private MetavariableDeliverer metavariableDeliverer;

    /** the environment of the proof with specs and java model*/
    private ProofEnvironment proofEnv;
    
    /** the environment of the proof with specs and java model*/
    private ProofCorrectnessMgt localMgt;

    private BasicTask task;

    private Strategy activeStrategy;

	/** when differente users load and save a proof this vector fills up with
     * Strings containing the user names.*/
    public Vector userLog;

	/** when load and save a proof with different versions of key this vector
     * fills up with Strings containing the prcs versions.*/
    public Vector keyVersionLog;
    
    private AutoModeConfiguration autoModeConfiguration =
        new AutoModeConfiguration ();
        
    private HashMap counters = new HashMap();


    /** constructs a new empty proof with name */
    public Proof(Name name) {
	this.name = name;
	metavariableDeliverer = new MetavariableDeliverer ( this );
	addConstraintListener ();
	addStrategyListener ();
    }

    /** constructs a new empty proof */
    public Proof() {
	this(new Name(""));
    }


    /** constructs a new empty proof with name */
    public Proof(String name) {
	this(new Name(name));
    }
    
    public Proof(String name, Term problem, TacletIndex rules,
		 BuiltInRuleIndex builtInRules) {

        this(name);

	localMgt = new DefaultProofCorrectnessMgt(this);

        ConstrainedFormula cf = new ConstrainedFormula(problem);
        Node rootNode = new Node(this, 
				 Sequent.createSuccSequent
				 (Semisequent.EMPTY_SEMISEQUENT.insert(0,cf).semisequent()));
        setRoot(rootNode);

	Goal firstGoal = new Goal(rootNode, 
                                  new RuleAppIndex(new TacletAppIndex(rules),
						   new BuiltInRuleAppIndex(builtInRules)));
	openGoals = openGoals.prepend(firstGoal);

	if (closed())
	    fireProofClosed();
    }
    
    
    public Proof(String name, Term problem, String header, TacletIndex rules,
		 BuiltInRuleIndex builtInRules) {
        this(name, problem, rules, builtInRules);
        problemHeader = header;
    }

    public Proof(String name, Term problem, String header, TacletIndex rules,
		 BuiltInRuleIndex builtInRules, SetOfChoice activatedChoices) {
        this(name, problem, rules, builtInRules);
        problemHeader = header;
	this.activatedChoices = activatedChoices;
    }

    /** 
     * returns the name of the proof. Describes in short what has to be proved.     
     * @return the name of the proof
     */
    public Name name() {
	return name;
    }
    
    public String header() {
       return problemHeader;
    }
    
    public void setHeader(String s) {
       problemHeader = s;
    }

    public ProofCorrectnessMgt mgt() {
	return localMgt;
    }
    
    /** returns an existing named counter, creates a new one otherwise */
    public Counter getCounter(String name) {
        Counter c = (Counter) counters.get(name);
        if (c!=null) return c;
        c = new Counter(name);
        counters.put(name, c);
        return c;
    }
    
    /** 
     * returns a collection of the namespaces valid for this proof
     */
    public NamespaceSet getNamespaces() {
	return namemap;
    }

    /** sets the ldts for this proof */
    public void setLDTSettings(LDTSettings ldts){
	this.ldts=ldts;
    }

    /** returns the activated choices for this proof */
    public SetOfChoice getChoices(){
	return activatedChoices;
    }

    public void setActivatedChoices(SetOfChoice activatedChoices){
    	this.activatedChoices = activatedChoices;
    }

    /** returns the LDT-settings for this proof */
    public LDTSettings getLDTSettings(){
	return ldts;
    }

    /** returns the JavaInfo with the java type information */
    public JavaInfo getJavaInfo() {
	return services.getJavaInfo();
    }

    /** returns the Services with the java service classes */
    public Services getServices() {
       return services;
    }

    /** sets the variable, function, sort, heuristics namespaces */
    public void setNamespaces(NamespaceSet ns) {
        namemap = ns;
        if (openGoals().size() > 1) 
            throw new IllegalStateException("Proof: ProgVars set too late");
        openGoals().head().addProgramVariables(ns.programVariables());
    }

    public void setBasicTask(BasicTask t) {
	task = t;
    }

    public BasicTask getBasicTask() {
	return task;
    }

    public AbbrevMap abbreviations(){
	return abbreviations;
    }

    /** sets the JavaServices with the java type information */
    public void setJavaServices(Services serv) {
	this.services = serv;
    }

    public Strategy getActiveStrategy () {
        if ( activeStrategy == null ) {
            // get strategy from AutoModeConfiguration
            Name localName = autoModeConfiguration.getStrategy();
            
            activeStrategy = StrategyFactory.create ( this,localName );
        }
        return activeStrategy;
    }
    
    private void setActiveStrategy ( Name name ) {
        activeStrategy = StrategyFactory.create ( this, name );
        
        final IteratorOfGoal it = openGoals ().iterator ();
        while ( it.hasNext () )
            it.next ().setGoalStrategy(activeStrategy);
    }
    
    /** counter required to give the taclets a unique name */
    public int getUniqueTacletNr(Node undoAnchor) {
        return getCounter("tacletNumber").getCountPlusPlus(undoAnchor);
    }
    
    /** 
     * returns the default simplifier to be used (may be overwritten by branch
     * specific simplifiers in the future)
     * @return the UpdateSimplifier to be used as default one
     */
    public UpdateSimplifier simplifier() {
	return upd_simplifier;
    }

    /** 
     * sets the default simplifier
     * @param upd_simplifier the UpdateSimplifier to be used as
     * default (may be overwritten by branch specific simplifiers in the future)
     */
    public void setSimplifier(UpdateSimplifier upd_simplifier) {
	this.upd_simplifier = upd_simplifier;
    }

    /** returns the user constraint (table model)
     * @return the user constraint
     */
    public ConstraintTableModel getUserConstraint() {
	return userConstraint;
    }

    private void addConstraintListener() {
	getUserConstraint ()
	    .addConstraintTableListener ( new ConstraintTableListener () {
		    public void constraintChanged ( ConstraintTableEvent e ) {
			updateGoals ();
		}
	    });
    }
    
    private void addStrategyListener () {
        getAutoModeConfiguration()
            .addListener ( new AutoModeConfigurationListener () {
                public void strategyChanged ( AutoModeConfiguration config ) {
                    setActiveStrategy( config.getStrategy() );
                }
                
                public void maxAutomaticStepsChanged ( AutoModeConfiguration config ) {
                }
            });
    }

    private void updateGoals () {
        // Taclet indices of the particular goals have to
        // be rebuilt
        final IteratorOfGoal it = openGoals ().iterator ();
        while ( it.hasNext () )
            it.next ().updateRuleAppIndex ();
    }
    
    /** @return Deliverer of new metavariables (with unique names)*/
    public MetavariableDeliverer getMetavariableDeliverer () {
	return metavariableDeliverer;
    }
    
    public JavaModel getJavaModel() {
        return proofEnv.getJavaModel();
    }

    public void setProofEnv(ProofEnvironment env) {
	proofEnv=env;
    }

    public ProofEnvironment env() {
	return proofEnv;
    }
        
    /**
     * Returns the AutoModeConfiguration of this proof.
     */
    public AutoModeConfiguration getAutoModeConfiguration () {
        return autoModeConfiguration;
    }

    /**
     * returns the root node of the proof
     */
    public Node root() {
	return root;
    }

    /** sets the root of the proof */
    public void setRoot(Node root) {
	if (this.root != null) {
	    throw new IllegalStateException
		("Tried to reset the root of the proof.");
	} else {
	    this.root = root;
	    fireProofStructureChanged();

	    if (closed())
		fireProofClosed();
	}
    }


    /** 
     * returns the list of open goals
     * @return list with the open goals
     */
    public ListOfGoal openGoals() {
	return openGoals;
    }

    /** 
     * removes the given goal and adds the new goals in list 
     * @param goal the old goal that has to be removed from list
     * @param newGoal the ListOfGoal with the new goals that were
     * result of a rule application on goal
     */
    public void replace(Goal oldGoal, ListOfGoal newGoals) {
	openGoals = openGoals.removeAll(oldGoal);

	if ( closed () )
	    fireProofClosed();
	else {
	    fireProofGoalRemoved(oldGoal);
	    add(newGoals);	
	}
    }

    /**
     * Add the given constraint to the closure constraint of the given
     * goal, i.e. the given goal is closed if p_c is satisfied.
     */
    public void closeGoal ( Goal p_goal, Constraint p_c ) {
	p_goal.addClosureConstraint ( p_c );

	removeClosedSubtree ();

	if ( closed () )
	    fireProofClosed();
    }

    /**
     * This is called by a Node that is the root of a subtree that is
     * closed
     */
    public void subtreeCompletelyClosed ( Node p_node ) {
	// This method will be called for nodes with increasing
	// distance to the root
	if ( closedSubtree == null )
	    closedSubtree = p_node;
    }

    /**
     * Use this information to remove the goals of the closed subtree
     */
    protected void removeClosedSubtree () {
	// give the information that a subtree is closed
	// to all nodes below this node
	if (closedSubtree != null) 
	    closedSubtree.setClosed();

	if ( !closed () && closedSubtree != null ) {

	    boolean        b    = false;
	    IteratorOfNode it   = closedSubtree.leavesIterator ();
	    Goal           goal;

	    while ( it.hasNext () ) {
		goal = getGoal ( it.next () );
		if ( goal != null ) {
		    b = true;
		    remove ( goal );
		}
	    }

	    if ( b )
		// For the moment it is necessary to fire the message ALWAYS
		// in order to detect branch closing.
		fireProofGoalsAdded ( SLListOfGoal.EMPTY_LIST );		
	}

	closedSubtree = null;
    }

    /** removes the given goal from the list of open goals. Take care
     * removing the last goal will fire the proofClosed event
     * @param goal the Goal to be removed
     */
    public void remove(Goal goal) {
	ListOfGoal newOpenGoals = openGoals.removeAll(goal);
	if (newOpenGoals != openGoals) {
	    openGoals = newOpenGoals;	
	    if (closed()) {
		fireProofClosed();
	    } else {
		fireProofGoalRemoved(goal);
	    }
	}
    }

    /** adds a new goal to the list of goals 
     * @param goal the Goal to be added 
     */
    public void add(Goal goal) {
	ListOfGoal newOpenGoals = openGoals.prepend(goal);
	if (openGoals != newOpenGoals) {
	    openGoals = newOpenGoals;
	    fireProofGoalsAdded(goal);
	}
    }

    /** adds a list with new goals to the list of open goals 
     * @param goals the ListOfGoal to be prepended 
     */
    public void add(ListOfGoal goals) {
	ListOfGoal newOpenGoals = openGoals.prepend(goals);
	if (openGoals != newOpenGoals) {
	    openGoals = newOpenGoals;
	}

	// For the moment it is necessary to fire the message ALWAYS
	// in order to detect branch closing.
	fireProofGoalsAdded(goals);
	
    }

    /** Return whether the remaining goals can be closed, i.e. whether
     * the conjunction of the constraints of the open goals is
     * satisfiable. In this case all remaining open goals are removed.
     */
    public boolean closed () {
	if ( root ().getRootSink ().isSatisfiable () ) {
	    Goal goal;
	    while ( openGoals != SLListOfGoal.EMPTY_LIST ) {
		goal      = openGoals.head ();
		openGoals = openGoals.tail ();
		fireProofGoalRemoved ( goal );
	    }
	    return true;
	}

	return false;
    }
    
    private void setBackCounters(Node n) {
        Iterator it = counters.values().iterator();
        while (it.hasNext()) ((Counter)it.next()).undo(n);
    }
    
    /** returns true iff sets back to the step that created the given
     * goal. If the undo operation was succesful true is returned.
     * @param goal the Goal desribing the location where to set back
     * @return true iff undo operation was succesfull.
     */
    public  boolean setBack(Goal goal) {		
	if (goal != null) {
	    Node parent = goal.node().parent();
	    if (parent != null) {
                setBackCounters(goal.node());
		openGoals = goal.setBack(openGoals);
		fireProofGoalsChanged();
		return true;
	    }
	}
	// root reached or proof closed
	return false;
    }

    /** Prunes away the subtree beneath <code>node</code>.
     *	<code>node</code> is going to be the last node on its
     * branch.
     * @param node the node desribing the location where to set back
     * @return true iff undo operation was succesfull.
     */
    public boolean setBack(Node node) {
	Goal goal = getGoal(node);
	if (goal!=null) {
	    return true;
	} else {
	    ListOfGoal goalList = getSubtreeGoals(node);
	    if (goalList != SLListOfGoal.EMPTY_LIST) {
		// The subtree goals (goalList) are scanned for common
		// direct ancestors (parents). Afterwards the remove
		// list is the greatest subset of the subtree goals such
		// that the parents of the goals are disjoint.
		HashSet parentSet = new HashSet();
		ListOfGoal removeList = SLListOfGoal.EMPTY_LIST;
		IteratorOfGoal goalIt = goalList.iterator();
		while (goalIt.hasNext()) {
		    Goal nextGoal = goalIt.next();
		    if (!parentSet.contains(nextGoal.node().parent())) {
			removeList = removeList.prepend(nextGoal);
			parentSet.add(nextGoal.node().parent());
		    }
		}
		//call setBack(Goal) on each element in the remove
		//list. The former parents become the new goals
		IteratorOfGoal removeIt = removeList.iterator();
		while (removeIt.hasNext()) {
		    setBack(removeIt.next());
		}
		return setBack(node);
	    }
	    return false;
	}
    }

    // ?? seems to be required for presentation uses
    // I think there is a mismatch between what the method does and the method's
    // name. Can the one who implemented this method check the name and write a
    // short comment about its purpose %%RB
    public void updateProof() {
	fireProofGoalsChanged();
    }


    /** fires the event that the proof has been expanded at the given node */
    protected void fireProofExpanded(Node node) {
	ProofTreeEvent e = new ProofTreeEvent(this, node);
	for (int i = 0; i<listenerList.size(); i++) {
	    ((ProofTreeListener)listenerList.get(i)).proofExpanded(e);
	}
    }

    /** fires the event that the proof has been pruned at the given node */
    protected void fireProofPruned(Node node, Node removedNode) {
	ProofTreeEvent e = new ProofTreeRemovedNodeEvent(this, node, removedNode);
	for (int i = 0; i<listenerList.size(); i++) {
	    ((ProofTreeListener)listenerList.get(i)).proofPruned(e);
	}
    } 

    /** fires the event that the proof has been restructured */
    protected void fireProofStructureChanged() {
	ProofTreeEvent e = new ProofTreeEvent(this);
	for (int i = 0; i<listenerList.size(); i++) {
	    ((ProofTreeListener)listenerList.get(i)).proofStructureChanged(e);
	}    
    }

    /** fires the event that a goal has been removed from the list of goals */
    protected void fireProofGoalRemoved(Goal goal) {
	ProofTreeEvent e = new ProofTreeEvent(this, goal);
	for (int i = 0; i<listenerList.size(); i++) {
	    ((ProofTreeListener)listenerList.get(i)).proofGoalRemoved(e);
	}	
    }

    /** fires the event that new goals have been added to the list of
     * goals 
     */
    protected void fireProofGoalsAdded(ListOfGoal goals) {
	ProofTreeEvent e = new ProofTreeEvent(this, goals);
	for (int i = 0; i<listenerList.size(); i++) {
	    ((ProofTreeListener)listenerList.get(i)).proofGoalsAdded(e);
	}	
    }

    /** fires the event that new goals have been added to the list of
     * goals 
     */
    protected void fireProofGoalsAdded(Goal goal) {
	fireProofGoalsAdded(SLListOfGoal.EMPTY_LIST.prepend(goal));
    }

    /** fires the event that the proof has been restructured */
    protected void fireProofGoalsChanged() {
	ProofTreeEvent e = new ProofTreeEvent(this, openGoals());
	for (int i = 0; i<listenerList.size(); i++) {
	    ((ProofTreeListener)listenerList.get(i)).proofGoalsChanged(e);
	}
    } 

    /** fires the event that the proof has closed. 
     * This event fired instead of the proofGoalRemoved event when
     * the last goal in list is removed.
     */
    protected void fireProofClosed() {
	ProofTreeEvent e = new ProofTreeEvent(this);
	for (int i = 0; i<listenerList.size(); i++) {
	    ((ProofTreeListener)listenerList.get(i)).proofClosed(e);
	}
    }

    /**
     * adds a listener to the proof 
     * @param listener the ProofTreeListener to be added
     */
    public synchronized void addProofTreeListener
	(ProofTreeListener listener) {
	synchronized(listenerList) {
	    listenerList.add(listener);
	}
    }

    /**
     * removes a listener from the proof 
     * @param listener the ProofTreeListener to be removed
     */
    public synchronized void removeProofTreeListener
	(ProofTreeListener listener) {
	synchronized(listenerList) {
	    listenerList.remove(listener);
	}
    }

    /** returns true if the given node is part of a Goal 
     * @return  true if the given node is part of a Goal 
     */
    public boolean isGoal(Node node) {	
	return getGoal(node) != null;
    }

    /** returns the goal that belongs to the given node or null if the
     * node is an inner one 
     * @return the goal that belongs to the given node or null if the
     * node is an inner one 
     */
    public Goal getGoal(Node node) {	
	Goal result = null;
	IteratorOfGoal it = openGoals.iterator();
	while (it.hasNext()) {
	    result = it.next();
	    if (result.node() == node) {
		return result;
	    }
	}
	return null;
    }

    /** returns the list of goals of the subtree starting with node 
     * @param node the Node where to start from
     * @return the list of goals of the subtree starting with node 
     */
    public ListOfGoal getSubtreeGoals(Node node) {	
	ListOfGoal result = SLListOfGoal.EMPTY_LIST;
	IteratorOfGoal goalsIt  = openGoals.iterator();
	while (goalsIt.hasNext()) {
	    Goal goal = goalsIt.next();
	    IteratorOfNode leavesIt = node.leavesIterator();
	    while (leavesIt.hasNext()) {
		if (leavesIt.next() == goal.node()) {
		    result = result.prepend(goal);
		}
	    }
	}
	return result;
    }

    /** returns true iff the given node is found in the proof tree 
     *	@param node the Node to search for
     *	@return true iff the given node is found in the proof tree 
    */
    public boolean find(Node node) {
	if (root == null) {
	    return false;
	}
	return root.find(node);
    }

    /**
     * retrieves number of nodes
     */
    public int countNodes() {
	return root.countNodes();
    }

    /**
     * Currently the rule app index can either operate in interactive mode (and
     * contain applications of all existing taclets) or in automatic mode (and
     * only contain a restricted set of taclets that can possibly be applied
     * automated). This distinction could be replaced with a more general way to
     * control the contents of the rule app index
     */
    public void setRuleAppIndexToAutoMode () {
	IteratorOfGoal it = openGoals.iterator ();
	while ( it.hasNext () ) {
	    it.next ().ruleAppIndex ().autoModeStarted ();
	}
    }

    public void setRuleAppIndexToInteractiveMode () {
	IteratorOfGoal it = openGoals.iterator ();
	while ( it.hasNext () ) {
	    it.next ().ruleAppIndex ().autoModeStopped ();
	}
    }

    /**
     * retrieves number of branches
     */
    public int countBranches() {
	return root.countBranches();
    }

    public String statistics() {
	return "Nodes:"  + countNodes() + "\n" +
	    "Branches: " + countBranches() + "\n";
    }

    /** toString */
    public String toString() {
	StringBuffer result = new StringBuffer();
	    result.append("Proof -- ");
	    if (!"".equals(name.toString())) {
		result.append(name.toString());
	    } else {
		result.append("unnamed");
	    }
	result.append("\nProoftree:\n");
	result.append(root.toString());
	return result.toString();
    }
}
