// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.proof.init;

import de.uka.ilkd.key.collection.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.proof.mgt.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.util.ProgressMonitor;
import de.uka.ilkd.key.parser.*;
import java.io.*;
import de.uka.ilkd.key.util.KeYExceptionHandler;
import de.uka.ilkd.key.util.ExtList;
import java.util.Vector;
import de.uka.ilkd.key.java.Recoder2KeY;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.parser.jml.*;
import de.uka.ilkd.key.gui.Main;
import de.uka.ilkd.key.gui.JMLSpecBrowser;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.jml.*;
import javax.swing.JOptionPane;
import java.util.*;
import java.io.*;

/** represents an input from a .key user problem file producing environment
 *  as well as a proof obligation.
 */
public class KeYUserProblemFile extends KeYFile implements ProofOblInput{

    private Term problemTerm = null;
    private String problemHeader = "";

    private String javaPath;
    private boolean javaPathAlreadyParsed=false;


    /** creates a new representation of a KeYUserFile with the given name,
     * a rule source representing the physical source of the input, and
     * a graphical representation to call back in order to report the progress
     * while reading.
     */
    public KeYUserProblemFile(String name, File file, 
			       ProgressMonitor monitor) {
	super(name, file, monitor);
    }

    public void readHelp(ModStrategy mod, boolean problemOnly) 
	throws ProofInputException {
	if(file == null) return;
	if (initConfig==null) {
	    throw new IllegalStateException("KeYFile: InitConfig not set.");
	}
	try {
	    CountingBufferedInputStream cinp = 
		new CountingBufferedInputStream
		    (getNewStream(),monitor,getNumberOfChars()/100);
            DeclPicker lexer = new DeclPicker(new KeYLexer(cinp,initConfig.getServices().getExceptionHandler()));

	    /*
	      two namespace sets which share all namespace except the
	      variable namespace 	     	      
	    */
	    NamespaceSet normal = initConfig.namespaces().copy();
	    NamespaceSet schema = 
		new NamespaceSet(normal.variables().copy(), 
				 normal.functions(),
				 normal.sorts(), 
				 normal.ruleSets(),
				 normal.choices(),
				 normal.programVariables());
	    Choice defaultChoice = initConfig.getDefaultChoice().copy();
	    ParserConfig normalConfig 
		= new ParserConfig(initConfig.getServices(),
				   normal, defaultChoice);

	    ParserConfig schemaConfig 
		= new ParserConfig(initConfig.getServices(), 
				   schema, defaultChoice);
	    problemParser = new ProblemParser(lexer, 
					      file.toString(), 
					      schemaConfig, 
					      normalConfig,
					      initConfig.getTaclet2Builder(),
					      initConfig.getTaclets()); 
	    if (problemOnly) {
		problemTerm = problemParser.parseProblem();
	    } else {
		problemTerm = problemParser.problem();
	    }
	    problemHeader = lexer.getText();
	    initConfig.setTaclets(problemParser.getTaclets());
	    initConfig.add(normalConfig.namespaces(), mod);
	    initConfig.setDefaultChoice(normalConfig.getDefaultChoice()); 
	    initConfig.getProofEnv().addContracts(problemParser.getContracts());
            problemHeader = lexer.getText();
            if(problemHeader != null && 
	       problemHeader.lastIndexOf("problem") != -1){
		problemHeader = problemHeader.substring(
		    0, problemHeader.lastIndexOf("problem"));
	    }
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
        }  
    }

    /** reads the whole .key file and modifies the initial configuration
     * assigned to this object according to the given modification strategy. 
     * Throws an exception if no initial configuration has been set yet.
     */
    public void read(ModStrategy mod) throws ProofInputException {
	readHelp(mod, false);
    }

    /** reads the the problem section of the .key file and modifies
     * the initial configuration assigned to this object according to
     * the given modification strategy.  Throws an exception if no
     * initial configuration has been set yet.
     */
    public void readProblem(ModStrategy mod) throws ProofInputException {
	readHelp(mod, true);
    }



    public ProofList getPO() {
	// if we only want to parse a java-model annotated with 
	// jml-specifications, we may not have a problem section.
        if (problemTerm == null){
	    String jp = null;
	    try{
		jp = getJavaPath();
	    }catch(ProofInputException e){
		e.printStackTrace();
	    }
	    JMLSpecBrowser sb = new JMLSpecBrowser(initConfig.getServices(),
						   jp);
	    JMLProofOblInput jmlPO = sb.getPOI();
	    while(jmlPO == null){
		JOptionPane.showConfirmDialog
		    (null,
		     "Please select a Proof Obligation",
		     "No Proof", 
		     JOptionPane.DEFAULT_OPTION);
		sb = new JMLSpecBrowser(initConfig.getServices(), 
					jp);
		jmlPO = sb.getPOI();
	    }
	    jmlPO.setConfig(getInitConfig());
	    problemHeader += "\n"+jmlPO.createProgramVarsSection();
	    try{
		//the ModStrategy doesn't matter here anyway.
		jmlPO.readProblem(ModStrategy.NO_FUNCS);
	    }catch(ProofInputException e){
		e.printStackTrace();
	    }
	    ProofList pl = jmlPO.getPO();
	    getInitConfig().getProofEnv().registerProof(jmlPO, pl);
	    return pl;
	}
	return new ProofList(
            new Proof(name(), problemTerm, problemHeader,
                      initConfig.createTacletIndex(), 
		      initConfig.createBuiltInRuleIndex(),
		      initConfig.getActivatedChoices()), name());
    }

    /** returns the <code>java.io.file</code> file encapsulated by
     * the <code>KeYUserProblemFile</code>.
     */
    public File getFile(){
	return file.file();
    }
    
    /**
     * @return Returns the problemHeader.
     */
    protected String getProblemHeader () {
        return problemHeader;
    }

    /** returns true, that is the input asks the user which
     * environment he prefers if there are multiple possibilities
     */
    public boolean askUserForEnvironment() {
	return true;
    }

    public void readSpecs(){
	try {
	    if(method2jmlspecs != null && !method2jmlspecs.isEmpty()){
		Iterator it = method2jmlspecs.keySet().iterator();
		while(it.hasNext()){
		    Object meth = it.next();
		    IteratorOfJMLMethodSpec sit = 
			((SetOfJMLMethodSpec) method2jmlspecs.get(meth)).
			iterator();
		    while(sit.hasNext()){
			JMLMethodSpec jmlspec = sit.next();
			createJMLMethodContract(meth, jmlspec);
		    }
		}
	    }
	    // library classes specifications
	    SpecExtractor spex = null;
	    String jpath = getJavaPath();
	    if (jpath==null) return;
	    spex = new OCLSpecExtractor(initConfig.getServices(),jpath);
	    initConfig.getProofEnv().addContracts(spex.getSpecsFromAllClasses());
	} catch (ProofInputException e) {
	    // do nothing
	    Debug.out("Exception thrown by class KeYFile at SpecExtraction");
	}
    }

    private void createJMLMethodContract(Object meth, JMLMethodSpec jmlspec){
	Modality[] allMod = new Modality[]{Op.DIA, Op.BOX};
	String jp = null;
	try{
	    jp = getJavaPath();
	}catch(ProofInputException e){
	    e.printStackTrace();
	}
	for (int i=0; i<allMod.length; i++) {
	    KeYJavaType returnType;
	    MethodContract mct = null;
	    if(meth instanceof ProgramMethod){
		returnType = 
		    ((ProgramMethod) meth).getKeYJavaType();
		if(returnType == null){
		    returnType = initConfig.getServices().
			getJavaInfo().getKeYJavaType("void");
		}
		mct = new JMLMethodContract(new JavaModelMethod(
						(ProgramMethod) meth,
						jp), 
					    jmlspec, 
					    allMod[i], "");
	    }else if(meth instanceof MethodDeclaration){
		if(((MethodDeclaration)meth).getTypeReference() == null){
		    returnType = initConfig.getServices().
			getJavaInfo().getKeYJavaType("void");
		}else{
		    returnType = ((MethodDeclaration)meth).
			getTypeReference().getKeYJavaType();
		}
		mct = new JMLMethodContract(
		    new JavaModelMethod(
			jmlspec.getProgramMethod(), 
			jp), 
		    jmlspec, allMod[i], "");
	    }
	    initConfig.getProofEnv().addContract(mct);
	}
    }

    public boolean equals(Object o){
	if(!(o instanceof KeYUserProblemFile)) return false;
	KeYUserProblemFile kf = (KeYUserProblemFile) o;
	if(file != null && kf.file != null){
	    return kf.file.file().getAbsolutePath().
		equals(file.file().getAbsolutePath());
	}
	if(file == null && kf.file == null){
	    try{
		return getJavaPath().equals(kf.getJavaPath());
	    }catch(ProofInputException e){
		return kf == this;
	    }
	}
	return false;
    }

    public int hashCode(){
	if(file == null){
	    try{
		return getJavaPath().hashCode();
	    }catch(ProofInputException e){
		return 0;
	    }
	}
	return file.file().getAbsolutePath().hashCode();
    }

    protected SetOfProof parseJMLSpecs(CompilationUnit[] cUnits)
	throws ProofInputException {
	SetOfProof p = SetAsListOfProof.EMPTY_SET;
	if(cUnits != null){
	    JavaInfo ji = initConfig.getServices().getJavaInfo();
	    method2jmlspecs = new HashMap();
	    Set builders = new HashSet();
	    for(int i=0; i<cUnits.length; i++){
		ArrayOfTypeDeclaration tds = cUnits[i].getDeclarations();
		for(int j=0; j<tds.size(); j++){
		    if(tds.getTypeDeclaration(j) instanceof InterfaceDeclaration
		       || tds.getTypeDeclaration(j) instanceof ClassDeclaration){
			JMLSpecBuilder builder = 
			    new JMLSpecBuilder(
				tds.getTypeDeclaration(j),
				initConfig.getServices(),
				initConfig.createNamespacesForActivatedChoices().copy(),
				TermFactory.DEFAULT
				);
			builders.add(builder);
		    }
		}
	    }
	    Iterator it = builders.iterator();
	    while(it.hasNext()){
		JMLSpecBuilder b = (JMLSpecBuilder) it.next();
		b.parseModelMethodDecls();
	    }
	    it = builders.iterator();
	    while(it.hasNext()){
		JMLSpecBuilder b = (JMLSpecBuilder) it.next();
		b.parseModelFieldDecls();
	    }
	    it = builders.iterator();
	    while(it.hasNext()){
		JMLSpecBuilder b = (JMLSpecBuilder) it.next();
		b.parseJMLClassSpec();
	    }
	    it = builders.iterator();
	    while(it.hasNext()){
		JMLSpecBuilder b = (JMLSpecBuilder) it.next();
		b.parseJMLMethodSpecs();
		method2jmlspecs.putAll(b.getMethod2Specs());
		method2jmlspecs.putAll(b.getModelMethod2Specs());
		initConfig.add(b.namespaces(), ModStrategy.MOD_ALL);
	    }
	    initConfig.getServices().getImplementation2SpecMap().
		createLoopAnnotations();
	}
	return p;
    }

    private Vector getClasses(String f) throws ProofInputException  {
	File cfile = new File(f);
	Vector v=new Vector();
	if (cfile.isDirectory()) {
	    String[] list=cfile.list();
	    for (int i=0; i<list.length; i++) {
		String fullName = cfile.getPath()+File.separator+list[i];
		File n=new File(fullName);
		if (n.isDirectory()) {		    
		    v.addAll(getClasses(fullName));
		} else if (list[i].endsWith(".java")) {
		    v.add(fullName);	
		}
	    }
	    return v;
	} else {
	    throw new ProofInputException("Java model path "+f+" not found.");
	}
	
    }

    private String[] getClassesInPath(String f) throws ProofInputException {
	return (String[]) getClasses(f).toArray(new String[]{});
    }

    public String getJavaPath() throws ProofInputException {
	if (javaPathAlreadyParsed) return javaPath;
	  problemParser = new ProblemParser(new KeYLexer(getNewStream()),
    	  file.toString(), 
	  null, null, null, null);
        try {
	    String discardPrefs = problemParser.preferences();
            javaPath=problemParser.java(); 
	    javaPathAlreadyParsed=true;
            if (javaPath==null) return null;
            File cfile = new File(javaPath);
	    if (!cfile.isAbsolute()) { // test relative pathname
	        File parent=file.file().getParentFile();
	        cfile=new File(parent,javaPath).getAbsoluteFile();
                javaPath = cfile.getAbsolutePath();
	    }
	    return javaPath;
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
        } catch(de.uka.ilkd.key.util.ExceptionHandlerException ehe){
	    throw new ProofInputException(ehe.getCause().getMessage());
	}
    }

    private boolean checkExceptions(ExtList l){
	Iterator it = l.iterator();
	while(it.hasNext()){
	    if(!(it.next() instanceof NotSupportedExpressionException)){
		return false;
	    }
	}
	return true;
    }

    public String readJava() throws ProofInputException {
	if (initConfig==null) {
	    throw new IllegalStateException("KeYFile: InitConfig not set.");
	}
	// no java is parsed at all just the java path
	javaPath = getJavaPath();
	if (javaPath != null) {
	    Recoder2KeY r2k = new Recoder2KeY(initConfig.getServices(),
					      initConfig.namespaces());
	    String[] cus = getClassesInPath(javaPath);
	    try { 
		parseJMLSpecs(r2k.readCompilationUnitsAsFiles(cus));
	    } catch (ConvertException ce) {
		throw new ProofInputException(ce);
	    } 
	    if(!checkExceptions(initConfig.getServices().getExceptionHandler().
				getExceptions())){
		throw new ProofInputException("Exceptions occured");
	    }
	    initConfig.getServices().getImplementation2SpecMap().setProgVarNS(
		initConfig.progVarNS());
	}

        return javaPath;
    }
}
