// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.proof.init;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.ldt.*;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.TypeConverter;
import de.uka.ilkd.key.java.Recoder2KeY;
import de.uka.ilkd.key.gui.LDTSettings;
import de.uka.ilkd.key.util.Debug;
import java.io.*;
import de.uka.ilkd.key.java.abstraction.ListOfType;
import de.uka.ilkd.key.java.abstraction.SLListOfType;
import de.uka.ilkd.key.java.abstraction.IteratorOfType;
import de.uka.ilkd.key.java.abstraction.Type;
import de.uka.ilkd.key.java.abstraction.PrimitiveType;
import de.uka.ilkd.key.java.declaration.ClassDeclaration;
import de.uka.ilkd.key.gui.Main;

/** represents the LDT .key files as a whole. Reading the LDT input means 
 * reading parts of associated .key files in a special order, namely first
 * all sort declarations then all function and predicate declarations and
 * third the rules. This procedure makes it possible to use all declared
 * sorts in all rules, e.g.
 * 
 */
public class LDTInput implements EnvInput {

    private InitConfig initConfig = null;
    private String name = "language data types";
    private KeYFile[] keyFile;
    private Main main;
    private LDTSettings settings;
    private Recoder2KeY r2k;

    /** creates a representation of the LDT files to be used as input
     * to the KeY prover. 
     * @param settings the LDT settings to be used 
     * @param main the main class used to report the progress of reading
     */
    public LDTInput(KeYFile[] keyFile, LDTSettings settings, Main main) {
	this.keyFile = keyFile;
	this.settings=settings;
	this.main=main;
    }

    /** sets the initial configuration that is used to add the read changes to.
     */
    public void addTo(InitConfig conf) {
	this.initConfig=conf;
	r2k = new Recoder2KeY(initConfig.getServices(),
			      initConfig.namespaces());
    }

    /** returns to number of chars to be read by the LDT input.
     */
    public int getNumberOfChars() {
	int sum=0;
	for (int i=0; i<keyFile.length; i++) {
	    sum=sum+keyFile[i].getNumberOfChars();
	}
	return sum;
    }

   /** returns the initial configuration that is used to read and
     * that may be modified by reading.
     */
    public InitConfig getInitConfig() {
	return initConfig;
    }

    /** reads all LDTs, i.e., all associated .key files with respect to 
     * the given modification strategy. Reading is done in a special order: first
     * all sort declarations then all function and predicate declarations and
     * third the rules. This procedure makes it possible to use all declared
     * sorts in all rules, e.g.
     */
    public void read(ModStrategy mod) throws ProofInputException {
	if (initConfig==null) {
	    throw new IllegalStateException("LDTInput: InitConfig not set.");
	}
	for (int i=0; i<keyFile.length; i++) {
	    keyFile[i].readSorts(mod);
	}
	//	r2k.parseSpecialClasses();
	for (int i=0; i<keyFile.length; i++) {
	    keyFile[i].readFuncAndPred(mod);
	}
	for (int i=0; i<keyFile.length; i++) {
	    initLDT(keyFile[i]);
	    if (main != null) {
		main.setStatusLine("Reading "+keyFile[i].name(), 
				   keyFile[i].getNumberOfChars());
	    }
	    keyFile[i].readRulesAndProblem(mod);
	}
    }


    /** returns a string telling that this is LDT input
     */
    public String name() {
	return name;
    }

    /** initializes the LDTs */
    private void initLDT(KeYFile file) {
	String localName = file.name().toString();
	int start = localName.lastIndexOf(File.separator) + 1;
	int end = localName.lastIndexOf(".key");
	end = (end == -1 ? localName.length() : end);
	localName = localName.substring(start, end);
	if(localName.equals("booleanRules")){
	    initConfig.initBooleanLDT();
	}else if(localName.equals("intRules")){
	    initConfig.initIntegerLDT();
	}else if(localName.equals("intRulesIgnoringOverflow")){
	    initConfig.initArithIntLDTIgnoringOF();
	}else if(localName.equals("intRulesArith")){
	    initConfig.initArithIntLDT();
	}else{
	    System.err.println("This LDT-name is a bit unexpected: "+localName);
	}
    }


    private void add(Namespace target, Namespace source) {
	IteratorOfNamed it=source.elements().iterator();
	while (it.hasNext()) {
	    target.add(it.next());
	}
	
    }

}
