// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.proof.mgt;

import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.rule.RuleSet;
import de.uka.ilkd.key.rule.Taclet;
import de.uka.ilkd.key.proof.init.ProofOblInput;
import de.uka.ilkd.key.proof.init.InitConfig;
import de.uka.ilkd.key.proof.init.ModStrategy;
import de.uka.ilkd.key.proof.init.ProofInputException;
import de.uka.ilkd.key.proof.init.Includes;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Namespace;
import de.uka.ilkd.key.logic.SetOfTerm;
import de.uka.ilkd.key.logic.SetOfChoice;
import de.uka.ilkd.key.logic.op.Modality;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.JavaBlock;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.java.Statement;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.statement.Desugarable;
import de.uka.ilkd.key.util.KeYExceptionHandler;

public class DLHoareTripleContract extends AbstractContract 
                                   implements HoareTripleContract {

    protected Term fma;
    private SetOfTerm modifies;
    protected Statement sta;
    protected Taclet lemma;
    private RuleSet ruleSet;
    private String name;
    private Services services;


    public DLHoareTripleContract(Term fma, SetOfTerm modifies,
				 Namespace heurNS, String name,
				 Services services) {
	JavaProgramElement pe = fma.sub(1).javaBlock().program();
	sta = (Statement) pe;//((NonTerminalProgramElement) pe).getChildAt(0);
	this.fma = fma;
	this.modifies = modifies;
	this.name = name;
	ruleSet = (RuleSet) heurNS.lookup(new Name("use_method_specs"));
	this.services = services;
    }


    public Term getPre() {
	return fma.sub(0);
    }

    public String getPreText(){
	return getPre().toString();
    }

    public Term getPost() {
	return fma.sub(1).sub(0);
    }

    public String getPostText(){
	return getPost().toString();
    }

    public Modality getModality() {
	return (Modality) fma.sub(1).op();
    }

    public SetOfTerm getModifies() {
	return modifies;
    }

    protected AbstractMethodLemmaFactory getMethodLemmaFactory() {
	return new DLMethodLemmaFactory(getPre(), getPost(), sta, 
					ruleSet, getModifies(), getModality(), name,
					services);
    }

    public Taclet getLemmaAndRegister() {
	if (lemma==null) {
	    AbstractMethodLemmaFactory ms = getMethodLemmaFactory();
	    lemma=ms.getTaclet();
	    if (lemma!=null) {
		env.registerRule(lemma, new RuleJustificationByMethodSpec(this));
	    } 
	}
	return lemma;
    }

    private String getName() {
	return (name==null ? "DL Hoare Triple": name)
	    +((lemma==null) ? "" : " "+lemma.displayName());
    }

    public ProofOblInput getProofOblInput() {
	return new DLHoareTriplePO(fma, getName());
    }

    public String toString() {
	return "DL Hoare Triple Contract "+getName();
    }
    
    public Object getObjectOfContract() {
	return getStatement();
    }

    public Statement getStatement() {
	return sta;
    }

    static class DLHoareTriplePO implements ProofOblInput {

	private ProofList res;
	private Term term;
	private InitConfig initConfig;
	private String name;

	public DLHoareTriplePO(Term t, String name) {
	    this.term = desugar(t);
	    this.name = name;
	}

	private static Term desugar(Term fma) {
	    TermFactory tf = TermFactory.DEFAULT;
	    Statement sta = ((StatementBlock)fma.sub(1).javaBlock().program())
		.getStatementAt(0);
	    if (sta instanceof Desugarable) {
		Statement desugaredSta 
		    = (Statement)((Desugarable)sta).desugar();
		return tf.createJunctorTerm
		    (Op.IMP, fma.sub(0),
		     tf.createProgramTerm((Modality)fma.sub(1).op(),
					  JavaBlock.createJavaBlock
					  (new StatementBlock(desugaredSta)), 
					  fma.sub(1).sub(0)));
	    } else {
		return fma;
	    }
	}

	/** returns the proof obligation term as result of the proof obligation
	 * input. If there is still no input available because nothing has
	 * been read yet null is returned.
	 */
	public ProofList getPO() {
	    return res;
	}

	/** starts reading the input and modifies the InitConfig of this 
	 * object with respect to the given modification
	 * strategy. 
	 */
	public void read(ModStrategy mod) throws ProofInputException {
	    Proof p = new Proof(name(), term, "",
				initConfig.createTacletIndex(), 
				initConfig.createBuiltInRuleIndex(),
				initConfig.getActivatedChoices());
	    res = new ProofList(p, name());

	}

	public void readProblem(ModStrategy mod) throws ProofInputException {
	    read(mod);
	}

	public void setExceptionHandler(KeYExceptionHandler keh){
	}

	public boolean askUserForEnvironment() {
	    return false;
	}

	/** reads the Java model that belongs to this input.
	 */
	public String readJava() throws ProofInputException {
	    return "";
	}

	/** returns the path to the Java model.
	 */
	public String getJavaPath() throws ProofInputException {
	    return "";
	}

	/** set the initial configuration used to read an input. It may become
	 * modified during reading depending on the modification strategy used
	 * for reading.
	 */
	public void setConfig(InitConfig i) {
	    this.initConfig = i;
	}

	public void readSpecs(){}

	public void readActivatedChoices() throws ProofInputException {}

	public SetOfChoice getActivatedChoices() throws ProofInputException {
	    return initConfig.getActivatedChoices();
	}
    
	/** reads the include section and returns an Includes object.  
	 */
	public Includes readIncludes() throws ProofInputException { return null;}
    
	/** returns the name of the proof obligation input.
	 */
	public String name() {
	    return name;
	}

	public Object getObjectOfContract() {
	    NonTerminalProgramElement pe 
		= (NonTerminalProgramElement) term.sub(1).javaBlock().program();
	    return (Statement) ((NonTerminalProgramElement) pe).getChildAt(0);
	}

	public boolean fitsToContract(Contract ct) {
	    if (!(ct.getObjectOfContract() instanceof Statement)) {
		return false;
	    }
	    return ((Statement)ct.getObjectOfContract())
		.equals(getObjectOfContract());
	}

	public void startProtocol() {
	    // do nothing
	}

    }


}
