// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.rule;

import java.io.File;
import java.util.LinkedList;

import junit.framework.*;
import de.uka.ilkd.key.java.StatementBlock;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.declaration.ClassDeclaration;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.gui.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.proof.ProofList;
import de.uka.ilkd.key.util.HelperClassForTests;


/**
 * Class used for testing the simultaneous update simplifier.
 * ATTENTION: If <code>assertSame</code> is used, do not replace this
 * test with <code>assertEquals</code> without thinking, as we test here
 * also if the memory resources are wasted or not.
 */
public class TestUpdateSimplifier extends TestCase {

    private Namespace variables = new Namespace();
    private Namespace functions = new Namespace();
    private Namespace sorts     = new Namespace();
    
    private ProgramVariable[] pv;
    private ProgramVariable spv;

    private TermFactory tf = TermFactory.DEFAULT;

    private Sort testSort0;
    private Sort testSort1;
    private Sort testSort2;
    private Sort cloneable;
    private Sort integerSort;


    // t : testSort1
    private Term t;
    // i : testSort1
    private Term i;
    // o : testSort1
    private Term o;
    // u : testSort1
    private Term u;
    // r: testSort2
    private Term r;
    // o.a 
    private Term oa;
    // u.a 
    private Term ua;
    // r.a 
    private Term ra;

    // o.spv (spv static attribute) 
    private Term ospv;
    // u.spv (spv static attribute) 
    private Term uspv;
    // r.spv (spv static attribute) 
    private Term rspv;
    
    // arrays
    private ArraySort arraySort1;
    private ArraySort arraySort2;

    private Term a;
    private Term b;
    private Term c;
    // integers
    private Term idx;
    private Term jdx;
    private Term mdx;
    // a[idx] : array of testSort 1
    private Term ai;
    // a[jdx] : array of testSort 1
    private Term aj;
    // a[mdx] : array of testSort 1
    private Term am;
    // b[idx] : array of testSort 1
    private Term bi;
    // b[jdx] : array of testSort 1
    private Term bj;
    // b[mdx] : array of testSort 1
    private Term bm;
    // c[idx] : array of testSort 2
    private Term ci;
    // c[jdx] : array of testSort 2
    private Term cj;
    // c[mdx] : array of testSort 2
    private Term cm;

    

    public TestUpdateSimplifier(String s) {
	super(s);
    }

    private Term createUpdateTerm(Term[] subs) {
	Location[] loc = new Location[(subs.length-1)/2];
	LinkedList newSubs = new LinkedList();
	int count = 0;
	for (int i = 0; i<subs.length-2; i+=2, count++) {
	    loc[count] = (Location) subs[i].op();
	    for (int j = 0; j<subs[i].arity(); j++) {
	        newSubs.add(subs[i].sub(j));	        
	    }   
	    newSubs.add(subs[i+1]);	  
	}
	newSubs.add(subs[subs.length-1]);
	return tf.createUpdateTerm(loc, (Term[])newSubs.toArray(new Term[0]));
    }

    public void setUp() {
	testSort0 = new ClassInstanceSortImpl(new Name("testSort0"));
	testSort1 = new ClassInstanceSortImpl(new Name("testSort1"), testSort0);
 	testSort2 = new ClassInstanceSortImpl(new Name("testSort2"), testSort0);
	cloneable =  new ClassInstanceSortImpl(new Name("cloneable"), testSort1);

	KeYJavaType kjt = 
	    new KeYJavaType(new ClassDeclaration(new ProgramElementName("Object"), 
						 new ProgramElementName("java.lang.Object")), 
			    testSort1);
	sorts.add(testSort0);
	sorts.add(testSort1);
	sorts.add(testSort2);
	sorts.add(cloneable);

	pv = new ProgramVariable[7];
	for (int i=0; i<pv.length; i++) {	    
	    ProgramElementName name;
	    switch (i) {
	    case 1: name = new ProgramElementName("t");
		break;
	    case 2: name = new ProgramElementName("i");
		break;
	    case 3: name = new ProgramElementName("o");
		break;
	    case 4: name = new ProgramElementName("u");
		break;
	    case 5: name = new ProgramElementName("a");
		break;
	    default:
		name = new ProgramElementName("pv"+i);
		break;
	    } 
	    pv[i] = new ProgramVariable(name, kjt);
	    variables.add(pv[i]);	    
	}

	spv = new ProgramVariable(new ProgramElementName("spv"), kjt, kjt, false, true);

	// just initialize the parser
 	parseTerm("{t:=i} o");
	// for the systematic tests

	t  = tf.createFunctionTerm(pv[1]);
	i  = tf.createFunctionTerm(pv[2]);
	o  = tf.createFunctionTerm(pv[3]);
	u  = tf.createFunctionTerm(pv[4]);
	ProgramVariable r_var = 
	    new ProgramVariable(new ProgramElementName("r"),
				testSort2);	
	r  = tf.createFunctionTerm(r_var);

	oa = tf.createAttributeTerm(pv[5], o);
	ua = tf.createAttributeTerm(pv[5], u);
	ra = tf.createAttributeTerm(pv[5], r);
	

	ospv = tf.createAttributeTerm(spv, o);
	uspv = tf.createAttributeTerm(spv, u);
	rspv = tf.createAttributeTerm(spv, r);
	
	arraySort1  = ArraySortImpl.getArraySort(testSort1, testSort0, cloneable);  
	arraySort2  = ArraySortImpl.getArraySort(testSort2, testSort0, cloneable); 

	   
	ProgramVariable a_var = 
	    new ProgramVariable(new ProgramElementName("_a"),
				arraySort1);
	ProgramVariable b_var = 
	    new ProgramVariable(new ProgramElementName("_b"),
				arraySort1);
	ProgramVariable c_var = 
	    new ProgramVariable(new ProgramElementName("_c"),
				arraySort2);

	a  = tf.createFunctionTerm(a_var);
	b  = tf.createFunctionTerm(b_var);
	c  = tf.createFunctionTerm(c_var);

	integerSort = ProofSettings.THE_SETTINGS.getLDTSettings().
	    getNaturalNumberLDT().targetSort();
	ProgramVariable idx_var = 
	    new ProgramVariable(new ProgramElementName("i"),
				integerSort);
	ProgramVariable jdx_var = 
	    new ProgramVariable(new ProgramElementName("j"),
				integerSort);

	ProgramVariable mdx_var = 
	    new ProgramVariable(new ProgramElementName("m"),
				integerSort);

	idx  = tf.createFunctionTerm(idx_var);
	jdx  = tf.createFunctionTerm(jdx_var);
	mdx  = tf.createFunctionTerm(mdx_var);


	ai = tf.createArrayTerm(ArrayOp.getArrayOp(a.sort()), a, idx);
	aj = tf.createArrayTerm(ArrayOp.getArrayOp(a.sort()), a, jdx);
	am = tf.createArrayTerm(ArrayOp.getArrayOp(a.sort()), a, mdx);

	bi = tf.createArrayTerm(ArrayOp.getArrayOp(b.sort()), b, idx);
	bj = tf.createArrayTerm(ArrayOp.getArrayOp(b.sort()), b, jdx);
	bm = tf.createArrayTerm(ArrayOp.getArrayOp(b.sort()), b, mdx);
	
	ci = tf.createArrayTerm(ArrayOp.getArrayOp(c.sort()), c, idx);
	cj = tf.createArrayTerm(ArrayOp.getArrayOp(c.sort()), c, jdx);	
	cm = tf.createArrayTerm(ArrayOp.getArrayOp(c.sort()), c, mdx);	
    }


    private Term parseTerm(String termstr) {
	return TacletForTests.parseTerm(termstr, 
					new NamespaceSet
					    (new Namespace(), functions, 
					     sorts, new Namespace(), 
					     new Namespace(), variables));
    }
    
    public void testBasicRules() {
	UpdateSimplifier simply=new UpdateSimplifier();

 	Term parsed = parseTerm("{t:=i} o");
	assertTrue(simply.simplify(parsed) == parsed.sub(parsed.arity()-1));

	parsed=parseTerm("{t:=i} t");
	assertTrue(simply.simplify(parsed)==parsed.sub(parsed.arity()-2));
	

	parsed=parseTerm("{t:=i, i:=o} t");
	assertTrue(simply.simplify(parsed).op()==pv[2]);
	

	Term[] subs=new Term[parsed.arity()];
	for (int i=0;i<parsed.arity()-1;i++) {
	    subs[i]=parsed.sub(i);
	}
	subs[parsed.arity()-1]=tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[4]));
		    
	// {t:=i, i:=o} u.a ~~~> u.a
	Term constr     = createUpdateTerm(new Term[]{t, i, i, o, ua});
	Term simplified = simply.simplify(constr);
	Term expected   = ua;
	assertEquals("Failed applying {t:=i, i:=o} u.a", 
		     expected, simplified);
	assertSame ("Failed applying  {t:=i, i:=o} u.a (wasted memory)", 
		    expected, simplified);
	
	// {i:=t, o.a:=i} u.a ~~~>(u?=o).a/->i
	constr  = createUpdateTerm(new Term[]{i, t, oa, i, ua});
	simplified = simply.simplify(constr);
	expected = tf.createIfElseTerm(u, o, i, ua);

	assertEquals("Failed applying {i:=t, o.a:=i} u.a",
		     expected, simplified);

	assertSame("Failed applying {i:=t, o.a:=i} u.a (memory wasted)",
		   u, simplified.sub(0));
	assertSame("Failed applying {i:=t, o.a:=i} u.a (memory wasted)",
		   o, simplified.sub(1));
	assertSame("Failed applying {i:=t, o.a:=i} u.a (memory wasted)",
		   i, simplified.sub(2));
	assertSame("Failed applying {i:=t, o.a:=i} u.a (memory wasted)",
		   ua, simplified.sub(3));
		    		   	
	// {o.a:=u, u.a:=i} t.a ---> (t ?= u) i : (t ?= o) u : t.a
	Term ta  = tf.createAttributeTerm(pv[5], t);
	constr   = createUpdateTerm(new Term[]{oa, u, ua, i, ta});
	simplified = simply.simplify(constr);
	expected = tf.createIfElseTerm
	    (t, u, i, tf.createIfElseTerm(t, o, u, ta));

	assertEquals("Failed applying {o.a:=u, u.a:=i} t.a",
		     expected, simplified);
	for (int i = 0; i<expected.arity(); i++) {
	    if (expected.sub(i).op() != Op.IF_ELSE) {
		assertSame("Memory waste detected", expected.sub(i),
			   simplified.sub(i));
	    }
	}

	// {t:=i} {i:=o, o:=t}<>true ~~> {t:=i, i:=o, o:=i}<>true
	parsed     = parseTerm("{t:=i} {i:=o, o:=t} <{}>true");
	expected   = parseTerm("{t:=i, i:=o, o:=i} <{}>true");
	simplified = simply.simplify(parsed);
	assertEquals("Failed applying {t:=i} {i:=o, o:=t}<>true",
		     expected, simplified);


	// {t:=i} {t:=o, o:=t}<>true ~~> {t:=o, o:=i}<>true
	parsed     = parseTerm("{t:=i} {t:=o, o:=t} <{}>true");
	expected   = parseTerm("{t:=o, o:=i} <{}>true");
	simplified = simply.simplify(parsed);
	assertEquals("Failed applying {t:=i} {t:=o, o:=t}<>true",
		     expected, simplified);
	
	// {i.a:=t, t.a:=i, u.a:=o} t.a = t ~~> 
	// ((t ?= u) o : i) = t
	subs = new Term[7];
	subs[0] = tf.createAttributeTerm(pv[5], 
					 tf.createFunctionTerm(pv[2])); 
	subs[1] = tf.createFunctionTerm(pv[1]); 
	subs[2] = tf.createAttributeTerm(pv[5], 
					 tf.createFunctionTerm(pv[1])); 
	subs[3] = tf.createFunctionTerm(pv[2]); 
	subs[4] = tf.createAttributeTerm(pv[5], 
					 tf.createFunctionTerm(pv[4])); 
	subs[5] = tf.createFunctionTerm(pv[3]); 
	subs[6] = tf.createEqualityTerm
	    (tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[1])),
	     tf.createFunctionTerm(pv[1]));

	constr  = createUpdateTerm(subs);
        expected = parseTerm("(t ?= u) o : i");

	Term o5 = tf.createAttributeTerm(pv[5], 
					   tf.createFunctionTerm(pv[3]));
	Term[] e_subs = new Term[expected.arity()];
	for (int i = 0; i < expected.arity(); i++) {
	    e_subs[i] = expected.sub(i);
	}	

	expected = tf.createEqualityTerm
	    (tf.createIfElseTerm(Op.IF_ELSE, e_subs),
	     tf.createFunctionTerm(pv[1]));
					 

	simplified = simply.simplify(constr);


	assertTrue("Expected:" + expected + ", but is:" + simplified,
		   simplified.equals(expected));
	

    }


    public void testApplyOnAttribute() {
	UpdateSimplifier simply=new UpdateSimplifier();
	// none-static
	// {o.a := pv6, t.a:=i} u.a ~~~>(u?=t) i : (u ?= o) pv6 : u.a
	Term ta = tf.createAttributeTerm(pv[5], t);
	Term pv6     = tf.createFunctionTerm(pv[6]);
	Term constr  = createUpdateTerm(new Term[]{oa, pv6, ta, i, ua});
	Term simplified = simply.simplify(constr);


	assertTrue("If-Else Operator expected, but is "+simplified.op(), 
		   simplified.op() == Op.IF_ELSE);

	assertTrue("Expected u, but is: "+simplified.sub(0).op(),
		   simplified.sub(0).op()==pv[4]);

	assertTrue("Expected t, but is: " + simplified.sub(1).op(),
		   simplified.sub(1).op()==pv[1]);

	assertTrue("Expected i, but is: "+simplified.sub(2).op(),
		   simplified.sub(2).op()==pv[2]);

	assertTrue("If-Else Operator expected, but is "+simplified.sub(3).op(), 
		   simplified.sub(3).op() == Op.IF_ELSE);
	
	assertTrue("Expected u, but is: "+simplified.sub(3).sub(0).op(),
		   simplified.sub(3).sub(0).op()==pv[4]);

	assertTrue("Expected o, but is: "+simplified.sub(3).sub(1).op(),
		   simplified.sub(3).sub(1).op()==pv[3]);

	assertTrue("Expected pv6, but is: "+simplified.sub(3).sub(2).op(),
		   simplified.sub(3).sub(2).op()==pv[6]);

	assertTrue("Expected u.a, but is: "+simplified.sub(3).sub(3),
		   simplified.sub(3).sub(3).op() instanceof AttributeOp &&
		   ((AttributeOp)simplified.sub(3).sub(3).op()).attribute() == pv[5] &&
		   simplified.sub(3).sub(3).sub(0).op()==pv[4]);

    }
    
    
    public void testDeletionStrategy() {
	
        UpdateSimplifier us = new UpdateSimplifier(true);
	
	Term parsed = parseTerm("{t:=i} <{}> o=o");
	Term expected = parseTerm("<{}>o=o");
	Term result = us.simplify(parsed);
	assertTrue("Expected:" + expected + "\n Is:" + result, 
	       result == parsed.sub(parsed.arity()-1));
	
	parsed = parseTerm("{t:=i, o:=a}<{}>t=t");	
	result = us.simplify(parsed);
	expected = parseTerm("{t:=i}<{}> t=t");
	assertTrue("Expected: " + expected + "\n Is: " + result, 
	       result.equals(expected));
	
	
	parsed = parseTerm("{t:=i, o:=a, u:=u}<{}> t=t");	
	result = us.simplify(parsed);
	expected = parseTerm("{t:=i}<{}>t=t");
	assertTrue("Expected: " + expected + "\n Is: " + result, 
	       result.equals(expected));


	parsed = parseTerm("{t:=i, o:=a, u:=u}<{}>u=u");	
	result = us.simplify(parsed);
	expected = parseTerm("<{}>u=u");
	assertTrue("Expected: " + expected + "\n Is: " + result, 
	       result.equals(expected));
	
	        
	// {t:=i, o:=a, u:=u}<{ o = o; }> t=t --> 
	//               {t:=i, o:=a} <{ o = o; }> t=t
	parsed = parseTerm("{t:=i, o:=a, u:=u}<{ o = o; }> t=t");	
 	result = us.simplify(parsed);
	expected = parseTerm("{t:=i, o:=a}<{ o = o; }> t=t");

 	assertEquals("Failed deletion of unused var (or simple updates) in "+
		   "{t:=i, o:=a, u:=u}<{ o = o; }> t=t", 
		     expected, result);

 		
	parsed = parseTerm("{t:=i, o:=a, u:=o} <{}> t=u");	
 	result = us.simplify(parsed);
	expected = parseTerm("{t:=i, u:=o}<{}> t=u");	
 	assertEquals
	    ("Failed deletion of unused var in "+
	     "{t:=i, o:=a, u:=o} t=u",
	     expected, result);
	
    }
    
    public void testSimultaneousUpdateEquality() {
	Term cmp1 = parseTerm("{t:=i, o:=a, u:=o} true");	
	Term cmp2 = parseTerm("{o:=a, u:=o, t:=i} true");	

	assertTrue("ProgramVariables commute.", cmp1.equals(cmp2));
    }


    public void testApplicationOnAttributeNoneSim() {

	UpdateSimplifier simply=new UpdateSimplifier();

	Term loc1 = tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[4]));
	Term loc2 = tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[3]));
	Term val  = tf.createFunctionTerm(pv[2]);

	// {p4.p5 := i} {p3.p5 := i} (p4.p5 = p3.p5)

	Term constr     = createUpdateTerm
	    (new Term[]{loc1, val,
	     createUpdateTerm
	    (new Term[]{loc2, val, tf.createEqualityTerm(loc2, loc1)})});
	Term simplified = simply.simplify(constr);
	Term expected   = tf.createEqualityTerm(val, val);						
	assertEquals("Error applying non-simultaneous updates on attributes.", 
		     expected, simplified);
    }

    public void testApplicationOnAttributeSim() {

	UpdateSimplifier simply=new UpdateSimplifier();

	Term loc1 = tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[4]));
	Term loc2 = tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[3]));
	Term val  = tf.createFunctionTerm(pv[2]);

	// {u.a := i} {p3.a := i} (p4.p5 = p3.a)

	Term constr     = createUpdateTerm
	    (new Term[]{loc1, val, loc2, val, tf.createEqualityTerm(loc2, loc1)});
	Term simplified = simply.simplify(constr);
	Term expected   = tf.createEqualityTerm(val, val);						
	assertEquals("Error applying simultaneous update on attributes.", 
		     expected, simplified);
    }

    public void testBugInUStarComputation() {
	UpdateSimplifier simply=new UpdateSimplifier();

	Term p4p5 = tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[4]));
	Term p1   = tf.createFunctionTerm(pv[1]);
	Term p2   = tf.createFunctionTerm(pv[2]);

	// {t:=u.a, u.a := i} {u.a := t} <>(i=i)

	Term constr     = createUpdateTerm
	    (new Term[]{p1, p4p5, p4p5, p2,  
			createUpdateTerm
			(new Term[]{p4p5, p1,
			tf.createDiamondTerm(JavaBlock.createJavaBlock(new StatementBlock()), 
					     tf.createEqualityTerm(p2, p2))})});
	Term simplified = simply.simplify(constr);
	Term expected   = createUpdateTerm
	    (new Term[]{p1, p4p5, p4p5, p4p5,
			tf.createDiamondTerm(JavaBlock.createJavaBlock(new StatementBlock()), 
					     tf.createEqualityTerm(p2, p2))});
	assertEquals("Error when merging updates.", expected, simplified);	
    }

    public void xtestBugInDeleteTrivialUpdates() {
	// deletion of updates has been wrong for the folowing case:
	//   {o1.a:=c; o2.a:=o2.a} phi
        // previously o2.a:=o2.a has been deleted, but that is wrong, if o1=o2
       
	UpdateSimplifier simply=new UpdateSimplifier();

	Term p3p5 = tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[3]));
	Term p4p5 = tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[4]));
	Term p1   = tf.createFunctionTerm(pv[1]);
	Term p2   = tf.createFunctionTerm(pv[2]);

	// {o.a:=t, u.a := u.a} <>(t=i)

	Term constr     = createUpdateTerm
	    (new Term[]{p3p5, p1, p4p5, p4p5,  
			tf.createDiamondTerm(JavaBlock.createJavaBlock(new StatementBlock()), 
					     tf.createEqualityTerm(p2, p2))});

	Term simplified = simply.simplify(constr);
	Term expected   = constr;
	assertEquals("Trivial updates may only be deleted if it is safe.", expected, simplified);	
    }

    // more systematic tests
    
    /**
     *  tests the application of <ul>
     *    <li> {o := t} o   </li> 
     *    <li> {o := t} u   </li> 
     *    <li> {o := t} o.a </li> 
     *    <li> {o := t} u.a </li> 
     *    <li> {o := t} r.a </li> 
     *    <li> {a := b} a[i] </li> 
     *    <li> {a := b} c[i] </li> 
     *    <li> {i := j} a[i] </li> 
     *    </ul>
     */
    public void testBaseLocalVariableApplications() {
	UpdateSimplifier simply=new UpdateSimplifier();

	// {o:=t} o
	Term constr     = createUpdateTerm(new Term[]{o, t, o});
	Term simplified = simply.simplify(constr);
	Term expected   = t;
	assertEquals("Failed applying {o := t} o ", expected,
		     simplified);		
	//     {o := t} u    
	constr     = createUpdateTerm(new Term[]{o, t, u});         	
	simplified = simply.simplify(constr);                            
	expected   = u;                                                  
	assertEquals("Failed applying {o := t} u (o,u compatible) ", 
		     expected, simplified);	

	//     {o := t} r    
	constr     = createUpdateTerm(new Term[]{o, t, r});         	
	simplified = simply.simplify(constr);                            
	expected   = r;
	assertEquals("Failed applying {o := t} r (o, r not compatible) ", 
		     expected, simplified);	

	//     {o := t} o.a  
	constr     = createUpdateTerm(new Term[]{o, t, oa});         	
	simplified = simply.simplify(constr);                            
	expected   = tf.createAttributeTerm(pv[5], t);
	assertEquals("Failed applying {o := t} o.a", 
		     expected, simplified);	

	//     {o := t} u.a  
	constr     = createUpdateTerm(new Term[]{o, t, ua});         	
	simplified = simply.simplify(constr);                            
	expected   = ua;                                                  
	assertEquals("Failed applying {o := t} u.a (o, u compatible) ", 
		     expected, simplified);	
	//     {o := t} r.a  
	constr     = createUpdateTerm(new Term[]{o, t, ra});         	
	simplified = simply.simplify(constr);                            
	expected   = ra;
	assertEquals("Failed applying {o := t} r.a (o, r not compatible) ", 
		     expected, simplified);	

	//     {a := b} a[i]  
	constr     = createUpdateTerm(new Term[]{a, b, ai});         	
	simplified = simply.simplify(constr);                            
	expected   = bi;
	assertEquals("Failed applying {a := b} a[i] ", 
		     expected, simplified);	

	//     {a := b} b[i]  
	constr     = createUpdateTerm(new Term[]{a, b, ai});         	
	simplified = simply.simplify(constr);                            
	expected   = bi;
	assertEquals("Failed applying {a := b} b[i] (a, b compatible)", 
		     expected, simplified);	

	//     {a := b} c[i]  
	constr     = createUpdateTerm(new Term[]{a, b, ci});         	
	simplified = simply.simplify(constr);                            
	expected   = ci;
	assertEquals("Failed applying {a := b} c[i] (a, c not compatible) ", 
		     expected, simplified);	

	//     {i := j} a[i]  
	constr     = createUpdateTerm(new Term[]{idx, jdx, ai});         	
	simplified = simply.simplify(constr);                            
	expected   = aj;
	assertEquals("Failed applying {i := j} a[i] (o, r not compatible) ", 
		     expected, simplified);	


    }

    /**
     *  tests the application of <ul>
     *    <li> {o.a := t} o.a </li> 
     *    <li> {o.a := t} u.a (u compatible to o)     </li> 
     *    <li> {o.a := t} r.a (r not compatible to o) </li> 
     *    <li> {o.a := t} i    </li> 
     *    <li> {o.a := t} a[i] </li> 
     *    </ul>
     */
    public void testBaseAttributeApplications() {
	UpdateSimplifier simply=new UpdateSimplifier();


	// {o.a:=t} o.a
	Term constr     = createUpdateTerm(new Term[]{oa, t, oa});
	Term simplified = simply.simplify(constr);
	Term expected   = t;
	assertEquals("Failed applying {o.a := t} o.a ", expected, simplified);	


	// {o.a:=t} u.a
	constr     = createUpdateTerm(new Term[]{oa, t, ua});
	simplified = simply.simplify(constr);
	expected   = tf.createIfElseTerm(u, o, t, ua);
	assertEquals("Failed applying {o.a := t} u.a (o, u compatible) ", 
		     expected, simplified);		

	// {o.a:=t} r.a
	constr     = createUpdateTerm(new Term[]{oa, t, ra});
	simplified = simply.simplify(constr);
	expected   = ra;
	assertEquals("Failed applying {o.a := t} r.a (o, r not compatible) ", 
		     expected, simplified);		

	// {o.a:=t} i
	constr     = createUpdateTerm(new Term[]{oa, t, i});
	simplified = simply.simplify(constr);
	expected   = i;
	assertEquals("Failed applying {o.a := t} i ", expected, simplified);		

	// {o.a:=t} a[i]
	constr     = createUpdateTerm(new Term[]{oa, t, ai});
	simplified = simply.simplify(constr);
	expected   = ai;
 	assertEquals("Failed applying {o.a := t} a[i] ", expected, simplified);

	// {o.a:=t} a[i]
	constr     = createUpdateTerm(new Term[]{oa, t, ai});
	simplified = simply.simplify(constr);
	expected   = ai;
 	assertEquals("Failed applying {o.a := t} a[i] ", expected, simplified);	
    }
    


    /**
     *  tests the application of <ul>
     *    <li> {a[i] := t} a[i] </li> 
     *    <li> {a[i] := t} a[j] </li> 
     *    <li> {a[i] := t} b[i] (a compatible to b)     </li> 
     *    <li> {a[i] := t} b[j] (a compatible to b)     </li> 
     *    <li> {a[i] := t} c[j] (a not compatible to c)     </li> 
     *    <li> {a[i] := t} o.a </li> 
     *    <li> {a[i] := t} i   </li> 
     *    </ul>
     */
    public void testBaseArrayApplications() {
	UpdateSimplifier simply=new UpdateSimplifier();

	// {a[i]:=t} a[i]
	Term constr     = createUpdateTerm(new Term[]{ai, t, ai});
	Term simplified = simply.simplify(constr);
	Term expected   = t;
 	assertEquals("Failed applying {a[i] := t} a[i] ", expected, simplified);

	// {a[i]:=t} a[j]
	constr     = createUpdateTerm(new Term[] {ai, t, aj});
	simplified = simply.simplify(constr);
	expected   = tf.createConjCondTerm(new Term[]{a, jdx, aj,
						      a, idx, t});
 	assertEquals("Failed applying {a[i] := t} a[j] ", expected, simplified);

	// {a[i]:=t} b[i]
	constr     = createUpdateTerm(new Term[] {ai, t, bi});
	simplified = simply.simplify(constr);
	expected   = tf.createConjCondTerm(new Term[]{b, idx, bi,
						      a, idx, t});
 	assertEquals("Failed applying {a[i] := t} b[i] " + 
		     "(a, b compatible) ", 
		     expected, simplified);

	// {a[i]:=t} b[j]
	constr     = createUpdateTerm(new Term[] {ai, t, bj});
	simplified = simply.simplify(constr);
	expected   = tf.createConjCondTerm(new Term[]{b, jdx, bj,
						      a, idx, t});
 	assertEquals("Failed applying {a[i] := t} b[i] " + 
		     "(a, b compatible) ", 
		     expected, simplified);

	// {a[i]:=t} c[j]
	constr     = createUpdateTerm(new Term[] {ai, t, cj});
	simplified = simply.simplify(constr);
	expected   = cj;
 	assertEquals("Failed applying {a[i] := t} c[j] " + 
		     "(a, c not compatible) ", 
		     expected, simplified);

	// {a[i]:=t} o.a
	constr     = createUpdateTerm(new Term[] {ai, t, oa});
	simplified = simply.simplify(constr);
	expected   = oa;
 	assertEquals("Failed applying {a[i] := t} o.a ", 
		     expected, simplified);

	// {a[i]:=t} t
	constr     = createUpdateTerm(new Term[] {ai, t, t});
	simplified = simply.simplify(constr);
	expected   = t;
 	assertEquals("Failed applying {a[i] := t} t ", 
		     expected, simplified);

	// {a[i]:=t} a
	constr     = createUpdateTerm(new Term[] {ai, t, a});
	simplified = simply.simplify(constr);
	expected   = a;
 	assertEquals("Failed applying {a[i] := t} a ", 
		     expected, simplified);
    }


    /**
     *  tests the application of <ul>
     *    <li> {a := b} {a[i] := o} <> true </li> 
     *    <li> {i := j} {a[i] := o} <> true </li> 
     *    <li> {a[i] := t} {a[i] := o} <> true </li> 
     *    <li> {a[i] := t} {b[i] := o} <> true (a, b compatible) </li> 
     *    <li> {a[i] := t} {c[i] := o} <> true (a, c not compatible)</li> 
     *    <li> {a[i] := t} {o := a[i]} <> true </li>
     *    <li> {a[i] := t} {o := a[j]} <> true </li>
     *    <li> {o.a := t} {o.a := o  }   <> true </li> 
     *    <li> {o.a := t} {o.a.a := o}   <> true </li> 
     *    <li> {o.a := t} {o.a.a := o.a} <> true </li> 
     *    <li> {o.a := t} {u.a := u.a}   <> true </li> 
     *    <li> {o.a := t} {r.a := r.a}   <> true </li> 
     *    <li> {o.a := t} {r.a.a := o}   <> true </li> 
     *    </ul>
     */
    public void testMergeSingleUpdates() {
	UpdateSimplifier simply=new UpdateSimplifier();
	Term diaTrue = tf.createDiamondTerm
	    (JavaBlock.createJavaBlock
	     (new StatementBlock()), tf.createJunctorTerm(Op.TRUE));

	// {a := b} {a[i] := o} <> true
	Term constr     = createUpdateTerm
	    (new Term[]{a, b, 
			createUpdateTerm
			(new Term[]{ai, o, diaTrue})});
	Term simplified = simply.simplify(constr);
	Term expected   = createUpdateTerm
	    (new Term[]{a, b, bi, o, diaTrue});;
 	assertEquals("Failed applying {a := b} {a[i] := o} <> true", 
		     expected, simplified);

	// {i := j} {a[i] := o} <> true 
	constr     = createUpdateTerm
	    (new Term[]{idx, jdx, 
			createUpdateTerm
			(new Term[]{ai, o, diaTrue})});
	simplified = simply.simplify(constr);
	expected   = createUpdateTerm
	    (new Term[]{idx, jdx, aj, o, diaTrue});;
 	assertEquals("Failed applying {i := j} {a[i] := o} <> true", 
		     expected, simplified);
       
	// {a[i] := t} {a[i] := o} <> true
	constr     = createUpdateTerm
	    (new Term[]{ai, t, 
			createUpdateTerm
			(new Term[]{ai, o, diaTrue})});
	simplified = simply.simplify(constr);
	expected   = createUpdateTerm
	    (new Term[]{ai, o, diaTrue});;
 	assertEquals("Failed applying {i := j} {a[i] := o} <> true", 
		     expected, simplified);

	//  {a[i] := t} {b[i] := o} <> true (a, b compatible)
	constr     = createUpdateTerm
	    (new Term[]{ai, t, 
			createUpdateTerm
			(new Term[]{bi, o, diaTrue})});
	simplified = simply.simplify(constr);
	/* expected   = createUpdateTerm
	    (new Term[]{ai, 
			tf.createConjCondTerm(new Term[]{a, idx, t, b,
							 idx, ai}),
			bi, o, diaTrue});*/ // skipped this kind of improvement
	expected = createUpdateTerm(new Term[]{ai, t, bi, o, diaTrue});
    
 	assertEquals("Failed applying {a[i] := t} {b[i] := o} <> true"+
		     "(a,b compatible)", expected, simplified);

	// {a[i] := t} {c[i] := o} <> true (a, c not compatible)
	constr     = createUpdateTerm
	    (new Term[]{ai, t, 
			createUpdateTerm
			(new Term[]{ci, o, diaTrue})});
	simplified = simply.simplify(constr);
	expected   = createUpdateTerm
	    (new Term[]{ai, t, ci, o, diaTrue});;
 	assertEquals("Failed applying {a[i] := t} {c[i] := o} <> true"+
		     "(a, c not compatible)", expected, simplified);

	// {a[i] := t} {o := a[i]} <> true 
	constr     = createUpdateTerm
	    (new Term[]{ai, t, 
			createUpdateTerm
			(new Term[]{o, ai, diaTrue})});
	simplified = simply.simplify(constr);
	expected   = createUpdateTerm
	    (new Term[]{ai, t, o, t, diaTrue});;
 	assertEquals("Failed applying {a[i] := t} {o:=a[i]} <> true",
		     expected, simplified);

	// {a[i] := t} {o := a[j]} <> true 
	constr     = createUpdateTerm
	    (new Term[]{ai, t, 
			createUpdateTerm
			(new Term[]{o, aj, diaTrue})});
	simplified = simply.simplify(constr);
	expected   = createUpdateTerm
	    (new Term[]{ai, t, o, 
			tf.createConjCondTerm(new Term[]{a, jdx, aj, a, idx, t})
			, diaTrue});
 	assertEquals("Failed applying {a[i] := t} {o:=a[j]} <> true",
		     expected, simplified);

	//     {o.a := t} {o.a := o}   <> true  
	constr     = createUpdateTerm
	    (new Term[]{oa, t, 
			createUpdateTerm
			(new Term[]{oa, o, diaTrue})});
	simplified = simply.simplify(constr);
	expected   = createUpdateTerm
	    (new Term[]{oa, o, diaTrue});
 	assertEquals("Failed applying  {o.a := t} {o.a := o}<> true", 
		     expected, simplified);

	//     {o.a := t} {o.a.a := o}   <> true 
	Term oaa = tf.createAttributeTerm(pv[5], oa);
	Term ta = tf.createAttributeTerm(pv[5], t);
	constr     = createUpdateTerm
	    (new Term[]{oa, t, 
			createUpdateTerm
			(new Term[]{oaa, o, diaTrue})});
	simplified = simply.simplify(constr);
	/* expected   = createUpdateTerm
	    (new Term[]{oa, tf.createIfElseTerm(o,t,oa, t), 
			ta, o, diaTrue});*/ //skipped this improvement
	expected = createUpdateTerm
	(new Term[]{oa, t, ta, o, diaTrue});
 	assertEquals("Failed applying {o.a := t} {o.a.a := o}<> true ", 
		     expected, simplified);

	//     {o.a := t} {o.a.a := o.a} <> true 
	constr     = createUpdateTerm
	    (new Term[]{oa, t, 
			createUpdateTerm
			(new Term[]{oaa, oa, diaTrue})});
	simplified = simply.simplify(constr);
	/*	expected   = createUpdateTerm
	    (new Term[]{oa, tf.createIfElseTerm(o,t,oa, t), ta, t, diaTrue});
        */  // this "optimization" is no longer performed
	expected = createUpdateTerm
        (new Term[]{oa, t, ta, t, diaTrue});
 	assertEquals
	    ("Failed applying {o.a := t} {o.a.a := o.a} <> true", 
	     expected, simplified);
	// {o.a := t} {u.a := u.a}   <> true 
	constr     = createUpdateTerm
	    (new Term[]{oa, t, 
			createUpdateTerm
			(new Term[]{ua, ua, diaTrue})});
	simplified = simply.simplify(constr);
	/*expected   = createUpdateTerm
	    (new Term[] 
		{ oa, tf.createIfElseTerm
		  (o,u,oa,t), ua, tf.createIfElseTerm
		  (u,o,t,ua), diaTrue });*/
        expected =  createUpdateTerm
        (new Term[] 
            { oa, t, ua, tf.createIfElseTerm
              (u,o,t,ua), diaTrue });
 	assertEquals("Failed applying {o.a := t} {u.a := u.a} <> true"+
		     "(o, u compatible)", expected, simplified);

	// {o.a := t} {r.a := r.a}   <> true 
	constr     = createUpdateTerm
	    (new Term[]{oa, t, 
			createUpdateTerm
			(new Term[]{ra, ra, diaTrue})});
	simplified = simply.simplify(constr);
	expected   = createUpdateTerm
	    (new Term[] { oa, t, ra, ra, diaTrue });
 	assertEquals("Failed applying {o.a := t} {r.a := r.a} <> true"+
		     "(o, r bot compatible)", expected, simplified);
	
	// {o.a := t} {r.a.a := o}   <> true	
	Term raa = tf.createAttributeTerm(pv[5], ra);
	constr     = createUpdateTerm
	    (new Term[]{oa, t, 
			createUpdateTerm
			(new Term[]{raa, o, diaTrue})});
	simplified = simply.simplify(constr);
/*	expected   = createUpdateTerm
	    (new Term[] { oa, tf.createIfElseTerm(o, ra, oa, t), 
			  raa, o, diaTrue });*/ // this "optimisation" is no longer performed
	expected = createUpdateTerm
            (new Term[] { oa, t, raa, o, diaTrue });
 	assertEquals("Failed applying {o.a := t} {r.a.a := o} <> true"+
		     "(o, r not compatible)", expected, simplified);

    }

    /**
     *  tests the application of <ul>
     *    <li> {o.spv:=u, t.spv:=i} u.spv </li> 
     *    </ul>
     */
    public void testStaticAttributes() {
	UpdateSimplifier simply=new UpdateSimplifier();
	// {o.spv:=u, t.spv:=i} u.spv ~~~> i
	Term tspv = tf.createAttributeTerm(spv, t);
 	Term constr = createUpdateTerm
	    (new Term[]{ospv, u, tspv, i, uspv});
 	Term simplified = simply.simplify(constr);
	Term expected   = i;
 	assertSame("Failed applying {o.spv:=pv6, t.spv:=i} u.spv", 
 		   expected, simplified);
    }


    /**
     *  tests the application of <ul>
     *    <li> {a[i] := t, a[j] :=u} a[i] </li> 
     *    <li> {a[j] := t, a[i] :=u} a[i] </li> 
     *    <li> {a[j] := t, a[i] := o, a[m] := u} a[i] </li>
     *    <li> {a[i] := u, a[m] := u} a[i] </li>
     *    </ul>
     */
    public void testSimultaneousArrayApplications() {
	UpdateSimplifier simply=new UpdateSimplifier();

	// {a[i] := t, a[j] :=u} a[i]
	Term constr     = createUpdateTerm(new Term[]{ai, t, aj, u, ai});
	Term simplified = simply.simplify(constr);
	Term expected   = tf.createConjCondTerm(new Term[]
	    {a, idx, t, a, jdx, u});
 	assertEquals("Failed applying {a[i] := t, a[j] :=u} a[i]", 
		     expected, simplified);

	// {a[j] := t, a[i] := u} a[i]
	constr     = createUpdateTerm(new Term[] {aj, t, ai, u, ai});
	simplified = simply.simplify(constr);
	expected   = u;
 	assertEquals("Failed applying {a[j] := t, a[i] := u} a[i]", expected, simplified);

	// {a[j] := t, a[m] := u} a[i]
	constr     = createUpdateTerm(new Term[] {aj, t, am, u, ai});
	simplified = simply.simplify(constr);
	expected   = tf.createConjCondTerm(new Term[]
	    {a, idx, tf.createConjCondTerm(new Term[]{
		a, idx, ai, a, jdx, t}), 
	     a, mdx, u});
 	assertEquals("Failed applying {a[j] := t, a[m] := u} a[i]", expected, simplified);

	// {a[I] := u, a[m] := u} a[i]
	// important to check simplification of conj cond
	constr     = createUpdateTerm(new Term[] {ai, u, am, u, ai});
	simplified = simply.simplify(constr);
	expected   = u;
 	assertEquals("Failed applying {a[i] := u, a[m] := u} a[i]", expected, simplified);

    }

    /**
     *  tests the application of updates on more complex terms.
     * In particular it is tested if conditional terms are simplified as
     * far as possible.
     * <ul>
     *  <li> {a[i] := u, c[j] :=u} (a * j : a[i] ?= b * m : c[j]) </li>
     *  <li> {a[i] := u, c[j] :=u} (a * j : a[i] ?= b * m : c[j]) </li>
     *  <li> {i := j} (a * j : a[i] ?= a * i : c[j]) </li>
     *  <li> {o.a := u, r.a :=u} (o ?= u) o.a : r.a </li>
     *  <li> {o := u} (o ?= u) u.a : o.a </li>
     * </ul>
     */
    public void testSimultaneousUpdateApplicationOnComposedTerms() {
	UpdateSimplifier simply=new UpdateSimplifier();

	
	Term constr     = createUpdateTerm(new Term[]
	    {ai, u, cj, u, tf.createConjCondTerm
	     (new Term[]{a, jdx, ai, b, mdx, cj})});
	Term simplified = simply.simplify(constr);
	Term expected   = u;
 	assertEquals("Failed applying {a[i] := u, c[j] :=u}" +
		     "(a * j : a[i] ?= b * m : c[j])  ", 
		     expected, simplified);

	// {a[i] := u, c[j] :=t} (a * j : a[i] ?= b * m : c[j]) 
	constr     = createUpdateTerm(new Term[]
	    {ai, u, cj, t, tf.createConjCondTerm
	     (new Term[]{a, jdx, ai, b, mdx, cj})});
	simplified = simply.simplify(constr);
	expected   = tf.createConjCondTerm
	    (new Term[]{a, jdx, u, b, mdx, t});
 	assertEquals("Failed applying {a[i] := u, c[j] :=t}" +
		     "(a * j : a[i] ?= b * m : c[j])  ", 
		     expected, simplified);

	// {i := j} (a * j : a[i] ?= a * i : c[j]) 
	constr     = createUpdateTerm(new Term[]
	    {idx, jdx, tf.createConjCondTerm
	     (new Term[]{a, jdx, ai, a, jdx, cj})});
	simplified = simply.simplify(constr);
	expected   = cj;
 	assertSame("Failed applying {i := j} (a * j : a[i] ?= a * i : c[j]) ",
		   expected, simplified);

	// {o.a := u, r.a :=u} (o ?= u) o.a : r.a 
	constr     = createUpdateTerm(new Term[]
	    {oa, u, ra, u, tf.createIfElseTerm
	     (o, u, oa, ra)});
	simplified = simply.simplify(constr);
	expected   = u;
 	assertSame("Failed applying  {o.a := u, r.a :=u} (o ?= u) o.a : r.a ",
		   expected, simplified);
	// {o := u} (o ?= u) u.a : o.a 
	constr     = createUpdateTerm(new Term[]
	    {o, u, tf.createIfElseTerm
	     (o, u, ua, oa)});
	simplified = simply.simplify(constr);
	expected   = ua;
 	assertSame("Failed applying {o := u} (o ?= u) u.a : o.a ",
		   expected, simplified);

	// {o := t} (o ?= u) u.a : o.a 
	Term ta = tf.createAttributeTerm(pv[5], t);
	constr     = createUpdateTerm(new Term[]
	    {o, t, tf.createIfElseTerm
	     (o, u, ua, oa)});
	simplified = simply.simplify(constr);
	expected   = tf.createIfElseTerm(t, u, ua, ta);
 	assertEquals("Failed applying {o := t} (o ?= u) u.a : o.a ",
		     expected, simplified);


	// {r := t} (o ?= u) u.a : o.a 
	constr     = createUpdateTerm(new Term[]
	    {r, t, tf.createIfElseTerm(o, u, ua, oa)});
	simplified = simply.simplify(constr);
	expected   = constr.sub(1);
 	assertSame("Failed applying {r := t} (o ?= u) u.a : o.a " +
		   "(r incompatible to o,u", expected, simplified);	
	
    }
    
    // new testing style here
    
    private static final HelperClassForTests helper = new HelperClassForTests();
    
    public static final String testRules = System.getProperty("key.home")+
        File.separator+"system"+File.separator+"proofExamples"+
        File.separator+"_testcase"+File.separator+"updatesimplification";
    
    
    public void testAttributeEvaluateSubsFirst() {
        ProofList proofList = helper.parse
            (new File(testRules+File.separator+"testAttributeRule1.key"));
        Term t = helper.extractProblemTerm(proofList.get(0));
        UpdateSimplifier us = new UpdateSimplifier(); 
        assertEquals("Evaluate attribute references under the update first", 
                t.sub(1), us.simplify(t.sub(0)));
    }
    
    
    public void testAttributeRule3() {
        ProofList proofList = helper.parse
            (new File(testRules+File.separator+"testAttributeRule3.key"));
        Term t = helper.extractProblemTerm(proofList.get(0));
        UpdateSimplifier us = new UpdateSimplifier(); 
        assertEquals(t.sub(1), us.simplify(t.sub(0)));
    }
    
    public void testAttributeRule4() {
        ProofList proofList = helper.parse
            (new File(testRules+File.separator+"testAttributeRule4.key"));
        Term t = helper.extractProblemTerm(proofList.get(0));
        UpdateSimplifier us = new UpdateSimplifier(); 
        assertEquals(t.sub(1), us.simplify(t.sub(0)));
    }
  
  
    public void testShadowedArraySimplificationRule() {
        ProofList proofList = helper.parse
            (new File(testRules+File.separator+"testShadowedArrayRule1.key"));
        Term t = helper.extractProblemTerm(proofList.get(0));
        UpdateSimplifier us = new UpdateSimplifier(); 
        assertEquals("Shadowed array are not aliased to " + 
                "their unshadowed version", t.sub(1), us.simplify(t.sub(0)));
    }
   
    public void testApplyArrayAccessOnShadowedArray() {
        ProofList proofList = helper.parse
            (new File(testRules+File.separator+"testShadowedArrayRule2.key"));
        Term t = helper.extractProblemTerm(proofList.get(0));
        UpdateSimplifier us = new UpdateSimplifier(true); 
        assertEquals("An array is aliased to " + 
                "its shadowed version", t.sub(1), us.simplify(t.sub(0)));
    }
    
    public void testApplyShadowedAttributeOnAttribute() {
        ProofList proofList = helper.parse
            (new File(testRules+File.separator+"testShadowedAttributeRule1.key"));
        Term t = helper.extractProblemTerm(proofList.get(0));
        UpdateSimplifier us = new UpdateSimplifier(); 
        assertEquals("Shadowed attributes are not aliased to " + 
                "their unshadowed version", t.sub(1), us.simplify(t.sub(0)));
    }
   
    public void testApplyAttributeOnShadowedAttribute() {
        ProofList proofList = helper.parse
            (new File(testRules+File.separator+"testShadowedAttributeRule2.key"));
        Term t = helper.extractProblemTerm(proofList.get(0));
        UpdateSimplifier us = new UpdateSimplifier(true); 
        assertEquals("An attribute is aliased to " + 
                "its shadowed version", t.sub(1), us.simplify(t.sub(0)));
    }
  
    public void testShadowOnShadowSameTransactionNr() {
        ProofList proofList = helper.parse
            (new File(testRules+File.separator+"testShadowOnShadowSameNr.key"));
        Term t = helper.extractProblemTerm(proofList.get(0));
        UpdateSimplifier us = new UpdateSimplifier(true); 
        assertEquals("Same number means shadows are aliased.", t.sub(1), 
                us.simplify(t.sub(0)));
    }
    
    
    public void testDeletion() {
        ProofList proofList = helper.parse
            (new File(testRules+File.separator+"testDeletion.key"));
        Term t = helper.extractProblemTerm(proofList.get(0));
        UpdateSimplifier us = new UpdateSimplifier(true); 
        assertEquals("Deletion is broken.", t.sub(1), 
                us.simplify(t.sub(0)));
    }
   
    
    public static void main(String[] args) {
	TestUpdateSimplifier tsus=new TestUpdateSimplifier("t");
	tsus.setUp();
        tsus.testDeletion();
// 	tsus.testBasicRules();
// 	tsus.testDeletionStrategy();
// 	tsus.testSimultaneousUpdateEquality();
// 	tsus.testApplyOnAttribute();
    }

   
}
