// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.rule.metaconstruct;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.expression.literal.*;
import de.uka.ilkd.key.java.statement.MethodFrame;
import de.uka.ilkd.key.java.visitor.ProgramContextAdder;
import de.uka.ilkd.key.java.declaration.ParameterDeclaration;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.sort.ProgramSVSort;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import de.uka.ilkd.key.rule.inst.ContextStatementBlockInstantiation;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.abstraction.Type;
import de.uka.ilkd.key.java.abstraction.ArrayType;
import de.uka.ilkd.key.java.statement.Break;
import de.uka.ilkd.key.java.statement.While;
import de.uka.ilkd.key.util.NotSupported;
import de.uka.ilkd.key.util.ExtList;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.ListIterator;

public class WhileInvRule extends AbstractMetaOperator {

    /** the outer label that is used to leave the while loop ('l1') */
    private final SchemaVariable outerLabel = 
	SchemaVariableFactory.createProgramSV(new ProgramElementName("outer_label"),
				       ProgramSVSort.LABEL, false);
    /** the inner label ('l2') */
    private final SchemaVariable innerLabel =
	SchemaVariableFactory.createProgramSV(new ProgramElementName("inner_label"),
				       ProgramSVSort.LABEL, false);
    /** list of the labels */
    private ListOfSchemaVariable instantiations  = null;

    /** list of breaks that lead to abrupt termination
     * of the loop to be transformed. Is initialised by
     * the method neededInstantiations that is invoked 
     * before calculate
     */
    private LinkedList breakList;

    /** The JavaInfo object which is handed over as
     * a parameter of calculate.
     */
    private JavaInfo javaInfo;
    private TypeConverter typeConv;
    private TermFactory tf;
    
    public WhileInvRule() {
	super(new Name("#whileInvRule"), 4);
    }


    /** Unlike other meta operators this one returns a formula
     * not a term.
     */
    public Sort sort(Term[] term) {
	return Sort.FORMULA;
    }


    /**
     * checks whether the top level structure of the given @link Term
     * is syntactically valid, given the assumption that the top level
     * operator of the term is the same as this Operator. The
     * assumption that the top level operator and the term are equal
     * is NOT checked.  
     * @return true iff the top level structure of
     * the @link Term is valid.
     */
    public boolean validTopLevel(Term term) {
	// a meta operator accepts almost everything
	return  term.arity()==arity();
    }


    /** calculates the resulting term. */
    public Term calculate(Term term, SVInstantiations svInst, Services services) {
	
	// some initialisations...
	javaInfo = services.getJavaInfo();
	tf = TermFactory.DEFAULT ;
	typeConv = services.getTypeConverter();
	ProgramElement body = ((StatementBlock)term.sub(0).javaBlock().program()).
	    getStatementAt(0);
	Term inv = term.sub(1);
	Term post = term.sub(2);
        JavaNonTerminalProgramElement context = 
	    (JavaNonTerminalProgramElement)
	    term.sub(3).javaBlock().program();		

	ArrayList stmnt = new ArrayList();
	ArrayList breakIfCascade = new ArrayList();
	
	KeYJavaType returnType = 
	    returnType((ProgramElement)context, body);

	ProgramVariable contFlag = KeYJavaASTFactory.localVariable
	    ("cont", javaInfo.getKeYJavaType("boolean"));
	ProgramVariable returnFlag = KeYJavaASTFactory.localVariable
	    ("rtrn", javaInfo.getKeYJavaType("boolean"));	 
	ProgramVariable breakFlag = KeYJavaASTFactory.localVariable
	    ("brk", javaInfo.getKeYJavaType("boolean"));	
	// xxx who to ensure that "exc" has not been used before??
	ProgramVariable excFlag = KeYJavaASTFactory.localVariable
	    ("exc", javaInfo.getKeYJavaType("boolean"));
	ProgramVariable excParam = KeYJavaASTFactory.localVariable
	    ("e", javaInfo.getKeYJavaType("java.lang.Throwable"));
	ProgramVariable thrownException = KeYJavaASTFactory.localVariable
	    ("thrownExc", javaInfo.getKeYJavaType("java.lang.Throwable"));		       
	ProgramVariable returnExpression = null;
	if (returnType != null)
	    returnExpression = KeYJavaASTFactory.localVariable
		("returnExpr",  returnType);	

	Term contFlagTerm = null;
	Term returnFlagTerm = null;
	Term breakFlagTerm = null;
	Term excFlagTerm = null;
	
	// end of initialisation............

	int breakCounter = 0;
	ListIterator it = breakList.listIterator(0);
	int numberOfBreaks = 0;
	while (it.hasNext()) {
	    BreakToBeReplaced b = (BreakToBeReplaced)it.next();
	    ProgramVariable newVar = KeYJavaASTFactory.localVariable
		("break_"+breakCounter++, 
		 javaInfo.getKeYJavaType("boolean"));
	    b.setProgramVariable(newVar);
	    stmnt.add(KeYJavaASTFactory.declare
		      (newVar, BooleanLiteral.FALSE,
		       javaInfo.getKeYJavaType("boolean")));
	    numberOfBreaks++;
	    Statement s;
	    if (b.getBreak().getLabel() != null) 
		s = KeYJavaASTFactory.breakStatement(b.getBreak().getLabel());
	    else 
		s = KeYJavaASTFactory.emptyStatement();
	    breakIfCascade.add(KeYJavaASTFactory.ifThen(newVar, s));
	}

	
	WhileInvariantTransformation w = 
	    new WhileInvariantTransformation((While)body, svInst,
					     (ProgramElementName)
					     svInst.getInstantiation(outerLabel),
					     (ProgramElementName)
					     svInst.getInstantiation(innerLabel),
					     contFlag, excFlag, excParam, 
					     thrownException, breakFlag, 
					     returnFlag, returnExpression,
					     breakList, services);
	w.start();

	ArrayList resultSubterms = new ArrayList();

	// normal case and continue
	if (w.continueOccurred()) {
	    stmnt.add(contFlagDecl(contFlag));
	    contFlagTerm = tf.createEqualityTerm
		(Op.EQUALS, 
		 typeConv.convertToLogicElement(contFlag), 
		 typeConv.getBooleanLDT().getTrueTerm());
	}
	
	// exception case
	resultSubterms.add(throwCase(excFlag, thrownException, 
				     post, svInst, context));
	
	// return case
	if (w.returnOccurred()) {
	    stmnt.add(returnFlagDecl(returnFlag, svInst));
	    returnFlagTerm = 
		typeConv.convertToLogicElement(returnFlag);
	    resultSubterms.add
		(returnCase(returnFlag, returnType,
			    returnExpression, post,
			    svInst,context));
	    
	    if (returnType != null) {
		stmnt.add(KeYJavaASTFactory.declare
			  (returnExpression, returnType));
	    }
	}


	// break case
	if (numberOfBreaks > 0) {
	    stmnt.add(breakFlagDecl(breakFlag));
	    breakFlagTerm = 
		typeConv.convertToLogicElement(breakFlag);
	    resultSubterms.add(breakCase(breakFlag, post, 
					 breakIfCascade,
					 svInst, context)); 
	}


	// we always catch all exceptions
	stmnt.add(KeYJavaASTFactory.declare
		  (excFlag, BooleanLiteral.FALSE,
		   javaInfo.getKeYJavaType("boolean")));
	excFlagTerm = 
	    typeConv.convertToLogicElement(excFlag);
	stmnt.add(KeYJavaASTFactory.declare
		  (thrownException,
		   javaInfo.getKeYJavaType("java.lang.Throwable"))); 


	resultSubterms.add
	    (normalCaseAndContinue(contFlagTerm, returnFlagTerm,
				   breakFlagTerm, excFlagTerm, inv));
	
	Term result = createLongJunctorTerm(Op.AND, resultSubterms); 

	stmnt.add(w.result());
	return tf.createProgramTerm
	    (Op.BOX, 
	     JavaBlock.createJavaBlock
	     (new StatementBlock
	      ((Statement[])stmnt.toArray(new Statement[0]))), 
	     result);
	
    }

    /** returns the schemavariables that are needed to transform the given
     * loop. The unwind-loop construct may need labels if unlabeled breaks 
     * and/or continues occur in the loop. Often there will be uninstantiated
     * Schemavariables in the loop that is why the found instantiations have to
     * be given.
     */
    public ListOfSchemaVariable neededInstantiations(ProgramElement originalLoop,
						     SVInstantiations svInst) {
	WhileInvariantTransformation w = 
	    new WhileInvariantTransformation(originalLoop, svInst);
	w.start();
	instantiations = SLListOfSchemaVariable.EMPTY_LIST;
	if (w.innerLabelNeeded()) {
	    instantiations = instantiations.prepend(innerLabel);
	}
	if (w.outerLabelNeeded()) {
	    instantiations = instantiations.prepend(outerLabel);
	}
	breakList = w.breakList();
	return instantiations;
    }

    //---------------------------------------------------------------
    // --- private helper methods to construct the result term
    //---------------------------------------------------------------


    private KeYJavaType returnType(ProgramElement root,
				   ProgramElement body) {
	KeYJavaType returnType = null;
	while (root != body && 
	       root instanceof NonTerminalProgramElement) {
	    if (root instanceof MethodFrame) {
		IProgramVariable result = ((MethodFrame)root).getProgramVariable();
		if (result != null)
		    returnType = result.getKeYJavaType();
		ProgramElement b = ((MethodFrame)root).getBody();
		if ((b instanceof NonTerminalProgramElement) &&
		    (((NonTerminalProgramElement)b).getChildCount()>0)) {
		    root = (ProgramElement)((NonTerminalProgramElement)b).getChildAt(0);
		}
		else
		    root = body;
	    } else if (((NonTerminalProgramElement)root).getChildCount()>0)
		root = (ProgramElement)((NonTerminalProgramElement)root).getChildAt(0);
	    else    
		root = body;
	}
	return returnType;
    }

    private Term createLongJunctorTerm(Junctor junctor, ArrayList terms) {
	if (terms.size() == 1)
	    return (Term)terms.get(0);
	else if (terms.size() == 2)
	    return tf.createJunctorTerm(junctor,
					(Term)terms.get(0),
					(Term)terms.get(1));
	else {
	    Term arg1 = (Term)terms.get(0);
	    terms.remove(0);
	    return 
		tf.createJunctorTerm(junctor, 
				     arg1, 
				     createLongJunctorTerm(junctor, terms));
	}
    }

    private  JavaBlock 
	addContext(JavaNonTerminalProgramElement context, 
		   StatementBlock block,  
		   SVInstantiations svInst) {
	return 
	    JavaBlock.createJavaBlock((StatementBlock)ProgramContextAdder.INSTANCE.start
				      (context, block, 
				       (ContextStatementBlockInstantiation)svInst.
				       getContextInstantiation().getInstantiation()));
    }


    private Statement returnFlagDecl(ProgramVariable returnFlag,
				     SVInstantiations svInst) {
	return KeYJavaASTFactory.
	    declare(returnFlag, BooleanLiteral.FALSE,
		    javaInfo.getKeYJavaType("boolean"));
    }

    private Term returnCase(ProgramVariable returnFlag,
			    KeYJavaType returnType,
			    ProgramVariable returnExpression,
			    Term post,
			    SVInstantiations svInst,
			    JavaNonTerminalProgramElement context) {
	Term executeReturn = tf.createProgramTerm
	    (Op.BOX, 
	     addContext(context, 
			new StatementBlock
			(KeYJavaASTFactory.returnClause(returnExpression)),
			svInst), 
	     post);
	
	return tf.createJunctorTerm
	    (Op.IMP, 
	     tf.createEqualityTerm(Op.EQUALS, 
				   typeConv.convertToLogicElement(returnFlag), 
				   typeConv.getBooleanLDT().getTrueTerm()), 
	     executeReturn);
	
    }


    private Statement breakFlagDecl(ProgramVariable breakFlag) {
	return KeYJavaASTFactory.
	    declare(breakFlag, BooleanLiteral.FALSE,
		    javaInfo.getKeYJavaType("boolean"));
    }

    private Statement contFlagDecl(ProgramVariable contFlag) {
	return KeYJavaASTFactory.
	    declare(contFlag, BooleanLiteral.FALSE,
		    javaInfo.getKeYJavaType("boolean"));
    }

    private Term breakCase(ProgramVariable breakFlag,
			   Term post,
			   ArrayList breakIfCascade,
			   SVInstantiations svInst,
			   JavaNonTerminalProgramElement context) {
	Term executeBreak = 
	    tf.createProgramTerm
	    (Op.BOX,
	     addContext(context, 
			new StatementBlock
			((Statement[])breakIfCascade.toArray(new Statement[0])),
			svInst),
	     post);
	return tf.createJunctorTerm
	    (Op.IMP, 
	     tf.createEqualityTerm(Op.EQUALS, 
				   typeConv.convertToLogicElement(breakFlag), 
				   typeConv.getBooleanLDT().getTrueTerm()), 
	     executeBreak); 
    }


    private Term  normalCaseAndContinue(Term contFlagTerm,
					Term returnFlagTerm,
					Term breakFlagTerm,
					Term excFlagTerm,
					Term inv) {

	final Term TRUE_TERM = typeConv.getBooleanLDT().getTrueTerm();

	ArrayList al = new ArrayList();

	if (returnFlagTerm != null)
	    al.add(tf.createEqualityTerm(Op.EQUALS, returnFlagTerm, TRUE_TERM));
	if (breakFlagTerm != null)
	    al.add(tf.createEqualityTerm(Op.EQUALS, breakFlagTerm, TRUE_TERM));
	if (excFlagTerm != null)
	    al.add(tf.createEqualityTerm(Op.EQUALS, excFlagTerm, TRUE_TERM));

	if (al.size() == 0) {
	    if (contFlagTerm == null)
		return inv;
	    else 
		return tf.createJunctorTerm(Op.IMP, contFlagTerm, inv);
	} else {
	    Term premiss = tf.createJunctorTerm(Op.NOT, 
						createLongJunctorTerm(Op.OR, al));
	    if (contFlagTerm != null)
		premiss = tf.createJunctorTerm(Op.OR, 
					       contFlagTerm,
					       premiss);	    
	    
	    return 
		tf.createJunctorTerm(Op.IMP, premiss, inv);
	}	
    }
    

    private Term throwCase(ProgramVariable excFlag,
			   ProgramVariable thrownException,
			   Term post,
			   SVInstantiations svInst,
			   JavaNonTerminalProgramElement context) {
	Term throwException = 
	    tf.createProgramTerm
	    (Op.BOX, addContext
	     (context, 
	      new StatementBlock(KeYJavaASTFactory.throwClause(thrownException)),
	      svInst), 
	     post);
	return tf.createJunctorTerm
	    (Op.IMP, 
	     tf.createEqualityTerm(Op.EQUALS, 
				   typeConv.convertToLogicElement(excFlag), 
				   typeConv.getBooleanLDT().getTrueTerm()), 
	     throwException);
    }
    
}

