// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.rule.updatesimplifier;

import de.uka.ilkd.key.logic.Constraint;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.rule.AbstractUpdateRule;
import de.uka.ilkd.key.rule.UpdateSimplifier;

/**
 * Abstract update simplification rule for access operators.
 * @author bubel
 */
public abstract class ApplyOnAccessTerm extends AbstractUpdateRule {
    
    /** three valued logics */ 
    protected static final int TRUE = 1;
    protected static final int UNKNOWN = 0;
    protected static final int FALSE = -1;
    /**
     * @param updateSimplifier
     */
    public ApplyOnAccessTerm(UpdateSimplifier updateSimplifier) {
        super(updateSimplifier);
    }

    /**
     * creates the result term of the update application
     * 
     * @param locSubs
     *            array of subterms describing the reference representant of the
     *            locations
     * @param mayBeAliases
     * @param simplifiedLocation
     *            the simplified locations ie. with the update being applied on
     *            its subterms (the subterms should be the same as the one of
     *            the first parameter)
     * @return the term describing the value of the original location evaluated
     *         under the update
     */
    protected abstract Term createAliasResolutionTerm(Term[] locSubs,
            ArrayOfAssignmentPair mayBeAliases, Term simplifiedLocation);

    /**
     * Tries to optimize the given array of may be aliases, i.e. it is tried to
     * retrieve if one of the given assignment pairs can be definitely told
     * aliased and to cut the array at this position.
     * 
     * @param subs
     * @param maybeAliasedAssignmentPairs
     *            the AssignmentPairs who are
     * @return optimized vmay be aliases chain
     */
    protected ArrayOfAssignmentPair optimize(Term[] targetSubs, ArrayOfAssignmentPair maybeAliasedAssignmentPairs) {
    
        ListOfAssignmentPair mayBeAliases = SLListOfAssignmentPair.EMPTY_LIST;
    
        for (int i = maybeAliasedAssignmentPairs.size() - 1; i >= 0; i--) {
            AssignmentPair pair = maybeAliasedAssignmentPairs
                    .getAssignmentPair(i);
            mayBeAliases = mayBeAliases.prepend(pair);
    
            int isDefinitelyAliased = isDefinitelyAliased(pair.locationSubs(),
                    targetSubs);
            if (isDefinitelyAliased == 1) {
                return new ArrayOfAssignmentPair(mayBeAliases.toArray());
            } else if (isDefinitelyAliased == -1) {
                mayBeAliases = mayBeAliases.tail();
            }
        }
        return new ArrayOfAssignmentPair(mayBeAliases.toArray());
    }

    /**
     * determines the alias relationship between two locations. The location
     * operator must have been checked before. This method assumes that the
     * given location sub terms belong to potential aliased locations.
     * 
     * @param aliasedSubs
     *            an array of Term with the to be tested to be aliased
     * @param reference
     * @return returns a three-valued truth value
     *         <ul>
     *         <li>TRUE, both locations describe the same location</li>
     *         <li>UNKNOWN, cannot determine aliasing by the available static
     *         information</li>
     *         <li>FALSE, the both locations describe definitely different
     *         locations</li>
     *         </ul>
     */
    protected int isDefinitelyAliased(Term[] aliasedSubs, Term[] originalSubs) {
    
        Constraint constrainingAlias = Constraint.BOTTOM;
        boolean compatibleSort = aliasedSubs[0].sort().extendsTrans(
                originalSubs[0].sort())
                || originalSubs[0].sort().extendsTrans(aliasedSubs[0].sort());
    
        if (!compatibleSort) {
            return FALSE;
        }
    
        // this works well for attribute and arrays inclusive their interference
        // with their shadowed versions
        for (int i = 0; i < aliasedSubs.length; i++) {
            constrainingAlias = constrainingAlias.unify(originalSubs[i],
                    aliasedSubs[i]);
        }
    
        return (constrainingAlias.isAsWeakAs(updateSimplifier()
                .formulaConstraint())) ? TRUE : UNKNOWN;
    }

    
  
}
