// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
/*
 * Created on 18.12.2004
 */
package de.uka.ilkd.key.rule.updatesimplifier;

import de.uka.ilkd.key.logic.Constraint;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.logic.op.AttributeOp;
import de.uka.ilkd.key.rule.UpdateSimplifier;

/**
 * An abstract update simplification rule for update application on attribute 
 * access terms. This class is inherited by the concrete ruels for shadowed 
 * and unshadowed attribute accesses.
 * 
 * @author bubel
 */
public abstract class ApplyOnAttributeAccess extends
        ApplyOnAccessTerm {

    
    
    /**
     * 
     * @param updateSimplifier 
     */
    public ApplyOnAttributeAccess(UpdateSimplifier updateSimplifier) {
        super(updateSimplifier);
    }

    /**
     * applies an update on a term representing an instance attribute
     * 
     * @param update
     *            the Update to be applied on target
     * @param target
     *            the Term representing an instance attribute
     * @return the simplified term describing the value of target under the
     *         given update
     * @requires update != null && target!=null && target.op() instanceof
     *           AttributeOp && !(target.op() instanceof ShadowedOperator)
     */
    public Term apply(Update update, Term target) {
        logEnter(update, target);

        final AttributeOp targetOp = (AttributeOp) target.op();

        // the field reference part (and if shadowed the transaction number)
        // of target evaluated in the state described by the update
        final PropagationResult pr = propagateUpdateToSubterms(update, target);
        final Term[] subs = pr.getSimplifiedSubterms();
        boolean changed = pr.hasChanged();

        // check if aliased relationship can be derived
        ArrayOfAssignmentPair mayBeAliases = optimize(subs, update
                .getAssignmentPairs(targetOp));

        // returns the conditional star term and simplifies as far
        // as possible:
        //        (Uu ?= l_i,...,l_(i+k)) t_i:...:t_(i+k):*.a
        // where
        //  * update = {l_1:=t_1,..,l_n:=t_n},
        //  * l_j with j in {i,..,i+k} are of shape "l_j.a" and
        //  * target = u.a
        final TermFactory tf = UpdateSimplifierTermFactory.DEFAULT
                .getBasicTermFactory();

        Term result = changed ? tf.createAttributeTerm(targetOp, subs) : target;

        result = createAliasResolutionTerm(subs, mayBeAliases, result);

        logExit(result);
        return result;
    }

}
