// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.rule.updatesimplifier;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;

import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.op.Location;
import de.uka.ilkd.key.logic.op.Operator;
import de.uka.ilkd.key.logic.op.UpdateOperator;
import de.uka.ilkd.key.rule.AbstractUpdateRule;
import de.uka.ilkd.key.rule.UpdateSimplifier;


/**
 * Applies an update on an array using rule ...
 */
public class ApplyOnUpdate extends AbstractUpdateRule {
       
    /**
     * creates an instance of this rule used by the given update
     * simplifier
     */
    public ApplyOnUpdate(UpdateSimplifier us) {
	super(us);
    }

    /**
     * determines whether this rule is applicable for the 
     * update x target (the term on which the update will be
     * applied) pair 
     * @param update the Update to be applied on target
     * @param target the Term on which the update is applied
     * @return true if the top level operator of target is an
     * attribute operator
     */
    public boolean isApplicable(Update update, Term target) {
	return target.op() instanceof UpdateOperator;
    }

    /**
     * determines whether this rule is applicable for the pair of update and
     * target (the term on which the update will be applied) term
     * 
     * @param update
     *            the Update to be applied on the given term
     * @param target
     *            the Term on which the update is applied
     * @return true if the rule can be applied on the update, target pair
     */
    public Term apply(Update update, Term target) {
        final UpdateOperator targetOp = (UpdateOperator) target.op();
        final ArrayList asgnPairs = new ArrayList();

        final ArrayOfAssignmentPair allUpdatePairs = 
	    update.getAllAssignmentPairs();
        
        for (int i = 0; i < allUpdatePairs.size(); i++) {
            asgnPairs.add(allUpdatePairs.getAssignmentPair(i));
        }

        for (int accessOpPos = 0, targetLocCount = targetOp.locationCount(); 
             accessOpPos < targetLocCount; accessOpPos++) {
            final Location accessOp = targetOp.location(accessOpPos);
            final int valuePos = targetOp.valuePos(accessOpPos);
            // apply update on location
            final Term[] locSubs = 
                applyOnLocation(update, target, accessOp, valuePos);
            // apply update on value
            final Term value = updateSimplifier().
                simplify(update, target.sub(valuePos));
            asgnPairs.add(new AssignmentPairImpl(accessOp, locSubs, value));
        }

        final AssignmentPair[] result = optimize(asgnPairs);

        return updateSimplifier().simplify(Update.createUpdate(result),
                    target.sub(target.arity()-1));
    }

    /**
     * TODO comment
     */
    private AssignmentPair[] optimize(ArrayList assignmentPairs) {
        LinkedHashMap mayBeAliasChains = new LinkedHashMap();
        final int locationCount = assignmentPairs.size();

        for (int i = locationCount-1; i >= 0; i--) {
            final AssignmentPair current = (AssignmentPair) assignmentPairs.get(i);
            LinkedHashMap chain = 
                  (LinkedHashMap) mayBeAliasChains.get(current.location());            
            if (chain == null) {
                chain = new LinkedHashMap();
                mayBeAliasChains.put(current.location(), chain);
            }
            final AssignmentPair.LocationAsKey key = 
		new AssignmentPair.LocationAsKey(current);
              if (!chain.containsKey(key)) {
                chain.put(key, current);
              } 
        }
        
        // get all assignment pairs
        LinkedList result = new LinkedList();
        Iterator it = mayBeAliasChains.values().iterator();
        while (it.hasNext()) {
            Iterator pairIt = ((LinkedHashMap)it.next()).values().iterator();
            while (pairIt.hasNext()) {
                result.addFirst(pairIt.next());
             }            
        }
       
        return (AssignmentPair[])
            result.toArray(new AssignmentPair[result.size()]) ;
    }

    /**
     * TODO comment
     * @param update
     * @param target
     * @param accessOp
     * @param valuePos
     * @return
     */
    private Term[] applyOnLocation(Update update, Term target, 
            final Operator accessOp, final int valuePos) {
        final Term[] locSubs = new Term[accessOp.arity()];
        // applies update on location
        for (int i = 0; i < locSubs.length; i++) {
            final Term simplified = updateSimplifier().simplify(update,
                    target.sub(valuePos - (i + 1)));
            locSubs[locSubs.length-(i+1)] = simplified;
        }
        return locSubs;
    }
}
