// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.rule.updatesimplifier;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.Location;
import de.uka.ilkd.key.logic.op.UpdateOperator;

/**
 * Models an assignment pair <code> l_i := t_i </code>  of an update.
 * This data structure is only used for update simplification.
 */
public class AssignmentPairLazy implements AssignmentPair {
    
    /** 
     * term with the update operator as top level operator to
     * which the modeled assignment pair belongs 
     */
    private final Term update;
    /**
     * the update operator of term <code>update</code>
     */
    private final UpdateOperator updateOp;
    /** the position <code> i </code> of the location in the update operator */
    private final int locationPos;

    /** caches the location <code> l_i </code> */
    private Location location;

    /** stores the location hash code */
    private int locationHash = 0;
    
    /**
     * creates the assignment pair <code> l_i := t_i </code>
     * @param update the Term with update as top level operator whose
     * locationPos-th assignment pair is modeled 
     * @param locationPos the position of the location of this assignment pair
     */
    public AssignmentPairLazy(Term update, int locationPos) {
	this.update      = update;
	this.updateOp    = (UpdateOperator) update.op();
	this.locationPos = locationPos;
    }

    /**
     * returns the location
     * @return the left side of the assignment
     */
    public Term locationAsTerm() {
	return updateOp.location(update, locationPos);
    }

    /**
     * returns the location operator
     * @return the location specifying operator
     */
    public Location location() {
	if (location == null) {
	    location = updateOp.location(locationPos);
	}
	return location;
    }

    /**
     * returns the locations sub terms
     * @return the left side of the assignment
     */
    public Term[] locationSubs() {
	final Term[] result = new Term[location().arity()];
	final int valuePos = updateOp.valuePos(locationPos);
	for (int i = result.length; i>0; i--) {
	    result[result.length-i] = update.sub(valuePos - i);
	}
	return result;
    }

    /**
     * returns an int fulfilling the usual hashcode properties but
     * without consideration of the value part of the assignment pair
     * @return an int as location hashcode 
     */
    public int locationHashCode() {
	if (locationHash == 0) {
	    locationHash = location().hashCode();
	    final Term[] locSubs = locationSubs();
	    for (int i = 0; i<locSubs.length; i++) {
		locationHash += 17*locSubs[i].hashCode();
	    }
	    if (locationHash == 0) { 
		locationHash = 1;
	    }
	}
	return locationHash;
    }

    /**
     * returns true if the location that is updated is equal to location
     * of the given assignment pair
     * @param p the AssignmentPair whose location is compared
     * @return true if the location that is updated is equal to location
     * of the given assignment pair
     */
    public boolean equalLocations(AssignmentPair p) {	
	if (p.location() != location()) {
	    return false;
	}

	final Term[] pLocSubs = p.locationSubs();
	final Term[] tLocSubs = this.locationSubs();
	
	if (pLocSubs.length != tLocSubs.length) {
	    return false;
	} 

	for (int i = 0; i<pLocSubs.length; i++) {
	    if (!(pLocSubs[i].equals(tLocSubs[i]))) {
		return false;
	    }
	}
	return true;
    }

    
    /**
     * returns true if the given object equals this one
     */
    public boolean equal(Object  o) {
	if (!(o instanceof AssignmentPair)) {
	    return false;
	}

	final AssignmentPair p = (AssignmentPair)o;
	
	return equalLocations(p) && p.value().equals(value());
    }

    /**
     * the hashCode of this assignment pair
     */
    public int hashCode() {
	return locationHashCode() + 17*value().hashCode();
    }

    /** 
     * returns the value that is assigned to the location
     * @return the right side of the assignment      
     */
    public Term value() {
	return update.sub(updateOp.valuePos(locationPos));
    }

    /** prints the assignment pair as location */
    public String toString() {
	return locationAsTerm() + " := " + value();
    }

}
