// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.rule.updatesimplifier;

import java.util.HashMap;
import java.util.HashSet;

import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.op.AnonymousUpdate;
import de.uka.ilkd.key.logic.op.Function;
import de.uka.ilkd.key.logic.op.Location;
import de.uka.ilkd.key.logic.op.UpdateOperator;
import de.uka.ilkd.key.logic.sort.AbstractSort;
import de.uka.ilkd.key.logic.sort.SetAsListOfSort;
import de.uka.ilkd.key.logic.sort.SetOfSort;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.util.Debug;

/**
 * represents an update and is used as basis data structure for the
 * updatesimplifier. Currently this is only a naive default implementation.
 */
public abstract class Update {
    
    // just for unique handling of anonymous updates
    // it as to be taken care that this sort does not escape
    private static final Sort SPECIAL_SORT = 
        new AbstractSort(new Name("special_sort")) {
            public SetOfSort extendsSorts() {              
                return SetAsListOfSort.EMPTY_SET;
            }
    	};
    
    /**
     * @author bubel
     */
    public static class UpdateInParts extends Update {

        private final HashMap loc2assignmentPairs;

        private final HashSet locationCache;

        private final ArrayOfAssignmentPair pairs;

        /**
         * @param pairs
         */
        public UpdateInParts(final ArrayOfAssignmentPair pairs) {
            this.pairs = pairs;
            loc2assignmentPairs = new HashMap();
            this.locationCache = new HashSet();

            for (int i = 0; i < pairs.size(); i++) {
                locationCache.add(pairs.getAssignmentPair(i).location());
            }

        }

        /*
         * (non-Javadoc)
         * 
         * @see de.uka.ilkd.key.rule.updatesimplifier.Update#getAllAssignmentPairs()
         */
        public ArrayOfAssignmentPair getAllAssignmentPairs() {
            return pairs;
        }

        /*
         * (non-Javadoc)
         * 
         * @see de.uka.ilkd.key.rule.updatesimplifier.Update#getAssignmentPairs(de.uka.ilkd.key.logic.op.Operator)
         */
        public ArrayOfAssignmentPair getAssignmentPairs(Location loc) {
            if (!loc2assignmentPairs.containsKey(loc)) {
                ListOfAssignmentPair result = SLListOfAssignmentPair.EMPTY_LIST;
                for (int i = pairs.size() - 1; i >= 0; i--) {
                    AssignmentPair assignmentPair = pairs.getAssignmentPair(i);
                    final Location op = assignmentPair.location();
                    if (loc.mayBeAliasedBy(op)) {
                        result = result.prepend(assignmentPair);
                    }
                }

                loc2assignmentPairs.put(loc, new ArrayOfAssignmentPair(result
                        .toArray()));
            }
            return (ArrayOfAssignmentPair) loc2assignmentPairs.get(loc);
        }

        /*
         * (non-Javadoc)
         * 
         * @see de.uka.ilkd.key.rule.updatesimplifier.Update#hasLocation(de.uka.ilkd.key.logic.op.Operator)
         */
        public boolean hasLocation(Location loc) {
            return locationCache.contains(loc);
        }

        /*
         * (non-Javadoc)
         * 
         * @see de.uka.ilkd.key.rule.updatesimplifier.Update#location(int)
         */
        public Location location(int n) {
            return pairs.getAssignmentPair(n).location();
        }

        /*
         * (non-Javadoc)
         * 
         * @see de.uka.ilkd.key.rule.updatesimplifier.Update#locationCount()
         */
        public int locationCount() {
            return pairs.size();
        }

    }

    private static class UpdateWithUpdateTerm extends Update {

        private final HashMap loc2assignmentPairs;

        private final HashSet locationCache;

        private final Term update;

        private final UpdateOperator updateOp;

        public UpdateWithUpdateTerm(Term update) {
            this.update = update;
            this.updateOp = (UpdateOperator) update.op();
            this.loc2assignmentPairs = new HashMap();
            this.locationCache = new HashSet();

            for (int i = 0; i < updateOp.locationCount(); i++) {
                locationCache.add(updateOp.location(i));            
            }   
            
            Debug.assertTrue(!locationCache.contains
                    (StarLocation.ALL), 
                    	"Anonymous update shall not be created using this " +
                    	"class. (at least as long no generalized terms are " +
                    	"supported)");
        }

        /**
         * @return
         */
        public ArrayOfAssignmentPair getAllAssignmentPairs() {
            final AssignmentPair[] result = 
                new AssignmentPair[locationCount()];
            for (int i = 0; i < locationCount(); i++) {
                result[i] = new AssignmentPairLazy(update, i);
            }
            return new ArrayOfAssignmentPair(result);
        }

        /**
         * determines and returns all assignment pairs whose location part may 
         * be an alias of <code>loc</code>
         */
        public ArrayOfAssignmentPair getAssignmentPairs(Location loc) {
            if (!loc2assignmentPairs.containsKey(loc)) {
                ListOfAssignmentPair result = SLListOfAssignmentPair.EMPTY_LIST;
                for (int i = updateOp.locationCount() - 1; i >= 0; i--) {
                    final Location op = updateOp.location(i);
                    if (loc.mayBeAliasedBy(op)) {
                        result = 
                            result.prepend
                            	(new AssignmentPairLazy(update, i));
                    }
                }

                loc2assignmentPairs.put(loc, 
                        new ArrayOfAssignmentPair(result.toArray()));
            }
            return (ArrayOfAssignmentPair) loc2assignmentPairs.get(loc);
        }

        public boolean hasLocation(Location loc) {
            return locationCache.contains(loc);
        }

        public Location location(int n) {
            return updateOp.location(n);
        }

        /**
         * returns the number of locations
         * 
         * @return the number of locations as int
         */
        public int locationCount() {
            return updateOp.locationCount();
        }

    }

    /**
     * creates an update representation out of the top level operator of the
     * given update term
     */
    public static Update createUpdate(AssignmentPair[] pairs) {
        return new UpdateInParts(new ArrayOfAssignmentPair(pairs));
    }

    /**
     * creates an update representation out of the top level operator of the
     * given update term
     */
    public static Update createUpdate(Term t) {        
        if (t.op() instanceof AnonymousUpdate) {
           Term valueTerm = UpdateSimplifierTermFactory.
           DEFAULT.getBasicTermFactory().
           	createFunctionTerm
           	(new Function(t.op().name(), SPECIAL_SORT, new Sort[0]));
           AssignmentPair pair = new AssignmentPairImpl
           	(StarLocation.ALL, new Term[0],valueTerm);
           return createUpdate
           	(new AssignmentPair[]{pair});
        }
        return new UpdateWithUpdateTerm(t);
    }

    /**
     * @return
     */
    public abstract ArrayOfAssignmentPair getAllAssignmentPairs();

    /**
     * determines and returns all assignment pairs whose location part has the
     * same top level operator as the given one
     */
    public abstract ArrayOfAssignmentPair getAssignmentPairs(Location loc);

    /**
     * returns true if the given location is updated by this update
     * 
     * @param location
     *            the Operator specifying the location which is updated
     * @return true if location occurs on the left side of an assignment pair in
     *         this update
     */
    public abstract boolean hasLocation(Location loc);

    /**
     * returns the n-th location operator
     * 
     * @param n
     *            an int specifying the position of the location operator to be
     *            retrieved
     * @return the n-tl location operator
     */
    public abstract Location location(int n);

    /**
     * returns the number of locations
     * 
     * @return the number of locations as int
     */
    public abstract int locationCount();
    
    /**
     * returns true if this update object describes an anonymous update
     */
    public boolean isAnonymousUpdate() {
        return hasLocation(StarLocation.ALL);
    }
    
    public String toString() {
        StringBuffer result = new StringBuffer("{");
        ArrayOfAssignmentPair pairs = getAllAssignmentPairs();
        for (int i = 0; i<pairs.size(); i++) {           
            result.append(pairs.getAssignmentPair(i).toString());
            if (i<pairs.size()-1) result.append(",");
        }
        result.append("}");
        return result.toString();
    }

    // these classes are used to unify treatment of anonymous and 
    // normal updates. May become obsolete with generalized terms 
    public static class StarLocation extends Function 
    	implements Location {

        public final static StarLocation ALL = new StarLocation();
        
        private final static Name name = new Name("*");
        
        private StarLocation() {
            super(name, SPECIAL_SORT, new Sort[0]);
        }
        
        /* (non-Javadoc)
         * @see de.uka.ilkd.key.logic.op.Location#mayBeAliasedBy(de.uka.ilkd.key.logic.op.Location)
         */
        public boolean mayBeAliasedBy(Location loc) {            
            return false;
        }

        /* (non-Javadoc)
         * @see de.uka.ilkd.key.logic.op.Operator#name()
         */
        public Name name() {            
            return name;
        }

        /* (non-Javadoc)
         * @see de.uka.ilkd.key.logic.op.Operator#validTopLevel(de.uka.ilkd.key.logic.Term)
         */
        public boolean validTopLevel(Term term) {
            return term.arity() == 0 && term.sort() != Sort.FORMULA;
        }

        /* (non-Javadoc)
         * @see de.uka.ilkd.key.logic.op.Operator#sort(de.uka.ilkd.key.logic.Term[])
         */
        public Sort sort(Term[] term) {            
            return SPECIAL_SORT;
        }
        
        /* (non-Javadoc)
         * @see de.uka.ilkd.key.logic.op.Operator#arity()
         */
        public int arity() {            
            return 0;
        }

        /* (non-Javadoc)
         * @see de.uka.ilkd.key.logic.op.Operator#isRigid(de.uka.ilkd.key.logic.Term)
         */
        public boolean isRigid(Term term) {          
            return true;
        }
    
    }    
}
