// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.rule.updatesimplifier;

import java.util.LinkedList;

import de.uka.ilkd.key.logic.IteratorOfTerm;
import de.uka.ilkd.key.logic.ListOfTerm;
import de.uka.ilkd.key.logic.SLListOfTerm;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.logic.op.Location;
import de.uka.ilkd.key.util.Debug;

public class UpdateSimplifierTermFactory {
    
    public final static UpdateSimplifierTermFactory DEFAULT = 
	new UpdateSimplifierTermFactory();


    private final static TermFactory tf = TermFactory.DEFAULT;

    
    /**
     * returns true if all then and else parts are equal 
     * @param conditionals an array of Term with the sub terms of the 
     * conditional operator 
     * @param branchStep an int describing the steps to perform between two 
     * branches  
     * @return true if the branches are all equal
     */
    private boolean checkSameBranches(Term[] conditionals, int branchStep) {	
	Term lastBranch = conditionals[branchStep];
	for (int i = 2*branchStep+1; i<conditionals.length; i+=(branchStep+1)) {
	    if (!lastBranch.equals(conditionals[i])) {
		return false;
	    }
	}
	return true;
    }
       
    private Term createConditionalStarTerm(Term comparisionTerm, 
					   ListOfTerm compareList,
					   ListOfTerm thenList,
					   Term starTerm) {
	return createConditionalTerm(comparisionTerm, compareList,
				     thenList, starTerm);
    }


    /**
     * creates a conditional star term and simplifies as far as
     * possible
     * @param compareThenPairs are in inverted order
     */
    public Term createConditionalStarTermAndSimplify
	(Term comparisionTerm,
	 ArrayOfAssignmentPair compareThenPairs, 
	 Term starTerm) {

	Debug.assertTrue(compareThenPairs != null && 
			 compareThenPairs.size() > 0, 
			 "Called createConditionalTermAndSimplify " + 
			 "with empty comparision list (or null).");
	

	Term elseTerm = starTerm;
	ListOfTerm compareList = SLListOfTerm.EMPTY_LIST;
	ListOfTerm thenList    = SLListOfTerm.EMPTY_LIST;

	// flag indicating if all branches are syntactical equal then
	// we can just return one of them
	boolean sameBranches = true;
	// flag that indicates at which position a compare term is
	// equal to the comparision term then we can stop building the
	// conditional term at this position
	Term lastBranch = null;

	for (int i = compareThenPairs.size()-1; i>=0; i--) {
	    AssignmentPair current = compareThenPairs.getAssignmentPair(i);
	    if (current.locationSubs()[0].equals(comparisionTerm)) {
		if (!sameBranches) {
		    // we have a situation like:
		    //   (u ?= o1,...,o_(i-1),u,o_i+1,....,on) t1:...:tn:*.a
		    // which can be simplified to
		    //   (u ?= o1,...o_(i-1)) t1:...:t_(i-1):t_i		
		    return createConditionalTerm(comparisionTerm, 
						 compareList, 
						 thenList,
						 current.value());
		} else {
		    // ensures that lastBranch ist non-null
		    elseTerm = current.value();
		    break;
		}	
	    }

	    compareList = compareList.prepend(current.locationSubs()[0]);
	    thenList    = thenList.prepend(current.value());		    	   
        
	    // check for same branch property
	    if (sameBranches) {
		if (lastBranch == null) {
		    lastBranch = current.value();
		} else {
		    sameBranches = lastBranch.equals(current.value());
		}		
	    }
	}
	
	return sameBranches && elseTerm.equals(lastBranch) ? 
	    lastBranch : createConditionalStarTerm(comparisionTerm,
						   compareList,
						   thenList,
						   elseTerm);
    }

    private Term createConditionalTerm(Term comparisionTerm, 
				       ListOfTerm compareList,
				       ListOfTerm thenBranches,
				       Term elseBranch) {
	Debug.assertTrue(compareList.size() == thenBranches.size(), 
			 "Tried to construct misformed conditional term.");

	final IteratorOfTerm compareIt = compareList.iterator();
	final IteratorOfTerm thenIt    = thenBranches.iterator();
	Term result = elseBranch;	

	while (compareIt.hasNext()) {
	    result = tf.createIfElseTerm(comparisionTerm, 
	                                 compareIt.next(), 
					 thenIt.next(), result);
	}

	return result;
    }


    /**
     * helper method for the creation conjunction conditional terms.
     * Called with <code>createConjCondTerm(s, 0)</code> 
     * the term
     * <code>
     *    s[0]*s[1] : ( s[0]*s[1] : s[2] ?= s[6]*s[7]:s[8] ) ?= s[3]*s[4]:s[5]
     * </code>
     * with length of <tt>s</tt> equals to nine.
     */
    private Term createConjCondTerm(Term[] subs, int start) {       
	if (start+5 >= subs.length) {
	    return subs[2];
	} else {
	    return tf.createConjCondTerm(new Term[]{subs[0], subs[1], createConjCondTerm(subs, start+3), 
						    subs[start+3], subs[start+4], subs[start+5]});
	}

    }


    /**
     */
    public Term createConjunctionConditionalStarTerm
	(Term[] fstComparators, 
	 ArrayOfAssignmentPair compareThenPairs, 
	 Term starTerm) {

	LinkedList conditionalSubs = new LinkedList();

	for (int i = 0; i<fstComparators.length; i++) {
	    conditionalSubs.add(fstComparators[i]);
	}
	conditionalSubs.add(starTerm);


        for (int j = compareThenPairs.size() - 1; j>=0; j--) {
	    final AssignmentPair asgnPair = compareThenPairs.getAssignmentPair(j);
	    final Term[] locSubs = asgnPair.locationSubs();

	    boolean bothComparatorsAreEquals = true;

	    for (int i = 0; i<locSubs.length; i++) {
		bothComparatorsAreEquals = bothComparatorsAreEquals && 
		    locSubs[i].equals(fstComparators[i]);
		conditionalSubs.add(locSubs[i]);
	    }

	    if (bothComparatorsAreEquals) {

		for (int i = 0; i<locSubs.length; i++) {
		    conditionalSubs.removeLast();
		}
		conditionalSubs.remove(fstComparators.length);
		conditionalSubs.add(fstComparators.length, asgnPair.value());
		break;

	    } else {
		conditionalSubs.add(asgnPair.value());
	    }
	}

	final Term[] conditionals = 
	    (Term[])conditionalSubs.toArray(new Term[conditionalSubs.size()]);
	
	return checkSameBranches(conditionals, 2) ? 
	    conditionals[fstComparators.length] : createConjCondTerm(conditionals, 0);
    }

    /**
     * creates an update term out of the given internal representation
     * used by the update simplifier, i.e. the assignmentpairs are made
     * explicit
     * @param update an ArrayOfAssignmentPair with the assignmentpairs
     * <tt>l:=r</tt> of a simulataneous update.
     * @param target the term evaluated under the update
     * @return the update term <tt>{l1:=r1,...,ln:=rn}target</tt>
     */
    public Term createUpdateTerm
	(ArrayOfAssignmentPair assignmentPairs, Term target) {
        final Location[] locations = new Location[assignmentPairs.size()];
        LinkedList subs = new LinkedList();       
        for (int i = 0; i<assignmentPairs.size(); i++) {            
            final AssignmentPair assignmentPair = 
                assignmentPairs.getAssignmentPair(i);            
            locations[i] = assignmentPair.location();
            if (locations[i] == Update.StarLocation.ALL) {
                Debug.assertTrue(locations.length==1,
                        "Special star operator tried to escape. " +
                        "(not allowed as long as no generalized terms)");              
                return tf.createAnonymousUpdateTerm
                (assignmentPair.value().op().name(), target);
            }
            final Term[] locationSubs = assignmentPair.locationSubs();
            for (int j = 0; j<locationSubs.length; j++) {                 
                subs.add(locationSubs[j]);            
            }
            subs.add(assignmentPair.value());
        }    
        subs.add(target);
        return tf.createUpdateTerm(locations, (Term[])subs.toArray
                (new Term[subs.size()]));      
    }  

    /**
     * creates an update term out of the given internal representation
     * used by the update simplifier, i.e. the assignmentpairs are made
     * explicit
     * @param update an array of AssignmentPairs <tt>l:=r</tt> of the
     * simulataneous update to be created
     * @param target the term evaluated under the update
     * @return the update term <tt>{l1:=r1,...,ln:=rn}target</tt>
     */
    public Term createUpdateTerm(AssignmentPair[] update, Term target) 
    {
        return createUpdateTerm(new ArrayOfAssignmentPair(update), 
                target);
    }

    /**
     * @return the default termfactory for JavaCardDL terms
     */
    public TermFactory getBasicTermFactory() {    
        return tf;
    }
					   

}
