// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.strategy;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.BooleanContainer;
import de.uka.ilkd.key.logic.Constraint;
import de.uka.ilkd.key.logic.PosInOccurrence;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.NewRuleListener;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.RuleAppIndex;
import de.uka.ilkd.key.rule.RuleApp;
import de.uka.ilkd.key.rule.TacletApp;


public class AutomatedRuleAppManager implements NewRuleListener {

    /** The goal this manager belongs to */
    private Goal goal = null;

    /** The priority queue containing all possible next rule applications,
     * ordered by the costs the strategy object has assigned to them */
    private HeapOfRuleAppContainer queue       = null;

    private HeapOfRuleAppContainer secQueue    = null;
    
    /** rule apps that have been deferred during the last call
     * of <code>next</code>, but that could be still relevant */
    private ListOfRuleAppContainer workingList = null;

    private static final int PRIMARY_QUEUE   = 0;
    private static final int SECONDARY_QUEUE = 1;
    private static final int WORKING_LIST    = 2;
    
  
    public void setGoal ( Goal p_goal ) {
	if ( goal != null )
	    goal.ruleAppIndex().removeNewRuleListener(this);

    	goal = p_goal;

	if ( goal != null )
	    goal.ruleAppIndex().addNewRuleListener(this);
    }


    /**
     * Clear the heap of applicable rules
     */
    public void clearCache () {
        queue       = null;
        secQueue    = null;
        workingList = null;
    }
    
    
    /**
     * Add all rules to the heap that are not reported via the <code>RuleListener</code>
     * connection
     */
    private void ensureQueueExists () {
	if ( queue != null )
	    return;

	if ( getGoal () == null ) {
            clearCache ();
	    return;
	}
		    
	queue       = LeftistHeapOfRuleAppContainer.EMPTY_HEAP;
        secQueue    = LeftistHeapOfRuleAppContainer.EMPTY_HEAP;
	workingList = SLListOfRuleAppContainer.EMPTY_LIST;

        getGoal ().ruleAppIndex ().reportAutomatedRuleApps ( this,
                                                             getServices (),
                                                             getUserConstraint () );
//        printQueue(queue);
    }


    /**
     * Implementation of the method from <code>NewRuleListener</code>. The new
     * rule app is added to the heap
     */
    public void ruleAdded(RuleApp rule, PosInOccurrence pos) {
        //System.out.println ( "Rule added: " + rule + "\n");
        
//	ensureQueueExists ();

        if ( queue == null )
            // then the heap has to be rebuilt completely anyway, and the new
            // rule app is not of interest for us
            return;
        
        push ( RuleAppContainer.createAppContainers
	           ( rule, pos, getGoal (), getStrategy () ).iterator (),
               PRIMARY_QUEUE );
    }


    /**
     * Add a number of new rule apps to the heap
     */
    private void push ( IteratorOfRuleAppContainer it, int target ) {
        while ( it.hasNext () )
            push ( it.next (), target );
    }


    /**
     * Add a new rule app to the heap, provided that the rule app is not
     * infinitely expensive
     */
    private void push (RuleAppContainer c, int target) {
        if ( c.getCost () instanceof TopRuleAppCost )
            return;

        switch ( target ) {
        case PRIMARY_QUEUE:
            queue = queue.insert ( c );
            break;
        case SECONDARY_QUEUE:
            secQueue = secQueue.insert ( c );
            break;
        case WORKING_LIST:
            workingList = workingList.prepend(c);
            break;
        }
    }

    /**
     * @return the first applicable rule app, i.e. the least expensive element
     * of the heap that is not obsolete
     */
    public RuleApp next() {
	ensureQueueExists ();

    	final RuleAppIndex index = getGoal ().ruleAppIndex ();
        index.fillCache ();
        
//        printQueue(queue);
//        printQueue(secQueue);
            
        createFurtherApps ();
        
        final RuleApp res = nextHelp();

//        System.out.println("Queue size: " + queue.size());
//        System.out.println("Secondary queue size: " + secQueue.size());
//        System.out.println("Working list size: " + workingList.size());

        queue = queue.insert ( secQueue );
        secQueue = LeftistHeapOfRuleAppContainer.EMPTY_HEAP;
        
    	return res;
    }


    private RuleApp nextHelp () {
        final BooleanContainer secondaryQueueUsed = new BooleanContainer ();        
        RuleApp res = null;
        
        while ( res == null && ( !queue.isEmpty() || !secQueue.isEmpty() ) ) {
            final RuleAppContainer c = getMinRuleApp ( secondaryQueueUsed );
//            printContainer ( "considering rule ", c );

            res = c.completeRuleApp(getGoal(), getStrategy());

            if ( res == null ) {
                if ( !secondaryQueueUsed.val () )
                    createFurtherRuleApps ( c, true );
                else
                    push ( c, WORKING_LIST );
            } else {
                // the found rule app will be re-evaluated when calling
                // <code>next()</code> the next time; all other considered rule
                // apps can be put into the primary queue immediately
                queue = queue.insert ( workingList.iterator () );
                workingList = SLListOfRuleAppContainer.EMPTY_LIST;
                push ( c, WORKING_LIST );
            }
            
//            if (res != null) {
//                printContainer ( "selected rule ", c );
//            }
        }
        
        return res;
    }

    private RuleAppContainer getMinRuleApp (BooleanContainer usedSecondary) {
        if ( queue.isEmpty () ) {
            usedSecondary.setVal ( true );
            final RuleAppContainer c = secQueue.findMin ();
            secQueue = secQueue.deleteMin ();
            return c;
        }

        if ( secQueue.isEmpty () ) {
            usedSecondary.setVal ( false );
            final RuleAppContainer c = queue.findMin ();
            queue = queue.deleteMin ();
            return c;
        }

        final RuleAppContainer priC = queue.findMin ();
        final RuleAppContainer secC = secQueue.findMin ();

        usedSecondary.setVal ( priC.compareTo ( secC ) > 0 );

        if ( usedSecondary.val () ) {
            secQueue = secQueue.deleteMin ();
            return secC;
        } else {
            queue = queue.deleteMin ();
            return priC;
        }
    }
    
    /**
     * Call the method <code>createFurtherApps</code> for all elements of
     * the list <code>consideredRecently</code>, and clear the list. The new
     * rule apps are added to the heap
     */
    private void createFurtherApps () {
    	final IteratorOfRuleAppContainer it = workingList.iterator();
    	workingList = SLListOfRuleAppContainer.EMPTY_LIST;
    
        while ( it.hasNext() )
            createFurtherRuleApps ( it.next (), true );
    }


    private void createFurtherRuleApps (RuleAppContainer app, boolean secondary) {
        push ( app.createFurtherApps ( getGoal (), getStrategy () ).iterator (),
               secondary ? SECONDARY_QUEUE : PRIMARY_QUEUE );
    }


    /**
     * The goal this manager belongs to
     */
    private Goal getGoal() {
        return goal;
    }

    private Services getServices() {
	return getProof ().getServices ();
    }

    private Constraint getUserConstraint() {
        return getProof ().getUserConstraint ().getConstraint ();
    }

    private Proof getProof() {
	return getGoal ().node ().proof ();
    }

    private Strategy getStrategy () {
    	return getGoal().getGoalStrategy();
    }

    public AutomatedRuleAppManager copy () {
    	return (AutomatedRuleAppManager)clone ();
    }

    public Object clone () {
    	AutomatedRuleAppManager res = new AutomatedRuleAppManager ();
    	res.queue                   = queue;
        res.secQueue                = secQueue;
    	res.workingList             = workingList;
    	return res;
    }
    
    void printQueue(HeapOfRuleAppContainer p_queue) {
        IteratorOfRuleAppContainer it = p_queue.sortedIterator();
        
        System.out.println("---- start of queue ----");
        
        int n = 0;
        while (it.hasNext ()) {
            n++;
            final RuleAppContainer container = it.next ();

            final String prefix = n + ") ";
            printContainer ( prefix, container );
        }
        
        System.out.println("---- end of queue ----");
    }


    private void printContainer ( String           prefix,
            			  RuleAppContainer container ) {
        final RuleApp ruleApp = container.getRuleApp ();
        String message = prefix + ruleApp.rule ().name ()
                + " with cost " + container.getCost ();
        
        if ( ruleApp instanceof TacletApp ) {
            TacletApp tacletApp = (TacletApp) ruleApp;
            if ( !tacletApp.ifInstsComplete() ) {
                message = message + " (unmatched if-formulas)";
            }
        }
        
        System.out.println ( message );
    }
}
