// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.strategy.feature;

import de.uka.ilkd.key.logic.ConstrainedFormula;
import de.uka.ilkd.key.logic.PosInOccurrence;
import de.uka.ilkd.key.logic.Sequent;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.Node;
import de.uka.ilkd.key.rule.IteratorOfIfFormulaInstantiation;
import de.uka.ilkd.key.rule.PosTacletApp;
import de.uka.ilkd.key.rule.RuleApp;
import de.uka.ilkd.key.rule.TacletApp;
import de.uka.ilkd.key.util.Debug;


/**
 * Binary feature that returns zero iff a certain Taclet app has not already
 * been performed. Contrary to <code>NonDuplicateAppFeature</code>, this feature
 * is also able to handle failing meta-constructs correctly (these constructs
 * return equal, but not identical formulas in case of a failure), but is less
 * efficient.
 */
public class EqNonDuplicateAppFeature extends BinaryTacletAppFeature {

    public static final Feature INSTANCE = new EqNonDuplicateAppFeature ();

    private static boolean equalRuleApps (TacletApp app0,
                                          RuleApp app1,
                                          PosInOccurrence pio) {

        // compare the rules
        if ( app0.rule () != app1.rule () ) return false;

        final TacletApp cmp = (TacletApp)app1;

        // compare the position of application
        if ( pio != null ) {
            if ( !( cmp instanceof PosTacletApp ) ) return false;
            final PosTacletApp cmp2 = (PosTacletApp)cmp;
            if ( !cmp2.posInOccurrence ().eqEquals ( pio ) ) return false;
        }

        // compare the if-sequent instantiations
        if ( app0.ifFormulaInstantiations () == null
             && cmp.ifFormulaInstantiations () == null ) return true;

        if ( app0.ifFormulaInstantiations () != null
             && cmp.ifFormulaInstantiations () != null ) {
            final IteratorOfIfFormulaInstantiation it0 =
                app0.ifFormulaInstantiations ().iterator ();
            final IteratorOfIfFormulaInstantiation it1 =
                cmp.ifFormulaInstantiations ().iterator ();

            while ( it0.hasNext () ) {
                // this test should be improved
                if ( it0.next ().getConstrainedFormula () !=
                     it1.next ().getConstrainedFormula () )
                return false;
            }

            return true;
        }

        return false;
    }

    public boolean filter (TacletApp app, PosInOccurrence pos, Goal goal) {
        Debug.assertFalse ( pos == null,
                            "Feature is only applicable to rules with find" );

        if ( !app.ifInstsComplete () ) return true;

        final ConstrainedFormula cfma = pos.constrainedFormula ();
        final boolean antec = pos.isInAntec ();

        Node node = goal.node ();

        while ( !node.root () ) {
            final Node par = node.parent ();
            final Sequent pseq = par.sequent ();

            if ( !( antec ? pseq.antecedent () : pseq.succedent () )
                                        .containsEqual ( cfma ) )
                return true;
                
            final RuleApp rapp = par.getAppliedRuleApp ();
                
            if ( equalRuleApps ( app, rapp, pos ) )
                return false;

            node = par;
        }

        return true;
    }
}
